/* vectors.h
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __GIRAM_VECTORS_H__
#define __GIRAM_VECTORS_H__

typedef double Vector[5];
typedef double Matrix[4][4];

typedef struct TransformStruct
{
  Matrix Direct;
  Matrix Inverse;
} TransformStruct;

#define VLength(V) sqrt((V)[0]*(V)[0]+(V)[1]*(V)[1]+(V)[2]*(V)[2])
#define VDot(a,b) ( (a)[0]*(b)[0] + (a)[1]*(b)[1] + (a)[2]*(b)[2] )
#define Veq(V,a,b,c) (V)[0]=(a); (V)[1]=(b); (V)[2]=(c);

void MidPoint(Vector Result, Vector a, Vector b);
void VCross(Vector a, Vector b, Vector c);
void MIdentity(Matrix Mat);
void MTimes(Matrix Result, Matrix M1, Matrix M2);
void MInverse(Matrix Result, Matrix M);  
void MEvaluatePoint(Vector Result, TransformStruct *Trans, Vector Source);
void MEvaluateVector(Vector Result, TransformStruct *Trans, Vector Source);  
void MInverseEvaluatePoint(Vector Result, TransformStruct *Trans, Vector Source);
void MInverseEvaluateVector(Vector Result, TransformStruct *Trans, Vector Source);  
void ComputeTranslateTrans(TransformStruct *Trans ,Vector Vect);
void ComputeRotateTrans(TransformStruct *Trans ,Vector Vect);
void ComputeScaleTrans(TransformStruct *Trans ,Vector Vect);
void ComposeTrans(TransformStruct *Trans, TransformStruct *NewTrans);

#endif
