/* LoadPov.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2001 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>  
#include <ctype.h>

#include "giram.h"
#include "utils.h"
#include "frame.h"
#include "box.h"
#include "cone.h"
#include "cylinder.h"
#include "disc.h"
#include "mesh.h"
#include "plane.h"
#include "sor.h"
#include "sphere.h"
#include "superellipsoid.h"
#include "torus.h"
#include "triangle.h"
#include "csg.h"
#include "object.h"
#include "camera.h"
#include "csgtree.h"
#include "view.h"
#include "view_camera.h"
#include "copy.h"
#include "LoadPov.h"

#include "giramrc.h"
#include "giramintl.h"

#define SMALL_VALUE    10e-10

#define LastWord       263
struct { int Token; char *String; } ReservedWordTable[LastWord] = {
  { Taa_level, "aa_level" },
  { Taa_threshold, "aa_threshold" },
  { Tabs, "abs" },
  { Tacos, "acos" },  
  { Tadaptive, "adaptive" },
  { Tadc_bailout, "adc_bailout" },
  { Tagate, "agate" },
  { Tagate_turb, "agate_turb" },
  { Talpha, "alpha" },
  { Tambient, "ambient" },
  { Tambient_light, "ambient_light" },
  { Tangle, "angle" },
  { Taperture, "aperture" },
  { Tarc_angle, "arc_angle" },
  { Tarea_light, "area_light" },
  { Tasc, "asc" },
  { Tasin, "asin" },
  { Tassumed_gamma, "assumed_gamma" },
  { Tatan2, "atan2" },
  { Tatmosphere, "atmosphere" },
  { Tatmospheric_attenuation, "atmospheric_attenuation" },
  { Tattenuating, "attenuating" },
  { Taverage, "average" },
  { Tbicubic_patch, "bicubic_patch" },
  { Tblob, "blob" },
  { Tblue, "blue" },
  { Tblur_samples, "blur_samples" },
  { Tbounded_by, "bounded_by" },
  { Tbox, "box" },
  { Tbox_mapping, "box_mapping" },
  { Tbozo, "bozo" },
  { Tbrick, "brick" },
  { Tbrick_size, "brick_size" },
  { Tbrightness, "brightness" },
  { Tbrilliance, "brilliance" },
  { Tbump_map, "bump_map" },
  { Tbump_size, "bump_size" },
  { Tbumps, "bumps" },
  { Tcamera, "camera" },
  { Tcaustics, "caustics" },
  { Tceil, "ceil" },
  { Tchecker, "checker" },
  { Tchr, "chr" },
  { Tclipped_by, "clipped_by" },
  { Tclock, "clock" },
  { Tcolor, "color" },
  { Tcolor, "colour" },
  { Tcolor_map, "color_map" },
  { Tcolor_map, "colour_map" },
  { Tconcat, "concat" },
  { Tcone, "cone" },
  { Tconfidence, "confidence" },
  { Tconstant, "constant" },
  { Tcontrol0, "control0" },
  { Tcontrol1, "control1" },
  { Tcos, "cos" },
  { Tcount, "count" },
  { Tcrackle, "crackle" },
  { Tcrand, "crand" },
  { Tcubic, "cubic" },
  { Tcylinder, "cylinder" },
  { Tcylindrical_mapping, "cylindrical_mapping" },
  { Tdegrees, "degrees" },
  { Tdents, "dents" },
  { Tdifference, "difference" },
  { Tdiffuse, "diffuse" },
  { Tdirection, "direction" },
  { Tdisc, "disc" },
  { Tdistance, "distance" },
  { Tdistance_maximum, "distance_maximum" },
  { Tdiv, "div" },
  { Tdust, "dust" },
  { Tdust_type, "dust_type" },
  { Teccentricity, "eccentricity" },
  { Temitting, "emitting" },
  { Terror_bound, "error_bound" },
  { Texp, "exp" },
  { Texponent, "exponent" },
  { Tfade_distance, "fade_distance" },
  { Tfade_power, "fade_power" },
  { Tfalloff, "falloff" },
  { Tfalloff_angle, "falloff_angle" },
  { Tfalse, "false" },
  { Tfile_exists, "file_exists" },
  { Tfilter, "filter" },
  { Tfinish, "finish" },
  { Tfisheye, "fisheye" },
  { Tfloor, "floor" },
  { Tfocal_point, "focal_point" },
  { Tfog, "fog" },
  { Tfog_alt, "fog_alt" },
  { Tfog_offset, "fog_offset" },
  { Tfog_type, "fog_type" },
  { Tfrequency, "frequency" },
  { Tglobal_settings, "global_settings" },
  { Tglowing, "glowing" },
  { Tgradient, "gradient" },
  { Tgranite, "granite" },
  { Tgray_threshold, "gray_threshold" },
  { Tgreen, "green" },
  { Thalo, "halo" },
  { Theight_field, "height_field" },
  { Thexagon, "hexagon" },
  { Thf_gray_16, "hf_gray_16" },
  { Thierarchy, "hierarchy" },
  { Thollow, "hollow" },
  { Timage_map, "image_map" },
  { Tint, "int" },
  { Tintersection, "intersection" },
  { Tinverse, "inverse" },
  { Tior, "ior" },
  { Tirid, "irid" },
  { Tirid_wavelength, "irid_wavelength" },
  { Tjitter, "jitter" },
  { Tjulia_fractal, "julia_fractal" },
  { Tlambda, "lambda" },
  { Tlathe, "lathe" },
  { Tleopard, "leopard" },
  { Tlight_source, "light_source" },
  { Tlinear, "linear" },
  { Tlocation, "location" },
  { Tlog, "log" },
  { Tlow_error_factor, "low_error_factor" },
  { Tlook_at, "look_at" },
  { Tlooks_like, "looks_like" },
  { Tmandel, "mandel" },
  { Tmarble, "marble" },
  { Tmatrix, "matrix" },
  { Tmax, "max" },
  { Tmax_intersections, "max_intersections" },
  { Tmax_trace_level, "max_trace_level" },
  { Tmax_value, "max_value" },
  { Tmerge, "merge" },
  { Tmesh, "mesh" },
  { Tmetallic, "metallic" },
  { Tmin, "min" },
  { Tminimum_reuse, "minimum_reuse" },
  { Tmod, "mod" },
  { Tmortar, "mortar" },
  { Tnearest_count, "nearest_count" },
  { Tno, "no" },
  { Tno_shadow, "no_shadow" },
  { Tnormal, "normal" },
  { Tnormal_map, "normal_map" },
  { Tnumber_of_waves, "number_of_waves" },
  { Tobject, "object" },
  { Toctaves, "octaves" },
  { Toff, "off" },
  { Tomega, "omega" },
  { Tomnimax, "omnimax" },
  { Ton, "on" },
  { Tonion, "onion" },
  { Topen, "open" },
  { Torthographic, "orthographic" },
  { Tpanoramic, "panoramic" },
  { Tperspective, "perspective" },
  { Tphase, "phase" },
  { Tphong, "phong" },
  { Tphong_size, "phong_size" },
  { Tpi, "pi" },
  { Tpigment, "pigment" },
  { Tpigment_map, "pigment_map" },
  { Tplanar_mapping, "planar_mapping" },
  { Tplane, "plane" },
  { Tpoint_at, "point_at" },
  { Tpoly, "poly" },
  { Tpolygon, "polygon" },
  { Tpow, "pow" },
  { Tprism, "prism" },
  { Tquadric, "quadric" },
  { Tquartic, "quartic" },
  { Tquick_color, "quick_color" },
  { Tquilted, "quilted" },
  { Tradial, "radial" },
  { Tradians, "radians" },
  { Tradiosity, "radiosity" },
  { Tradius, "radius" },
  { Trainbow, "rainbow" },
  { Tramp_wave, "ramp_wave" },
  { Trand, "rand" },
  { Trecursion_limit, "recursion_limit" },
  { Tred, "red" },
  { Treflection, "reflection" },
  { Trefraction, "refraction" },
  { Trgb, "rgb" },
  { Trgbf, "rgbf" },
  { Trgbft, "rgbft" },
  { Trgbt, "rgbt" },
  { Tright, "right" },
  { Tripples, "ripples" },
  { Trotate, "rotate" },
  { Troughness, "roughness" },
  { Tsamples, "samples" },
  { Tscale, "scale" },
  { Tscallop_wave, "scallop_wave" },
  { Tscattering, "scattering" },
  { Tseed, "seed" },
  { Tshadowless, "shadowless" },
  { Tsin, "sin" },
  { Tsine_wave, "sine_wave" },
  { Tsky, "sky" },
  { Tsky_sphere, "sky_sphere" },
  { Tslope_map, "slope_map" },
  { Tsmooth, "smooth" },
  { Tsmooth_triangle, "smooth_triangle" },
  { Tsor, "sor" },
  { Tspecular, "specular" },
  { Tsphere, "sphere" },
  { Tspherical_mapping, "spherical_mapping" },
  { Tspiral1, "spiral1" },
  { Tspiral2, "spiral2" },
  { Tspotlight, "spotlight" },
  { Tspotted, "spotted" },
  { Tsqrt, "sqrt" },
  { Tstr, "str" },
  { Tstrcmp, "strcmp" },
  { Tstrlen, "strlen" },
  { Tstrlwr, "strlwr" },
  { Tstrupr, "strupr" },
  { Tsturm, "sturm" },
  { Tsubstr, "substr" },
  { Tsuperellipsoid, "superellipsoid" },
  { Tt, "t" },
  { Ttan, "tan" },
  { Ttext, "text" },
  { Ttexture, "texture" },
  { Tthickness, "thickness" },
  { Ttightness, "tightness" },
  { Ttorus, "torus" },
  { Ttransform, "transform" },
  { Ttranslate, "translate" },
  { Ttransmit, "transmit" },
  { Ttriangle, "triangle" },
  { Ttriangle_wave, "triangle_wave" },
  { Ttrue, "true" },
  { Tttf, "ttf" },
  { Tturb_depth, "turb_depth" },
  { Tturbulence, "turbulence" },
  { Ttype, "type" },
  { Tu, "u" },
  { Tunion, "union" },
  { Tup, "up" },
  { Tultra_wide_angle, "ultra_wide_angle" },
  { Tv, "v" },
  { Tval, "val" },
  { Tvariance, "variance" },
  { Tvaxis_rotate, "vaxis_rotate" },
  { Tvcross, "vcross" },
  { Tvdot, "vdot" },
  { Tversion, "version" },
  { Tvlength, "vlength" },
  { Tvnormalize, "vnormalize" },
  { Tvrotate, "vrotate" },
  { Twarp, "warp" },
  { Twater_level, "water_level" },
  { Twaves, "waves" },
  { Twidth, "width" },
  { Twood, "wood" },
  { Twrinkles, "wrinkles" },
  { Tx, "x" },
  { Ty, "y" },
  { Tyes, "yes" },
  { Tz, "z" }
};

ListFile *PovFile;
int OldChar, IsOldChar;
int OldToken, IsOldToken;
double PovVersionLangage;
double LastFloat;
char *LastString = NULL;
FloatConstantStruct       *LastFloatConstant;
VectorConstantStruct      *LastVectorConstant;
ColorConstantStruct       *LastColorConstant;
StringConstantStruct      *LastStringConstant;
ObjectConstantStruct      *LastObjectConstant;
TextureConstantStruct     *LastTextureConstant;
PigmentConstantStruct     *LastPigmentConstant;
CameraConstantStruct      *LastCameraConstant;
LightSourceConstantStruct *LastLightSourceConstant;
double PovClock = 0.0;
static unsigned int NbPovRandomTable;
static unsigned long int *PovRandomTable;
int ParsingDirective;
int DontParse;

typedef struct ParsedVector
{
  Vector Val;
  int nbTerms;
} ParsedVector;

#define IF_TRUE_IF_ELSE           1
#define IF_TRUE_ELSE_END          2
#define IF_FALSE_IF_ELSE          3
#define IF_FALSE_ELSE_END         4
#define WHILE                     5
#define SWITCH_SWITCH_CASE        6
#define SWITCH_CASE_BREAK_TRUE    7
#define SWITCH_CASE_BREAK_FALSE   8
#define SWITCH_BREAK_END          9
#define UNPARSED                 10

typedef struct CondDirective
{
  struct CondDirective *Next;
  int Style;
  double SwitchValue;
  int WhileLineNo;
  ListFile *WhileFile;
  long WhilePos;
} CondDirective;

CondDirective *ConditionalStack;
FrameStruct *CurrentFrame;

static int next_token(int *pToken);
static int parse_float(double *pLocalFloat);
static int parse_color(Vector TmpVector);
static int parse_pigment(PigmentStruct **pLocalPigment);
static int parse_rel_factor(ParsedVector *TmpParsedVector);
static int parse_object(ObjectStruct **pLocalObject);
static int parse_light_source(LightSourceStruct **pLocalLightSource);
static int parse_atmosphere(AtmosphereStruct **pLocalAtmosphere);
static int parse_fog(FogStruct **pLocalFog);
static int parse_texture(TextureStruct **pLocalTexture);
static int parse_camera(CameraStruct **pLocalCamera);

/*****************************************************************************
*  next_char
******************************************************************************/
static int next_char(void)
{
  if (IsOldChar)
  {
    IsOldChar = 0;
    return OldChar;
  }
  return getc(PovFile->handle);
}

/*****************************************************************************
*  ignore_space
******************************************************************************/
static int ignore_space(void)
{
  int c;
  
  do
  {
    c = next_char(); 
    if (c=='\n')
    {
      while (c=='\n')
      {
        PovFile->Line++;
        c = next_char();
        if (c=='\r') c = next_char();
      }
    } else if (c=='\r')
    {
      while (c=='\r')
      {
        PovFile->Line++;
        c = next_char();
        if (c=='\n') c = next_char();
      }
    }
  } while (c==' ' || c=='\t' || c=='' || c=='\r' || c=='\n');
  OldChar = c; IsOldChar = 1;
  return 0;
}

/*****************************************************************************
*  parse_comma
******************************************************************************/
static int parse_comma(void)
{
  int t;
  
  if (next_token(&t))
    return 1;
  if (t != TComma)
  {
    OldToken = t ; IsOldToken = 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_float_literal
******************************************************************************/
static int parse_float_literal(double *pReturnedFloat)
{
  char *FloatText;
  guint FloatTextLength, FloatTextMaxLength;
  int c;

  FloatTextMaxLength = 20;
  FloatText = g_malloc(21);
  FloatTextLength = 0;

  c = next_char();
  if (c == '.')
  {
    strcpy(FloatText,"0.");
    FloatTextLength = 2;
    c = next_char();
    if (!isdigit(c))
    {
      g_message(_("%s, line %d\nMisformed float\n"), PovFile->Name, PovFile->Line);
      return 1;
    }
    while (isdigit(c))
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = next_char();
    }
    OldChar = c; IsOldChar = 1;
  } else
  { /* c = [0-9] */
    while (isdigit(c))
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = next_char();
    }
    if (c=='.')
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = next_char();
      if (!isdigit(c))
      {
        FloatText[FloatTextLength++] = '0';
        if (FloatTextLength==FloatTextMaxLength)
        {
          FloatTextMaxLength *= 2;
          FloatText = g_realloc(FloatText, FloatTextMaxLength);
        }
      }
      while (isdigit(c))
      {
        FloatText[FloatTextLength++] = (char) c;
        if (FloatTextLength==FloatTextMaxLength)
        {
          FloatTextMaxLength *= 2;
          FloatText = g_realloc(FloatText, FloatTextMaxLength);
        }
        c = next_char();
      }
    }
    OldChar = c; IsOldChar = 1;
  }
  /* At this point, FloatText is either 
     0.[0-9]+, [0-9]+, [0-9]+.0 or [0-9]+.[0-9]+ */
  c = next_char();
  if (c=='e' || c=='E')
  {
    FloatText[FloatTextLength++] = 'e';
    if (FloatTextLength==FloatTextMaxLength)
    {
      FloatTextMaxLength *= 2;
      FloatText = g_realloc(FloatText, FloatTextMaxLength);
    }
    c = next_char();
    if (c=='+' || c=='-')
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = next_char();
    }
    if (!isdigit(c))
    {
      g_message(_("%s, line %d\nMisformed float\n"),
                PovFile->Name, PovFile->Line);
      return 1;
    }
    while (isdigit(c))
    {
      FloatText[FloatTextLength++] = (char) c;
      if (FloatTextLength==FloatTextMaxLength)
      {
        FloatTextMaxLength *= 2;
        FloatText = g_realloc(FloatText, FloatTextMaxLength);
      }
      c = next_char();
    }
  }
  OldChar = c; IsOldChar = 1;
  FloatText[FloatTextLength]=0; 
  *pReturnedFloat = atof(FloatText);
  g_free(FloatText);
  return 0;
}

/*****************************************************************************
*  paren_error
******************************************************************************/
static int paren_error(char *Function, int Token)
{
  int t;
  if (next_token(&t))
    return 1;
  if (t != Token)
  {
    g_message(_("%s, line %d\nThe argument of '%s' must be enclosed between parentheses...\n"),
              PovFile->Name, PovFile->Line, Function);
    return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_string_literal
******************************************************************************/
static int parse_string_literal(char **pString)
{
  guint StringLength, StringMaxLength;
  int c;
  
  StringMaxLength = 20;
  *pString = g_malloc(21);
  StringLength = 0;
  c = next_char();
  while(c != '"')
  {
    if (c == EOF) return 1;
    (*pString)[StringLength++] = (char)c;
    if (StringLength==StringMaxLength)
    {
      StringMaxLength *= 2;
      *pString = g_realloc(*pString, StringMaxLength);
    }
    c = next_char();
  }
  (*pString)[StringLength] = 0;
  return 0;
}

/*****************************************************************************
*  parse_string
******************************************************************************/
static int parse_string(char **pLocalString)
{
  char TmpString1[100], TmpString2[100];
  char *AppendString;
  int t,i;
  double Val1, Val2, Val3;

  if (next_token(&t))
    return 1;
  switch (t)
  {
    case TStringLiteral:
      *pLocalString = LastString;
      break;
    case Tstr:
      if (paren_error("str", TLeftParen))
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val2))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val3))
        return 1;
      if (paren_error("str", TRightParen))
        return 1;
      if (Val3>=0)
      {
        if (Val2>0)
          sprintf(TmpString1, "%%%d.%df", (int)Val2, (int)Val3);
        else if (Val2 < 0)
          sprintf(TmpString1, "%%0%d.%df", (int)-Val2, (int)Val3);
        else
          sprintf(TmpString1, "%%.%df", (int)Val3);
      } else
      {
        if (Val2>0)
          sprintf(TmpString1, "%%%df",(int)Val2);
        else if (Val2 < 0)
          sprintf(TmpString1, "%%0%df",(int)-Val2);
        else
          strcpy(TmpString1, "%f");
      }
      sprintf(TmpString2, TmpString1, Val1);
      *pLocalString = g_malloc(strlen(TmpString2)+1);
      strcpy(*pLocalString, TmpString2);
      break;
    case Tconcat:
      if (paren_error("concat",TLeftParen))
        return 1;
      if (parse_string(pLocalString))
        return 1;
      if (next_token(&t))
        return 1;
      while(t != TRightParen)
      {
        if (t != TComma)
        {
          OldToken = t; IsOldToken = 1;
        }
        parse_string(&AppendString);
        *pLocalString = g_realloc(*pLocalString, strlen(*pLocalString)+strlen(AppendString)+1);
        strcat(*pLocalString, AppendString);
        g_free(AppendString);
        if (next_token(&t))
          return 1;
      }
      break;
    case Tchr:
      if (paren_error("chr", TLeftParen))
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (paren_error("chr", TRightParen))
        return 1;
      if (Val1 < 0.0 || Val1 >= 256.0)
      {
        g_message(_("%s, line %d\nBad value for chr() : %g...\n"),
                  PovFile->Name, PovFile->Line, Val1);
        return 1;
      }
      *pLocalString = g_malloc(2);
      *pLocalString[0] = (char)((int)Val1);
      *pLocalString[1] = 0;
      break;
    case Tsubstr:
      if (paren_error("substr", TLeftParen))
        return 1;
      if (parse_string(&AppendString))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val2))
        return 1;
      if (paren_error("substr", TRightParen))
        return 1;
      if (((Val1+Val2-1.0) > strlen(AppendString)) ||
         (Val1<0.0) ||
         (Val2<1.0))
      {
        g_message(_("%s, line %d\nBad params in substr(%s,%g,%g)...\n"),
                  PovFile->Name, PovFile->Line, AppendString, Val1, Val2);
        return 1;
      }
      *pLocalString = g_malloc((guint)Val2+1);
      strncpy(*pLocalString,&(AppendString[(int)Val1-1]),(guint)Val2);
      *pLocalString[(int)Val2]='\0';
      g_free(AppendString);
      break;
    case Tstrupr:
      if (paren_error("strupr", TLeftParen))
        return 1;
      if (parse_string(pLocalString))
        return 1;
      if (paren_error("strupr", TRightParen))
        return 1;
      for (i=0 ; i<(gint)strlen(*pLocalString) ; i++)
        if (*pLocalString[i]>='a' && *pLocalString[i] <='z')
          *pLocalString[i]-=32;
      break;
    case Tstrlwr:
      if (paren_error("strlwr", TLeftParen))
        return 1;
      if (parse_string(pLocalString))
        return 1;
      if (paren_error("strlwr", TRightParen))
        return 1;
      for (i=0 ; i<(gint)strlen(*pLocalString) ; i++)
        if (*pLocalString[i]>='A' && *pLocalString[i] <='Z')
          *pLocalString[i]+=32;
      break;
    case TStringID:
      *pLocalString = g_strdup(LastStringConstant->String);
      break;
    default:
      g_message(_("%s, line %d\nA String was expected here...\n"),
                PovFile->Name, PovFile->Line);
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_vector
******************************************************************************/
static int parse_vector(ParsedVector *TmpParsedVector)
{
  if (parse_rel_factor(TmpParsedVector))
    return 1;
  if (TmpParsedVector->nbTerms>3)
  {
    g_message(_("%s, line %d\nExpecting a vector but got a color expression...\n"),
              PovFile->Name, PovFile->Line);
    return 1;
  }
  if (TmpParsedVector->nbTerms==1)
  {
    TmpParsedVector->Val[1] = TmpParsedVector->Val[2] = TmpParsedVector->Val[0];
  } else if (TmpParsedVector->nbTerms==2)
  {
    TmpParsedVector->Val[2] = 0.0;
  }
  TmpParsedVector->nbTerms = 3;
  return 0;
}

/*****************************************************************************
*  parse_rel_term
******************************************************************************/
static int parse_rel_term(ParsedVector *TmpParsedVector)
{
  ParsedVector SecondFactor;
  int t;
  
  if (parse_rel_factor(TmpParsedVector))
    return 1;
  if (next_token(&t))
    return 1;
  while ((t==TLeftAngleBracket) || (t==TLowerOrEqual) ||
         (t==TEqual) || (t==TUnequal) ||
         (t==TUpperOrEqual) || (t==TRightAngleBracket))
  {
    if (parse_rel_factor(&SecondFactor))
      return 1;
    TmpParsedVector->nbTerms = 1;
    switch(t)
    {
      case TLeftAngleBracket:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] < SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TLowerOrEqual:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] <= SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TEqual:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] == SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TUnequal:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] != SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TUpperOrEqual:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] >= SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
      case TRightAngleBracket:
        TmpParsedVector->Val[0] = 
               (TmpParsedVector->Val[0] > SecondFactor.Val[0]) ? 1.0 : 0.0;
        break;
    }
    if (next_token(&t))
      return 1;
  }
  OldToken = t;
  IsOldToken = 1;
  return 0;
}

/*****************************************************************************
*  parse_logical
******************************************************************************/
static int parse_logical(ParsedVector *TmpParsedVector)
{
  ParsedVector SecondTerm;
  int t;
  
  if (parse_rel_term(TmpParsedVector))
    return 1;
  if (next_token(&t))
    return 1;
  while ((t==TAmpersand) || (t==TBar))
  {
    if (parse_rel_term(&SecondTerm))
      return 1;
    TmpParsedVector->nbTerms = 1;
    if (t==TAmpersand)
    {
      TmpParsedVector->Val[0] = TmpParsedVector->Val[0] && SecondTerm.Val[0];
    } else
    {
      TmpParsedVector->Val[0] = TmpParsedVector->Val[0] || SecondTerm.Val[0];
    }
    if (next_token(&t))
      return 1;
  }
  OldToken = t;
  IsOldToken = 1;
  return 0;
}

/*****************************************************************************
*  parse_express
******************************************************************************/
static int parse_express(ParsedVector *TmpParsedVector)
{
  ParsedVector LocalVector;
  ParsedVector LocalVector1;
  ParsedVector LocalVector2;
  int t,i;

  if (parse_logical(&LocalVector))
    return 1;
  if (next_token(&t))
    return 1;
  if (t == TQuestion)
  {
    if (LocalVector.nbTerms != 1)
    {
      g_message(_("%s, line %d\nConditional must evaluate to a float...\n"),
                PovFile->Name, PovFile->Line);
      return 1;
    }
    if (parse_express(&LocalVector1))
      return 1;
    if (next_token(&t))
      return 1;
    if (t != TColon)
    {
      g_message(_("%s, line %d\nConditional must be followed by a colon...\n"),
                PovFile->Name, PovFile->Line);
      return 1;
    }
    if (parse_express(&LocalVector2))
      return 1;
    if (LocalVector.Val[0])
    {
      TmpParsedVector->nbTerms = LocalVector1.nbTerms;
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] = LocalVector1.Val[i];
    } else
    {
      TmpParsedVector->nbTerms = LocalVector2.nbTerms;
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] = LocalVector2.Val[i];
    }
  } else
  {
    OldToken = t; IsOldToken = 1;
    TmpParsedVector->nbTerms = LocalVector.nbTerms;
    for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
      TmpParsedVector->Val[i] = LocalVector.Val[i];
  }
  return 0;
}

/*****************************************************************************
*  parse_num_factor
******************************************************************************/
static int parse_num_factor(ParsedVector *TmpParsedVector)
{
  int t, c, i;
  ParsedVector LocalVector;
  ParsedVector LocalVector2;
  double Val1, Val2;
  FILE *f;
  char *LocalString, *LocalString2;

  if (next_token(&t))
    return 1;
  while (t == TPlus)
  {
    if (next_token(&t))
      return 1;
  }
  switch (t)
  {
    case Tabs:
      if (paren_error("abs", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'abs' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = fabs(LocalVector.Val[0]);
      if (paren_error("abs", TRightParen))
        return 1;
      break;
    case Tacos:
      if (paren_error("acos", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'acos' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      if (LocalVector.Val[0]>1.0 || LocalVector.Val[0]<-1.0)
      {
        g_message(_("%s, line %d\nThe argument of 'acos' must be between -1 and 1...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = acos(LocalVector.Val[0]);
      if (paren_error("acos", TRightParen))
        return 1;
      break;
    case Tval:
      if (paren_error("val", TLeftParen))
        return 1;
      if (parse_string(&LocalString))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = atof(LocalString);
      g_free(LocalString);
      if (paren_error("val", TRightParen))
        return 1;
      break;
    case Tasc:
      if (paren_error("asc", TLeftParen))
        return 1;
      if (parse_string(&LocalString))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LocalString[0];
      g_free(LocalString);
      if (paren_error("asc", TRightParen))
        return 1;
      break;
    case Tasin:
      if (paren_error("asin", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'asin' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      if (LocalVector.Val[0]>1.0 || LocalVector.Val[0]<-1.0)
      {
        g_message(_("%s, line %d\nThe argument of 'asin' must be between -1 and 1...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = asin(LocalVector.Val[0]);
      if (paren_error("asin", TRightParen))
        return 1;
      break;
    case Tatan2:
      if (paren_error("atan2", TLeftParen))
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = atan2(Val1,Val2);
      if (paren_error("atan2", TRightParen))
        return 1;
      break;
    case Tceil:
      if (paren_error("ceil", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'ceil' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = ceil(LocalVector.Val[0]);
      if (paren_error("ceil", TRightParen))
        return 1;
      break;
    case Tclock: /* FIXME: should be special... */
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = PovClock;
      break;
    case Tcos:
      if (paren_error("cos", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'cos' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = cos(LocalVector.Val[0]);
      if (paren_error("cos", TRightParen))
        return 1;
      break;
    case Tdegrees:
      if (paren_error("degrees", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'degrees' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LocalVector.Val[0]/M_PI*180.0;
      if (paren_error("degrees", TRightParen))
        return 1;
      break;
    case Tdiv:
      if (paren_error("div", TLeftParen))
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val2))
        return 1;
      if (fabs(Val2) <= 1.0e-6)
      {
        g_message(_("%s, line %d\nThe second argument of 'div' mustn't be nul...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = (double)((int)(Val1 / Val2));
      if (paren_error("div", TRightParen))
        return 1;
      break;
    case Texp:
      if (paren_error("exp", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'exp' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = exp(LocalVector.Val[0]);
      if (paren_error("exp", TRightParen))
        return 1;
      break;
    case Tfile_exists:
      if (paren_error("file_exists", TLeftParen))
        return 1;
      if (parse_string(&LocalString))
        return 1;
      TmpParsedVector->nbTerms = 1;
      f = LibOpen(LocalString);
      if (f)
      {
        TmpParsedVector->Val[0] = 1.0;
        fclose(f);
      }
      else
        TmpParsedVector->Val[0] = 0.0;
      g_free(LocalString);
      if (paren_error("file_exists", TRightParen))
        return 1;
      break;
    case TFloatID:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LastFloatConstant->Float;
      break;
    case TFloatLiteral:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LastFloat;
      break;
    case Tfloor:
      if (paren_error("floor", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'floor' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = floor(LocalVector.Val[0]);
      if (paren_error("floor", TRightParen))
        return 1;
      break;
    case Tint:
      if (paren_error("int", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'int' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = (double)((int)LocalVector.Val[0]);
      if (paren_error("int", TRightParen))
        return 1;
      break;
    case Tlog:
      if (paren_error("log", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'log' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      if (TmpParsedVector->Val[0] <= 0.0)
      {
        g_message(_("%s, line %d\nThe argument of 'log' must be strictly positive...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = log(LocalVector.Val[0]);
      if (paren_error("log", TRightParen))
        return 1;
      break;
    case Tmin:
      if (paren_error("min", TLeftParen))
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = Val1>Val2?Val2:Val1;
      if (paren_error("min", TRightParen))
        return 1;
      break;
    case Tmax:
      if (paren_error("max", TLeftParen))
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = Val1<Val2?Val2:Val1;
      if (paren_error("max", TRightParen))
        return 1;
      break;
    case Tmod:
      if (paren_error("mod", TLeftParen))
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val2))
        return 1;
      if (fabs(Val2) <= 1.0e-6)
      {
        g_message(_("%s, line %d\nThe second argument of 'mod' mustn't be nul...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = fmod(Val1,Val2);
      if (paren_error("mod", TRightParen))
        return 1;
      break;
    case Tpi:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = M_PI;
      break;
    case Tpow:
      if (paren_error("pow", TLeftParen))
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val2))
        return 1;
      if (Val1 <= 0.0 )
        TmpParsedVector->Val[0] = pow(Val1, Val2);
      else /* FIXME ??? */
        TmpParsedVector->Val[0] = pow(Val1, Val2);
      TmpParsedVector->nbTerms = 1;
      if (paren_error("pow", TRightParen))
        return 1;
      break;
    case Tradians:
      if (paren_error("radians", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'radians' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = LocalVector.Val[0] * M_PI / 180.0;
      if (paren_error("radians", TRightParen))
        return 1;
      break;
    case Tsin:
      if (paren_error("sin", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'sin' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = sin(LocalVector.Val[0]);
      if (paren_error("sin", TRightParen))
        return 1;
      break;
    case Tsqrt:
      if (paren_error("sqrt", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'sqrt' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      if (LocalVector.Val[0] < 0.0)
      {
        g_message(_("%s, line %d\nThe argument of 'sqrt' must be postive...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = sqrt(LocalVector.Val[0]);
      if (paren_error("sqrt", TRightParen))
        return 1;
      break;
    case Tstrcmp:
      if (paren_error("strcmp", TLeftParen))
        return 1;
      if (parse_string(&LocalString))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_string(&LocalString2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = strcmp(LocalString, LocalString2);
      g_free(LocalString);
      g_free(LocalString2);
      if (paren_error("strcmp", TRightParen))
        return 1;
      break;
    case Tstrlen:
      if (paren_error("strlen", TLeftParen))
        return 1;
      if (parse_string(&LocalString))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = strlen(LocalString);
      g_free(LocalString);
      if (paren_error("strlen", TRightParen))
        return 1;
      break;
    case Ttan:
      if (paren_error("tan", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'tan' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = tan(LocalVector.Val[0]);
      if (paren_error("tan", TRightParen))
        return 1;
      break;
    case Tvdot:
      if (paren_error("vdot", TLeftParen))
        return 1;
      if (parse_vector(&LocalVector))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_vector(&LocalVector2))
        return 1;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = VDot(LocalVector.Val, LocalVector2.Val);
      if (paren_error("vdot", TRightParen))
        return 1;
      break;
    case Tvlength:
      if (paren_error("vlength", TLeftParen))
        return 1;
      parse_vector(&LocalVector);
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = VLength(LocalVector.Val);
      if (paren_error("vlength", TRightParen))
        return 1;
      break;
    case Tversion:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = PovVersionLangage;
      break;
    case Ttrue:
    case Tyes:
    case Ton:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = 1.0;
      break;
    case Tfalse:
    case Tno:
    case Toff:
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = 0.0;
      break;
    case Tseed:
      if (paren_error("seed", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (paren_error("seed", TRightParen))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'seed' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      PovRandomTable = g_realloc(PovRandomTable, 
                                 (NbPovRandomTable+1)*sizeof(unsigned long));
      PovRandomTable[NbPovRandomTable] = (unsigned long)LocalVector.Val[0];
      NbPovRandomTable++;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = NbPovRandomTable-1;
      break;
    case Trand:
      if (paren_error("rand", TLeftParen))
        return 1;
      if (parse_express(&LocalVector))
        return 1;
      if (paren_error("rand", TRightParen))
        return 1;
      if (LocalVector.nbTerms != 1)
      {
        g_message(_("%s, line %d\nThe argument of 'rand' must be a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      PovRandomTable[(int)LocalVector.Val[0]] = 
             PovRandomTable[(int)LocalVector.Val[0]] * 1812433253L + 12345L;
      TmpParsedVector->nbTerms = 1;
      TmpParsedVector->Val[0] = 
             ((double)(PovRandomTable[(int)LocalVector.Val[0]] & 0xFFFFFFFFUL) 
              / 0xFFFFFFFFUL);
      break;
    case Tvaxis_rotate:
      if (paren_error("vaxis_rotate", TLeftParen))
        return 1;
      if (parse_vector(&LocalVector))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_vector(&LocalVector2))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_float(&Val1))
        return 1;
      if (paren_error("vaxis_rotate", TRightParen))
        return 1;
      /* FIXME */
      g_message(_("%s, line %d\nThe function 'vaxis_rotate' not yet implemented\n"),
                PovFile->Name, PovFile->Line);
      return 1;
      break;
    case Tvcross:
      if (paren_error("vcross", TLeftParen))
        return 1;
      if (parse_vector(&LocalVector))
        return 1;
      if (parse_comma())
        return 1;
      if (parse_vector(&LocalVector2))
        return 1;
      if (paren_error("vcross", TRightParen))
        return 1;
      TmpParsedVector->nbTerms = 3;
      VCross(TmpParsedVector->Val, LocalVector.Val, LocalVector2.Val);
      break;
    case TVectorID:
      TmpParsedVector->nbTerms = 3;
      TmpParsedVector->Val[0] = LastVectorConstant->Vector[0];
      TmpParsedVector->Val[1] = LastVectorConstant->Vector[1];
      TmpParsedVector->Val[2] = LastVectorConstant->Vector[2];
      break;
    case Tvnormalize:
      if (paren_error("vnormalize", TLeftParen))
        return 1;
      if (parse_vector(TmpParsedVector))
        return 1;
      if (paren_error("vnormalize", TRightParen))
        return 1;
      Val1 = VLength(TmpParsedVector->Val);
      if (Val1 != 0.0)
      {
        TmpParsedVector->Val[0] /= Val1;
        TmpParsedVector->Val[1] /= Val1;
        TmpParsedVector->Val[2] /= Val1;
      }
      break;
    case Tvrotate:
      g_message(_("%s, line %d\nThe function 'vrotate' not yet implemented\n"),
                PovFile->Name, PovFile->Line);
      return 1;
      /* FIXME */
      break;
    case Tx:
      TmpParsedVector->nbTerms = 3;
      TmpParsedVector->Val[0] = 1.0;
      TmpParsedVector->Val[1] = 0.0;
      TmpParsedVector->Val[2] = 0.0;
      break;
    case Ty:
      TmpParsedVector->nbTerms = 3;
      TmpParsedVector->Val[0] = 0.0;
      TmpParsedVector->Val[1] = 1.0;
      TmpParsedVector->Val[2] = 0.0;
      break;
    case Tz:
      TmpParsedVector->nbTerms = 3;
      TmpParsedVector->Val[0] = 0.0;
      TmpParsedVector->Val[1] = 0.0;
      TmpParsedVector->Val[2] = 1.0;
      break;
    case TColorID:
      TmpParsedVector->nbTerms = 5;
      TmpParsedVector->Val[0] = LastColorConstant->Color[0];
      TmpParsedVector->Val[1] = LastColorConstant->Color[1];
      TmpParsedVector->Val[2] = LastColorConstant->Color[2];
      TmpParsedVector->Val[3] = LastColorConstant->Color[3];
      TmpParsedVector->Val[4] = LastColorConstant->Color[4];
      break;
    case Tt:
      TmpParsedVector->nbTerms = 4;
      TmpParsedVector->Val[0] = 0.0;
      TmpParsedVector->Val[1] = 0.0;
      TmpParsedVector->Val[2] = 0.0;
      TmpParsedVector->Val[3] = 1.0;
      break;
    case Tu:
      TmpParsedVector->nbTerms = 2;
      TmpParsedVector->Val[0] = 1.0;
      TmpParsedVector->Val[1] = 0.0;
      break;
    case Tv:
      TmpParsedVector->nbTerms = 2;
      TmpParsedVector->Val[0] = 0.0;
      TmpParsedVector->Val[1] = 1.0;
      break;
    case TMinus:
      if (parse_num_factor(TmpParsedVector))
        return 1;
      for (i=0 ; i<5 ; i++)
        TmpParsedVector->Val[i] = -TmpParsedVector->Val[i];
      break;
    case TExclamation:
      if (parse_num_factor(TmpParsedVector))
        return 1;
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] = fabs(TmpParsedVector->Val[i])>10e-6?0.0:1.0;
      break;
    case TLeftParen:
      if (parse_express(TmpParsedVector))
        return 1;
      if (paren_error(_("Left Parenthese"), TRightParen))
        return 1;
      break;
    case TLeftAngleBracket:
      if (parse_float(&(TmpParsedVector->Val[0])))
        return 1;
      if (parse_comma())
        return 1;
      if (next_token(&t))
        return 1;
      if (t == TRightAngleBracket)
      {
        TmpParsedVector->nbTerms = 1;
      } else
      { 
        IsOldToken = 1; OldToken = t;
        if (parse_float(&(TmpParsedVector->Val[1])))
          return 1;
        if (parse_comma())
          return 1;
        if (next_token(&t))
          return 1;
        if (t == TRightAngleBracket)
        {
          TmpParsedVector->nbTerms = 2;
        } else
        {
          IsOldToken = 1; OldToken = t;
          if (parse_float(&(TmpParsedVector->Val[2])))
            return 1;
          if (parse_comma())
            return 1;
          if (next_token(&t))
            return 1;
          if (t == TRightAngleBracket)
          {
            TmpParsedVector->nbTerms = 3;
          } else
          {
            IsOldToken = 1; OldToken = t;
            if (parse_float(&(TmpParsedVector->Val[3])))
              return 1;
            if (parse_comma())
              return 1;
            if (next_token(&t))
              return 1;
            if (t == TRightAngleBracket)
            {
              TmpParsedVector->nbTerms = 4;
            } else
            {
              IsOldToken = 1; OldToken = t;
              if (parse_float(&(TmpParsedVector->Val[4])))
                return 1;
              if (parse_comma())
                return 1;
              if (next_token(&t))
                return 1;
              if (t == TRightAngleBracket)
              {
                TmpParsedVector->nbTerms = 5;
              } else
              {
                g_message(_("%s, line %d\nVectors can only have 5 terms or less...\n"),
                          PovFile->Name, PovFile->Line);
                return 1;
              }
            }
          }
        }
      }
      break;
    default:
      g_message(_("%s, line %d\nNumeric Expression expected (%d)...\n"), 
                PovFile->Name, PovFile->Line, t);
      return 1;
  }
  if (ignore_space())
    return 1;
  c = next_char();
  if (c != '.')
  {
    IsOldChar = 1; OldChar = c;
    return 0;
  }
  if (next_token(&t))
    return 1;
  switch (t)
  {
    case Tx:
      if (TmpParsedVector->nbTerms != 3)
      {
        g_message(_("%s, line %d\nThe 'x' suffix can only be used with vectors...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
      }
      break;
    case Ty:
      if (TmpParsedVector->nbTerms != 3)
      {
        g_message(_("%s, line %d\nThe 'y' suffix can only be used with vectors...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[1];
      }
      break;
    case Tz:
      if (TmpParsedVector->nbTerms != 3)
      {
        g_message(_("%s, line %d\nThe 'z' suffix can only be used with vectors...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[2];
      }
      break;
    case Tred:
      if (TmpParsedVector->nbTerms != 5)
      {
        g_message(_("%s, line %d\nThe 'red' suffix can only be used with colors...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
      }
      break;
    case Tgreen:
      if (TmpParsedVector->nbTerms != 5)
      {
        g_message(_("%s, line %d\nThe 'green' suffix can only be used with colors...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[1];
      }
      break;
    case Tblue:
      if (TmpParsedVector->nbTerms != 5)
      {
        g_message(_("%s, line %d\nThe 'blue' suffix can only be used with colors...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[2];
      }
      break;
    case Tfilter:
      if (TmpParsedVector->nbTerms != 5)
      {
        g_message(_("%s, line %d\nThe 'filter' suffix can only be used with colors...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[3];
      }
      break;
    case Ttransmit:
      if (TmpParsedVector->nbTerms != 5)
      {
        g_message(_("%s, line %d\nThe 'transmit' suffix can only be used with colors...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[4];
      }
      break;
    case Tu:
      TmpParsedVector->nbTerms = 1;
      break;
    case Tv:
      if (TmpParsedVector->nbTerms == 1)
      {
        g_message(_("%s, line %d\nThe 'v' suffix can't be used with float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[1];
      }
      break;
    case Tt:
      if (TmpParsedVector->nbTerms < 4)
      {
        g_message(_("%s, line %d\nThe 't' suffix can't be used there...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      } else
      {
        TmpParsedVector->nbTerms = 1;
        TmpParsedVector->Val[0] = TmpParsedVector->Val[3];
      }
      break;
  }
  return 0;
}

/*****************************************************************************
*  parse_num_term
******************************************************************************/
static int parse_num_term(ParsedVector *TmpParsedVector)
{
  ParsedVector SecondFactor;
  int t,i;
  
  if (parse_num_factor(TmpParsedVector))
    return 1;
  if (next_token(&t))
    return 1;
  while ((t==TSlash) || (t==TStar))
  {
    if (parse_num_factor(&SecondFactor))
      return 1;
    if (SecondFactor.nbTerms != TmpParsedVector->nbTerms)
    {
      if (SecondFactor.nbTerms > TmpParsedVector->nbTerms)
      {
        if (TmpParsedVector->nbTerms == 1)
          for (i=1 ; i<SecondFactor.nbTerms ; i++)
            TmpParsedVector->Val[i] = TmpParsedVector->Val[0];
        else
          for (i=TmpParsedVector->nbTerms ; i<SecondFactor.nbTerms ; i++)
            TmpParsedVector->Val[i] = 0.0;
        TmpParsedVector->nbTerms = SecondFactor.nbTerms;
      } else
      {
        if (SecondFactor.nbTerms == 1)
          for (i=1 ; i<TmpParsedVector->nbTerms ; i++)
            SecondFactor.Val[i] = SecondFactor.Val[0];
        else
          for (i=SecondFactor.nbTerms ; i<TmpParsedVector->nbTerms ; i++)
            SecondFactor.Val[i] = 0.0;
      }
    }
    if (t==TStar)
    {
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] *= SecondFactor.Val[i];
    } else
    {
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] /= SecondFactor.Val[i];
    }
    if (next_token(&t))
      return 1;
  }
  OldToken = t;
  IsOldToken = 1;
  return 0;
}

/*****************************************************************************
*  parse_rel_factor
******************************************************************************/
static int parse_rel_factor(ParsedVector *TmpParsedVector)
{
  ParsedVector SecondTerm;
  int t, i;
  
  if (parse_num_term(TmpParsedVector))
    return 1;
  if (next_token(&t))
    return 1;
  while ((t==TPlus) || (t==TMinus))
  {
    if (parse_num_term(&SecondTerm))
      return 1;
    if (SecondTerm.nbTerms != TmpParsedVector->nbTerms)
    {
      if (SecondTerm.nbTerms > TmpParsedVector->nbTerms)
      {
        if (TmpParsedVector->nbTerms == 1)
          for (i=1 ; i<SecondTerm.nbTerms ; i++)
            TmpParsedVector->Val[i] = TmpParsedVector->Val[0];
        else
          for (i=TmpParsedVector->nbTerms ; i<SecondTerm.nbTerms ; i++)
            TmpParsedVector->Val[i] = 0.0;
        TmpParsedVector->nbTerms = SecondTerm.nbTerms;
      } else
      {
        if (SecondTerm.nbTerms == 1)
          for (i=1 ; i<TmpParsedVector->nbTerms ; i++)
            SecondTerm.Val[i] = SecondTerm.Val[0];
        else
          for (i=SecondTerm.nbTerms ; i<TmpParsedVector->nbTerms ; i++)
            SecondTerm.Val[i] = 0.0;
      }
    }
    if (t==TPlus)
    {
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] += SecondTerm.Val[i];
    } else
    {
      for (i=0 ; i<TmpParsedVector->nbTerms ; i++)
        TmpParsedVector->Val[i] -= SecondTerm.Val[i];
    }
    if (next_token(&t))
      return 1;
  }
  OldToken = t;
  IsOldToken = 1;
  return 0;
}

/*****************************************************************************
*  parse_float
******************************************************************************/
static int parse_float(double *pLocalFloat)
{
  ParsedVector TmpParsedVector;

  parse_rel_factor(&TmpParsedVector);
  if (TmpParsedVector.nbTerms != 1)
  {
    g_message(_("%s, line %d\nExpecting a Float but got a vector...\n"),
              PovFile->Name, PovFile->Line);
    return 1;
  }
  *pLocalFloat = TmpParsedVector.Val[0];
  return 0;
}

/*****************************************************************************
*  parse_word
******************************************************************************/
static int parse_word(int *pToken)
{
  char *Word;
  guint WordLength, WordMaxLength;
  int c, i;
  FloatConstantStruct       *LocalFloatConstant;
  VectorConstantStruct      *LocalVectorConstant;
  ColorConstantStruct       *LocalColorConstant;
  StringConstantStruct      *LocalStringConstant;
  ObjectConstantStruct      *LocalObjectConstant;
  TextureConstantStruct     *LocalTextureConstant;
  PigmentConstantStruct     *LocalPigmentConstant;
  CameraConstantStruct      *LocalCameraConstant;
  LightSourceConstantStruct *LocalLightSourceConstant;
  
  WordMaxLength = 20;
  Word = g_malloc(21);
  WordLength = 0;
  c = next_char();
  while((c>='a' && c<='z') || (c>='A' && c<='Z') || (c>='0' && c<='9') 
      || c=='_')
  {
    Word[WordLength++] = (char)c;
    if (WordLength == WordMaxLength)
    {
      WordMaxLength *= 2;
      Word = g_realloc(Word, WordMaxLength);
    }
    c = next_char();
  }
  OldChar = c; IsOldChar = 1;
  Word[WordLength] = 0;
 
  for (i=0 ; i<LastWord ; i++)
  {
    if (!strcmp(ReservedWordTable[i].String, Word))
    {
      *pToken = ReservedWordTable[i].Token;
      g_free(Word);
      return 0;
    }
  }
  for (LocalFloatConstant = CurrentFrame->FloatConstantList ;
       LocalFloatConstant ;
       LocalFloatConstant=LocalFloatConstant->Next)
  {
    if (!strcmp(LocalFloatConstant->Name, Word))
    {
      *pToken = TFloatID;
      LastFloatConstant = LocalFloatConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalVectorConstant = CurrentFrame->VectorConstantList ;
       LocalVectorConstant ;
       LocalVectorConstant=LocalVectorConstant->Next)
  {
    if (!strcmp(LocalVectorConstant->Name, Word))
    {
      *pToken = TVectorID;
      LastVectorConstant = LocalVectorConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalColorConstant = CurrentFrame->ColorConstantList ;
       LocalColorConstant ;
       LocalColorConstant=LocalColorConstant->Next)
  {
    if (!strcmp(LocalColorConstant->Name, Word))
    {
      *pToken = TColorID;
      LastColorConstant = LocalColorConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalStringConstant = CurrentFrame->StringConstantList ;
       LocalStringConstant ;
       LocalStringConstant = LocalStringConstant->Next)
  {
    if (!strcmp(LocalStringConstant->Name, Word))
    {
      *pToken = TStringID;
      LastStringConstant = LocalStringConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalObjectConstant = CurrentFrame->ObjectConstantList ;
       LocalObjectConstant ;
       LocalObjectConstant = LocalObjectConstant->Next)
  {
    if (!strcmp(LocalObjectConstant->Name, Word))
    {
      *pToken = TObjectID;
      LastObjectConstant = LocalObjectConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalTextureConstant = CurrentFrame->TextureConstantList ;
       LocalTextureConstant ;
       LocalTextureConstant = LocalTextureConstant->Next)
  {
    if (!strcmp(LocalTextureConstant->Name, Word))
    {
      *pToken = TTextureID;
      LastTextureConstant = LocalTextureConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalPigmentConstant = CurrentFrame->PigmentConstantList ;
       LocalPigmentConstant ;
       LocalPigmentConstant = LocalPigmentConstant->Next)
  {
    if (!strcmp(LocalPigmentConstant->Name, Word))
    {
      *pToken = TPigmentID;
      LastPigmentConstant = LocalPigmentConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalCameraConstant = CurrentFrame->CameraConstantList;
       LocalCameraConstant ;
       LocalCameraConstant = LocalCameraConstant->Next)
  {
    if (!strcmp(LocalCameraConstant->Name, Word))
    {
      *pToken = TCameraID;
      LastCameraConstant = LocalCameraConstant;
      g_free(Word);
      return 0;
    }
  }
  for (LocalLightSourceConstant = CurrentFrame->LightSourceConstantList;
       LocalLightSourceConstant ;
       LocalLightSourceConstant = LocalLightSourceConstant->Next)
  {
    if (!strcmp(LocalLightSourceConstant->Name, Word))
    {
      *pToken = TLightSourceID;
      LastLightSourceConstant = LocalLightSourceConstant;
      g_free(Word);
      return 0;
    }
  }
  if (*pToken == -1)
  {
    g_message(_("%s, line %d\nUnknown keyword : '%s'...\n"), 
              PovFile->Name, PovFile->Line, Word);
    return 1;
  }
  g_free(Word);
  return 0;
}

/*****************************************************************************
*  get_type_id
******************************************************************************/
static int get_type_id(char *Name)
{
  FloatConstantStruct *LocalFloatConstant;
  VectorConstantStruct *LocalVectorConstant;
  ColorConstantStruct *LocalColorConstant;
  StringConstantStruct *LocalStringConstant;
  ObjectConstantStruct *LocalObjectConstant;
  TextureConstantStruct *LocalTextureConstant;
  PigmentConstantStruct *LocalPigmentConstant;
  CameraConstantStruct *LocalCameraConstant;
  LightSourceConstantStruct *LocalLightSourceConstant;

  for (LocalFloatConstant = CurrentFrame->FloatConstantList ;
       LocalFloatConstant ;
       LocalFloatConstant=LocalFloatConstant->Next)
    if (!strcmp(LocalFloatConstant->Name, Name))
      return TFloatID;
  for (LocalVectorConstant = CurrentFrame->VectorConstantList ;
       LocalVectorConstant ;
       LocalVectorConstant=LocalVectorConstant->Next)
    if (!strcmp(LocalVectorConstant->Name, Name))
      return TVectorID;
  for (LocalColorConstant = CurrentFrame->ColorConstantList ;
       LocalColorConstant ;
       LocalColorConstant=LocalColorConstant->Next)
    if (!strcmp(LocalColorConstant->Name, Name))
      return TColorID;
  for (LocalStringConstant = CurrentFrame->StringConstantList ;
       LocalStringConstant ;
       LocalStringConstant = LocalStringConstant->Next)
    if (!strcmp(LocalStringConstant->Name, Name))
      return TStringID;
  for (LocalObjectConstant = CurrentFrame->ObjectConstantList ;
       LocalObjectConstant ;
       LocalObjectConstant = LocalObjectConstant->Next)
    if (!strcmp(LocalObjectConstant->Name, Name))
      return TObjectID;
  for (LocalTextureConstant = CurrentFrame->TextureConstantList ;
       LocalTextureConstant ;
       LocalTextureConstant = LocalTextureConstant->Next)
    if (!strcmp(LocalTextureConstant->Name, Name))
      return TTextureID;
  for (LocalPigmentConstant = CurrentFrame->PigmentConstantList ;
       LocalPigmentConstant ;
       LocalPigmentConstant = LocalPigmentConstant->Next)
    if (!strcmp(LocalPigmentConstant->Name, Name))
      return TPigmentID;
  for (LocalCameraConstant = CurrentFrame->CameraConstantList;
       LocalCameraConstant ;
       LocalCameraConstant = LocalCameraConstant->Next)
    if (!strcmp(LocalCameraConstant->Name, Name))
      return TCameraID;
  for (LocalLightSourceConstant = CurrentFrame->LightSourceConstantList;
       LocalLightSourceConstant ;
       LocalLightSourceConstant = LocalLightSourceConstant->Next)
    if (!strcmp(LocalLightSourceConstant->Name, Name))
      return TLightSourceID;
  return 0;
}

/*****************************************************************************
*  parse_declare
*  Parsing of "#declare Name = blah" sort of things 
******************************************************************************/
static int parse_declare(void)
{
  char *ConstantName;
  guint ConstantNameMaxLength, ConstantNameLength;
  int c, t, i;
  ParsedVector LocalExpress;
  Vector LocalColor;
  char *LocalString;
  ObjectStruct *LocalObject = NULL;
  TextureStruct *LocalTexture = NULL;
  PigmentStruct *LocalPigment = NULL;
  CameraStruct *LocalCamera = NULL;
  LightSourceStruct *LocalLightSource = NULL;
  int PreviousTypeId;
  FloatConstantStruct       *TmpFloatConstant;
  VectorConstantStruct      *TmpVectorConstant;
  ColorConstantStruct       *TmpColorConstant;
  StringConstantStruct      *TmpStringConstant;
  ObjectConstantStruct      *TmpObjectConstant;
  TextureConstantStruct     *TmpTextureConstant;
  PigmentConstantStruct     *TmpPigmentConstant;
  CameraConstantStruct      *TmpCameraConstant;
  LightSourceConstantStruct *TmpLightSourceConstant;
 
  /* First, get the Constant Name ([a-zA-Z0-9_]+) before the '=' */
  if (ignore_space())
    return 1;
  ConstantNameMaxLength = 20;
  ConstantName = g_malloc(21);
  ConstantNameLength = 0;
  c = next_char();
  while((c>='a' && c<='z') || (c>='A' && c<='Z') || 
        (c>='0' && c<='9') || c=='_')
  {
    ConstantName[ConstantNameLength++] = (char)c;
    if (ConstantNameLength == ConstantNameMaxLength)
    {
      ConstantNameMaxLength *= 2;
      ConstantName = g_realloc(ConstantName, ConstantNameMaxLength);
    }
    c = next_char();
  }
  OldChar = c; IsOldChar = 1;
  ConstantName[ConstantNameLength] = 0;
  PreviousTypeId = get_type_id(ConstantName);
  if (next_token(&t))
    return 1;
  if (t != TEqual)
  {
    g_message(_("%s, line %d\nA '=' sign was expected here...\n"),
               PovFile->Name, PovFile->Line);
    return 1;
  }
  /* Get the first Token */
  if (next_token(&t))
    return 1;
  OldToken = t; IsOldToken = 1;
  /* And guess the type of the declaration with the first token */
  switch (t)
  {
    case TFloatID: case TVectorID: case TPlus: case TMinus: case Tabs: 
    case Tacos: case Tval: case Tasc: case Tasin: case Tatan2: case Tceil:
    case Tclock: case Tcos: case Tdegrees: case Tdiv: case Texp: 
    case Tfile_exists: case TFloatLiteral: case Tfloor: case Tint: case Tlog:
    case Tmin: case Tmax: case Tmod: case Tpi: case Tpow: case Tradians:
    case Tsin: case Tsqrt: case Tstrcmp: case Tstrlen: case Ttan: case Tvdot:
    case Tvlength: case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse:
    case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
    case Tvcross: case Tvnormalize: case Tvrotate: case Tx: case Ty: case Tz:
    case Tt: case Tu: case Tv: case TExclamation: case TLeftParen:
    case TLeftAngleBracket:
      /* Expression Constant */
      if (parse_express(&LocalExpress))
        return 1;
      if (LocalExpress.nbTerms == 1)
      { /* Float */
        switch(PreviousTypeId)
        {
          case 0: /* New Float Constant */
            TmpFloatConstant = g_malloc(sizeof(FloatConstantStruct));
            TmpFloatConstant->Next = CurrentFrame->FloatConstantList;
            TmpFloatConstant->Name = g_malloc(strlen(ConstantName)+1);
            strcpy(TmpFloatConstant->Name, ConstantName);
            TmpFloatConstant->Float = LocalExpress.Val[0];
            CurrentFrame->FloatConstantList = TmpFloatConstant;
            break;
          case TFloatID: /* Redefinition of a Float Constant */
            for (TmpFloatConstant = CurrentFrame->FloatConstantList ;
                 TmpFloatConstant ;
                 TmpFloatConstant = TmpFloatConstant->Next)
              if (!strcmp(TmpFloatConstant->Name, ConstantName))
                TmpFloatConstant->Float = LocalExpress.Val[0];
            break;
          default: /* Redefinition of a non-Float Constant to a Float Constant */
            /* FIXME */
            g_message(_("%s, line %d\nCan't deals (yet) with redefinitions...(%s)\n"),
                      PovFile->Name, PovFile->Line, ConstantName);
            return 1;
        }
      } else
      { /* Vector */
        switch (PreviousTypeId)
        {
          case 0: /* New Vector Constant */
            TmpVectorConstant = g_malloc(sizeof(VectorConstantStruct));
            TmpVectorConstant->Next = CurrentFrame->VectorConstantList;
            TmpVectorConstant->Name = g_malloc(strlen(ConstantName)+1);
            strcpy(TmpVectorConstant->Name, ConstantName);
            for (i=0; i < LocalExpress.nbTerms ; i++)
              TmpVectorConstant->Vector[i] = LocalExpress.Val[i];
            CurrentFrame->VectorConstantList = TmpVectorConstant;
          case TVectorID: /* Redefinition of a Vector Constant */
            for (TmpVectorConstant = CurrentFrame->VectorConstantList ;
                 TmpVectorConstant ;
                 TmpVectorConstant = TmpVectorConstant->Next)
              if (!strcmp(TmpVectorConstant->Name, ConstantName))
              {
                TmpVectorConstant->Vector[0] = LocalExpress.Val[0];
                TmpVectorConstant->Vector[1] = LocalExpress.Val[1];
                TmpVectorConstant->Vector[2] = LocalExpress.Val[2];
                TmpVectorConstant->Vector[3] = LocalExpress.Val[3];
                TmpVectorConstant->Vector[4] = LocalExpress.Val[4];
              }
            return 1;
            break;
          default: /* Redefinition of a non-vector constant to a vector constant */
            /* FIXME */
            g_message(_("%s, line %d\nCan't deals (yet) with redefinitions...(%s)\n"),
                      PovFile->Name, PovFile->Line);
            return 1;
        }
      }
      break;
    case TColorID: case Tred: case Tgreen: case Tblue: case Tfilter: 
    case Ttransmit: case Tcolor: case Talpha: case Trgb: case Trgbt:
    case Trgbf: case Trgbft:
      /* Color Constant */
      if (parse_color(LocalColor))
        return 1;
      switch(PreviousTypeId)
      {
        case 0: /* New Color Constant */
          {
            gdouble *tmp_color;
            TmpColorConstant = g_malloc(sizeof(ColorConstantStruct));
            TmpColorConstant->Next = CurrentFrame->ColorConstantList;
            TmpColorConstant->Name = g_malloc(strlen(ConstantName)+1);
            strcpy(TmpColorConstant->Name, ConstantName);
            for (i=0 ; i<5 ; i++)
              TmpColorConstant->Color[i] = LocalColor[i];
            CurrentFrame->ColorConstantList = TmpColorConstant;
          
            tmp_color = g_malloc(5*sizeof(gdouble));
            for (i=0 ; i<5 ; i++)
              tmp_color[i] = TmpColorConstant->Color[i];
            g_hash_table_insert(colors_hash,
                                TmpColorConstant->Name,
                                tmp_color);
          }
          break;
        case TColorID: /* Redefinition of a Color Constant */
          for (TmpColorConstant = CurrentFrame->ColorConstantList ;
               TmpColorConstant ;
               TmpColorConstant = TmpColorConstant->Next)
            if (!strcmp(TmpColorConstant->Name, ConstantName))
            {
              TmpColorConstant->Color[0] = LocalColor[0];
              TmpColorConstant->Color[1] = LocalColor[1];
              TmpColorConstant->Color[2] = LocalColor[2];
              TmpColorConstant->Color[3] = LocalColor[3];
              TmpColorConstant->Color[4] = LocalColor[4];
            }
          break;
        default: /* Redefinition of a non-Color Constant to a Color Constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deals (yet) with redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;
      }
      break;
    case TStringID: case Tstr: case Tconcat: case Tchr: case Tsubstr:
    case Tstrupr: case Tstrlwr: case TStringLiteral:
      /* String Constant */
      if (parse_string(&LocalString))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New String Constant */
          TmpStringConstant = g_malloc(sizeof(StringConstantStruct));
          TmpStringConstant->Next = CurrentFrame->StringConstantList;
          TmpStringConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpStringConstant->Name, ConstantName);
          TmpStringConstant->String = g_malloc(strlen(LocalString)+1);
          strcpy(TmpStringConstant->String, LocalString);
          CurrentFrame->StringConstantList = TmpStringConstant;
          break;
        case TStringID: /* Redefinition of a String Constant */
          for (TmpStringConstant = CurrentFrame->StringConstantList ;
               TmpStringConstant ;
               TmpStringConstant = TmpStringConstant->Next)
            if (!strcmp(TmpStringConstant->Name, ConstantName))
            {
              g_free(TmpStringConstant->String);
              TmpStringConstant->String = g_strdup(LocalString);
            }
          break;
        default: /* Redefinition of a non-String Constant to a String Constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deals (yet) with redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;
      }
      break;
    case TObjectID: case Tobject: case Tbicubic_patch: case Tblob: case Tbox:
    case Tcone: case Tcubic: case Tcylinder: case Tdifference: case Tdisc:
    case Theight_field: case Tintersection: case Tjulia_fractal: case Tlathe:
    case Tmerge: case Tmesh: case Tplane: case Tpoly: case Tpolygon:
    case Tprism: case Tquadric: case Tquartic: case Tsmooth_triangle: case Tsor:
    case Tsphere: case Tsuperellipsoid: case Ttext: case Ttorus: case Ttriangle:
    case Tunion:
      /* Object Constant */
      if (parse_object(&LocalObject))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New Object Constant */
          TmpObjectConstant = g_malloc(sizeof(ObjectConstantStruct));
          TmpObjectConstant->Next = CurrentFrame->ObjectConstantList;
          TmpObjectConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpObjectConstant->Name, ConstantName);
          TmpObjectConstant->Object = CopyObject(LocalObject);
          CurrentFrame->ObjectConstantList = TmpObjectConstant;
          break;
        case TObjectID: /* Redefinition of an object constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with object redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-Object Constant to an object constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;
      }
      break;
    case TTextureID: case Ttexture:
      /* Texture Constant */
      IsOldToken = 0;
      if (parse_texture(&LocalTexture))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New Texture Constant */
          TmpTextureConstant = g_malloc(sizeof(TextureConstantStruct));
          TmpTextureConstant->Next = CurrentFrame->TextureConstantList;
          TmpTextureConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpTextureConstant->Name, ConstantName);
          TmpTextureConstant->Texture = CopyTexture(LocalTexture);
          CurrentFrame->TextureConstantList = TmpTextureConstant;
          break;
        case TTextureID: /* Redefinition of a texture constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with texture redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-Texture Constant to a texture constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;
      }
      break;
    case TPigmentID: case Tpigment:
      /* Pigment Constant */
      IsOldToken = 0;
      if (parse_pigment(&LocalPigment))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New Pigment Constant */
          TmpPigmentConstant = g_malloc(sizeof(PigmentConstantStruct));
          TmpPigmentConstant->Next = CurrentFrame->PigmentConstantList;
          TmpPigmentConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpPigmentConstant->Name, ConstantName);
          TmpPigmentConstant->Pigment = CopyPigment(LocalPigment);
          CurrentFrame->PigmentConstantList = TmpPigmentConstant;
          break;
        case TPigmentID: /* Redefinition of a pigment constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with pigment redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-pigment Constant to a pigment constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Name, ConstantName);
          return 1;
      }
      break;
    case TCameraID: case Tcamera:
      /* Camera Constant */
      IsOldToken = 0;
      if (parse_camera(&LocalCamera))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New camera constant */
          TmpCameraConstant = g_malloc(sizeof(CameraConstantStruct));
          TmpCameraConstant->Next = CurrentFrame->CameraConstantList;
          TmpCameraConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpCameraConstant->Name, ConstantName);
          TmpCameraConstant->Camera = CopyCamera(LocalCamera);
          CurrentFrame->CameraConstantList = TmpCameraConstant;
          break;
        case TCameraID: /* Redefinition of a camera constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with camera redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-camera constant to a camera constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;
      }
      break;
    case TLightSourceID: case Tlight_source:
      /* Light Source Constant */
      IsOldToken = 0;
      if (parse_light_source(&LocalLightSource))
        return 1;
      switch (PreviousTypeId)
      {
        case 0: /* New lightsource constant */
          TmpLightSourceConstant = g_malloc(sizeof(LightSourceConstantStruct));
          TmpLightSourceConstant->Next = CurrentFrame->LightSourceConstantList;
          TmpLightSourceConstant->Name = g_malloc(strlen(ConstantName)+1);
          strcpy(TmpLightSourceConstant->Name, ConstantName);
          TmpLightSourceConstant->LightSource = CopyLightSource(LocalLightSource);
          CurrentFrame->LightSourceConstantList = TmpLightSourceConstant;
          break;
        case TLightSourceID: /* Redefinition of a lightsource constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with light source redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;

          break;
        default: /* Redefinition of a non-lightsource constant to a lightsource constant */
          /* FIXME */
          g_message(_("%s, line %d\nCan't deal (yet) with redefinitions...(%s)\n"),
                    PovFile->Name, PovFile->Line, ConstantName);
          return 1;
      }
      break;
    default:
      g_message(_("%s, line %d\nParse Error...\n"), PovFile->Name, PovFile->Line);
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_directive
******************************************************************************/
static int parse_directive(void)
{
  char *Directive;
  guint DirectiveLength, DirectiveMaxLength;
  int c;
  char *FileName;
  ParsedVector LocalExpress;
  CondDirective *TmpCond; 
  int LocalToken;
  char *LocalString;
 
  if (ignore_space()) /* Are Spaces really allowed between Hash and Directive ?!? */
    return 1;
  ParsingDirective = 1;
  DirectiveMaxLength = 20;
  Directive = g_malloc(21);
  DirectiveLength = 0;
  c = next_char();
  while(c>='a' && c<='z')
  {
    Directive[DirectiveLength++]=(char)c;
    if (DirectiveLength==DirectiveMaxLength)
    {
      DirectiveMaxLength *= 2;
      Directive = g_realloc(Directive, DirectiveMaxLength);
    }
    c = next_char();
  }
  OldChar = c; IsOldChar = 1;
  Directive[DirectiveLength] = 0;

  if (!strcmp(Directive, "version"))
  {
    if (parse_float(&PovVersionLangage))
      return 1;
  } else if (!strcmp(Directive, "include"))
  {
    if (parse_string(&FileName))
      return 1;
    PovFile->Next = g_malloc(sizeof(ListFile));
    PovFile->Next->Prev = PovFile;
    PovFile = PovFile->Next;
    PovFile->Next = NULL;
    PovFile->handle = LibOpen(FileName);
    if (PovFile->handle == NULL)
    {
      g_message(_("%s, line %d\nCan't open file \"%s\"...\n"),
                PovFile->Prev->Name, PovFile->Prev->Line, FileName);
      return 1;
    }
    PovFile->Name = FileName;
    PovFile->Line = 1;
  } else if (!strcmp(Directive, "ifdef"))
  {
    TmpCond = g_malloc(sizeof(CondDirective));
    TmpCond->Next = ConditionalStack;
    ConditionalStack = TmpCond;
    if (DontParse)
    {
      ConditionalStack->Style = UNPARSED;
    } else
    {
      if (paren_error("#ifdef", TLeftParen))
        return 1;
      LocalToken = 0;
      if (parse_word(&LocalToken))
        return 1;
      if (paren_error("#ifdef", TRightParen))
        return 1;
      if (LocalToken)
      {
        ConditionalStack->Style = IF_TRUE_IF_ELSE;
      } else
      {
        ConditionalStack->Style = IF_FALSE_IF_ELSE;
        DontParse = TRUE;
      }
    }
  } else if (!strcmp(Directive, "ifndef"))
  {
    TmpCond = g_malloc(sizeof(CondDirective));
    TmpCond->Next = ConditionalStack;
    ConditionalStack = TmpCond;
    if (DontParse)
    {
      ConditionalStack->Style = UNPARSED;
    } else
    {
      if (paren_error("#ifndef", TLeftParen))
        return 1;
      LocalToken = 0;
      if (parse_word(&LocalToken))
        return 1;
      if (paren_error("#ifndef", TRightParen))
        return 1;
      if (!LocalToken)
      {
        ConditionalStack->Style = IF_TRUE_IF_ELSE;
      } else
      {
        ConditionalStack->Style = IF_FALSE_IF_ELSE;
        DontParse = TRUE;
      }
    }
  } else if (!strcmp(Directive, "if"))
  {
    TmpCond = g_malloc(sizeof(CondDirective));
    TmpCond->Next = ConditionalStack;
    ConditionalStack = TmpCond;
    if (DontParse)
    {
      ConditionalStack->Style = UNPARSED;
    } else
    {
      if (paren_error("#if", TLeftParen))
        return 1;
      if (parse_express(&LocalExpress))
        return 1;
      if (paren_error("#if", TRightParen))
        return 1;
      if (LocalExpress.nbTerms != 1)
      {
        g_message(_("%s, line %d\nConditions must evaluate to a float...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
      }
      if (fabs(LocalExpress.Val[0]) > SMALL_VALUE)
      {
        ConditionalStack->Style = IF_TRUE_IF_ELSE;
      } else
      {
        ConditionalStack->Style = IF_FALSE_IF_ELSE;
        DontParse = TRUE;
      }
    }
  } else if (!strcmp(Directive, "else"))
  {
    if (!ConditionalStack)
    {
      g_message(_("%s, line %d\nMisplaced #else...\n"), PovFile->Name, PovFile->Line);
      return 1;
    }
    switch (ConditionalStack->Style)
    {
      case IF_TRUE_IF_ELSE :
        DontParse = TRUE;
        ConditionalStack->Style = IF_TRUE_ELSE_END;
        break;
      case IF_TRUE_ELSE_END :
      case IF_FALSE_ELSE_END :
      case WHILE :
      case SWITCH_SWITCH_CASE :
        g_message(_("%s, line %d\nMisplaced #else...\n"), PovFile->Name, PovFile->Line);
        return 1;
      case IF_FALSE_IF_ELSE :
      case SWITCH_CASE_BREAK_FALSE :
        DontParse = FALSE;
        ConditionalStack->Style = IF_FALSE_ELSE_END;
        break;
      case SWITCH_CASE_BREAK_TRUE :
        DontParse = TRUE;
        ConditionalStack->Style = SWITCH_BREAK_END;
        break;
      case SWITCH_BREAK_END:
      case UNPARSED:
        /* Nothing */
        break;
    }
  } else if (!strcmp(Directive, "end"))
  {
    if (!ConditionalStack)
    {
      g_message(_("%s, line %d\nMisplaced #end...\n"), PovFile->Name, PovFile->Line);
      return 1;
    }
    switch (ConditionalStack->Style)
    {
      case IF_TRUE_IF_ELSE:
      case IF_TRUE_ELSE_END:
      case IF_FALSE_IF_ELSE:
      case IF_FALSE_ELSE_END:
      case SWITCH_BREAK_END:
      case SWITCH_CASE_BREAK_TRUE:
      case SWITCH_CASE_BREAK_FALSE:
      case SWITCH_SWITCH_CASE:
        TmpCond = ConditionalStack->Next;
        g_free(ConditionalStack);
        ConditionalStack = TmpCond;
        DontParse = FALSE;
        break;
      case WHILE:
        if (ConditionalStack->WhileFile != PovFile)
        {
          g_message(_("%s, line %d\n#while loop didn't end in file where it started...\n"),
                    PovFile->Name, PovFile->Line);
          return 1;
        }
        fseek(PovFile->handle, ConditionalStack->WhilePos, SEEK_SET);
        PovFile->Line = ConditionalStack->WhileLineNo;
        if (paren_error("#while", TLeftParen))
          return 1;
        if (parse_express(&LocalExpress))
          return 1;
        if (paren_error("#while", TRightParen))
          return 1;
        if (fabs(LocalExpress.Val[0]) > SMALL_VALUE)
        {
          /* Nothing ! ? */
        } else
        {
          DontParse = TRUE;
          ConditionalStack->Style = IF_TRUE_ELSE_END;
        }
        break;
      case UNPARSED:
        TmpCond = ConditionalStack->Next;
        g_free(ConditionalStack);
        ConditionalStack = TmpCond;
        break;
    }
  } else if (!strcmp(Directive, "declare"))
  {
    if (parse_declare())
      return 1;
  } else if (!strcmp(Directive, "debug"))
  {
    if (!DontParse)
    {
      if (parse_string(&LocalString))
        return 1;
      printf(LocalString); /* FIXME Should be redirected on the debug stream */
      g_free(LocalString);
    }
  } else if (!strcmp(Directive, "error"))
  {
    if (parse_string(&LocalString))
      return 1;
    if (!DontParse)
    {
      g_message(_("%s, line %d\nError: %s"),
                PovFile->Name, PovFile->Line, LocalString);
      return 1;
    }
    g_free(LocalString);
  } else if (!strcmp(Directive, "render"))
  {
    if (parse_string(&LocalString))
      return 1;
    if (!DontParse)
    {
      printf(LocalString); /* FIXME */
    }
    g_free(LocalString);
  } else if (!strcmp(Directive, "statistics"))
  {
    if (parse_string(&LocalString))
      return 1;
    if (!DontParse)
    {
      printf(LocalString); /* FIXME */
    }
    g_free(LocalString);
  } else if (!strcmp(Directive, "warning"))
  {
    if (parse_string(&LocalString))
      return 1;
    if (!DontParse)
    {
      printf(LocalString); /* FIXME */
    }
    g_free(LocalString);
  } else 
  {
    g_message(_("%s, line %d\nUnknown directive : %s\n"),
              PovFile->Name, PovFile->Line, Directive);
    return 1;
  }
  g_free(Directive);
  ParsingDirective = 0;
  return 0;
}

/*****************************************************************************
*  next_token
******************************************************************************/
static int next_token(int *pToken)
{
  int c, c1, c2;

  *pToken = 0;
  while (*pToken == 0)
  {
    if (IsOldToken)
    {
      IsOldToken = 0;
      if (OldToken != -1)
      {
        *pToken = OldToken;
        return 0; 
      }  
    }
    if (ignore_space())
      return 1;
    c = next_char();
    switch (c)
    {
      case EOF :
        if (PovFile->Prev)
        {
          fclose(PovFile->handle);
          g_free(PovFile->Name);
          PovFile = PovFile->Prev;
          g_free(PovFile->Next);
        } else *pToken = TEOF;
        break;
      case '!' :
        c2 = next_char();
        if (c2 == '=')
        {
          *pToken = TUnequal; 
        } else
        {
          *pToken = TExclamation;
          IsOldChar = 1 ; OldChar = c2;
        }
        break;
      case '"' :
        if (parse_string_literal(&LastString))
          return 1;
        *pToken = TStringLiteral;
        if (DontParse)
          g_free(LastString);
        break;
      case '#' :
        if (ParsingDirective)
        {
          IsOldChar = 1; OldChar = c;  
          *pToken = -1;
        } else
        {
          if (parse_directive())
            return 1;
          if (OldToken == -1)
            IsOldToken = 0;
        }
        break;
      case '&' :
        *pToken = TAmpersand;
        break;
      case '(' :
        *pToken = TLeftParen;
        break;
      case ')' :
        *pToken = TRightParen;
        break;
      case '*' :
        *pToken = TStar;
        break;
      case '+' :
        *pToken = TPlus;
        break;
      case ',' :
        *pToken = TComma;
        break;
      case '-' :
        *pToken = TMinus;
        break;
      case '0' ... '9' :
      case '.' :
        OldChar = c; IsOldChar = 1;
        if (parse_float_literal(&LastFloat))
          return 1;
        *pToken = TFloatLiteral;
        break;
      case '/' :
        c = next_char();
        if (c == '/')
        { /* comments ala C++ */
          while (c!='\r' && c!='\n')
          {
            c = next_char();
          }
          if (c=='\r')
          {
            c = next_char();
            if (c=='\n') c = next_char();
          }
          if (c=='\n')
          {
            c = next_char();
            if (c=='\r') c = next_char();
          }
          PovFile->Line++;
          OldChar = c; IsOldChar = 1;
          break;
        }
        if (c == '*')
        { /* comments ala C */
          c1 = c2 = next_char();
          do
          {
            if (c2=='\n')
            {
              PovFile->Line++;
              c2 = next_char();
              if (c2=='\r') c2 = next_char();
            } else if (c2=='\r')
            {
              PovFile->Line++;
              c2 = next_char();
              if (c2=='\n') c2 = next_char();
            } else
            {
              c1 = c2;
              c2 = next_char();
            }
          } while (c1!='*' || c2!='/');
          break;
        }
        OldChar = c; IsOldChar = 1;
        *pToken = TSlash;
        break;
      case ':' :
        *pToken = TColon;
        break;
      case '<' :
        c2 = next_char();
        if (c2 == '=')
        {
          *pToken = TLowerOrEqual;
        } else
        {
          IsOldChar = 1; OldChar = c2;
          *pToken = TLeftAngleBracket;
        }
        break;
      case '=' :
        *pToken = TEqual;
        break;
      case '>' :
        c2 = next_char();
        if (c2 == '=')
        {
          *pToken = TUpperOrEqual;
        } else
        {
          IsOldChar = 1; OldChar = c2;
          *pToken = TRightAngleBracket;
        }
        break;
      case '?' :
        *pToken = TQuestion;
        break;
      case 'A' ... 'Z' :
      case 'a' ... 'z' :
      case '_' :
        OldChar = c; IsOldChar = 1;
        if (!DontParse)
          *pToken = -1;
        if (parse_word(pToken))
          return 1;
        break;
      case '{' :
        *pToken = TLeftCurlyBracket;
        break;
      case '|' :
        *pToken = TBar;
        break;
      case '}' :
        *pToken = TRightCurlyBracket;
        break;
      case '[' :
        *pToken = TLeftSquareBracket;
        break;
      case ']' :
        *pToken = TRightSquareBracket;
        break;
      case ';' :
        *pToken = Tsemicolon;
        break;
      default :
        g_message(_("%s, line %d\nUnexpected character : '%c'\n"),
                  PovFile->Name, PovFile->Line, c);
        return 1;
    }
    if (DontParse)
      *pToken = 0;
  }
  return 0;
}

/*****************************************************************************
*  parse_color
******************************************************************************/
static int parse_color(Vector LocalColor)
{
  int t, i, Done;
  ParsedVector TmpColor;
  
  if (next_token(&t))
    return 1;
  if (t != Tcolor)
  {
    IsOldToken = 1; OldToken = t;
  }
  Done = FALSE;
  TmpColor.nbTerms = 5;
  for (i=0 ; i<5 ; i++)
    LocalColor[i] = TmpColor.Val[i]=0.0;
  while (!Done)
  {
    if (next_token(&t))
      return 1;
    switch (t)
    {
      case Tred:
        if (parse_float(&(TmpColor.Val[0])))
          return 1;
        LocalColor[0] = TmpColor.Val[0];
        break;
      case Tgreen:
        if (parse_float(&(TmpColor.Val[1])))
          return 1;
        LocalColor[1] = TmpColor.Val[1];
        break;
      case Tblue:
        if (parse_float(&(TmpColor.Val[2])))
          return 1;
        LocalColor[2] = TmpColor.Val[2];
        break;
      case Talpha: case Tfilter:
        if (parse_float(&(TmpColor.Val[3])))
          return 1;
        LocalColor[3] = TmpColor.Val[3];
        break;
      case Ttransmit:
        if (parse_float(&(TmpColor.Val[4])))
          return 1;
        LocalColor[4] = TmpColor.Val[4];
        break;
      case Trgb:
        if (parse_express(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[2] = TmpColor.Val[1] = TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 3)
        {
          for (i=TmpColor.nbTerms ; i<3 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<3 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        LocalColor[3] = LocalColor[4] = 0.0;
        break;
      case Trgbf:
        if (parse_express(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[3] = TmpColor.Val[2] = 
                            TmpColor.Val[1] = 
                            TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 4)
        {
          for (i=TmpColor.nbTerms ; i<4 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<4 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        LocalColor[4] = 0.0;
        break;
      case Trgbft:
        if (parse_express(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[4] = TmpColor.Val[3] = TmpColor.Val[2] = 
                            TmpColor.Val[1] = TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 5)
        {
          for (i=TmpColor.nbTerms ; i<5 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<5 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        break;
      case Trgbt:
        if (parse_express(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[3] = TmpColor.Val[2] = 
                            TmpColor.Val[1] = 
                            TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 4)
        {
          for (i=TmpColor.nbTerms ; i<4 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<3 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        LocalColor[3] = 0.0;
        LocalColor[4] = TmpColor.Val[3];
        break;
      case TPlus: case TMinus:
      case Tabs: case Tacos: case Tval: case Tasc: case Tasin: case Tatan2:
      case Tceil: case Tclock: case Tcos: case Tdegrees: case Tdiv: 
      case Texp: case Tfile_exists: case TFloatID: case TFloatLiteral:
      case Tfloor: case Tint: case Tlog: case Tmin: case Tmax: case Tmod:
      case Tpi: case Tpow: case Tradians: case Tsin: case Tsqrt: 
      case Tstrcmp: case Tstrlen: case Ttan: case Tvdot: case Tvlength:
      case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse: 
      case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
      case Tvcross: case TVectorID: case Tvnormalize: case Tvrotate:
      case Tx: case Ty: case Tz: case TColorID: case Tt: case Tu: case Tv:
      case TExclamation: case TLeftParen: case TLeftAngleBracket:
        OldToken = t; IsOldToken = 1;
        if (parse_express(&TmpColor))
          return 1;
        if (TmpColor.nbTerms == 1)
        {
          TmpColor.Val[4] = TmpColor.Val[3] = TmpColor.Val[2] = 
                            TmpColor.Val[1] = TmpColor.Val[0];
        } else if (TmpColor.nbTerms < 5)
        {
          for (i=TmpColor.nbTerms ; i<5 ; i++)
            TmpColor.Val[i] = 0.0;
        }
        for (i=0 ; i<5 ; i++)
          LocalColor[i] = TmpColor.Val[i];
        break;
      default:
        IsOldToken = 1; OldToken = t;
        Done = TRUE;
    }
  }
  return 0;
}

/*****************************************************************************
*  parse_pigment_or_color
******************************************************************************/
static int parse_pigment_or_color(PigmentStruct **pLocalPigment)
{
  int t;
  
  if (next_token(&t))
    return 1;
  if (t == Tpigment)
  {
    if (parse_pigment(pLocalPigment))
      return 1;
  } else
  {
    (*pLocalPigment)->Type = PAT_SOLID_COLOR;
    (*pLocalPigment)->Trans = NULL;
    if (parse_color((*pLocalPigment)->Color))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_matrix
******************************************************************************/
static int parse_matrix(TransformStruct *pTrans)
{
  int t;

  pTrans->Direct[0][3] = 0.0;
  pTrans->Direct[1][3] = 0.0;
  pTrans->Direct[2][3] = 0.0;
  pTrans->Direct[3][3] = 1.0;
  if (next_token(&t))
    return 1;
  if (t != TLeftAngleBracket)
  {
    g_message(_("%s, line %d\nA '<' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  if (parse_float(&(pTrans->Direct[0][0])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[0][1])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[0][2])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[1][0])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[1][1])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[1][2])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[2][0])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[2][1])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[2][2])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[3][0])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[3][1])))
    return 1;
  if (parse_comma()) return 1;
  if (parse_float(&(pTrans->Direct[3][2])))
    return 1;
  if (next_token(&t))
    return 1;
  if (t != TRightAngleBracket)
  {
    g_message(_("%s, line %d\nA '>' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  MInverse(pTrans->Inverse, pTrans->Direct);
  return 0;
}

/*****************************************************************************
*  parse_pigment
******************************************************************************/
static int parse_pigment(PigmentStruct **pLocalPigment)
{
  int t;
  ParsedVector LocalVector;
  TransformStruct Trans;
  double LocalFloat;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (next_token(&t))
    return 1;
  if (t == TPigmentID)
  {
    *pLocalPigment = CopyPigment(LastPigmentConstant->Pigment);
  } else
  {
    if (*pLocalPigment == NULL)
    {
      (*pLocalPigment) = g_malloc(sizeof(PigmentStruct));
      (*pLocalPigment)->Type = PAT_SOLID_COLOR;
      (*pLocalPigment)->Trans = NULL;
      (*pLocalPigment)->Color[0] = 0.0;
      (*pLocalPigment)->Color[1] = 0.0;
      (*pLocalPigment)->Color[2] = 0.0;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (next_token(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case TColorID: case Tred: case Tgreen: case Tblue: case Tfilter: 
      case Ttransmit: case Tcolor: case Talpha: case Trgb: case Trgbt:
      case Trgbf: case Trgbft:
        OldToken = t;IsOldToken = 1;
        if (parse_color((*pLocalPigment)->Color))
          return 1;
        (*pLocalPigment)->Type = PAT_SOLID_COLOR;
        break;
      case Tbrick:
/* FIXME LocalPigment->PigmentList = g_malloc(sizeof(PigmentListStruct));
        LocalPigment->PigmentList->NbPigment = 2;
        LocalPigment->PigmentList->PigmentArray = g_malloc(2*sizeof(PigmentStruct));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[0]));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[1]));
        LocalPigment->Type = PAT_BRICK;*/
        break;
      case Tchecker:
/* FIXME LocalPigment->PigmentList = g_malloc(sizeof(PigmentListStruct));
        LocalPigment->PigmentList->NbPigment = 2;
        LocalPigment->PigmentList->PigmentArray = g_malloc(2*sizeof(PigmentStruct));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[0]));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[1]));
        LocalPigment->Type = PAT_CHECKER;*/
        break;
      case Thexagon:
/* FIXME LocalPigment->PigmentList = g_malloc(sizeof(PigmentListStruct));
        LocalPigment->PigmentList->NbPigment = 3;
        LocalPigment->PigmentList->PigmentArray = g_malloc(3*sizeof(PigmentStruct));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[0]));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[1]));
        ParsePigmentOrColor(&(LocalPigment->PigmentList->PigmentArray[2]));
        LocalPigment->Type = PAT_HEXAGON;*/
        break;
      case Ttranslate:
        if (parse_vector(&LocalVector))
          return 1;
        if ((*pLocalPigment)->Trans)
        {
          ComputeTranslateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalPigment)->Trans, &Trans);
        } else
        {
          (*pLocalPigment)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeTranslateTrans((*pLocalPigment)->Trans, LocalVector.Val);
        }
        break;
      case Trotate:
        if (parse_vector(&LocalVector))
          return 1;
        if ((*pLocalPigment)->Trans)
        {
          ComputeRotateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalPigment)->Trans, &Trans);
        } else
        {
          (*pLocalPigment)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeRotateTrans((*pLocalPigment)->Trans, LocalVector.Val);
        }
        break;
      case Tscale:
        if (parse_vector(&LocalVector))
          return 1;
        if ((*pLocalPigment)->Trans)
        {
          ComputeScaleTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalPigment)->Trans, &Trans);
        } else
        {
          (*pLocalPigment)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeScaleTrans((*pLocalPigment)->Trans, LocalVector.Val);
        }
        break;
      case Tmatrix:
        if ((*pLocalPigment)->Trans)
        {
          if (parse_matrix(&Trans))
            return 1;
          ComposeTrans((*pLocalPigment)->Trans, &Trans);
        } else
        {
          (*pLocalPigment)->Trans = g_malloc(sizeof(TransformStruct));
          if (parse_matrix((*pLocalPigment)->Trans))
            return 1;
        }
        break;
      default:
        g_message(_("%s, line %d\nA pigment modifier was expected here...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
    }
    if (next_token(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_texture
******************************************************************************/
static int parse_texture(TextureStruct **pLocalTexture)
{
  int t;
  ParsedVector LocalVector;
  TransformStruct Trans;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (next_token(&t))
    return 1;
  if (t == TTextureID)
  {
    *pLocalTexture = CopyTexture(LastTextureConstant->Texture);
  } else
  {
    if (*pLocalTexture == NULL)
    {
      (*pLocalTexture) = g_malloc(sizeof(TextureStruct));
      (*pLocalTexture)->Pigment = NULL;
      (*pLocalTexture)->Trans = NULL;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (next_token(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case Tpigment:
        if (parse_pigment(&((*pLocalTexture)->Pigment)) )
          return 1;
        break;
      case Ttranslate:
        if (parse_vector(&LocalVector))
          return 1;
        if ((*pLocalTexture)->Trans)
        {
          ComputeTranslateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalTexture)->Trans, &Trans);
        } else
        {
          (*pLocalTexture)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeTranslateTrans((*pLocalTexture)->Trans, LocalVector.Val);
        }
        break;
      case Trotate:
        if (parse_vector(&LocalVector))
          return 1;
        if ((*pLocalTexture)->Trans)
        {
          ComputeRotateTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalTexture)->Trans, &Trans);
        } else
        {
          (*pLocalTexture)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeRotateTrans((*pLocalTexture)->Trans, LocalVector.Val);
        }
        break;
      case Tscale:
        if (parse_vector(&LocalVector))
          return 1;
        if ((*pLocalTexture)->Trans)
        {
          ComputeScaleTrans(&Trans, LocalVector.Val);
          ComposeTrans((*pLocalTexture)->Trans, &Trans);
        } else
        {
          (*pLocalTexture)->Trans = g_malloc(sizeof(TransformStruct));
          ComputeScaleTrans((*pLocalTexture)->Trans, LocalVector.Val);
        }
        break;
      case Tmatrix:
        if ((*pLocalTexture)->Trans)
        {
          if (parse_matrix(&Trans))
            return 1;
          ComposeTrans((*pLocalTexture)->Trans, &Trans);
        } else
        {
          (*pLocalTexture)->Trans = g_malloc(sizeof(TransformStruct));
          if (parse_matrix((*pLocalTexture)->Trans))
            return 1;
        }
        break;
      default:
        g_message(_("%s, line %d\nA texture modifier was expected here...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
    }
    if (next_token(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_object_modifier
******************************************************************************/
static int parse_object_modifier(ObjectStruct *LocalObject)
{
  int t;
  ParsedVector LocalVector;
  ObjectStruct *TmpObject;
  TransformStruct Trans;

  if (next_token(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch (t)
    {
      case Ttranslate:
        if (parse_vector(&LocalVector))
          return 1;
        giram_object_translate(LocalObject, LocalVector.Val);
        break;
      case Trotate:
        if (parse_vector(&LocalVector))
          return 1;
        giram_object_rotate(LocalObject, LocalVector.Val);
        break;
      case Tscale:
        if (parse_vector(&LocalVector))
          return 1;
        if (LocalVector.Val[0] == 0.0)
        { 
          g_message(_("%s, line %d\nYou can't scale by 0.0\nchanged to 1.0"),
                    PovFile->Name, PovFile->Line);
          LocalVector.Val[0] = 1.0;
        }
        if (LocalVector.Val[1] == 0.0)
        {
          g_message(_("%s, line %d\nYou can't scale by 0.0\nchanged to 1.0"),
                    PovFile->Name, PovFile->Line);
          LocalVector.Val[1] = 1.0;
        }
        if (LocalVector.Val[2] == 0.0)
        {
          g_message(_("%s, line %d\nYou can't scale by 0.0\nchanged to 1.0"),
                    PovFile->Name, PovFile->Line);
          LocalVector.Val[2] = 1.0;
        }
        giram_object_scale(LocalObject, LocalVector.Val);
        break;
      case Ttransform:
        /* FIXME */
        break;
      case Tmatrix:
        if (LocalObject->Trans)
        {
          if (parse_matrix(&Trans))
            return 1;
          ComposeTrans(LocalObject->Trans, &Trans);
        } else
        {
          LocalObject->Trans = g_malloc(sizeof(TransformStruct));
          if (parse_matrix(LocalObject->Trans))
            return 1;
        }
        break;
      case Ttexture:
        if (parse_texture(&(LocalObject->Texture)))
          return 1;
        break;
      case Tpigment:
        if (LocalObject->Texture == NULL)
        {
          LocalObject->Texture = g_malloc(sizeof(TextureStruct));
          LocalObject->Texture->Pigment = NULL;
          LocalObject->Texture->Trans = NULL;
        }
        if (parse_pigment(&(LocalObject->Texture->Pigment)))
          return 1;
        break;
      case Tinverse: /* A test for Flat Object could be welcomed here */
        LocalObject->Inverse = ! (LocalObject->Inverse);
        break;
      case Tno_shadow:
        LocalObject->NoShadow = 1;
        break;
      case Thollow:
        LocalObject->Hollow = 1;
        break;
      case Topen:
        switch (LocalObject->Type) {
	  case CYLINDER_OBJECT:
            ((CylinderStruct *)LocalObject)->Open = TRUE;
	    break;
          case SOR_OBJECT:
            ((SorStruct *)LocalObject)->Open = TRUE;
	    break;
	  case CONE_OBJECT:
	    ((ConeStruct *)LocalObject)->Open = TRUE;
	    break;
	  default:
            g_message(_("%s, line %d\nThe 'open' keyword can only be used with some objects\n"),
                      PovFile->Name, PovFile->Line);
            return 1;
        }
        break;
      default:
        g_message(_("%s, line %d\nAn object modifier was expected here...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
    }
    if (next_token(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_box
*  Parsing of box object
******************************************************************************/
static int parse_box(ObjectStruct **pBox)
{
  ParsedVector LocalVector1, LocalVector2;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_vector(&LocalVector1))
    return 1;
  if (parse_comma())
    return 1;
  if (parse_vector(&LocalVector2))
    return 1;
  *pBox = giram_box_new(LocalVector1.Val, LocalVector2.Val);
  if (parse_object_modifier(*pBox))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_cone
*  Parsing of cone object
******************************************************************************/
static int parse_cone(ObjectStruct **pCone)
{
  ParsedVector LocalVector1, LocalVector2;
  double Radius1, Radius2;
  TransformStruct Transform;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_vector(&LocalVector1))
    return 1;
  if (parse_comma())
    return 1;
  if (parse_float(&Radius1))
    return 1;
  if (parse_comma())
    return 1;
  if (parse_vector(&LocalVector2))
    return 1;
  if (parse_comma())
    return 1;
  if (parse_float(&Radius2))
    return 1;
  *pCone = giram_cone_new(LocalVector1.Val, Radius1, LocalVector2.Val, Radius2);

  if (parse_object_modifier(*pCone))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_cylinder
*  Parsing of cylinder object
******************************************************************************/
static int parse_cylinder(ObjectStruct **pCylinder)
{
  ParsedVector LocalVector1, LocalVector2;
  double Radius;
  Vector TmpVect;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_vector(&LocalVector1))
    return 1;
  if (parse_comma())
    return 1;
  if (parse_vector(&LocalVector2))
    return 1;
  if (parse_comma())
    return 1;
  if (parse_float(&Radius))
    return 1;
  if (Radius <= 0.0)
  {
    g_message(_("%s, line %d\nThe Radius of cylinder must be strictly positive.\n"),
              PovFile->Name, PovFile->Line);
    return 1;
  }
  TmpVect[0] = LocalVector2.Val[0] - LocalVector1.Val[0];
  TmpVect[1] = LocalVector2.Val[1] - LocalVector1.Val[1];
  TmpVect[2] = LocalVector2.Val[2] - LocalVector1.Val[2];
  if (fabs(VLength(TmpVect)) < 10e-6)
  {
    g_message(_("%s, line %d\nThe Length of cylinder mustn't be nul.\n"),
              PovFile->Name, PovFile->Line);
    return 1;
  }
  *pCylinder = giram_cylinder_new(LocalVector1.Val, LocalVector2.Val, Radius);
  if (parse_object_modifier(*pCylinder))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_disc
*  Parsing of disc object
******************************************************************************/
static int parse_disc(ObjectStruct **pDisc)
{
  ParsedVector LocalVector;
  Vector Center, Normal;
  double Radius, HoleRadius = 0.0;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_vector(&LocalVector))
    return 1;
  Center[0] = LocalVector.Val[0];
  Center[1] = LocalVector.Val[1];
  Center[2] = LocalVector.Val[2];
  if (parse_comma())
    return 1;
  if (parse_vector(&LocalVector))
    return 1;
  Normal[0] = LocalVector.Val[0];
  Normal[1] = LocalVector.Val[1];
  Normal[2] = LocalVector.Val[2];
  if (parse_comma())
    return 1;
  if (parse_float(&Radius))
    return 1;
  if (parse_comma())
    return 1;
  if (next_token(&t))
    return 1;
  OldToken = t; IsOldToken = 1;
  switch (t)
  {
    case TFloatID: case TVectorID: case TPlus: case TMinus: case Tabs: 
    case Tacos: case Tval: case Tasc: case Tasin: case Tatan2: case Tceil:
    case Tclock: case Tcos: case Tdegrees: case Tdiv: case Texp: 
    case Tfile_exists: case TFloatLiteral: case Tfloor: case Tint: case Tlog:
    case Tmin: case Tmax: case Tmod: case Tpi: case Tpow: case Tradians:
    case Tsin: case Tsqrt: case Tstrcmp: case Tstrlen: case Ttan: case Tvdot:
    case Tvlength: case Tversion: case Ttrue: case Tyes: case Ton: case Tfalse:
    case Tno: case Toff: case Tseed: case Trand: case Tvaxis_rotate:
    case Tvcross: case Tvnormalize: case Tvrotate: case Tx: case Ty: case Tz:
    case Tt: case Tu: case Tv: case TExclamation: case TLeftParen:
    case TLeftAngleBracket:
      if (parse_float(&HoleRadius))
        return 1;
      break;
  }
  *pDisc = giram_disc_new(Center, Normal, Radius, HoleRadius);
  
  if (parse_object_modifier(*pDisc))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_mesh
*  Parsing of mesh object
******************************************************************************/
static int parse_mesh(ObjectStruct **pMesh)
{
  MeshStruct *MMesh;
  ParsedVector LocalVector;
  TriangleListStruct *TmpTri;
  Vector V1, V2;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  *pMesh = giram_mesh_new();
  MMesh = (MeshStruct *)(*pMesh);
  if (next_token(&t))
    return 1;
  while (t != TRightCurlyBracket)
  { 
    switch(t)
    {
      case Tsmooth_triangle:
        if (next_token(&t))
          return 1;
        if (t != TLeftCurlyBracket)
        {
          g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        TmpTri = g_malloc(sizeof(TriangleListStruct));
        TmpTri->Next = MMesh->FirstTriangle;
        MMesh->FirstTriangle = TmpTri;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->P1[0] = LocalVector.Val[0];
        TmpTri->P1[1] = LocalVector.Val[1];
        TmpTri->P1[2] = LocalVector.Val[2];
        if (parse_comma()) return 1;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->N1[0] = LocalVector.Val[0];
        TmpTri->N1[1] = LocalVector.Val[1];
        TmpTri->N1[2] = LocalVector.Val[2];
        if (parse_comma()) return 1;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->P2[0] = LocalVector.Val[0];
        TmpTri->P2[1] = LocalVector.Val[1];
        TmpTri->P2[2] = LocalVector.Val[2];
        if (parse_comma()) return 1;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->N2[0] = LocalVector.Val[0];
        TmpTri->N2[1] = LocalVector.Val[1];
        TmpTri->N2[2] = LocalVector.Val[2];
        if (parse_comma()) return 1;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->P3[0] = LocalVector.Val[0];
        TmpTri->P3[1] = LocalVector.Val[1];
        TmpTri->P3[2] = LocalVector.Val[2];
        if (parse_comma()) return 1;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->N3[0] = LocalVector.Val[0];
        TmpTri->N3[1] = LocalVector.Val[1];
        TmpTri->N3[2] = LocalVector.Val[2];
        if (next_token(&t))
          return 1;
        if (t != TRightCurlyBracket)
        {
          g_message(_("%s, line %d\nA '}' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        break;
      case Ttriangle:
        if (next_token(&t))
          return 1;
        if (t != TLeftCurlyBracket)
        {
          g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        TmpTri = g_malloc(sizeof(TriangleListStruct));
        TmpTri->Next = MMesh->FirstTriangle;
        MMesh->FirstTriangle = TmpTri;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->P1[0] = LocalVector.Val[0];
        TmpTri->P1[1] = LocalVector.Val[1];
        TmpTri->P1[2] = LocalVector.Val[2];
        if (parse_comma()) return 1;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->P2[0] = LocalVector.Val[0];
        TmpTri->P2[1] = LocalVector.Val[1];
        TmpTri->P2[2] = LocalVector.Val[2];
        if (parse_comma()) return 1;
        if (parse_vector(&LocalVector)) return 1;
        TmpTri->P3[0] = LocalVector.Val[0];
        TmpTri->P3[1] = LocalVector.Val[1];
        TmpTri->P3[2] = LocalVector.Val[2];
        if (next_token(&t))
          return 1;
        if (t != TRightCurlyBracket)
        {
          g_message(_("%s, line %d\nA '}' was expected here...\n"), PovFile->Name, PovFile->Line); 
          return 1;
        }
        V1[0] = TmpTri->P2[0] - TmpTri->P1[0];
        V1[1] = TmpTri->P2[1] - TmpTri->P1[1];
        V1[2] = TmpTri->P2[2] - TmpTri->P1[2];
        V2[0] = TmpTri->P3[0] - TmpTri->P1[0];
        V2[1] = TmpTri->P3[1] - TmpTri->P1[1];
        V2[2] = TmpTri->P3[2] - TmpTri->P1[2];
        VCross(TmpTri->N1, V1, V2);
        VCross(TmpTri->N2, V1, V2);
        VCross(TmpTri->N3, V1, V2);
        break;
      default:
        OldToken = t; IsOldToken = 1;
        if (parse_object_modifier(*pMesh))
          return 1;
        IsOldToken = 1;
        OldToken = TRightCurlyBracket;
    }
    if (next_token(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_plane
*  Parsing of plane object
******************************************************************************/
static int parse_plane(ObjectStruct **pPlane)
{
  ParsedVector LocalVector;
  double Distance;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_vector(&LocalVector))
    return 1;
  if (parse_comma())
    return 1;
  if (parse_float(&(Distance)))
    return 1;
  (*pPlane) = giram_plane_new(LocalVector.Val, Distance);
  if (parse_object_modifier(*pPlane))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_smooth_triangle
*  Parsing of smooth triangles
******************************************************************************/
static int parse_smooth_triangle(ObjectStruct **pTriangle)
{
  TriangleStruct *TTriangle;
  ParsedVector LocalP1, LocalN1, LocalP2, LocalN2, LocalP3, LocalN3;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_vector(&LocalP1)) return 1;
  if (parse_comma()) return 1;
  if (parse_vector(&LocalN1)) return 1;
  if (parse_comma()) return 1;
  if (parse_vector(&LocalP2)) return 1;
  if (parse_comma()) return 1;
  if (parse_vector(&LocalN2)) return 1;
  if (parse_comma()) return 1;
  if (parse_vector(&LocalP3)) return 1;
  if (parse_comma()) return 1;
  if (parse_vector(&LocalN3)) return 1;
  *pTriangle = giram_triangle_new(LocalP1.Val, LocalP2.Val, LocalP3.Val);
  TTriangle = (TriangleStruct *)(*pTriangle);
  TTriangle->N1[0] = LocalN1.Val[0];
  TTriangle->N1[1] = LocalN1.Val[1];
  TTriangle->N1[2] = LocalN1.Val[2];
  TTriangle->N2[0] = LocalN2.Val[0];
  TTriangle->N2[1] = LocalN2.Val[1];
  TTriangle->N2[2] = LocalN2.Val[2];
  TTriangle->N3[0] = LocalN3.Val[0];
  TTriangle->N3[1] = LocalN3.Val[1];
  TTriangle->N3[2] = LocalN3.Val[2];
  if (parse_object_modifier(*pTriangle))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_sor
*  Parsing of sor object
******************************************************************************/
static int parse_sor(ObjectStruct **pSor)
{
  SorStruct *SSor;
  double LocalFloat;
  int t;
  guint i;
  ParsedVector LocalVector;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  *pSor = giram_sor_new(0, NULL);
  SSor = (SorStruct *)(*pSor);
  if (parse_float(&(LocalFloat)))
    return 1;
  SSor->NumberOfPoints = (int)LocalFloat;
  if (SSor->NumberOfPoints < 4)
  {
    g_message(_("%s, line %d\nIllegal number of point in a SOR object...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  SSor->Point = g_malloc(SSor->NumberOfPoints*sizeof(Vector));
  for (i=0 ; i<SSor->NumberOfPoints ; i++)
  {
    if (parse_comma())
      return 1;
    if (parse_vector(&LocalVector))
      return 1;
    SSor->Point[i][0] = LocalVector.Val[0];
    SSor->Point[i][1] = LocalVector.Val[1];
  }
  if (parse_object_modifier(*pSor))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_sphere
*  Parsing of sphere object
******************************************************************************/
static int parse_sphere(ObjectStruct **pSphere)
{
  double Radius;
  ParsedVector LocalVector;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_vector(&LocalVector)) return 1;
  if (parse_comma()) return 1;
  if (parse_float(&Radius)) return 1;
  *pSphere = giram_sphere_new(LocalVector.Val, Radius);
  if (parse_object_modifier(*pSphere))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_super_ellipsoid
*  Parsing of superellipsoid object
******************************************************************************/
static int parse_super_ellipsoid(ObjectStruct **pSuperEllipsoid)
{
  int t;
  double N, E;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (next_token(&t))
    return 1;
  if (t != TLeftAngleBracket)
  {
    g_message(_("%s, line %d\nA '<' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  if (parse_float(&E))
    return 1;
  if (parse_comma())
    return 1;
  if (parse_float(&N))
    return 1;
  if (next_token(&t))
    return 1;
  if (t != TRightAngleBracket)
  {
    g_message(_("%s, line %d\nA '>' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  *pSuperEllipsoid = giram_superellipsoid_new(N, E);
  if (parse_object_modifier(*pSuperEllipsoid))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_torus
*  Parsing of torus object
******************************************************************************/
static int parse_torus(ObjectStruct **pTorus)
{
  double Major, Minor;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_float(&Major)) return 1;
  if (parse_comma()) return 1;
  if (parse_float(&Minor)) return 1;
  *pTorus = giram_torus_new(Major, Minor);
  if (parse_object_modifier(*pTorus))
    return 1;
  return 0;
}

/*****************************************************************************
*  parse_triangle
*  Parsing of triangle object
******************************************************************************/
static int parse_triangle(ObjectStruct **pTriangle)
{
  TriangleStruct *TTriangle;
  ParsedVector LocalP1, LocalP2, LocalP3;
  Vector V1, V2;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (parse_vector(&LocalP1)) return 1;
  if (parse_comma()) return 1;
  if (parse_vector(&LocalP2)) return 1;
  if (parse_comma()) return 1;
  if (parse_vector(&LocalP3)) return 1;
  *pTriangle = giram_triangle_new(LocalP1.Val, LocalP2.Val, LocalP3.Val);
  if (parse_object_modifier(*pTriangle)) return 1;
  return 0;
}

/*****************************************************************************
*  add_object_to_csg
******************************************************************************/
static void add_object_to_csg(ObjectStruct *CSG, ObjectStruct *LocalObject)
{
  CSGStruct *CCSG = (CSGStruct *)CSG;

  CCSG->all_objects = g_slist_append(CCSG->all_objects, LocalObject);
}

/*****************************************************************************
*  parse_csg
*  Parsing of CSG object
******************************************************************************/
static int parse_csg(int Type, ObjectStruct **pCSG)
{
  CSGStruct *CCSG;
  ObjectStruct *TmpObject;
  int t;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  *pCSG = giram_csg_new(Type);
  if (next_token(&t))
    return 1;
  while (t != TRightCurlyBracket)
  { 
    switch(t)
    {
      case Tbox:
        if (parse_box(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tcone:
        if (parse_cone(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tcylinder:
        if (parse_cylinder(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tdifference:
        if (parse_csg(CSG_DIFFERENCE, &TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tdisc:
        if (parse_disc(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tintersection:
        if (parse_csg(CSG_INTERSECTION, &TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tmerge:
        if (parse_csg(CSG_MERGE, &TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tmesh:
        if (parse_mesh(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tobject:
        OldToken = t; IsOldToken = 1;
        TmpObject = NULL;
        if (parse_object(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tplane:
        if (parse_plane(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tsmooth_triangle:
        if (parse_smooth_triangle(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tsor:
        if (parse_sor(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tsphere:
        if (parse_sphere(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tsuperellipsoid:
        if (parse_super_ellipsoid(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Ttorus:
        if (parse_torus(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Ttriangle:
        if (parse_triangle(&TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      case Tunion:
        if (parse_csg(CSG_UNION, &TmpObject))
          return 1;
        add_object_to_csg(*pCSG, TmpObject);
        break;
      default:
        OldToken = t; IsOldToken = 1;
        if (parse_object_modifier(*pCSG))
          return 1;
        IsOldToken = 1;
        OldToken = TRightCurlyBracket;
    }
    if (next_token(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_object
******************************************************************************/
static int parse_object(ObjectStruct **pLocalObject)
{
  int t;

  if (next_token(&t))
    return 1;
  switch (t)
  {
    case TObjectID:
      *pLocalObject = CopyObject(LastObjectConstant->Object);
      break;
    case Tobject:
      if (next_token(&t))
        return 1;
      if (t != TLeftCurlyBracket)
      {
        g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
        return 1;
      }
      if (parse_object(pLocalObject))
        return 1;
      if (parse_object_modifier(*pLocalObject))
        return 1;
      break;
    case Tbox:
      if (parse_box(pLocalObject))
        return 1;
      break;
    case Tcone:
      if (parse_cone(pLocalObject))
        return 1;
      break;
    case Tcylinder:
      if (parse_cylinder(pLocalObject))
        return 1;
      break;
    case Tdifference:
      if (parse_csg(CSG_DIFFERENCE, pLocalObject))
        return 1;
      break;
    case Tdisc:
      if (parse_disc(pLocalObject))
        return 1;
      break;
    case Tintersection:
      if (parse_csg(CSG_INTERSECTION, pLocalObject))
        return 1;
      break;
    case Tmerge:
      if (parse_csg(CSG_MERGE, pLocalObject))
        return 1;
      break;
    case Tmesh:
      if (parse_mesh(pLocalObject))
        return 1;
      break;
    case Tplane:
      if (parse_plane(pLocalObject))
        return 1;
      break;
    case Tsmooth_triangle:
      if (parse_smooth_triangle(pLocalObject))
        return 1;
      break;
    case Tsor:
      if (parse_sor(pLocalObject))
        return 1;
      break;
    case Tsphere:
      if (parse_sphere(pLocalObject))
        return 1;
      break;
    case Tsuperellipsoid:
      if (parse_super_ellipsoid(pLocalObject))
        return 1;
      break;
    case Ttorus:
      if (parse_torus(pLocalObject))
        return 1;
      break;
    case Ttriangle:
      if (parse_triangle(pLocalObject))
        return 1;
      break;
    case Tunion:
      if (parse_csg(CSG_UNION, pLocalObject))
        return 1;
      break;
    default:
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_light_source
******************************************************************************/
static int parse_light_source(LightSourceStruct **pLocalLightSource)
{
  int t;
  ParsedVector LocalVector;
  
  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line);
    return 1;
  }
  /* Creation & initialization */
  if (next_token(&t))
    return 1;
  if (t == TLightSourceID)
  {
    *pLocalLightSource = CopyLightSource(LastLightSourceConstant->LightSource);
    return 0; /* XXX: Is it the 'Right Thing To Do' ? */
  } else
  {
    (*pLocalLightSource) = g_malloc(sizeof(LightSourceStruct));
    (*pLocalLightSource)->Next = NULL;
    (*pLocalLightSource)->Location[0] = 0.0;
    (*pLocalLightSource)->Location[1] = 0.0;
    (*pLocalLightSource)->Location[2] = 0.0;
    (*pLocalLightSource)->Color[0] = 1.0;
    (*pLocalLightSource)->Color[1] = 1.0;
    (*pLocalLightSource)->Color[2] = 1.0;
    (*pLocalLightSource)->Color[3] = 0.0;
    (*pLocalLightSource)->Color[4] = 0.0;
    (*pLocalLightSource)->Type = POINT_LIGHTSOURCE;
    (*pLocalLightSource)->PointAt[0] = 0.0;
    (*pLocalLightSource)->PointAt[1] = 0.0;
    (*pLocalLightSource)->PointAt[2] = 1.0;
    (*pLocalLightSource)->Radius = 0.35;
    (*pLocalLightSource)->FallOff = 0.35;
    (*pLocalLightSource)->Tightness = 10.0;
    (*pLocalLightSource)->AreaLight = FALSE;
    (*pLocalLightSource)->Area1[0] = 0.0;
    (*pLocalLightSource)->Area1[1] = 0.0;
    (*pLocalLightSource)->Area1[2] = 1.0;
    (*pLocalLightSource)->Area2[0] = 0.0;
    (*pLocalLightSource)->Area2[1] = 1.0;
    (*pLocalLightSource)->Area2[2] = 0.0;
    (*pLocalLightSource)->AreaSize1 = 0.0;
    (*pLocalLightSource)->AreaSize2 = 0.0;
    (*pLocalLightSource)->Adaptive = 100.0;
    (*pLocalLightSource)->Jitter = FALSE;
    (*pLocalLightSource)->LooksLike = NULL;
    (*pLocalLightSource)->FadeDistance = 0.0;
    (*pLocalLightSource)->FadePower = 0.0;
    (*pLocalLightSource)->AtmosphericAttenuation = FALSE;
    (*pLocalLightSource)->Atmosphere = FALSE;
    OldToken = t; IsOldToken = 1;
  }
  /* Parsing the mandatory parameters */
  if (parse_vector(&LocalVector))
    return 1;
  (*pLocalLightSource)->Location[0] = LocalVector.Val[0];
  (*pLocalLightSource)->Location[1] = LocalVector.Val[1];
  (*pLocalLightSource)->Location[2] = LocalVector.Val[2];
  if (parse_comma())
    return 1;
  if (parse_color((*pLocalLightSource)->Color))
    return 1;
  /* Parsing the optionnal parameters */
  if (next_token(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch (t)
    {
      case Tspotlight:
        (*pLocalLightSource)->Type = SPOTLIGHT_LIGHTSOURCE;
        break;
      case Tcylinder:
        (*pLocalLightSource)->Type = CYLINDER_LIGHTSOURCE;
        break;
      case Tshadowless:
        (*pLocalLightSource)->Type |= SHADOWLESS_LIGHTSOURCE;
        break;
      case Tpoint_at:
        if (parse_vector(&LocalVector))
          return 1;
        (*pLocalLightSource)->PointAt[0] = LocalVector.Val[0];
        (*pLocalLightSource)->PointAt[1] = LocalVector.Val[1];
        (*pLocalLightSource)->PointAt[2] = LocalVector.Val[2];
        break;
      case Tradius:
        if (parse_float(&((*pLocalLightSource)->Radius)))
          return 1;
        break;
      case Tfalloff:
        if (parse_float(&((*pLocalLightSource)->FallOff)))
          return 1;
        break;
      case Ttightness:
        if (parse_float(&((*pLocalLightSource)->Tightness)))
          return 1;
        break;
      case Tarea_light:
        /* FIXME */
        break;
      case Tadaptive:
        if (parse_float(&((*pLocalLightSource)->Adaptive)))
          return 1;
        break;
      case Tjitter:
        if (parse_float(&((*pLocalLightSource)->Jitter)))
          return 1;
        break;
      case Tlooks_like:
        /* FIXME */
        break;
      case Tfade_distance:
        if (parse_float(&((*pLocalLightSource)->FadeDistance)))
          return 1;
        break;
      case Tfade_power:
        if (parse_float(&((*pLocalLightSource)->FadePower)))
          return 1;
        break;
      case Tatmospheric_attenuation:
        /* FIXME */
        break;
      case Tatmosphere:
        /* FIXME */
        break;
      case Ttranslate:
        if (parse_vector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Trotate:
        if (parse_vector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Tscale:
        if (parse_vector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Ttransform:
        /* FIXME */
        break;
      case Tmatrix:
        /* FIXME */
        break;
      default:
        g_message(_("%s, line %d\nA Light Source modifier was expected here...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
    }
    if (next_token(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  parse_camera
******************************************************************************/
static int parse_camera(CameraStruct **pCamera)
{
  int t;
  double LocalFloat;
  ParsedVector LocalVector;
  Vector Vect;
  double k1, k2;
  double Norme;
  double DirectionLength, UpLength, RightLength;
  double Handedness;

  if (next_token(&t))
    return 1;
  if (t != TLeftCurlyBracket)
  {
    g_message(_("%s, line %d\nA '{' was expected here...\n"), PovFile->Name, PovFile->Line); 
    return 1;
  }
  if (next_token(&t))
    return 1;
  if (t == TCameraID)
  {
    *pCamera = CopyCamera(LastCameraConstant->Camera);
  } else
  {
    if (*pCamera == NULL)
    {
      (*pCamera) = g_malloc(sizeof(CameraStruct));
      (*pCamera)->Type = CAMERA_PERSPECTIVE;
      (*pCamera)->CylinderType = 1;
      (*pCamera)->Location[0] = 0.0;
      (*pCamera)->Location[1] = 0.0;
      (*pCamera)->Location[2] = 0.0;
      (*pCamera)->LookAt[0] = 0.0;
      (*pCamera)->LookAt[1] = 0.0;
      (*pCamera)->LookAt[2] = 1.0;
      (*pCamera)->Right[0] = 1.33;
      (*pCamera)->Right[1] = 0.0;
      (*pCamera)->Right[2] = 0.0;
      (*pCamera)->Up[0] = 0.0;
      (*pCamera)->Up[1] = 1.0;
      (*pCamera)->Up[2] = 0.0;
      (*pCamera)->Direction[0] = 0.0;
      (*pCamera)->Direction[1] = 0.0;
      (*pCamera)->Direction[2] = 1.0;
      (*pCamera)->Sky[0] = 0.0;
      (*pCamera)->Sky[1] = 1.0;
      (*pCamera)->Sky[2] = 0.0;
      (*pCamera)->Angle = 90.0;
      (*pCamera)->BlurSamples = 0;
      (*pCamera)->Aperture = 0.0;
      (*pCamera)->FocalPoint[0] = 0.0;
      (*pCamera)->FocalPoint[1] = 0.0;
      (*pCamera)->FocalPoint[2] = 1.0;
      (*pCamera)->Confidence = 0.9;
      (*pCamera)->Variance = 1.0/128.0;
    }
    OldToken = t; IsOldToken = 1;
  }
  if (next_token(&t))
    return 1;
  while (t != TRightCurlyBracket)
  {
    switch(t)
    {
      case Tperspective:
        (*pCamera)->Type = CAMERA_PERSPECTIVE;
        break;
      case Torthographic:
        (*pCamera)->Type = CAMERA_ORTHOGRAPHIC;
        Vect[0] = (*pCamera)->LookAt[0] - (*pCamera)->Location[0];
        Vect[1] = (*pCamera)->LookAt[1] - (*pCamera)->Location[1];
        Vect[2] = (*pCamera)->LookAt[2] - (*pCamera)->Location[2];
        k1 = VLength(Vect);
        k2 = VLength((*pCamera)->Direction);
        if ((k1 > 10e-6) && (k2 > 10e-6))
        {
          (*pCamera)->Right[0] *= k1/k2;
          (*pCamera)->Right[1] *= k1/k2;
          (*pCamera)->Right[2] *= k1/k2;
          (*pCamera)->Up[0] *= k1/k2;
          (*pCamera)->Up[1] *= k1/k2;
          (*pCamera)->Up[2] *= k1/k2;
        }
        break;
      case Tfisheye:
        (*pCamera)->Type = CAMERA_FISHEYE;
        break;
      case Tultra_wide_angle:
        (*pCamera)->Type = CAMERA_ULTRA_WIDE_ANGLE;
        break;
      case Tomnimax:
        (*pCamera)->Type = CAMERA_OMNIMAX;
        break;
      case Tpanoramic:
        (*pCamera)->Type = CAMERA_PANORAMIC;
        break;
      case Tcylinder:
        (*pCamera)->Type = CAMERA_CYLINDER;
        if (parse_float(&LocalFloat))
          return 1;
        (*pCamera)->CylinderType = (int)LocalFloat;
        if (((*pCamera)->CylinderType < 1) ||
            ((*pCamera)->CylinderType > 4))
        {
          g_message(_("%s, line %d\nThere's only 4 cylinder type of camera...\n"),
                    PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Tlocation:
        if (parse_vector(&LocalVector))
          return 1;
        (*pCamera)->Location[0] = LocalVector.Val[0];
        (*pCamera)->Location[1] = LocalVector.Val[1];
        (*pCamera)->Location[2] = LocalVector.Val[2];
        break;
      case Tlook_at:
        DirectionLength = VLength((*pCamera)->Direction);
        UpLength = VLength((*pCamera)->Up);
        RightLength = VLength((*pCamera)->Right);
        VCross(Vect, (*pCamera)->Up, (*pCamera)->Direction);
        Handedness = VDot(Vect, (*pCamera)->Right);
        if (parse_vector(&LocalVector))
          return 1;
        (*pCamera)->LookAt[0] = (*pCamera)->Direction[0] = LocalVector.Val[0];
        (*pCamera)->LookAt[1] = (*pCamera)->Direction[1] = LocalVector.Val[1];
        (*pCamera)->LookAt[2] = (*pCamera)->Direction[2] = LocalVector.Val[2];
        (*pCamera)->Direction[0] -= (*pCamera)->Location[0];
        (*pCamera)->Direction[1] -= (*pCamera)->Location[1];
        (*pCamera)->Direction[2] -= (*pCamera)->Location[2];
        /* Check for zero length direction vector. */
        if (VLength((*pCamera)->Direction) < 10e-6)
        {
          g_message(_("%s, line %d\nCamera location and look_at point must be different.\n"),
                    PovFile->Name, PovFile->Line);
          return 1;
        }
        Norme = VLength((*pCamera)->Direction);
        (*pCamera)->Direction[0] /= Norme;
        (*pCamera)->Direction[1] /= Norme;
        (*pCamera)->Direction[2] /= Norme;
        VCross((*pCamera)->Right, (*pCamera)->Sky, (*pCamera)->Direction);
        Norme = VLength((*pCamera)->Right);
        (*pCamera)->Right[0] /= Norme;
        (*pCamera)->Right[1] /= Norme;
        (*pCamera)->Right[2] /= Norme;
        VCross((*pCamera)->Up, (*pCamera)->Direction, (*pCamera)->Right);
        (*pCamera)->Direction[0] *= DirectionLength;
        (*pCamera)->Direction[1] *= DirectionLength;
        (*pCamera)->Direction[2] *= DirectionLength;
        if (Handedness > 0.0)
        {
          (*pCamera)->Right[0] *= RightLength;
          (*pCamera)->Right[1] *= RightLength;
          (*pCamera)->Right[2] *= RightLength;
        }
        else
        {
          (*pCamera)->Right[0] *= -RightLength;
          (*pCamera)->Right[1] *= -RightLength;
          (*pCamera)->Right[2] *= -RightLength;
        }
        (*pCamera)->Up[0] *= UpLength;
        (*pCamera)->Up[1] *= UpLength;
        (*pCamera)->Up[2] *= UpLength;
        break;
      case Tright:
        if (parse_vector(&LocalVector))
          return 1;
        (*pCamera)->Right[0] = LocalVector.Val[0];
        (*pCamera)->Right[1] = LocalVector.Val[1];
        (*pCamera)->Right[2] = LocalVector.Val[2];
        break;
      case Tup:
        if (parse_vector(&LocalVector))
          return 1;
        (*pCamera)->Up[0] = LocalVector.Val[0];
        (*pCamera)->Up[1] = LocalVector.Val[1];
        (*pCamera)->Up[2] = LocalVector.Val[2];
        break;
      case Tdirection:
        if (parse_vector(&LocalVector))
          return 1;
        (*pCamera)->Direction[0] = LocalVector.Val[0];
        (*pCamera)->Direction[1] = LocalVector.Val[1];
        (*pCamera)->Direction[2] = LocalVector.Val[2];
        break;
      case Tsky:
        if (parse_vector(&LocalVector))
          return 1;
        (*pCamera)->Sky[0] = LocalVector.Val[0];
        (*pCamera)->Sky[1] = LocalVector.Val[1];
        (*pCamera)->Sky[2] = LocalVector.Val[2];
        break;
      case Tangle:
        if (parse_float(&((*pCamera)->Angle)))
          return 1;
        if ((*pCamera)->Angle < 0.0)
        {
          g_message(_("%s, line %d\nNegative viewing angle."), 
                    PovFile->Name, PovFile->Line);
          return 1;
        }
        if ((*pCamera)->Type == CAMERA_PERSPECTIVE)
        {
          if ((*pCamera)->Angle >= 180.0)
          {
            g_message(_("%s, line %d\nViewing angle has to be smaller than 180 degrees."),
                      PovFile->Name, PovFile->Line);
          }
          Norme = VLength((*pCamera)->Direction);
          (*pCamera)->Direction[0] /= Norme;
          (*pCamera)->Direction[1] /= Norme;
          (*pCamera)->Direction[2] /= Norme;
          Norme = VLength((*pCamera)->Right) / tan((*pCamera)->Angle/2.0 * M_PI/180.0)/2.0;
          (*pCamera)->Direction[0] *= Norme;
          (*pCamera)->Direction[1] *= Norme;
          (*pCamera)->Direction[2] *= Norme;
        }
        break;
      case Tblur_samples:
        if (parse_float(&LocalFloat))
          return 1;
        (*pCamera)->BlurSamples = (int)LocalFloat;
        if ((*pCamera)->BlurSamples <=0 )
        {
          g_message(_("%s, line %d\nThe number of focal blur samples must be strictly positive...\n"),
                    PovFile->Name, PovFile->Line);
          return 1;
        }
        break;
      case Taperture:
        if (parse_float(&((*pCamera)->Aperture)))
          return 1;
        break;
      case Tfocal_point:
        if (parse_vector(&LocalVector))
          return 1;
        (*pCamera)->FocalPoint[0] = LocalVector.Val[0];
        (*pCamera)->FocalPoint[1] = LocalVector.Val[1];
        (*pCamera)->FocalPoint[2] = LocalVector.Val[2];
        break;
      case Tconfidence:
        if (parse_float(&LocalFloat))
          return 1;
        if ((LocalFloat > 0.0) && (LocalFloat < 1.0))
          (*pCamera)->Confidence = LocalFloat;
        else
          g_message(_("%s, line %d\nIllegal value of confidence last one will be use...\n"),
                    PovFile->Name, PovFile->Line);
        break;
      case Tvariance:
        if (parse_float(&LocalFloat))
          return 1;
        if ((LocalFloat >= 0.0) && (LocalFloat <= 1.0))
          (*pCamera)->Variance = LocalFloat;
        else
          g_message(_("%s, line %d\nIllegal value of variance last one will be use...\n"),
                    PovFile->Name, PovFile->Line);
        break;
      case Tscale:
        if (parse_vector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Trotate:
        if (parse_vector(&LocalVector))
          return 1;
        /* FIXME */
        break;
      case Ttranslate:
        parse_vector(&LocalVector);
        (*pCamera)->Location[0] += LocalVector.Val[0];
        (*pCamera)->Location[1] += LocalVector.Val[1];
        (*pCamera)->Location[2] += LocalVector.Val[2];
        (*pCamera)->LookAt[0] += LocalVector.Val[0];
        (*pCamera)->LookAt[1] += LocalVector.Val[1];
        (*pCamera)->LookAt[2] += LocalVector.Val[2];
        break;
      case Tmatrix:
        /* FIXME */
        break;
      case Ttransform:
        /* FIXME */
        break;
      default:
        g_message(_("%s, line %d\nA Camera modifier was expected here...\n"),
                  PovFile->Name, PovFile->Line);
        return 1;
    }
    if (next_token(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  init_load_pov
******************************************************************************/
static void init_load_pov(char *Name)
{
  IsOldChar = 0;
  IsOldToken = 0;
  PovVersionLangage = 3.0;
  LastString = NULL;
  PovClock = 0.0;
  NbPovRandomTable = 0;
  DontParse = 0;
  ParsingDirective = 0;
  PovRandomTable = NULL;
  CurrentFrame = NewFrame(Name);
  ConditionalStack = NULL;
}

/*****************************************************************************
*  parse_pov
******************************************************************************/
static int parse_pov(void)
{
  LightSourceStruct *TmpLightSource;
  ObjectStruct *TmpObject;
  int t;

  if (next_token(&t))
    return 1;
  while (t != TEOF)
  {
    switch (t)
    {
      case Tcamera:
/*        if (CurrentFrame->Camera)
        {
          WarningBox(_("Error"), _("%s:%d : Only one camera is allowed per frame...\n"),
                     PovFile->Name, PovFile->Line);
          return 1;
        }*/
        if (parse_camera(&(CurrentFrame->Camera)))
          return 1;
        break;
      case Tlight_source:
        if (parse_light_source(&TmpLightSource))
          return 1;
        TmpLightSource->Next = CurrentFrame->FirstLightSource;
        CurrentFrame->FirstLightSource = TmpLightSource;
        break;
      case Tbox:
        if (parse_box(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tcone:
        if (parse_cone(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tcylinder:
        if (parse_cylinder(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tdifference:
        if (parse_csg(CSG_DIFFERENCE, &TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tdisc:
        if (parse_disc(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tintersection:
        if (parse_csg(CSG_INTERSECTION, &TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tmerge:
        if (parse_csg(CSG_MERGE, &TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tmesh:
        if (parse_mesh(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tplane:
        if (parse_plane(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tsmooth_triangle:
        if (parse_smooth_triangle(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tsor:
        if (parse_sor(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tsphere:
        if (parse_sphere(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tsuperellipsoid:
        if (parse_super_ellipsoid(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Ttorus:
        if (parse_torus(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Ttriangle:
        if (parse_triangle(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tunion:
        if (parse_csg(CSG_UNION, &TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
      case Tobject:
        OldToken = t; IsOldToken = 1;
        if (parse_object(&TmpObject))
          return 1;
        CurrentFrame->all_objects = g_slist_append(CurrentFrame->all_objects,
                                                   TmpObject);
        break;
    }
    if (next_token(&t))
      return 1;
  }
  return 0;
}

/*****************************************************************************
*  LoadPov
******************************************************************************/
void LoadPov(GtkWidget *w, GtkFileSelection *FS)
{
  ObjectStruct *TmpObject;
  GtkWidget *menu_item;
  char *Name;

  Name = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS)));
  
  init_load_pov(Name);
  PovFile = g_malloc(sizeof(ListFile));
  PovFile->Prev = PovFile->Next = NULL;
  PovFile->Line = 1;
  PovFile->handle = LibOpen(Name);
  PovFile->Name = g_malloc(strlen(Name)+1);
  strcpy(PovFile->Name, Name);

  if (parse_pov())
  { /* There's an error in the parsing -> we must free the memory. */
    /* FIXME: DestroyFrame(CurrentFrame); */
    return;
  }

  g_slist_foreach(CurrentFrame->all_objects,
                  (GFunc)giram_object_build_triangle_mesh,
                  NULL);
  ComputeCamera(CurrentFrame->Camera);
  if (CSGTreeDialog)
  {
    menu_item = gtk_radio_menu_item_new_with_label(CSGTreeMenuGroup,
                                                   CurrentFrame->Name);
    gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
                       (GtkSignalFunc)RebuildCSGTree, CurrentFrame);
    if (!CSGTreeMenuGroup)
      CSGTreeMenuGroup = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(menu_item));
    gtk_menu_append(GTK_MENU(CSGTreeMenu), menu_item);
    gtk_object_set_data(GTK_OBJECT(menu_item), "CSGMenu", CurrentFrame);
    gtk_widget_show(menu_item);
  }
  if (show_xy_view)
    NewView(CurrentFrame, ORTHO_XY_CAMERA);
  if (show_xz_view)
    NewView(CurrentFrame, ORTHO_XZ_CAMERA);
  if (show_zy_view)
    NewView(CurrentFrame, ORTHO_ZY_CAMERA);
  if (show_camera_view)
    NewViewCamera(CurrentFrame);
}

