/* LoadPov.h 
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>

/* The Tokens */
#define TEOF                       1
#define TPlus                      2
#define TMinus                     3
#define TStar                      4
#define TSlash                     5
#define Tabs                       6
#define Tacos                      7
#define Tval                       8
#define Tasc                       9
#define Tasin                     10
#define Tatan2                    11
#define Tceil                     12
#define Tclock                    13
#define Tcos                      14
#define Tdegrees                  15
#define Tdiv                      16
#define Texp                      17
#define Tfile_exists              18
#define TFloatID                  19
#define TFloatLiteral             20
#define Tfloor                    21
#define Tint                      22
#define Tlog                      23
#define Tmin                      24
#define Tmax                      25
#define Tmod                      26
#define Tpi                       27
#define Tpow                      28
#define Tradians                  29
#define Tsin                      30
#define Tsqrt                     31
#define Tstrcmp                   32
#define Tstrlen                   33
#define Ttan                      34
#define Tvdot                     35
#define Tvlength                  36
#define Tversion                  37
#define Ttrue                     38
#define Tyes                      39
#define Ton                       40
#define Tfalse                    41
#define Tno                       42
#define Toff                      43
#define Tseed                     44
#define Trand                     45
#define Tvaxis_rotate             46
#define Tvcross                   47
#define TVectorID                 48
#define Tvnormalize               49
#define Tvrotate                  50
#define Tx                        51
#define Ty                        52
#define Tz                        53
#define TColorID                  54
#define Tt                        55
#define Tu                        56
#define Tv                        57
#define TExclamation              58
#define TLeftParen                59
#define TComma                    60
#define TRightParen               61
#define TLeftAngleBracket         62
#define TRightAngleBracket        63
#define Tred                      64
#define Tgreen                    65
#define Tblue                     66
#define Tfilter                   67
#define Ttransmit                 68
#define TColon                    69
#define TQuestion                 70
#define Tstr                      71
#define Tconcat                   72
#define Tchr                      73
#define Tsubstr                   74
#define Tstrupr                   75
#define Tstrlwr                   76
#define TStringLiteral            77
#define TStringID                 78
#define TAmpersand                79
#define TBar                      80
#define TEqual                    81
#define TUnequal                  82
#define TLowerOrEqual             83
#define TUpperOrEqual             84
#define THash                     85
#define Tglobal_settings          86
#define Tassumed_gamma            87
#define Thf_gray_16               88
#define Tambient_light            89
#define Tradiosity                90
#define Tnumber_of_waves          91
#define Tadc_bailout              92
#define Tirid_wavelength          93
#define Tmax_intersections        94
#define Tmax_trace_level          95
#define Tbrightness               96
#define Tcount                    97
#define Tdistance_maximum         98
#define Terror_bound              99
#define Tgray_threshold          100
#define Tlow_error_factor        101
#define Tminimum_reuse           102
#define Tnearest_count           103
#define Trecursion_limit         104
#define TLeftCurlyBracket        105
#define TRightCurlyBracket       106
#define Tcolor                   107
#define Talpha                   108
#define Trgb                     109
#define Trgbt                    110
#define Trgbf                    111
#define Trgbft                   112
#define TObjectID                113
#define TTextureID               114
#define TPigmentID               115
#define TNormalID                116
#define TFinishID                117
#define THaloID                  118
#define TColorMapID              119
#define TPigmentMapID            120
#define TSlopeMapID              121
#define TNormalMapID             122
#define TCameraID                123
#define TLightSourceID           124
#define TAtmosphereID            125
#define TFogID                   126
#define TRainbowID               127
#define TSkySphereID             128
#define TTransformID             129
#define Tcamera                  130
#define Tperspective             131
#define Torthographic            132
#define Tfisheye                 133
#define Tultra_wide_angle        134
#define Tomnimax                 135
#define Tpanoramic               136
#define Tcylinder                137
#define Tlocation                138
#define Tlook_at                 139
#define Tright                   140
#define Tup                      141
#define Tdirection               142
#define Tsky                     143
#define Tangle                   144
#define Tblur_samples            145
#define Taperture                146
#define Tfocal_point             147
#define Tnormal                  148
#define Tconfidence              150
#define Tvariance                151
#define Tscale                   152
#define Ttransform               153
#define Trotate                  154
#define Tmatrix                  155
#define Ttranslate               156
#define Tlight_source            157
#define Tspotlight               158
#define Tshadowless              159
#define Tpoint_at                160
#define Tradius                  161
#define Tfalloff                 162
#define Ttightness               163
#define Tarea_light              164
#define Tadaptive                165
#define Tjitter                  166
#define Tlooks_like              167
#define Tfade_distance           168
#define Tfade_power              169
#define Tatmospheric_attenuation 170
#define Tatmosphere              171
#define Tsky_sphere              172
#define Tpigment                 173
#define Tplane                   174
#define Tsphere                  175
#define Tbox                     176
#define Tbounded_by              177
#define Tclipped_by              178
#define Ttexture                 179
#define Tfinish                  180
#define Thalo                    181
#define Tinverse                 182
#define Tsturm                   183
#define Twater_level             184
#define Tsmooth                  185
#define Tno_shadow               186
#define Thierarchy               187
#define Thollow                  188
#define Tambient                 189
#define Tdiffuse                 190
#define Tbrilliance              191
#define Tphong                   192
#define Tphong_size              193
#define Tspecular                194
#define Troughness               195
#define Tmetallic                196
#define Treflection              197
#define Trefraction              198
#define Tior                     199
#define Tcaustics                200
#define Tirid                    201
#define Tcrand                   202
#define Tdifference              203
#define Tintersection            204
#define Tmerge                   205
#define Tunion                   206
#define Topen                    207
#define Ttorus                   208
#define Tbrick                   209
#define Tchecker                 210
#define Thexagon                 211
#define Tcolor_map               212
#define Tpigment_map             213
#define Timage_map               214
#define Tquick_color             215
#define Tturbulence              216
#define Tbrick_size              217
#define Tmortar                  218
#define Tagate                   219
#define Tagate_turb              220
#define Taverage                 221
#define Tbozo                    222
#define Tbumps                   223
#define Tcrackle                 224
#define Tdents                   225
#define Tgradient                226
#define Tgranite                 227
#define Tleopard                 228
#define Tmandel                  229
#define Tmarble                  230
#define Tonion                   231
#define Tquilted                 232
#define Tcontrol0                233
#define Tcontrol1                234
#define Tradial                  235
#define Tripples                 236
#define Tspiral1                 237
#define Tspiral2                 238
#define Tspotted                 239
#define Twaves                   240
#define Tfrequency               241
#define Tphase                   242
#define Twood                    243
#define Twrinkles                244
#define Tramp_wave               245
#define Tsine_wave               246
#define Tscallop_wave            247
#define Toctaves                 248
#define Tlambda                  249
#define Tomega                   250
#define Twarp                    251
#define TLeftSquareBracket       252
#define TRightSquareBracket      253
#define Tobject                  254
#define Tslope_map               255
#define Tnormal_map              256
#define Tfog                     257
#define Trainbow                 258
#define Ttype                    259
#define Tdistance                260
#define Tscattering              261
#define Teccentricity            262
#define Tsamples                 263
#define Taa_threshold            264
#define Taa_level                265
#define Tfog_type                266
#define Tturb_depth              267
#define Tfog_offset              268
#define Tfog_alt                 269
#define Twidth                   270
#define Tarc_angle               271
#define Tfalloff_angle           272
#define Tcone                    273
#define Tquadric                 274
#define Tbump_map                275
#define Ttriangle_wave           276
#define Tbump_size               277
#define Tthickness               278
#define Tattenuating             279
#define Temitting                280
#define Tglowing                 281
#define Tdust                    282
#define Tconstant                283
#define Tlinear                  284
#define Tcubic                   285
#define Tpoly                    286
#define Tplanar_mapping          287
#define Tspherical_mapping       288
#define Tcylindrical_mapping     289
#define Tbox_mapping             290
#define Tdust_type               291
#define Tmax_value               292
#define Texponent                293
#define Theight_field            294
#define Ttriangle                295
#define Tsmooth_triangle         296
#define Tsuperellipsoid          297
#define Tbicubic_patch           298
#define Tblob                    299
#define Tdisc                    300
#define Tjulia_fractal           301
#define Tlathe                   302
#define Tmesh                    303
#define Tpolygon                 304
#define Tprism                   305
#define Tquartic                 306
#define Tsor                     307
#define Ttext                    308
#define Tttf                     309
#define Tsemicolon               310

typedef struct ListFile
{
  int Line;
  FILE *handle;
  char *Name;
  struct ListFile *Prev;
  struct ListFile *Next;
} ListFile;

void LoadPov(GtkWidget *w, GtkFileSelection *fs);

gboolean show_xy_view, show_xz_view, show_zy_view, show_camera_view;
gchar *pov_include_path;
