/* SavePov.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include "giram.h"

#include "giramintl.h"

/*****************************************************************************
*  SavePovTrans
******************************************************************************/
static void SavePovTrans(FILE *OutFile, TransformStruct *Trans, guint indentation)
{
  guint i;
  char *s;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i]=' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%smatrix  <%g, %g, %g,\n", s,
          Trans->Direct[0][0], Trans->Direct[0][1], Trans->Direct[0][2]);
  fprintf(OutFile, "%s   %g, %g, %g,\n", s,
          Trans->Direct[1][0], Trans->Direct[1][1], Trans->Direct[1][2]);
  fprintf(OutFile, "%s   %g, %g, %g,\n", s,
          Trans->Direct[2][0], Trans->Direct[2][1], Trans->Direct[2][2]);
  fprintf(OutFile, "%s   %g, %g, %g>\n", s,
          Trans->Direct[3][0], Trans->Direct[3][1], Trans->Direct[3][2]);
}

/*****************************************************************************
*  SavePovPigment
******************************************************************************/
static void SavePovPigment(FILE *OutFile, PigmentStruct *Pigment, guint indentation)
{
  guint i;
  char *s;
  
  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i]=' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%spigment\n%s{\n", s, s);
  switch (Pigment->Type)
  {
    case PAT_SOLID_COLOR:
      fprintf(OutFile, "%s  color rgbft <%g, %g, %g,  %g, %g>\n", s,
              Pigment->Color[0], Pigment->Color[1], Pigment->Color[2],
              Pigment->Color[3], Pigment->Color[4]);
      break;
    case PAT_BRICK:
      /* FIXME */
      break;
    case PAT_CHECKER:
      /* FIXME */
      break;
    case PAT_HEXAGON:
      /* FIXME */
      break;
    case PAT_LAST:
    default:
      break;
  }
  if (Pigment->Trans)
    SavePovTrans(OutFile, Pigment->Trans, indentation+1);
  fprintf(OutFile, "%s}\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovTexture
******************************************************************************/
static void SavePovTexture(FILE *OutFile, TextureStruct *Texture, guint indentation)
{
  guint i;
  char *s;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i]=' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%stexture\n%s{\n", s, s);
  if (Texture->Pigment)
  {
    SavePovPigment(OutFile, Texture->Pigment, indentation+1);
  }
  if (Texture->Trans)
    SavePovTrans(OutFile, Texture->Trans, indentation+1);
  fprintf(OutFile, "%s}\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovBox
******************************************************************************/
static void SavePovBox(FILE *OutFile, ObjectStruct *Box, guint indentation)
{
  guint i;
  char *s;
  BoxStruct *BBox = (BoxStruct *)Box;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%sbox\n%s{\n%s  <%g, %g, %g>, <%g, %g, %g>\n", s, s, s,
                   BBox->MinCorner[0], BBox->MinCorner[1], BBox->MinCorner[2],
                   BBox->MaxCorner[0], BBox->MaxCorner[1], BBox->MaxCorner[2]);
  if (Box->Texture)
  {
    SavePovTexture(OutFile, Box->Texture, indentation+1);
  }
  if (Box->Trans)
    SavePovTrans(OutFile, Box->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovCone
******************************************************************************/
static void SavePovCone(FILE *OutFile, ObjectStruct *Cone, guint indentation)
{
  guint i;
  char *s;
  ConeStruct *CCone = (ConeStruct *)Cone;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%scone\n%s{\n%s  <0, %g, 0>, %g, <0, %g, 0>, %g\n",
          s, s, s, CCone->Radius1, CCone->Radius1,
                   CCone->Radius2, CCone->Radius2);
  if (CCone->Open)
    fprintf(OutFile, "%s  open\n",s);
  if (Cone->Texture)
  {
    SavePovTexture(OutFile, Cone->Texture, indentation+1);
  }
  if (Cone->Trans)
    SavePovTrans(OutFile, Cone->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovCylinder
******************************************************************************/
static void SavePovCylinder(FILE *OutFile, ObjectStruct *Cylinder, guint indentation)
{
  guint i;
  char *s;
  CylinderStruct *CCylinder = (CylinderStruct *)Cylinder;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%scylinder\n%s{\n%s  <%g, %g, %g>, <%g, %g, %g>, %g\n", s, s, s,
                   0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0);
  if (CCylinder->Open)
    fprintf(OutFile, "%s  open\n",s);
  if (Cylinder->Texture)
  {
    SavePovTexture(OutFile, Cylinder->Texture, indentation+1);
  }
  if (Cylinder->Trans)
    SavePovTrans(OutFile, Cylinder->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovDisc
******************************************************************************/
static void SavePovDisc(FILE *OutFile, ObjectStruct *Disc, guint indentation)
{
  guint i;
  char *s;
  DiscStruct *DDisc = (DiscStruct *)Disc;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%sdisc\n%s{\n%s  <0, 0, 0>, <0, 1, 0>, %g", 
          s, s, s, DDisc->Radius);
  if (DDisc->HoleRadius != 0.0)
    fprintf(OutFile, ", %g\n", DDisc->HoleRadius);
  else
    fprintf(OutFile, "\n");
  if (Disc->Texture)
  {
    SavePovTexture(OutFile, Disc->Texture, indentation+1);
  }
  if (Disc->Trans)
    SavePovTrans(OutFile, Disc->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovMesh
******************************************************************************/
static void SavePovMesh(FILE *OutFile, ObjectStruct *Mesh, guint indentation)
{
  guint i;
  char *s;
  MeshStruct *MMesh = (MeshStruct *)Mesh;
  TriangleListStruct *TmpTri;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%smesh\n%s{\n", s, s);
  for (TmpTri = MMesh->FirstTriangle ; TmpTri ; TmpTri=TmpTri->Next)
  {
    fprintf(OutFile, "%s  smooth_triangle { <%g, %g, %g>, <%g, %g, %g>,\n", s,
            TmpTri->P1[0], TmpTri->P1[1], TmpTri->P1[2],
            TmpTri->N1[0], TmpTri->N1[1], TmpTri->N1[2]);
    fprintf(OutFile, "%s                  <%g, %g, %g>, <%g, %g, %g>,\n", s,
            TmpTri->P2[0], TmpTri->P2[1], TmpTri->P2[2],
            TmpTri->N2[0], TmpTri->N2[1], TmpTri->N2[2]);
    fprintf(OutFile, "%s                  <%g, %g, %g>, <%g, %g, %g> }\n", s,
            TmpTri->P3[0], TmpTri->P3[1], TmpTri->P3[2],
            TmpTri->N3[0], TmpTri->N3[1], TmpTri->N3[2]);
    
  }
  if (Mesh->Texture)
  {
    SavePovTexture(OutFile, Mesh->Texture, indentation+1);
  }
  if (Mesh->Trans)
    SavePovTrans(OutFile, Mesh->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovPlane
******************************************************************************/
static void SavePovPlane(FILE *OutFile, ObjectStruct *Plane, guint indentation)
{
  guint i;
  char *s;
  PlaneStruct *PPlane = (PlaneStruct *)Plane;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%splane\n%s{\n%s  <%g, %g, %g>, %g\n", s, s, s,
                   PPlane->Normal[0], PPlane->Normal[1], PPlane->Normal[2],
                   PPlane->Distance);
  if (Plane->Texture)
  {
    SavePovTexture(OutFile, Plane->Texture, indentation+1);
  }
  if (Plane->Trans)
    SavePovTrans(OutFile, Plane->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovSor
******************************************************************************/
static void SavePovSor(FILE *OutFile, ObjectStruct *Sor, guint indentation)
{
  guint i;
  char *s;
  SorStruct *SSor = (SorStruct *)Sor;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%ssor\n%s{\n%s  %d,\n", s, s, s, SSor->NumberOfPoints);
  for (i=0 ; i<SSor->NumberOfPoints-1 ; i++)
    fprintf(OutFile, "%s  <%g, %g>,\n", s, SSor->Point[i][0], SSor->Point[i][1]);
  fprintf(OutFile, "%s  <%g, %g>\n", s, SSor->Point[i][0], SSor->Point[i][1]);
  if (Sor->Texture)
  {
    SavePovTexture(OutFile, Sor->Texture, indentation+1);
  }
  if (Sor->Trans)
    SavePovTrans(OutFile, Sor->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovSphere
******************************************************************************/
static void SavePovSphere(FILE *OutFile, ObjectStruct *Sphere, guint indentation)
{
  guint i;
  char *s;
  SphereStruct *SSphere = (SphereStruct *)Sphere;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%ssphere\n%s{\n%s  <%g, %g, %g>, %g\n", s, s, s,
                   SSphere->Center[0], SSphere->Center[1], SSphere->Center[2],
                   SSphere->Radius);
  if (Sphere->Texture)
  {
    SavePovTexture(OutFile, Sphere->Texture, indentation+1);
  }
  if (Sphere->Trans)
    SavePovTrans(OutFile, Sphere->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovSuperEllipsoid
******************************************************************************/
static void SavePovSuperEllipsoid(FILE *OutFile, ObjectStruct *SuperEllipsoid, guint indentation)
{
  guint i;
  char *s;
  SuperEllipsoidStruct *SSuperEllipsoid = (SuperEllipsoidStruct *)SuperEllipsoid;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%ssuperellipsoid\n%s{\n%s  <%g, %g>\n", s, s, s,
          SSuperEllipsoid->E, SSuperEllipsoid->N);
  if (SuperEllipsoid->Texture)
  {
    SavePovTexture(OutFile, SuperEllipsoid->Texture, indentation+1);
  }
  if (SuperEllipsoid->Trans)
    SavePovTrans(OutFile, SuperEllipsoid->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovTorus
******************************************************************************/
static void SavePovTorus(FILE *OutFile, ObjectStruct *Torus, guint indentation)
{
  guint i;
  char *s;
  TorusStruct *TTorus = (TorusStruct *)Torus;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%storus\n%s{\n%s  %g, %g\n", s, s, s,
                   TTorus->Major, TTorus->Minor);
  if (Torus->Texture)
  {
    SavePovTexture(OutFile, Torus->Texture, indentation+1);
  }
  if (Torus->Trans)
    SavePovTrans(OutFile, Torus->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovTriangle
******************************************************************************/
static void SavePovTriangle(FILE *OutFile, ObjectStruct *Triangle, guint indentation)
{
  guint i;
  char *s;
  TriangleStruct *TTriangle = (TriangleStruct *)Triangle;

  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i] = ' ';
  s[indentation*2] = 0;
  fprintf(OutFile, "%ssmooth_triangle\n%s{\n", s, s);
  fprintf(OutFile, "%s  <%g, %g, %g>, <%g, %g, %g>,\n", s,
          TTriangle->P1[0], TTriangle->P1[1], TTriangle->P1[2],
          TTriangle->N1[0], TTriangle->N1[1], TTriangle->N1[2]);
  fprintf(OutFile, "%s  <%g, %g, %g>, <%g, %g, %g>,\n", s,
          TTriangle->P2[0], TTriangle->P2[1], TTriangle->P2[2],
          TTriangle->N2[0], TTriangle->N2[1], TTriangle->N2[2]);
  fprintf(OutFile, "%s  <%g, %g, %g>, <%g, %g, %g>\n", s,
          TTriangle->P3[0], TTriangle->P3[1], TTriangle->P3[2],
          TTriangle->N3[0], TTriangle->N3[1], TTriangle->N3[2]);
  if (Triangle->Texture)
  {
    SavePovTexture(OutFile, Triangle->Texture, indentation+1);
  }
  if (Triangle->Trans)
    SavePovTrans(OutFile, Triangle->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovCSG
******************************************************************************/
static void SavePovCSG(FILE *OutFile, ObjectStruct *CSG, guint indentation)
{
  char *s;
  guint i;
  ObjectStruct *TmpObject;
  CSGStruct *CCSG = (CSGStruct *)CSG;
  GSList *tmp_list;
  
  s = g_malloc(indentation*2+1);
  for (i=0 ; i<indentation*2 ; i++)
    s[i]=' ';
  s[indentation*2] = 0;
  switch (CCSG->Type)
  {
    case CSG_DIFFERENCE:
      fprintf(OutFile, "%sdifference\n%s{\n", s, s);
      break;
    case CSG_INTERSECTION:
      fprintf(OutFile, "%sintersection\n%s{\n", s, s);
      break;
    case CSG_MERGE:
      fprintf(OutFile, "%smerge\n%s{\n", s, s);
      break;
    case CSG_UNION:
      fprintf(OutFile, "%sunion\n%s{\n", s, s);
      break;
  }
  for (tmp_list = CCSG->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    TmpObject = (ObjectStruct*)(tmp_list->data);
    switch (TmpObject->Type)
    {
      case BOX_OBJECT:
        SavePovBox(OutFile, TmpObject, indentation+1);
        break;
      case CONE_OBJECT:
        SavePovCone(OutFile, TmpObject, indentation+1);
        break;
      case CSG_OBJECT:
        SavePovCSG(OutFile, TmpObject, indentation+1);
        break;
      case CYLINDER_OBJECT:
        SavePovCylinder(OutFile, TmpObject, indentation+1);
        break;
      case DISC_OBJECT:
        SavePovDisc(OutFile, TmpObject, indentation+1);
        break;
      case MESH_OBJECT:
        SavePovMesh(OutFile, TmpObject, indentation+1);
        break;
      case PLANE_OBJECT:
        SavePovPlane(OutFile, TmpObject, indentation+1);
        break;
      case SOR_OBJECT:
        SavePovSor(OutFile, TmpObject, indentation+1);
        break;
      case SPHERE_OBJECT:
        SavePovSphere(OutFile, TmpObject, indentation+1);
        break;
      case SUPERELLIPSOID_OBJECT:
        SavePovSuperEllipsoid(OutFile, TmpObject, indentation+1);
        break;
      case TORUS_OBJECT:
        SavePovTorus(OutFile, TmpObject, indentation+1);
        break;
      case TRIANGLE_OBJECT:
        SavePovTriangle(OutFile, TmpObject, indentation+1);
        break;
    }
  }
  if (CSG->Texture)
  {
    SavePovTexture(OutFile, CSG->Texture, indentation+1);
  }
  if (CSG->Trans)
    SavePovTrans(OutFile, CSG->Trans, indentation+1);
  fprintf(OutFile, "%s}\n\n", s);
  g_free(s);
}

/*****************************************************************************
*  SavePovLightSource
******************************************************************************/
static void SavePovLightSource(FILE *OutFile, LightSourceStruct *FirstLS)
{ /* FIXME: far from complete... */
  LightSourceStruct *TmpLS;

  for (TmpLS=FirstLS ; TmpLS ; TmpLS=TmpLS->Next)
  {
    fprintf(OutFile, "light_source\n{\n");
    fprintf(OutFile, "  <%g, %g, %g>\n", TmpLS->Location[0],
                                         TmpLS->Location[1],
                                         TmpLS->Location[2]);
    fprintf(OutFile, "  color rgb <%g, %g, %g>\n", TmpLS->Color[0],
                                               TmpLS->Color[1],
                                               TmpLS->Color[2]);
    fprintf(OutFile, "}\n");
  }
  fprintf(OutFile, "\n");
}

/*****************************************************************************
*  SavePovCamera
******************************************************************************/
static void SavePovCamera(FILE *OutFile, CameraStruct *Camera)
{

  if (Camera == NULL)
    return;
  fprintf(OutFile, "camera\n{\n");
  switch (Camera->Type)
  {
    case CAMERA_PERSPECTIVE:
      fprintf(OutFile, "  perspective\n");
      break;
    case CAMERA_ORTHOGRAPHIC:
      fprintf(OutFile, "  orthographic\n");
      break;
    case CAMERA_FISHEYE:
      fprintf(OutFile, "  fisheye\n");
      break;
    case CAMERA_ULTRA_WIDE_ANGLE:
      fprintf(OutFile, "  ultra_wide_angle\n");
      break;
    case CAMERA_OMNIMAX:
      fprintf(OutFile, "  omnimax\n");
      break;
    case CAMERA_PANORAMIC:
      fprintf(OutFile, "  panoramic\n");
      break;
    case CAMERA_CYLINDER:
      fprintf(OutFile, "  cylinder %d\n", Camera->CylinderType);
    default:
      /* panic */
  }
  fprintf(OutFile, "  location <%g, %g, %g>\n", Camera->Location[0],
                                                Camera->Location[1],
                                                Camera->Location[2]);
  fprintf(OutFile, "  right <%.12g, %.12g, %.12g>\n", Camera->Right[0],
                                             Camera->Right[1],
                                             Camera->Right[2]);
  fprintf(OutFile, "  up <%.12g, %.12g, %.12g>\n", Camera->Up[0],
                                          Camera->Up[1],
                                          Camera->Up[2]);
  fprintf(OutFile, "  direction <%.12g, %.12g, %.12g>\n", Camera->Direction[0],
                                                 Camera->Direction[1],
                                                 Camera->Direction[2]);
  fprintf(OutFile, "  sky <%g, %g, %g>\n", Camera->Sky[0],
                                           Camera->Sky[1],
                                           Camera->Sky[2]);
  if (Camera->BlurSamples)
  {
    fprintf(OutFile, "  blur_samples %d\n", Camera->BlurSamples);
    fprintf(OutFile, "  aperture %g\n", Camera->Aperture);
    fprintf(OutFile, "  focal_point <%g, %g, %g>\n", Camera->FocalPoint[0],
                                                     Camera->FocalPoint[1],
                                                     Camera->FocalPoint[2]);
    fprintf(OutFile, "  confidence %g\n", Camera->Confidence);
    fprintf(OutFile, "  variance %g\n", Camera->Variance);
  }
  fprintf(OutFile, "}\n\n");
}

/*****************************************************************************
*  SavePovByName
******************************************************************************/
void SavePovByName(FrameStruct *LocalFrame, char *FileName)
{
  GSList *tmp_list;
  ObjectStruct *TmpObject;
  FILE *OutFile;

  OutFile = fopen(FileName, "w");

  fprintf(OutFile, _("// Persistence of Vision 3.0 File.\n"));
  fprintf(OutFile, _("// This file has been generated by giram.\n"));
  fprintf(OutFile, _("// If you found any error, please contact David.Odin@bigfoot.com\n\n"));
  fprintf(OutFile, _("// The Global Settings:\n"));
  fprintf(OutFile, _("// The Camera:\n"));
  SavePovCamera(OutFile, LocalFrame->Camera);
  fprintf(OutFile, _("// The Scene Light Sources:\n"));
  SavePovLightSource(OutFile, LocalFrame->FirstLightSource);
  fprintf(OutFile, _("// The Scene Objects:\n"));
  for (tmp_list = LocalFrame->all_objects ; tmp_list ; tmp_list = g_slist_next(tmp_list))
  {
    TmpObject = (ObjectStruct*)(tmp_list->data);
    switch (TmpObject->Type)
    {
      case BOX_OBJECT:
        SavePovBox(OutFile, TmpObject, 0);
        break;
      case CONE_OBJECT:
        SavePovCone(OutFile, TmpObject, 0);
        break;
      case CSG_OBJECT:
        SavePovCSG(OutFile, TmpObject, 0);
        break;
      case CYLINDER_OBJECT:
        SavePovCylinder(OutFile, TmpObject, 0);
        break;
      case DISC_OBJECT:
        SavePovDisc(OutFile, TmpObject, 0);
        break;
      case MESH_OBJECT:
        SavePovMesh(OutFile, TmpObject, 0);
        break;
      case PLANE_OBJECT:
        SavePovPlane(OutFile, TmpObject, 0);
        break;
      case SOR_OBJECT:
        SavePovSor(OutFile, TmpObject, 0);
        break;
      case SPHERE_OBJECT:
        SavePovSphere(OutFile, TmpObject, 0);
        break;
      case SUPERELLIPSOID_OBJECT:
        SavePovSuperEllipsoid(OutFile, TmpObject, 0);
        break;
      case TORUS_OBJECT:
        SavePovTorus(OutFile, TmpObject, 0);
        break;
      case TRIANGLE_OBJECT:
        SavePovTriangle(OutFile, TmpObject, 0);
        break;
    }
  }
  fclose(OutFile);
}

/*****************************************************************************
*  SavePov
******************************************************************************/
void SavePov(GtkWidget *w, GtkFileSelection *FS)
{
  FrameStruct     *LocalFrame;
  GSList          *tmp_list;
  FrameViewStruct *tmp_view;
  char            *FileName;

  FileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS));
  LocalFrame = (FrameStruct *)gtk_object_get_data(GTK_OBJECT(FS), "frame");
  SavePovByName(LocalFrame, FileName);

  /* update the title of every windows */
  for (tmp_list = LocalFrame->all_views ; tmp_list ; tmp_list = tmp_list->next)
  {
    GtkWidget *drawing_area, *window;
    ViewUserDataStruct *user_data;

    tmp_view = tmp_list->data;
    drawing_area = tmp_view->drawing_area;
    user_data = gtk_object_get_data(GTK_OBJECT(drawing_area), "ViewData");
    g_free(tmp_view->name);
    switch (user_data->CameraStyle)
    {
      case ORTHO_XY_CAMERA:
        tmp_view->name = g_strdup_printf("(X-Y) %s", FileName);
        break;
      case ORTHO_XZ_CAMERA:
        tmp_view->name = g_strdup_printf("(X-Z) %s", FileName);
        break;
      case ORTHO_ZY_CAMERA:
        tmp_view->name = g_strdup_printf("(Z-Y) %s", FileName);
        break;
      case FRAME_CAMERA:
        tmp_view->name = g_strdup_printf("Camera %s", FileName);
        break;
    }
    window = gtk_widget_get_toplevel(drawing_area);
    gtk_window_set_title(GTK_WINDOW(window), tmp_view->name);
  }
}

