/* csgtree.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include "giram.h"
#include "tool_select.h"
#include "csgtree.h"

#include "giramintl.h"

GtkWidget *CSGTreeDialog;
GtkWidget *CSGTreeMenu;
GSList *CSGTreeMenuGroup;
FrameStruct *CSGTreeCurrentFrame;
static char * book_open_xpm[] =
{
  "16 16 4 1",
  "       c None s None",
  ".      c black",
  "X      c #808080",
  "o      c white",
  "                ",
  "  ..            ",
  " .Xo.    ...    ",
  " .Xoo. ..oo.    ",
  " .Xooo.Xooo...  ",
  " .Xooo.oooo.X.  ",
  " .Xooo.Xooo.X.  ",
  " .Xooo.oooo.X.  ",
  " .Xooo.Xooo.X.  ",
  " .Xooo.oooo.X.  ",
  "  .Xoo.Xoo..X.  ",
  "   .Xo.o..ooX.  ",
  "    .X..XXXXX.  ",
  "    ..X.......  ",
  "     ..         ",
  "                "
};
  
static char * book_closed_xpm[] =
{
  "16 16 6 1",
  "       c None s None",
  ".      c black",
  "X      c red",
  "o      c yellow",
  "O      c #808080",
  "#      c white",
  "                ",
  "       ..       ",
  "     ..XX.      ",
  "   ..XXXXX.     ",
  " ..XXXXXXXX.    ",
  ".ooXXXXXXXXX.   ",
  "..ooXXXXXXXXX.  ",
  ".X.ooXXXXXXXXX. ",
  ".XX.ooXXXXXX..  ",
  " .XX.ooXXX..#O  ",
  "  .XX.oo..##OO. ",
  "   .XX..##OO..  ",
  "    .X.#OO..    ",
  "     ..O..      ",
  "      ..        ",
  "                "
};

static char * mini_page_xpm[] = 
{
  "16 16 4 1",
  "       c None s None",
  ".      c black",
  "X      c white",
  "o      c #808080",
  "                ",
  "   .......      ",
  "   .XXXXX..     ",
  "   .XoooX.X.    ",
  "   .XXXXX....   ",
  "   .XooooXoo.o  ",
  "   .XXXXXXXX.o  ",
  "   .XooooooX.o  ",
  "   .XXXXXXXX.o  ",
  "   .XooooooX.o  ",
  "   .XXXXXXXX.o  ",
  "   .XooooooX.o  ",
  "   .XXXXXXXX.o  ",
  "   ..........o  ",
  "    oooooooooo  ",
  "                "
};

static char *sphere_xpm[] =
{
  "16 16 131 2",
  ".. c #1f1f1f",
  ".# c #202020",
  ".a c #262626",
  ".b c #292929",
  ".c c #2a2a2a",
  ".d c #2b2b2b",
  ".e c #2d2d2d",
  ".f c #2e2e2e",
  ".g c #2f2f2f",
  ".h c #323232",
  ".i c #333333",
  ".j c #343434",
  ".k c #353535",
  ".l c #363636",
  ".m c #383838",
  ".n c #393939",
  ".o c #3a3a3a",
  ".p c #3b3b3b",
  ".q c #3c3c3c",
  ".r c #3d3d3d",
  ".s c #3f3f3f",
  ".t c #414141",
  ".u c #424242",
  ".v c #464646",
  ".w c #474747",
  ".x c #484848",
  ".y c #4a4a4a",
  ".z c #4b4b4b",
  ".A c #4c4c4c",
  ".B c #4d4d4d",
  ".C c #4e4e4e",
  ".D c #4f4f4f",
  ".E c #505050",
  ".F c #545454",
  ".G c #555555",
  ".H c #565656",
  ".I c #585858",
  ".J c #595959",
  ".K c #5a5a5a",
  ".L c #5b5b5b",
  ".M c #5c5c5c",
  ".N c #5d5d5d",
  ".O c #5e5e5e",
  ".P c #606060",
  ".Q c #626262",
  ".R c #636363",
  ".S c #646464",
  ".T c #666666",
  ".U c #676767",
  ".V c #686868",
  ".W c #696969",
  ".X c #6a6a6a",
  ".Y c #6b6b6b",
  ".Z c #6c6c6c",
  ".0 c #6d6d6d",
  ".1 c #6e6e6e",
  ".2 c #707070",
  ".3 c #717171",
  ".4 c #737373",
  ".5 c #747474",
  ".6 c #757575",
  ".7 c #767676",
  ".8 c #777777",
  ".9 c #787878",
  "#. c #797979",
  "## c #7a7a7a",
  "#a c #7b7b7b",
  "#b c #7c7c7c",
  "#c c #7e7e7e",
  "#d c #808080",
  "#e c #818181",
  "#f c #828282",
  "#g c #838383",
  "#h c #848484",
  "#i c #858585",
  "#j c #868686",
  "#k c #878787",
  "#l c #888888",
  "#m c #898989",
  "#n c #8a8a8a",
  "#o c #8b8b8b",
  "#p c #8c8c8c",
  "#q c #909090",
  "#r c #929292",
  "#s c #939393",
  "#t c #959595",
  "#u c #969696",
  "#v c #979797",
  "#w c #989898",
  "#x c #999999",
  "#y c #9a9a9a",
  "#z c #9c9c9c",
  "#A c #9d9d9d",
  "#B c #9f9f9f",
  "#C c #a0a0a0",
  "#D c #a2a2a2",
  "#E c #a4a4a4",
  "#F c #a5a5a5",
  "#G c #a6a6a6",
  "#H c #a7a7a7",
  "#I c #a8a8a8",
  "#J c #a9a9a9",
  "#K c #ababab",
  "#L c #b0b0b0",
  "#M c #b1b1b1",
  "#N c #b2b2b2",
  "#O c #b3b3b3",
  "#P c #b4b4b4",
  "#Q c #b5b5b5",
  "#R c #b6b6b6",
  "#S c #b8b8b8",
  "#T c #bcbcbc",
  "#U c #bdbdbd",
  "#V c #bebebe",
  "#W c #bfbfbf",
  "#X c #c0c0c0",
  "#Y c #c1c1c1",
  "#Z c #c2c2c2",
  "#0 c #c8c8c8",
  "#1 c #c9c9c9",
  "#2 c #d2d2d2",
  "#3 c #d7d7d7",
  "#4 c #d8d8d8",
  "#5 c #e0e0e0",
  "#6 c #ebebeb",
  "#7 c #ececec",
  "#8 c #f6f6f6",
  "#9 c #f7f7f7",
  "a. c #f8f8f8",
  "a# c #fbfbfb",
  "aa c #ffffff",
  "aaaaaaaa#5#G#l#d#n#C#Y#6aaaaaaaa",
  "aaaaa.#w.Q.1#a#l#s#z#D#F#0a#aaaa",
  "aa#9.7.I.T.5#f#q#A#I#L#P#L#Va#aa",
  "aa#o.A.L.X.9#k#u#E#N#U#W#T#M#1aa",
  "#3.s.B.M.Y###m#w#H#R#X#Y#X#Q#H#7",
  "#i.r.A.L.X#.#l#v#F#P#W#X#V#O#F#Z",
  ".A.p.y.J.U.7#h#r#B#K#P#S#P#K#B#D",
  ".f.m.v.F.Q.2#c#n#u#C#H#J#G#B#t#p",
  ".b.i.t.C.M.W.6#e#o#s#w#y#w#r#n#g",
  ".t.e.o.w.F.P.Y.6#c#i#m#o#m#i#c#n",
  ".9.a.h.s.z.H.P.W.3.8#a#b##.8.3#H",
  "#2...c.k.t.z.F.N.R.W.Z.0.Z.V.S#5",
  "aa.4.#.d.l.s.x.D.G.K.N.O.N.K#xaa",
  "aa#8.E.#.c.i.p.u.w.z.C.D.C.9a.aa",
  "aaaa#8.4...a.f.j.n.q.s.t#pa.aaaa",
  "aaaaaaaa#2.9.u.c.g.B#j#4aaaaaaaa"
};

#include "icons/eye.xpm"
#include "icons/noeye.xpm"

/* Blue items are selected, white ones not. XXX should be settable too*/
static GdkColor CSGSelectedColour;
static GdkColor CSGUnselectedColour;

static GdkPixmap *pixmap1;
static GdkPixmap *pixmap2;
static GdkPixmap *pixmap3;
static GdkPixmap *SpherePixmap;
static GdkBitmap *mask1;
static GdkBitmap *mask2;
static GdkBitmap *mask3;
static GdkBitmap *SphereMask;
static GdkPixmap *EyePixmap;
static GdkBitmap *EyeMask;
static GdkPixmap *NoEyePixmap;
static GdkBitmap *NoEyeMask;
GtkCTree *CSGTreeCTree = NULL;
 
/*****************************************************************************
*  CSGTreeSelect - Marks an object as selected in the CSG Tree
*  May be called if the CSG Tree dialog is not opened.
******************************************************************************/
void CSGTreeSelect(ObjectStruct *Object)
{
  if ((CSGTreeDialog!=NULL) && (CSGTreeCTree!=NULL))
  {
    gtk_ctree_select(GTK_CTREE(CSGTreeCTree), Object->CSGTreeNode);
  }
}

/*****************************************************************************
*  CSGTreeUnselect - Marks an object as not selected in the CSG Tree
*  May be called if the CSG Tree dialog is not opened.
******************************************************************************/
void CSGTreeUnselect(ObjectStruct *Object)
{
  if ((CSGTreeDialog!=NULL) && (CSGTreeCTree!=NULL))
  {
    gtk_ctree_unselect(GTK_CTREE(CSGTreeCTree), Object->CSGTreeNode);
  }
}

/*****************************************************************************
*  CSGRecursiveTree
******************************************************************************/
static void CSGRecursiveTree(ObjectStruct *CSG, GtkCTreeNode *parent)
{
  GtkCTreeNode *sibling;
  CSGStruct *CCSG = (CSGStruct *)CSG;
  GSList *tmp_list;
  ObjectStruct *TmpObject;
  gchar *text[3];
  gchar buf1[60];
  gchar buf2[60];
 
  text[0] = "";
  text[1] = buf1;
  text[2] = buf2;
  
  sibling = NULL;
  for (tmp_list = CCSG->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    TmpObject = (ObjectStruct*)(tmp_list->data);
    strcpy(buf2, _("hello"));
    switch (TmpObject->Type)
    {
      case BOX_OBJECT:
        {
          BoxStruct *BBox = (BoxStruct *)TmpObject;
          strcpy(buf1, "box");
          sprintf(buf2, "<%g, %g, %g>, <%g, %g, %g>", 
                  BBox->MinCorner[0], BBox->MinCorner[1], BBox->MinCorner[2],
                  BBox->MaxCorner[0], BBox->MaxCorner[1], BBox->MaxCorner[2]);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case CONE_OBJECT:
        {
          ConeStruct *CCone = (ConeStruct *)TmpObject;
          strcpy(buf1, "cone");
          sprintf(buf2, "Radius 1: %g, Radius 2: %g",
                  CCone->Radius1, CCone->Radius2);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                          pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case CYLINDER_OBJECT:
          strcpy(buf1, "cylinder");
          sprintf(buf2, "<0, 0, 0>, <0, 1, 0>, 1"); /* XXX: not that useful... */ 
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        break;
      case DISC_OBJECT:
        {
          DiscStruct *DDisc = (DiscStruct *)TmpObject;
          strcpy(buf1, "disc");
          sprintf(buf2, "%g, %g", DDisc->Radius, DDisc->HoleRadius);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case MESH_OBJECT:
        {
          strcpy(buf1, "mesh");
          sprintf(buf2, "%s", TmpObject->Name);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case PLANE_OBJECT:
        {
          PlaneStruct *PPlane = (PlaneStruct *)TmpObject;
          strcpy(buf1, "plane");
          sprintf(buf2, "<%g, %g, %g>, %g",
                  PPlane->Normal[0], PPlane->Normal[1], PPlane->Normal[2],
                  PPlane->Distance);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case SOR_OBJECT:
        {
          strcpy(buf1, "Sor");
          strcpy(buf2, "");
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case SPHERE_OBJECT:
        {
          SphereStruct *SSphere = (SphereStruct *)TmpObject;
          strcpy(buf1, "sphere");
          sprintf(buf2, "<%g, %g, %g>, %g",
                  SSphere->Center[0], SSphere->Center[1], SSphere->Center[2],
                  SSphere->Radius);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               SpherePixmap, SphereMask, NULL, NULL, TRUE, FALSE);
        }
        break;
      case SUPERELLIPSOID_OBJECT:
        {
          SuperEllipsoidStruct *SSuperEllipsoid = (SuperEllipsoidStruct *)TmpObject;
          strcpy(buf1, "superellipsoid");
          sprintf(buf2, "<%g, %g>", SSuperEllipsoid->E, SSuperEllipsoid->N);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                          pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case TORUS_OBJECT:
        {
          TorusStruct *TTorus = (TorusStruct *)TmpObject;
          strcpy(buf1, "torus");
          sprintf(buf2, "%g, %g", TTorus->Major, TTorus->Minor);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case TRIANGLE_OBJECT:
        {
          strcpy(buf1, "triangle");
          strcpy(buf2, "");
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case CSG_OBJECT:
        switch (((CSGStruct *)TmpObject)->Type)
        {
          case CSG_DIFFERENCE:
            strcpy(buf1, "difference");
            break;
          case CSG_INTERSECTION:
            strcpy(buf1, "intersection");
            break;
          case CSG_MERGE:
            strcpy(buf1, "merge");
            break;
          case CSG_UNION:
            strcpy(buf1, "union");
            break;
        }
        strcpy(buf2, "");
        sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                             pixmap1, mask1, pixmap2, mask2,
                             FALSE, TRUE);
        CSGRecursiveTree(TmpObject, sibling);
        break;
    }
    /* Mark object as selected if apropriate. */
    if (TmpObject->Selected!=0) {
      gtk_ctree_select(GTK_CTREE(CSGTreeCTree), sibling);
    }
    TmpObject->CSGTreeNode = sibling;
  }
}

/*****************************************************************************
*  RebuildCSGTree
******************************************************************************/
void RebuildCSGTree(GtkWidget *Menu, FrameStruct *TmpFrame)
{
  GtkCTreeNode *parent;
  GtkCTreeNode *sibling;
  gchar *text[3];
  GSList *tmp_list;
  ObjectStruct *TmpObject;
  LightSourceStruct *TmpLight;
  gchar buf1[60];
  gchar buf2[60];

  /* Colours of selected & unselected items. XXX Should be settable */
  CSGSelectedColour.red = 0;
  CSGSelectedColour.green = 0;
  CSGSelectedColour.blue = 65535;        /* Blue */
  
  CSGUnselectedColour.red = 65535;
  CSGUnselectedColour.green = 65535;
  CSGUnselectedColour.blue = 65535;      /* White */

  CSGTreeCurrentFrame = TmpFrame;
  if (!CSGTreeCTree)
    return;
  
  gtk_clist_freeze(GTK_CLIST(CSGTreeCTree));
  gtk_clist_clear(GTK_CLIST(CSGTreeCTree));

  text[0] = "";
  text[1] = "Lights";
  text[2] = "";

  parent = gtk_ctree_insert_node(CSGTreeCTree, NULL, NULL, text, 5,
                                 pixmap1, mask1,
                                 pixmap2, mask2,
                                 FALSE, TRUE);
  text[1] = buf1;
  text[2] = buf2;
  sibling = NULL;
  for (TmpLight = TmpFrame->FirstLightSource ;
       TmpLight ;
       TmpLight = TmpLight->Next)
  {
    sprintf(buf1, "center <%g, %g, %g>", TmpLight->Location[0],
                                         TmpLight->Location[1],
                                         TmpLight->Location[2]);
    sprintf(buf2, "color <%g, %g, %g>", TmpLight->Color[0],
                                        TmpLight->Color[1],
                                        TmpLight->Color[2]);
    sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                    pixmap3, mask3, NULL, NULL,
                                    TRUE, FALSE);
  }

  text[1] = "Objects";
  text[2] = "";
      
  parent = gtk_ctree_insert_node(CSGTreeCTree, NULL, NULL, text, 5, 
                                 pixmap1, mask1,
                                 pixmap2, mask2, 
                                 FALSE, TRUE);
  text[1] = buf1;
  text[2] = buf2;
  sibling = NULL;
  for (tmp_list = TmpFrame->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    TmpObject = (ObjectStruct *)(tmp_list->data);
    switch (TmpObject->Type)
    {
      case BOX_OBJECT:
        {
          BoxStruct *BBox = (BoxStruct *)TmpObject;
          
          strcpy(buf1, "box");
          sprintf(buf2, "<%g, %g, %g>, <%g, %g, %g>", 
                  BBox->MinCorner[0], BBox->MinCorner[1], BBox->MinCorner[2],
                  BBox->MaxCorner[0], BBox->MaxCorner[1], BBox->MaxCorner[2]);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case CONE_OBJECT:
        {
          ConeStruct *CCone = (ConeStruct *)TmpObject;
          strcpy(buf1, "cone");
          sprintf(buf2, "Radius 1: %g, Radius 2: %g",
                  CCone->Radius1, CCone->Radius2);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                          pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case CYLINDER_OBJECT:
        strcpy(buf1, "cylinder");
        sprintf(buf2, "<0, 0, 0>, <0, 1, 0>, 1"); /* XXX: not that useful */ 
        sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                             pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        break;
      case DISC_OBJECT:
        {
          DiscStruct *DDisc = (DiscStruct *)TmpObject;
          strcpy(buf1, "disc");
          sprintf(buf2, "%g, %g", DDisc->Radius, DDisc->HoleRadius);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case MESH_OBJECT:
        {
          strcpy(buf1, "mesh");
          sprintf(buf2, "%s", TmpObject->Name);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case PLANE_OBJECT:
        {
          PlaneStruct *PPlane = (PlaneStruct *)TmpObject;

          strcpy(buf1, "plane");
          sprintf(buf2, "<%g, %g, %g>, %g",
                  PPlane->Normal[0], PPlane->Normal[1], PPlane->Normal[2],
                  PPlane->Distance);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case SOR_OBJECT:
        {
          strcpy(buf1, "Sor");
          strcpy(buf2, "");
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case SPHERE_OBJECT:
        {
          SphereStruct *SSphere = (SphereStruct *)TmpObject;
          strcpy(buf1, "sphere");
          sprintf(buf2, "<%g, %g, %g>, %g",
                  SSphere->Center[0], SSphere->Center[1], SSphere->Center[2],
                  SSphere->Radius);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               SpherePixmap, SphereMask, NULL, NULL, TRUE, FALSE);
        }
        break;
      case SUPERELLIPSOID_OBJECT:
        {
          SuperEllipsoidStruct *SSuperEllipsoid = (SuperEllipsoidStruct *)TmpObject;
          strcpy(buf1, "superellipsoid");
          sprintf(buf2, "<%g, %g>", SSuperEllipsoid->E, SSuperEllipsoid->N);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                          pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case TORUS_OBJECT:
        {
          TorusStruct *TTorus = (TorusStruct *)TmpObject;
          strcpy(buf1, "torus");
          sprintf(buf2, "%g, %g", TTorus->Major, TTorus->Minor);
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                                pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case TRIANGLE_OBJECT:
        {
          strcpy(buf1, "triangle");
          strcpy(buf2, "");
          sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                               pixmap3, mask3, NULL, NULL, TRUE, FALSE);
        }
        break;
      case CSG_OBJECT:
        switch (((CSGStruct *)TmpObject)->Type)
        {
          case CSG_DIFFERENCE:
            strcpy(buf1, "difference");
            break;
          case CSG_INTERSECTION:
            strcpy(buf1, "intersection");
            break;
          case CSG_MERGE:
            strcpy(buf1, "merge");
            break;
          case CSG_UNION:
            strcpy(buf1, "union");
            break;
        }
        strcpy(buf2, "");
        sibling = gtk_ctree_insert_node(CSGTreeCTree, parent, sibling, text, 5,
                             pixmap1, mask1, pixmap2, mask2, FALSE, TRUE);
        CSGRecursiveTree(TmpObject, sibling);
        break;
    }
    TmpObject->CSGTreeNode = sibling;
    /* Mark object as selected if apropriate. */
    if (TmpObject->Selected!=0) {
      gtk_ctree_select(GTK_CTREE(CSGTreeCTree), sibling);
    }
    /* Mark object as visible if appropriate */
    if (TmpObject->Visible)
      gtk_ctree_node_set_pixmap(GTK_CTREE(CSGTreeCTree), sibling, 0, EyePixmap, EyeMask);
    gtk_ctree_node_set_row_data(GTK_CTREE(CSGTreeCTree), sibling, 
                                (gpointer)TmpObject);
  }
  /***********************************************************/
  gtk_clist_thaw(GTK_CLIST(CSGTreeCTree));
}

/*****************************************************************************
*  CSGTreeEventDispatcher
******************************************************************************/
gint CSGTreeEventDispatcher(GtkWidget *CTree, GdkEvent *event)
{
  GtkCTreeNode    *Node;
  ObjectStruct    *Object;
  GSList          *tmp_list;
  FrameViewStruct *TmpView;

  GdkEventButton *bevent = (GdkEventButton *) event;

  switch (event->type)
  {
    case GDK_2BUTTON_PRESS:
	/* Fake handler to trap double clicks */
	if (bevent->button == 1) return(TRUE);
	break;

    case GDK_BUTTON_PRESS:
      {
        GdkEventButton *bevent = (GdkEventButton *) event;
/*        if (bevent->button == 3) *//* Right Button ==> Popup Menu */
/*        gtk_menu_popup(GTK_MENU(Menu), NULL, NULL, NULL, NULL,
                         bevent->button, bevent->time);
        else */if (bevent->button == 1)
        {
          gint row, column;
          if (gtk_clist_get_selection_info(GTK_CLIST(CTree), 
                                           bevent->x, bevent->y, 
                                           &row, &column))
          {
            Node = gtk_ctree_node_nth(GTK_CTREE(CTree), row);
            Object = (ObjectStruct *)gtk_ctree_node_get_row_data(GTK_CTREE(CTree), Node);
            if (column == 0)
            {
              if (Object)
              {
                Object->Visible = !(Object->Visible);
                if (Object->Visible)
                {
                  gtk_ctree_node_set_pixmap(GTK_CTREE(CSGTreeCTree), Node, 0,
                                            EyePixmap, EyeMask);
                } else
                {
                  gtk_ctree_node_set_pixmap(GTK_CTREE(CSGTreeCTree), Node, 0,
                                            NoEyePixmap, NoEyeMask);
                }
              }
              for (tmp_list = CSGTreeCurrentFrame->all_views ; 
                   tmp_list ; 
                   tmp_list = tmp_list->next)
              {
                TmpView = tmp_list->data;
                gtk_widget_queue_draw(TmpView->drawing_area);
              }
              return TRUE;
            } else if ((column >= 1) && Object)
            {
              /* We have an object in the CSG tree. If it is clicked on,
                 we select the object and redraw the views. */
              Object->Selected = !(Object->Selected);
              if (Object->Selected)
                AddObjectToSelection(&(CSGTreeCurrentFrame->Selection), Object);
              else
                RemoveObjectFromSelection(&(CSGTreeCurrentFrame->Selection),
                                          Object);
              for (tmp_list = CSGTreeCurrentFrame->all_views ; 
                   tmp_list ; 
                   tmp_list = tmp_list->next)
              {
                TmpView = tmp_list->data;
                gtk_widget_queue_draw(TmpView->drawing_area);
              }
              return TRUE;
            }
          }
        }
      }
      break;
    default:
     return FALSE;
  }
  return FALSE;
}

/*****************************************************************************
*  ShowCSGTree
******************************************************************************/
void ShowCSGTree(GtkWidget *DrawingArea)
{
  FrameStruct *TmpFrame;
  ViewUserDataStruct *UserData = NULL;
  GtkWidget *button;
  GtkWidget *omenu;
  GtkWidget *menu_item;
  GtkWidget *Separator;
  char *title[] = { "Visible", "Tree" , "Info" };
  GtkWidget *ScrolledWin;  
  GdkColor transparent;
 
  if (CSGTreeDialog)
  {
    gdk_window_raise(CSGTreeDialog->window);
    return;
  }
  if (DrawingArea)
  {
    UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  } 
  CSGTreeDialog = gtk_dialog_new();
  /* Make sure CSGTreeDialog is set back to NULL when it gets destroyed */
  gtk_signal_connect(GTK_OBJECT(CSGTreeDialog), "destroy",
                     GTK_SIGNAL_FUNC(gtk_widget_destroyed), 
                     &CSGTreeDialog);
  gtk_window_set_title(GTK_WINDOW(CSGTreeDialog), _("CSG Tree"));
  gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(CSGTreeDialog)->vbox), 5);
  gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(CSGTreeDialog)->action_area), 0);
  button = gtk_button_new_with_label(_("Close"));
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(CSGTreeDialog));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(CSGTreeDialog)->action_area), button, 
                     TRUE, TRUE, 0);
  gtk_widget_show(button);
  /**********************************/
  CSGTreeCTree = GTK_CTREE(gtk_ctree_new_with_titles(3, 1, title));
  gtk_widget_set_usize(GTK_WIDGET(CSGTreeCTree),340,220);
  omenu = gtk_option_menu_new();
  CSGTreeMenu = gtk_menu_new();
  CSGTreeMenuGroup = NULL;
  menu_item = NULL;
  for (TmpFrame = FirstFrame ; TmpFrame ; TmpFrame=TmpFrame->Next)
  {
    menu_item = gtk_radio_menu_item_new_with_label(CSGTreeMenuGroup,
                                                   TmpFrame->Name);
    gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
                       (GtkSignalFunc)RebuildCSGTree, TmpFrame);
    if (!CSGTreeMenuGroup)
      CSGTreeMenuGroup = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(
                                                                  menu_item));
    gtk_menu_append(GTK_MENU(CSGTreeMenu), menu_item);
    
    if ((DrawingArea != NULL) && (TmpFrame == UserData->FrameView))
      gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menu_item), TRUE);
    
    if ((DrawingArea == NULL) && (TmpFrame == FirstFrame))
      gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menu_item), TRUE);
    gtk_object_set_data(GTK_OBJECT(menu_item), "CSGMenu", TmpFrame);
    gtk_widget_show(menu_item);
  }
  gtk_option_menu_set_menu(GTK_OPTION_MENU(omenu), CSGTreeMenu);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(CSGTreeDialog)->vbox), omenu, FALSE, 
                     FALSE, 0);
  gtk_widget_show(omenu);
  Separator = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(CSGTreeDialog)->vbox), Separator, 
                     FALSE, FALSE, 5);
  gtk_widget_show(Separator);
    
  /* The ctree */
  gtk_signal_connect(GTK_OBJECT(CSGTreeCTree), "event",
                     GTK_SIGNAL_FUNC(CSGTreeEventDispatcher), 
                     NULL);
    
  gtk_clist_set_column_auto_resize(GTK_CLIST(CSGTreeCTree), 0, TRUE);
  gtk_clist_set_column_auto_resize(GTK_CLIST(CSGTreeCTree), 1, TRUE);
  gtk_clist_set_column_auto_resize(GTK_CLIST(CSGTreeCTree), 2, TRUE);
   
  gtk_ctree_set_line_style(CSGTreeCTree, GTK_CTREE_LINES_DOTTED);
  gtk_clist_set_row_height(GTK_CLIST(CSGTreeCTree), 20);
  ScrolledWin = gtk_scrolled_window_new(NULL, NULL);
  gtk_container_add(GTK_CONTAINER(ScrolledWin), GTK_WIDGET(CSGTreeCTree));
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(ScrolledWin),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_container_set_border_width(GTK_CONTAINER(ScrolledWin), 5);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(CSGTreeDialog)->vbox), ScrolledWin, TRUE, TRUE, 0);
  gtk_clist_set_selection_mode(GTK_CLIST(CSGTreeCTree), GTK_SELECTION_MULTIPLE);
  gtk_clist_set_column_auto_resize(GTK_CLIST(CSGTreeCTree), 0, TRUE);
  
  gtk_widget_realize(GTK_WIDGET(CSGTreeCTree));
  pixmap1 = gdk_pixmap_create_from_xpm_d(GTK_WIDGET(CSGTreeCTree)->window,
                &mask1, &transparent, book_closed_xpm);
  pixmap2 = gdk_pixmap_create_from_xpm_d(GTK_WIDGET(CSGTreeCTree)->window,
                &mask2, &transparent, book_open_xpm);
  pixmap3 = gdk_pixmap_create_from_xpm_d(GTK_WIDGET(CSGTreeCTree)->window,
                &mask3, &transparent, mini_page_xpm);
  SpherePixmap = gdk_pixmap_create_from_xpm_d(GTK_WIDGET(CSGTreeCTree)->window,
                &SphereMask, &transparent, sphere_xpm);
  EyePixmap = gdk_pixmap_create_from_xpm_d(GTK_WIDGET(CSGTreeCTree)->window,
                &EyeMask, &transparent, eye_icon);
  NoEyePixmap = gdk_pixmap_create_from_xpm_d(GTK_WIDGET(CSGTreeCTree)->window,
                &NoEyeMask, &transparent, noeye_icon);
 
  if (DrawingArea)
    RebuildCSGTree(menu_item, UserData->FrameView);
  else if (FirstFrame)
    RebuildCSGTree(menu_item, FirstFrame);

  gtk_widget_show_all(CSGTreeDialog);
}

