/* edit.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include "giram.h"
#include "object.h"
#include "destroy.h"
#include "copy.h"
#include "csgtree.h"
#include "tool_select.h"
#include "edit.h"
#include "utils.h"

static GList *CutBuffer = NULL;

/*****************************************************************************
*  DestroyCutBuffer
******************************************************************************/
void DestroyCutBuffer(void)
{
  GList *TmpSel;

  for (TmpSel = CutBuffer; TmpSel ; TmpSel = g_list_next(TmpSel) )
    DestroyObject((ObjectStruct *)(TmpSel->data));
  g_list_free(CutBuffer);
  CutBuffer = NULL;
}

/*****************************************************************************
*  CutSelection
******************************************************************************/
void CutSelection()
{
  ViewUserDataStruct *UserData;
  ObjectSelection    *TmpSel, *TmpSel1, *TmpSel2;
  ObjectStruct       *TmpObject;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;

  UserData = get_view_user_data();
  /* Destroy the current cut buffer */
  DestroyCutBuffer();
  if (UserData->FrameView->Selection == NULL)
    return;
  for (TmpSel = UserData->FrameView->Selection ;
       TmpSel->Next != UserData->FrameView->Selection ;
       TmpSel = TmpSel->Next)
  {
    /* Remove the object from the frame. */
    g_slist_remove(UserData->FrameView->all_objects,
                   TmpSel->Object);
    /* Add the object to the current cut buffer. */
    CutBuffer = g_list_prepend(CutBuffer, CopyObject(TmpSel->Object));
    /* Destroy the object */
    DestroyObject(TmpSel->Object);
  }
  /* Remove the object from the frame. */
  g_slist_remove(UserData->FrameView->all_objects, TmpSel->Object);
  /* Add the object to the current cut buffer. */
  CutBuffer = g_list_prepend(CutBuffer, CopyObject(TmpSel->Object));
  /* Destroy the object */
  DestroyObject(TmpSel->Object);
  /* Destroy the Current Selection */
  TmpSel1 = UserData->FrameView->Selection;
  while (TmpSel1->Next != UserData->FrameView->Selection)
  {
    TmpSel2 = TmpSel1->Next;
    g_free(TmpSel1);
    TmpSel1 = TmpSel2;
  }
  g_free(TmpSel1);
  UserData->FrameView->Selection = NULL;
  /* Redraw all view plus CSG Tree */
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
}

/*****************************************************************************
*  CopySelection
******************************************************************************/
void CopySelection()
{
  ViewUserDataStruct *UserData;
  ObjectSelection    *TmpSel;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;

  UserData = get_view_user_data();
  /* Destroy the current cut buffer */
  DestroyCutBuffer();
  if (UserData->FrameView->Selection == NULL)
    return;
  for (TmpSel = UserData->FrameView->Selection ;
       TmpSel->Next != UserData->FrameView->Selection ;
       TmpSel = TmpSel->Next)
  {
    /* Add the object to the current cut buffer. */
    CutBuffer = g_list_prepend(CutBuffer, CopyObject(TmpSel->Object));
  }
  /* Add the object to the current cut buffer. */
  CutBuffer = g_list_prepend(CutBuffer, CopyObject(TmpSel->Object));
  /* Redraw all view plus CSG Tree */
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
}

/*****************************************************************************
*  PasteSelection
******************************************************************************/
void PasteSelection()
{
  ViewUserDataStruct *UserData;
  ObjectSelection    *TmpSel1, *TmpSel2;
  GList              *TmpSel;
  ObjectStruct       *TmpObject;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;

  UserData = get_view_user_data();
  /* Destroy the Current Selection */
  if (UserData->FrameView->Selection)
  {
    TmpSel1 = UserData->FrameView->Selection;
    while (TmpSel1->Next != UserData->FrameView->Selection)
    {
      TmpSel2 = TmpSel1->Next;
      TmpSel1->Object->Selected = FALSE;
      g_free(TmpSel1);
      TmpSel1 = TmpSel2;
    }
    TmpSel1->Object->Selected = FALSE;
    g_free(TmpSel1);
    UserData->FrameView->Selection = NULL;
  }

  if (CutBuffer)
  {
    for (TmpSel = CutBuffer ; TmpSel ; TmpSel = g_list_next(TmpSel))
    {
      ObjectStruct *Tmp1;

      TmpObject = CopyObject((ObjectStruct *)(TmpSel->data));
      UserData->FrameView->all_objects = g_slist_append(UserData->FrameView->all_objects,
                                                        TmpObject);
      AddObjectToSelection(&(UserData->FrameView->Selection), TmpObject);
    }
  }
  
  /* Redraw all view plus CSG Tree */
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
}

