/* giram.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2001 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/wait.h>

#ifndef  WAIT_ANY
#define  WAIT_ANY (-1)
#endif   /*  WAIT_ANY  */

#include "giram.h"
#include "about_dialog.h"
#include "help_dialog.h"
#include "frame.h"
#include "csgtree.h"
#include "LoadPov.h"
#include "dxf.h"
#ifdef WITH_S3D
#include "s3d.h"
#endif
#include "SavePov.h"
#ifdef USE_LIB3DS
#include "3ds.h"
#endif

#include "cone.h"

#include "tool_rotate.h"
#include "tool_scale.h"
#include "tool_options.h"
#include "tools.h"
#include "view.h"
#include "view_camera.h"
#include "plugins.h"
#include "preferences.h"
#include "utils.h"
#include "copy.h"
#include "giramrc.h"
#include "giramenv.h"
#include "giramintl.h"
#include "appenv.h"
#include "errors.h"
#include "giramsignal.h"
#include "user_install.h"
#include "app_procs.h"
#include "giramcolorlist.h"

GiramModeToolType ModeTool;
gint number_of_shapes;

static void giram_sigfatal_handler(gint sig_num);
static void giram_sigchld_handler(gint sig_num);

static void init(void);
static void giram_error_handler(const gchar    *domain,
                                GLogLevelFlags  flags,
                                const gchar    *msg,
                                gpointer        user_data);

/* GLOBAL data */
gboolean be_verbose        = FALSE;
gboolean use_debug_handler = FALSE;
gboolean console_messages  = FALSE;

MessageHandlerType message_handler = CONSOLE;

gchar *prog_name                = NULL; /* The path name we are invoked with */
gchar *alternate_giramrc        = NULL;
gchar *alternate_system_giramrc = NULL;

/*****************************************************************************
*  CreateDefaultTexture
******************************************************************************/
void CreateDefaultTexture(void)
{
  DefaultPigment = g_malloc(sizeof(PigmentStruct));
  DefaultPigment->refcount = 1;
  DefaultPigment->Type = PAT_SOLID_COLOR;
  DefaultPigment->Trans = NULL;
  DefaultPigment->Color[0] = 1.0;
  DefaultPigment->Color[1] = 1.0;
  DefaultPigment->Color[2] = 1.0;
  DefaultPigment->Color[3] = 0.0;
  DefaultPigment->Color[4] = 0.0;
  DefaultPigment->Color2[0] = 1.0;
  DefaultPigment->Color2[1] = 1.0;
  DefaultPigment->Color2[2] = 1.0;
  DefaultPigment->Color2[3] = 0.0;
  DefaultPigment->Color2[4] = 0.0;
  DefaultPigment->Color3[0] = 1.0;
  DefaultPigment->Color3[1] = 1.0;
  DefaultPigment->Color3[2] = 1.0;
  DefaultPigment->Color3[3] = 0.0;
  DefaultPigment->Color3[4] = 0.0;
  DefaultTexture = g_malloc(sizeof(TextureStruct));
  DefaultTexture->Pigment = CopyPigment(DefaultPigment);
  DefaultTexture->Trans = NULL;
  DefaultTexture->refcount = 1;
}

/*****************************************************************************
*  giramInit
******************************************************************************/
void giramInit(void)
{
#ifdef HAVE_GL
  Display *dpy;
  XVisualInfo *vi;
  int attrib[] =
  {
    GLX_RGBA,
    GLX_DOUBLEBUFFER,
    None
  };
#endif
  FirstFrame = NULL;
  CSGTreeDialog = NULL;
  EvenlyScaleFlag = TRUE;
  ConeFunnelFlag = TRUE;
  SnapToAngleFlag = FALSE;
  SnapAngle = 15.0;

  InitPlugins();
#ifdef HAVE_GL
  ReallyHaveGL = FALSE;
  if (glXQueryExtension(GDK_DISPLAY(),NULL,NULL))
  {
/*    printf("It seems you actually have OpenGL installed, Good!\n");*/
    dpy = GDK_DISPLAY();
    vi = glXChooseVisual(dpy,DefaultScreen(dpy), attrib);
    if (vi)
    {
      GLVisual = gdkx_visual_get(vi->visualid);
      if (GLVisual)
        ReallyHaveGL = TRUE;
      XFree(vi);
    }
  } else
  {
    g_message("OpenGL wasn't found, too bad...\n");
  }
#endif
  CreateDefaultTexture();
  CreateToolOptionsDialog();
  ModeTool = MT_SELECT;
  ShowOptionsForTool(ModeTool);
  colors_hash = g_hash_table_new((GHashFunc)g_str_hash, (GCompareFunc)g_str_equal);
}

/*****************************************************************************
*  comparTri FIXME: this had nothing to do here. -> view.c
******************************************************************************/
int comparTri(const void *Tri1, const void *Tri2)
{
  return (((Triangle2D *)Tri2)->distance - ((Triangle2D *)Tri1)->distance);
}

typedef struct RasterLine
{
  int minX, maxX;
  double MinColor[3];
  double MaxColor[3];
  double MinDistance, MaxDistance;
} RasterLine;

/*****************************************************************************
*  DrawGouraudTriangleZBuffer FIXME: this had nothing to do here. -> view.c
******************************************************************************/
void DrawGouraudTriangleZBuffer(GtkWidget *View, double *Zbuffer,
                                Triangle2DNormCol *Tri, guchar *rgb_buf)
{
  RasterLine *ListLine;
  int minY, maxY;
  int deltaX, deltaY;
  int offx,offy;
  register int i,j;
  int x,y;
  register int bx;
  double distance;
  GdkColor Color;
  GdkColormap *colormap;
  GdkGC *gc;

  gc = gdk_gc_new(View->window);
  colormap = gdk_window_get_colormap(View->window);
  maxY = minY = Tri->P[0][1];
  if (maxY < Tri->P[1][1]) maxY = Tri->P[1][1];
  if (minY > Tri->P[1][1]) minY = Tri->P[1][1];
  if (maxY < Tri->P[2][1]) maxY = Tri->P[2][1];
  if (minY > Tri->P[2][1]) minY = Tri->P[2][1];
  if (minY < 0) return;
  if (maxY >= View->allocation.height) return;
  ListLine = g_malloc((maxY-minY+1)*sizeof(RasterLine));
  for(i=minY;i<=maxY;i++)
  {
    ListLine[i-minY].minX = 2<<29;
    ListLine[i-minY].maxX = -(2<<29);
  }
  /* First, Bresenham algorithm */
  /* Line P[0] -> P[1] */
  deltaX = abs(Tri->P[1][0]-Tri->P[0][0]);
  deltaY = abs(Tri->P[1][1]-Tri->P[0][1]);
  if (Tri->P[1][0] > Tri->P[0][0]) offx=1; else offx=-1;
  if (Tri->P[1][1] > Tri->P[0][1]) offy=1; else offy=-1;
  /* "Trace" du premier point */
  ListLine[Tri->P[0][1]-minY].minX = Tri->P[0][0];
  ListLine[Tri->P[0][1]-minY].maxX = Tri->P[0][0];
  ListLine[Tri->P[0][1]-minY].MinColor[0] = Tri->Color[0][0];
  ListLine[Tri->P[0][1]-minY].MinColor[1] = Tri->Color[0][1];
  ListLine[Tri->P[0][1]-minY].MinColor[2] = Tri->Color[0][2];
  ListLine[Tri->P[0][1]-minY].MaxColor[0] = Tri->Color[0][0];
  ListLine[Tri->P[0][1]-minY].MaxColor[1] = Tri->Color[0][1];
  ListLine[Tri->P[0][1]-minY].MaxColor[2] = Tri->Color[0][2];
  ListLine[Tri->P[0][1]-minY].MinDistance =
  ListLine[Tri->P[0][1]-minY].MaxDistance = Tri->Distance[0];
  if (deltaX > deltaY)
  { /* Plutot horizontal */
    x = Tri->P[0][0];
    y = Tri->P[0][1];
    bx = 2*deltaY-deltaX;
    while(x != Tri->P[1][0])
    {
      x+=offx;
      if (bx<0)
        bx+=2*deltaY;
      else
      {
        bx+=2*(deltaY-deltaX);
        y+=offy;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[0][0]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][0]-Tri->Color[0][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[0][1]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][1]-Tri->Color[0][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[0][2]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][2]-Tri->Color[0][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[0]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Distance[1]-Tri->Distance[0]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[0][0]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][0]-Tri->Color[0][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[0][1]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][1]-Tri->Color[0][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[0][2]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Color[1][2]-Tri->Color[0][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[0]+((double)x-Tri->P[0][0])/(Tri->P[1][0]-Tri->P[0][0])*(Tri->Distance[1]-Tri->Distance[0]);
      }
    }
  } else
  { /* Plutot vertical */
    x = Tri->P[0][0];
    y = Tri->P[0][1];
    bx = 2*deltaX-deltaY;
    while (y != Tri->P[1][1])
    {
      y+=offy;
      if (bx<0) bx+=2*deltaX;
      else
      {
        bx+=2*(deltaX-deltaY);
        x+=offx;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[0][0]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][0]-Tri->Color[0][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[0][1]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][1]-Tri->Color[0][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[0][2]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][2]-Tri->Color[0][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[0]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Distance[1]-Tri->Distance[0]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[0][0]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][0]-Tri->Color[0][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[0][1]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][1]-Tri->Color[0][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[0][2]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Color[1][2]-Tri->Color[0][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[0]+((double)y-Tri->P[0][1])/(Tri->P[1][1]-Tri->P[0][1])*(Tri->Distance[1]-Tri->Distance[0]);
      }
    }
  }
  /* Line P[1] -> P[2] */
  deltaX = abs(Tri->P[2][0]-Tri->P[1][0]);
  deltaY = abs(Tri->P[2][1]-Tri->P[1][1]);
  if (Tri->P[2][0] > Tri->P[1][0]) offx=1; else offx=-1;
  if (Tri->P[2][1] > Tri->P[1][1]) offy=1; else offy=-1;
  /* "Trace" du premier point */
  if (ListLine[Tri->P[1][1]-minY].minX > Tri->P[1][0])
  {
    ListLine[Tri->P[1][1]-minY].minX=Tri->P[1][0];
    ListLine[Tri->P[1][1]-minY].MinColor[0]=Tri->Color[1][0];
    ListLine[Tri->P[1][1]-minY].MinColor[1]=Tri->Color[1][1];
    ListLine[Tri->P[1][1]-minY].MinColor[2]=Tri->Color[1][2];
    ListLine[Tri->P[1][1]-minY].MinDistance=Tri->Distance[1];
  }
  if (ListLine[Tri->P[1][1]-minY].maxX < Tri->P[1][0])
  {
    ListLine[Tri->P[1][1]-minY].maxX=Tri->P[1][0];
    ListLine[Tri->P[1][1]-minY].MaxColor[0]=Tri->Color[1][0];
    ListLine[Tri->P[1][1]-minY].MaxColor[1]=Tri->Color[1][1];
    ListLine[Tri->P[1][1]-minY].MaxColor[2]=Tri->Color[1][2];
    ListLine[Tri->P[1][1]-minY].MaxDistance=Tri->Distance[1];
  }
  if (deltaX > deltaY)
  { /* Plutot horizontal */
    x = Tri->P[1][0];
    y = Tri->P[1][1];
    bx = 2*deltaY-deltaX;
    while(x != Tri->P[2][0])
    {
      x+=offx;
      if (bx<0) bx+=2*deltaY;
      else
      {
        bx+=2*(deltaY-deltaX);
        y+=offy;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[1][0]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][0]-Tri->Color[1][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[1][1]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][1]-Tri->Color[1][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[1][2]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][2]-Tri->Color[1][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[1]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Distance[2]-Tri->Distance[1]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[1][0]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][0]-Tri->Color[1][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[1][1]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][1]-Tri->Color[1][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[1][2]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Color[2][2]-Tri->Color[1][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[1]+((double)x-Tri->P[1][0])/(Tri->P[2][0]-Tri->P[1][0])*(Tri->Distance[2]-Tri->Distance[1]);
      }
    }
  } else
  { /* Plutot vertical */
    x = Tri->P[1][0];
    y = Tri->P[1][1];
    bx = 2*deltaX-deltaY;
    while(y != Tri->P[2][1])
    {
      y+=offy;
      if (bx<0) bx+=2*deltaX;
      else
      {
        bx+=2*(deltaX-deltaY);
        x+=offx;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[1][0]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][0]-Tri->Color[1][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[1][1]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][1]-Tri->Color[1][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[1][2]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][2]-Tri->Color[1][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[1]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Distance[2]-Tri->Distance[1]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[1][0]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][0]-Tri->Color[1][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[1][1]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][1]-Tri->Color[1][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[1][2]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Color[2][2]-Tri->Color[1][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[1]+((double)y-Tri->P[1][1])/(Tri->P[2][1]-Tri->P[1][1])*(Tri->Distance[2]-Tri->Distance[1]);
      }
    }
  }
  /* Line P[2] -> P[0] */
  deltaX = abs(Tri->P[0][0]-Tri->P[2][0]);
  deltaY = abs(Tri->P[0][1]-Tri->P[2][1]);
  if (Tri->P[0][0] > Tri->P[2][0]) offx=1; else offx=-1;
  if (Tri->P[0][1] > Tri->P[2][1]) offy=1; else offy=-1;
  /* "Trace" du premier point */
  if (ListLine[Tri->P[2][1]-minY].minX > Tri->P[2][0])
  {
    ListLine[Tri->P[2][1]-minY].minX=Tri->P[2][0];
    ListLine[Tri->P[2][1]-minY].MinColor[0]=Tri->Color[2][0];
    ListLine[Tri->P[2][1]-minY].MinColor[1]=Tri->Color[2][1];
    ListLine[Tri->P[2][1]-minY].MinColor[2]=Tri->Color[2][2];
    ListLine[Tri->P[2][1]-minY].MinDistance=Tri->Distance[2];
  }
  if (ListLine[Tri->P[2][1]-minY].maxX < Tri->P[2][0])
  {
    ListLine[Tri->P[2][1]-minY].maxX=Tri->P[2][0];
    ListLine[Tri->P[2][1]-minY].MaxColor[0]=Tri->Color[2][0];
    ListLine[Tri->P[2][1]-minY].MaxColor[1]=Tri->Color[2][1];
    ListLine[Tri->P[2][1]-minY].MaxColor[2]=Tri->Color[2][2];
    ListLine[Tri->P[2][1]-minY].MaxDistance=Tri->Distance[2];
  }
  if (deltaX > deltaY)
  { /* Plutot horizontal */
    x = Tri->P[2][0];
    y = Tri->P[2][1];
    bx = 2*deltaY-deltaX;
    while(x != Tri->P[0][0])
    {
      x+=offx;
      if (bx<0) bx+=2*deltaY;
      else
      {
        bx+=2*(deltaY-deltaX);
        y+=offy;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[2][0]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][0]-Tri->Color[2][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[2][1]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][1]-Tri->Color[2][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[2][2]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][2]-Tri->Color[2][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[2]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Distance[0]-Tri->Distance[2]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[2][0]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][0]-Tri->Color[2][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[2][1]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][1]-Tri->Color[2][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[2][2]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Color[0][2]-Tri->Color[2][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[2]+((double)x-Tri->P[2][0])/(Tri->P[0][0]-Tri->P[2][0])*(Tri->Distance[0]-Tri->Distance[2]);
      }
    }
  } else
  { /* Plutot vertical */
    x = Tri->P[2][0];
    y = Tri->P[2][1];
    bx = 2*deltaX-deltaY;
    while(y != Tri->P[0][1])
    {
      y+=offy;
      if (bx<0) bx+=2*deltaX;
      else
      {
        bx+=2*(deltaX-deltaY);
        x+=offx;
      }
      /* "Trace" du point */
      if (ListLine[y-minY].minX > x)
      {
        ListLine[y-minY].minX = x;
        ListLine[y-minY].MinColor[0]=Tri->Color[2][0]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][0]-Tri->Color[2][0]);
        ListLine[y-minY].MinColor[1]=Tri->Color[2][1]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][1]-Tri->Color[2][1]);
        ListLine[y-minY].MinColor[2]=Tri->Color[2][2]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][2]-Tri->Color[2][2]);
        ListLine[y-minY].MinDistance=Tri->Distance[2]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Distance[0]-Tri->Distance[2]);
      }
      if (ListLine[y-minY].maxX < x)
      {
        ListLine[y-minY].maxX = x;
        ListLine[y-minY].MaxColor[0]=Tri->Color[2][0]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][0]-Tri->Color[2][0]);
        ListLine[y-minY].MaxColor[1]=Tri->Color[2][1]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][1]-Tri->Color[2][1]);
        ListLine[y-minY].MaxColor[2]=Tri->Color[2][2]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Color[0][2]-Tri->Color[2][2]);
        ListLine[y-minY].MaxDistance=Tri->Distance[2]+((double)y-Tri->P[2][1])/(Tri->P[0][1]-Tri->P[2][1])*(Tri->Distance[0]-Tri->Distance[2]);
      }
    }
  }
  /* Then, Zbuffer drawing */
  for (i=minY;i<=maxY;i++)
  {
    if ((ListLine[i-minY].minX < View->allocation.width) && (ListLine[i-minY].maxX >= 0))
      for (j=ListLine[i-minY].minX ; j<=ListLine[i-minY].maxX ; j++)
      {
        if ((i>=0) && (i<View->allocation.height) && 
            (j>=0) && (j<View->allocation.width))
        {
          if (ListLine[i-minY].maxX != ListLine[i-minY].minX)
            distance = ListLine[i-minY].MinDistance + (j-ListLine[i-minY].minX)/(ListLine[i-minY].maxX-ListLine[i-minY].minX)*(ListLine[i-minY].MaxDistance-ListLine[i-minY].MinDistance);
          else
            distance = ListLine[i-minY].MinDistance;
          if (Zbuffer[i*View->allocation.width+j] > distance)
          {
            Zbuffer[i*View->allocation.width+j] = distance;
            rgb_buf[(i*View->allocation.width+j)*3+0] = 
                (guchar)(255.0*(ListLine[i-minY].MinColor[0]+((double)j-(double)(ListLine[i-minY].minX))/((double)(ListLine[i-minY].maxX)-(double)(ListLine[i-minY].minX))*(ListLine[i-minY].MaxColor[0]-ListLine[i-minY].MinColor[0])));
            rgb_buf[(i*View->allocation.width+j)*3+1] = 
                (guchar)(255.0*(ListLine[i-minY].MinColor[1]+((double)j-(double)(ListLine[i-minY].minX))/((double)(ListLine[i-minY].maxX)-(double)(ListLine[i-minY].minX))*(ListLine[i-minY].MaxColor[1]-ListLine[i-minY].MinColor[1])));
            rgb_buf[(i*View->allocation.width+j)*3+2] = 
                (guchar)(255.0*(ListLine[i-minY].MinColor[2]+((double)j-(double)(ListLine[i-minY].minX))/((double)(ListLine[i-minY].maxX)-(double)(ListLine[i-minY].minX))*(ListLine[i-minY].MaxColor[2]-ListLine[i-minY].MinColor[2])));
          }
        }
      }
  }
  g_free(ListLine);
  gdk_gc_destroy(gc);
}

/*****************************************************************************
*  NewFrameFunc
******************************************************************************/
void NewFrameFunc(void)
{
  NewFrame(_("Untitled"));
}

/*****************************************************************************
*  NewViewOK
******************************************************************************/
FrameStruct *NewViewFrame;
void NewViewOK(GtkObject *Dialog)
{
  GtkToggleButton **Data;

  if (NewViewFrame == NULL)
  {
    NewFrame(_("Untitled"));
    NewViewFrame = FirstFrame;
  }
  Data = gtk_object_get_data(Dialog, "WidgetData");
  if (gtk_toggle_button_get_active(Data[0]))
    NewView(NewViewFrame, ORTHO_XY_CAMERA);
  if (gtk_toggle_button_get_active(Data[1]))
    NewView(NewViewFrame, ORTHO_XZ_CAMERA);
  if (gtk_toggle_button_get_active(Data[2]))
    NewView(NewViewFrame, ORTHO_ZY_CAMERA);
  if (gtk_toggle_button_get_active(Data[3]))
    NewViewCamera(NewViewFrame);
}

/*****************************************************************************
*  NewViewSetFrame
******************************************************************************/
void NewViewSetFrame(GtkWidget *dummy, FrameStruct *LocalFrame)
{
  NewViewFrame = LocalFrame;
}

/*****************************************************************************
*  NewViewFunc
******************************************************************************/
void NewViewFunc(GtkWidget *dummy, FrameStruct *LocalFrame)
{
  GtkWidget *Dialog;
  GtkWidget *Button;
  GtkWidget *OptionMenu;
  GtkWidget *FrameMenu;
  GtkWidget *MenuItem;
  GSList *MenuGroup;
  FrameStruct *TmpFrame;
  GtkWidget *Separator;
  GtkWidget *Option;
  GSList *OptionGroup;
  GtkWidget **Data;
 
  Dialog = gtk_dialog_new();
  Data = g_malloc(4*sizeof(GtkWidget*));
  gtk_object_set_data(GTK_OBJECT(Dialog), "WidgetData", Data);
  gtk_window_set_title(GTK_WINDOW(Dialog), _("New View"));

  /* The Ok Button */
  Button = gtk_button_new_with_label(_("OK"));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)NewViewOK,
                            GTK_OBJECT(Dialog));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE,
                     TRUE, 0);
  gtk_widget_grab_default(Button);
  /* The Close Button */
  Button = gtk_button_new_with_label(_("Close"));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button);

  /* The Frame Option Menu */
  OptionMenu = gtk_option_menu_new();
  FrameMenu = gtk_menu_new();
  MenuGroup = NULL;
  NewViewFrame = NULL;
  MenuItem = gtk_radio_menu_item_new_with_label(MenuGroup, _("New Frame"));
  MenuGroup = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(MenuItem));
  if (LocalFrame == NULL)
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(MenuItem), TRUE);
  else
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(MenuItem), FALSE);
  gtk_signal_connect(GTK_OBJECT(MenuItem), "activate",
                     GTK_SIGNAL_FUNC(NewViewSetFrame), NULL);
  gtk_menu_append(GTK_MENU(FrameMenu), MenuItem);
  gtk_widget_show(MenuItem);
  for (TmpFrame = FirstFrame ; TmpFrame ; TmpFrame = TmpFrame->Next)
  {
    MenuItem = gtk_radio_menu_item_new_with_label(MenuGroup, TmpFrame->Name);
    MenuGroup = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(MenuItem));
    if (TmpFrame == LocalFrame)
    {
      NewViewFrame = LocalFrame;
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(MenuItem), TRUE);
    } else
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(MenuItem), FALSE);
    gtk_signal_connect(GTK_OBJECT(MenuItem), "activate",
                       GTK_SIGNAL_FUNC(NewViewSetFrame),
                       TmpFrame);
    gtk_menu_append(GTK_MENU(FrameMenu), MenuItem);
    gtk_widget_show(MenuItem);
  }
  gtk_option_menu_set_menu(GTK_OPTION_MENU(OptionMenu), FrameMenu);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), OptionMenu, FALSE,
                     FALSE, 0);
  /* Separator */
  Separator = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Separator, FALSE,
                     FALSE, 5);
  /* View Options */ 
  OptionGroup = NULL;
  Option = gtk_radio_button_new_with_label(OptionGroup, _("View X-Y"));
  Data[0] = Option;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Option), TRUE);
  OptionGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Option));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Option, FALSE, FALSE,
                     0);
  Option = gtk_radio_button_new_with_label(OptionGroup, _("View X-Z"));
  Data[1] = Option;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Option), FALSE);
  OptionGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Option));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Option, FALSE, FALSE,
                     0);
  Option = gtk_radio_button_new_with_label(OptionGroup, _("View Z-Y"));
  Data[2] = Option;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Option), FALSE);
  OptionGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Option));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Option, FALSE, FALSE,
                     0);
  Option = gtk_radio_button_new_with_label(OptionGroup, _("Camera View"));
  Data[3] = Option;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Option), FALSE);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), Option, FALSE, FALSE,
                     0);

  gtk_widget_show_all(Dialog);
}

/*****************************************************************************
*  LoadPovFunc
******************************************************************************/
void LoadPovFunc(void)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Load Pov File"));
  /* Connect the ok_button to the LoadPov function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)LoadPov, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

#ifdef USE_LIB3DS
/*****************************************************************************
*  Load3DSFunc
******************************************************************************/
void Load3DSFunc(void)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Load 3DStudio 3DS File"));
  /* Connect the ok_button to the Load3DS function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)giram_load_3ds, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  Insert3DSFunc
******************************************************************************/
void Insert3DSFunc(void)
{
  GtkWidget *FileSelector;
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Insert 3DS objects"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", UserData->FrameView);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)giram_insert_3ds, FileSelector);

  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}
#endif

/*****************************************************************************
*  LoadDXFFunc
******************************************************************************/
void LoadDXFFunc(void)
{
  GtkWidget *FileSelector;

  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Load AutoCAD DXF File"));
  /* Connect the ok_button to the LoadPov function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)LoadDXF, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  InsertDXFFunc
******************************************************************************/
void InsertDXFFunc(void)
{
  GtkWidget *FileSelector;
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Insert DXF objects"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", UserData->FrameView);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)InsertDXF, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  SavePovFunc
******************************************************************************/
void SavePovFunc(void)
{
  GtkWidget *FileSelector;
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save Pov File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", UserData->FrameView);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SavePov, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  SaveDXFFunc
******************************************************************************/
void SaveDXFFunc(void)
{
  GtkWidget *FileSelector;
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save DXF File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", UserData->FrameView);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SaveDXF, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  SaveSelectedDXFFunc
******************************************************************************/
void SaveSelectedDXFFunc(void)
{
  GtkWidget *FileSelector;
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save DXF File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", UserData->FrameView);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SaveSelectedDXF, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

#ifdef WITH_S3D
/*****************************************************************************
*  SaveS3DFunc
******************************************************************************/
void SaveS3DFunc(void)
{
  GtkWidget *FileSelector;
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save S3D File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", UserData->FrameView);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SaveS3D, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}

/*****************************************************************************
*  SaveSelectedS3DFunc
******************************************************************************/
void SaveSelectedS3DFunc(void)
{
  GtkWidget *FileSelector;
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  /* Create a new file selection widget */
  FileSelector = gtk_file_selection_new(_("Save S3D File"));
  gtk_object_set_data(GTK_OBJECT(FileSelector), "frame", UserData->FrameView);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                     "clicked", (GtkSignalFunc)SaveSelectedS3D, FileSelector);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  /* Connect the ok_button to destroy the widget */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(FileSelector)->ok_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(FileSelector));
  gtk_widget_show(FileSelector);
}
#endif

/*****************************************************************************
*  ExitFunc
******************************************************************************/
void ExitFunc(void)
{
  gtk_main_quit();
}

/*****************************************************************************
*  show_color_list
******************************************************************************/
void show_color_list(void)
{
  GtkWidget *gcl;

  gcl = giram_color_list_new(colors_hash);
  
  gtk_widget_show(gcl);
}

/* LOCAL data */
static gint    giram_argc = 0;
static gchar **giram_argv = NULL;

/*****************************************************************************
*  main
******************************************************************************/
int main(int argc, char *argv[])
{
  GtkWidget *ToolsWindow;
  GtkWidget *Box;
  GtkWidget *menu_bar;
  GtkWidget *HandleBox;
  GtkAccelGroup *AccelGroup;
  GtkItemFactory *Factory;

static GtkItemFactoryEntry MainMenu[] =
{
  {"/_File",                   NULL, NULL,                      0, "<Branch>"},
  {"/File/New _Frame",         NULL, NewFrameFunc,              0, "<Item>"},
  {"/File/New _View",          NULL, NewViewFunc,               0, "<Item>"},
  {"/File/Load _Pov",          NULL, LoadPovFunc,               0, "<Item>"},
  {"/File/Load AutoCAD _DXF",  NULL, LoadDXFFunc,               0, "<Item>"},
#ifdef USE_LIB3DS
  {"/File/Load 3DStudio _3DS", NULL, Load3DSFunc,               0, "<Item>"},
#endif
  {"/File/_About...",          NULL, about_dialog_create,       0, "<Item>"},
  {"/File/_Help...",           NULL, ShowHelpDialog,            0, "<Item>"},
  {"/File/_Preferences...",    NULL, preferences_dialog_create, 0, "<Item>"},
  {"/File/Dum",                NULL, NULL,                      0, "<Separator>"},
  {"/File/_Quit",              NULL, ExitFunc,                  0, "<Item>"},
  {"/_Dialogs",                NULL, NULL,                      0, "<Branch>"},
  {"/Dialogs/_CSG Tree",       NULL, ShowCSGTree,               0, "<Item>"},
  {"/Dialogs/_Tool Options",   NULL, ShowToolOptionsDialog,     0, "<Item>"},
  {"/Dialogs/Colors",          NULL, show_color_list,           0, "<Item>"}
};

static guint NbEntries = sizeof(MainMenu) /
                         sizeof(GtkItemFactoryEntry);
  GtkWidget *ToolsTable;
  gboolean show_version = FALSE;
  gboolean show_help = FALSE;
  gint i;

#ifdef HAVE_PUTENV
  gchar *display_env;
#endif

  g_atexit(g_mem_profile);
    
  /* Initialize variables */

  prog_name = argv[0];

  /* Initialize i18n support */
  INIT_LOCALE("giram");

#ifdef ENABLE_NLS
  bindtextdomain ("giram-libgiram", LOCALEDIR);
#endif

  gtk_init(&argc, &argv);
  gdk_rgb_init();

  setlocale(LC_NUMERIC, "C");  /* gtk seems to zap this during init.. */

#ifdef HAVE_PUTENV
  display_env = g_strconcat("DISPLAY=", gdk_get_display(), NULL);
  putenv(display_env);
#endif
  
  for (i=1 ; i<argc ; i++)
  {
    if (strcmp(argv[i], "--system-giramrc") == 0)
    {
      argv[i] = NULL;
      if (argc <= ++i)
      {
        show_help = TRUE;
      } else
      {
        alternate_system_giramrc = argv[i];
        argv[i] = NULL;
      }
    } else if ((strcmp(argv[i], "--giramrc") == 0) ||
               (strcmp(argv[i], "-g") == 0))
    {
      argv[i] = NULL;
      if (argc <= ++i)
      {
        show_help = TRUE;
      } else
      {
        alternate_giramrc = argv[i];
        argv[i] = NULL;
      }
    } else if ((strcmp(argv[i], "--help") == 0) ||
               (strcmp(argv[i], "-h") == 0))
    {
      show_help = TRUE;
      argv[i] = NULL;
    } else if ((strcmp(argv[i], "--version") == 0) ||
               (strcmp (argv[i], "-v") == 0))
    {
      show_version = TRUE;
      argv[i] = NULL;
    } else if (strcmp(argv[i], "--verbose") == 0)
    {
      be_verbose = TRUE;
      argv[i] = NULL;
    } else if (strcmp(argv[i], "--debug-handlers") == 0)
    {
      use_debug_handler = TRUE;
      argv[i] = NULL;
    } else if ((strcmp(argv[i], "--console-messages") == 0) ||
               (strcmp(argv[i], "-c") == 0))
    {
      console_messages = TRUE;
      argv[i] = NULL;
    } else if (strcmp(argv[i], "--enable-stack-trace") == 0)
    {
      argv[i] = NULL;
      if (argc <= ++i)
      {
        show_help = TRUE;
      } else
      {
        if (!strcmp(argv[i], "never"))
          stack_trace_mode = STACK_TRACE_NEVER;
        else if (!strcmp(argv[i], "query"))
          stack_trace_mode = STACK_TRACE_QUERY;
        else if (!strcmp(argv[i], "always"))
          stack_trace_mode = STACK_TRACE_ALWAYS;
        else
          show_help = TRUE;

        argv[i] = NULL;
      }
    }
    /*
     * ANYTHING ELSE starting with a '-' is an error.
     */
    else if (argv[i][0] == '-')
    {
      g_print(_("\nInvalid option.\n"));
      show_help = TRUE;
    }
  }

  if (show_version)
    g_print("%s %s\n", _("Giram version"), VERSION);

  if (show_help)
  {
    g_print(_("\nUsage: %s [option ... ] [file ... ]\n\n"), argv[0]);
    g_print(_("Options:\n"));
    g_print(_("  -c, --console-messages   Display warnings to console instead of a dialog box.\n"));
    g_print(_("  -g, --giramrc <giramrc>  Use an alternate giramrc file.\n"));
    g_print(_("  -h, --help               Output this help.\n"));
    g_print(_("  -v, --version            Output version information.\n"));
    g_print(_("  --verbose                Show startup messages.\n"));
    g_print(_("  --debug-handlers         Enable non-fatal debugging signal handlers.\n"));
    g_print(_("  --display <display>      Use the designated X display.\n"));
    g_print(_("  --system-gimprc <gimprc> Use an alternate system gimprc file.\n"));
    g_print("  --enable-stack-trace <never | query | always>\n");
    g_print(_("                           Debugging mode for fatal signals.\n\n"));
  }

  if (show_version || show_help)
  {
    exit(0);
  }

  g_log_set_handler("Giram", G_LOG_LEVEL_MESSAGE,
                    giram_message_func, NULL);

  /* Handle fatal signals */

  giram_signal_private(SIGHUP,  giram_sigfatal_handler, 0);
  giram_signal_private(SIGINT,  giram_sigfatal_handler, 0);
  giram_signal_private(SIGQUIT, giram_sigfatal_handler, 0);
  giram_signal_private(SIGABRT, giram_sigfatal_handler, 0);
  giram_signal_private(SIGBUS,  giram_sigfatal_handler, 0);
  giram_signal_private(SIGSEGV, giram_sigfatal_handler, 0);
  giram_signal_private(SIGTERM, giram_sigfatal_handler, 0);
  giram_signal_private(SIGFPE,  giram_sigfatal_handler, 0);

  /* Ignore SIGPIPE because plug_in.c handles broken pipes (well, it will) */

  giram_signal_private(SIGPIPE, SIG_IGN, 0);

  /* Collect dead children */

  giram_signal_private(SIGCHLD, giram_sigchld_handler, SA_RESTART);

  g_log_set_handler(NULL,
                    G_LOG_LEVEL_ERROR | G_LOG_FLAG_FATAL,
                    giram_error_handler, NULL);

  /* Keep the command line arguments--for use in gimp_init */
  giram_argc = argc - 1;
  giram_argv = argv + 1;

  /* Check the user_installation */
  user_install_verify(init);

  giramInit();

  g_print("povray include path: '%s'\n'", pov_include_path);
  
  ToolsWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(ToolsWindow), _("Giram Tools"));
  /* Here we just set a handler for delete_event that immediately
   * exits GTK. */
  gtk_signal_connect(GTK_OBJECT(ToolsWindow), "delete_event",
                     GTK_SIGNAL_FUNC(ExitFunc), NULL);
  gtk_container_border_width(GTK_CONTAINER(ToolsWindow), 0);
  gtk_window_set_policy(GTK_WINDOW(ToolsWindow), FALSE, FALSE, TRUE);

  Box = gtk_vbox_new(FALSE, 0);
  /* Put the box into the main window. */
  gtk_container_add(GTK_CONTAINER(ToolsWindow), Box);
  

  HandleBox = gtk_handle_box_new();
  gtk_box_pack_start(GTK_BOX(Box), HandleBox, FALSE, FALSE, 2);
  gtk_widget_show(HandleBox);

  /* New Keyboard accel group */
  AccelGroup = gtk_accel_group_new();

  /* Creation of the item factory */
  Factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<Giram>", AccelGroup);

  gtk_item_factory_create_items(Factory, NbEntries, MainMenu, NULL);
  gtk_accel_group_attach(AccelGroup, GTK_OBJECT(ToolsWindow));

  /* We get the menubar from the factory */
  menu_bar = gtk_item_factory_get_widget(Factory,"<Giram>");
  gtk_widget_show(menu_bar);

  gtk_container_add(GTK_CONTAINER(HandleBox), menu_bar);

    
  /* The Tool Table */
  ToolsTable = CreateTools(Box);

  CreateViewMenu();
    
  gtk_widget_show(ToolsTable);
  gtk_widget_show(Box);
  gtk_widget_show(ToolsWindow);
  gtk_main();
  return 0;
}

static void init(void)
{
  /*  Continue initializing  */
  giram_init(giram_argc, giram_argv);
}

static void giram_error_handler(const gchar    *domain,
                                GLogLevelFlags  flags,
                                const gchar    *msg,
                                gpointer        user_data)
{
  giram_fatal_error("%s", msg);
}

/* giram core signal handler for fatal signals */
static void giram_sigfatal_handler(gint sig_num)
{
  switch (sig_num)
  {
    case SIGHUP:
    case SIGINT:
    case SIGQUIT:
    case SIGABRT:
    case SIGTERM:
      giram_terminate(g_strsignal(sig_num));
      break;
    case SIGBUS:
    case SIGSEGV:
    case SIGFPE:
    default:
      giram_fatal_error(g_strsignal(sig_num));
      break;
  }
}

/* giram core signal handler for death-of-child signals */

static void giram_sigchld_handler(gint sig_num)
{
  gint pid;
  gint status;

  while (TRUE)
  {
    pid = waitpid(WAIT_ANY, &status, WNOHANG);

    if (pid <= 0)
      break;
  }
}

