/* help_dialog.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include "giram.h"
#include "help_dialog.h"

#include "giramintl.h"

/*****************************************************************************
*  ShowHelpDialog
******************************************************************************/
void ShowHelpDialog(void)
{
  static GtkWidget *HelpDialog = NULL;
  GtkWidget *Button;
  GtkWidget *ScrolledWindow;
  GtkWidget *Text;
  FILE *HelpFile;

  if (HelpDialog)
  {
    if (!GTK_WIDGET_VISIBLE(HelpDialog))
    {
      gtk_widget_show(HelpDialog);
    }
    else
    {
      gdk_window_raise(HelpDialog->window);
    }
    return;
  }

  HelpDialog = gtk_dialog_new();
  gtk_window_set_wmclass(GTK_WINDOW(HelpDialog), "help_dialog", "Giram");
  gtk_window_set_title(GTK_WINDOW(HelpDialog), _("Help Giram"));
  gtk_window_set_policy(GTK_WINDOW(HelpDialog), TRUE, TRUE, FALSE);
  gtk_widget_set_usize(HelpDialog, 500, 500);

  gtk_signal_connect(GTK_OBJECT(HelpDialog), "destroy",
                     GTK_SIGNAL_FUNC(gtk_widget_destroyed),
                     &HelpDialog);
  /* The Close Button */ 
  Button = gtk_button_new_with_label(_("Close"));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(HelpDialog)->action_area),
                              Button);
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            GTK_SIGNAL_FUNC(gtk_widget_destroy),
                            GTK_OBJECT(HelpDialog));
  gtk_widget_grab_default(Button);
  /* The Scrolled window */
  ScrolledWindow = gtk_scrolled_window_new(NULL, NULL);
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(HelpDialog)->vbox),
                              ScrolledWindow);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(ScrolledWindow),
                                 GTK_POLICY_NEVER,
                                 GTK_POLICY_ALWAYS);
  /* The text widget */
  Text = gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(Text), FALSE);
  gtk_container_add(GTK_CONTAINER(ScrolledWindow), Text);
  gtk_widget_grab_focus(Text);

  gtk_text_freeze(GTK_TEXT(Text));
  HelpFile = fopen(HELPFILE, "r");
  if (HelpFile)
  {
    char *buffer;
    int nbytes_read, nbytes_alloc;

    nbytes_read = 0;
    nbytes_alloc = 1024;
    buffer = g_new(char, nbytes_alloc);
    while (1)
    {
      int len;

      if (nbytes_alloc < nbytes_read+1024)
      {
        nbytes_alloc *= 2;
        buffer = g_realloc(buffer, nbytes_alloc);
      }
      len = fread(buffer+nbytes_read, 1, 1024, HelpFile);
      nbytes_read += len;
      if (len < 1024)
        break;
    }
    gtk_text_insert(GTK_TEXT(Text), NULL, NULL, NULL, buffer, nbytes_read);
    g_free(buffer);
    fclose (HelpFile);
  } else
  {
    char *buffer = "The help file was not found\n\n"
                   "Please do a 'make install' from the Giram directory\n";
    gtk_text_insert(GTK_TEXT(Text), NULL, NULL, NULL, buffer, strlen(buffer));
  }
  gtk_text_thaw(GTK_TEXT(Text));
  
  gtk_widget_show_all(HelpDialog);
}

