/* object.h
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __GIRAM_OBJECT_H__
#define __GIRAM_OBJECT_H__

typedef struct _ObjectStruct ObjectStruct;
typedef struct _GiramObjectClass GiramObjectClass;

/* THE structure defining an object */
struct _ObjectStruct
{
  guint Type; /* The shape */
  GiramObjectClass *klass;
  guint refcount;
  char *Name;
  TriangleListStruct *FirstTriangle; /* The first triangle of the list */
/*  ObjectStruct *Next;  */              /* A pointer to the next object of the frame */
                                                 /* FIXME: we should use a GList there */
  TransformStruct *Trans;            /* The transformation matrix */
  TextureStruct *Texture;
  gboolean Inverse;                  /* Is the object inverted ? (CSG) */
  gboolean NoShadow;
  gboolean Hollow;
  gboolean Selected;               /* Is the object currently selected? */
  gboolean Visible;                /* Is the object currently visible? */
  GtkCTreeNode *CSGTreeNode;       /* A pointer to the node of the object in the CSGTree:
                                      looks very ugly... */
  gint DetailLevel;                /* NCU: should be used to let the user choose how
                                      detailed is the
                                      mesh representation of the objects.
                                      This also could be used for the
                                      object->mesh converter. */
  guint Id;
};

struct _GiramObjectClass
{
  void (* translate)(ObjectStruct *object, Vector vect);
  void (* rotate)(ObjectStruct *object, Vector vect);
  void (* scale)(ObjectStruct *object, Vector vect);
  void (* build_triangle_mesh)(ObjectStruct *object);
  gboolean (* inside)(ObjectStruct *object, double x, double y, double z);
  gboolean (* is_intersection)(ObjectStruct *object, Vector origin, Vector direction);
  gboolean (* click_in_XY)(ObjectStruct *object, double x, double y);
  gboolean (* click_in_XZ)(ObjectStruct *object, double x, double z);
  gboolean (* click_in_ZY)(ObjectStruct *object, double z, double y);
  gboolean (* find_intersection_segment)(ObjectStruct *object,
                                         Vector in_point, Vector out_point,
                                         Vector inter_point, Vector inter_norm);
};

void InitObject(ObjectStruct *Object);
void giram_object_translate(ObjectStruct *object, Vector vect);
void giram_object_rotate(ObjectStruct *object, Vector vect);
void giram_object_scale(ObjectStruct *object, Vector vect);
void giram_object_build_triangle_mesh(ObjectStruct *object);
gboolean giram_object_inside(ObjectStruct *object, double x, double y, double z);
gboolean giram_object_is_intersection(ObjectStruct *object, Vector origin, Vector direction);
gboolean giram_object_click_in_XY(ObjectStruct *object, double x, double y);
gboolean giram_object_click_in_XZ(ObjectStruct *object, double x, double z);
gboolean giram_object_click_in_ZY(ObjectStruct *object, double z, double y);
gboolean giram_object_find_intersection_segment(ObjectStruct *object,
                                                Vector in_point, Vector out_point,
                                                Vector inter_point, Vector inter_norm);
int PDBTransformObject(int *FrameId, guint *ObjectId, TransformStruct *Trans);
GiramObjectClass *giram_object_class_new(void);
#endif /* __GIRAM_OBJECT_H__ */
