/* snap.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include "giram.h"
#include "utils.h"
#include "snap.h"

#include "giramintl.h"

GtkWidget *SnapDialog = NULL;

/****************************************************************************
*  SnapDialogCancel
*****************************************************************************/
void SnapDialogCancel(GtkWidget *Button, GtkWidget **Data)
{
  g_free(Data);
  gtk_widget_destroy(SnapDialog);
}

/****************************************************************************
*  SnapDialogOK
*****************************************************************************/
void SnapDialogOK(GtkWidget *Button, GtkWidget **Data)
{
  ViewUserDataStruct *UserData;
  FrameStruct        *Frame;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;

  UserData = gtk_object_get_data(GTK_OBJECT(Data[0]), "ViewData");
  Frame = UserData->FrameView;
  Frame->SnapGrid[0] = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[1]));
  Frame->SnapGrid[1] = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[2]));
  Frame->SnapGrid[2] = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[3]));
  Frame->SnapToGrid = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(Data[4]));
  Frame->ShowGrid = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(Data[5]));
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  g_free(Data);
  gtk_widget_destroy(SnapDialog);
}

/****************************************************************************
*  SnapX10
* Multiply all grid values by 10
*****************************************************************************/
void SnapX10(GtkWidget *Button, GtkWidget **Data)
{
  ViewUserDataStruct *UserData;
  FrameStruct *Frame;
  FrameViewStruct *TmpView;

  UserData = gtk_object_get_data(GTK_OBJECT(Data[0]), "ViewData");
  Frame = UserData->FrameView;
  /* Get the grid spin button values and multiply by 10 */
  Frame->SnapGrid[0] = 10.0*gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[1]));
  Frame->SnapGrid[1] = 10.0*gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[2]));
  Frame->SnapGrid[2] = 10.0*gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[3]));
  /* Now set the values back */
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(Data[1]),Frame->SnapGrid[0]);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(Data[2]),Frame->SnapGrid[1]);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(Data[3]),Frame->SnapGrid[2]);
}

/****************************************************************************
*  SnapDiv10
* Divide all grid values by 10
*****************************************************************************/
void SnapDiv10(GtkWidget *Button, GtkWidget **Data)
{
  ViewUserDataStruct *UserData;
  FrameStruct *Frame;
  FrameViewStruct *TmpView;

  UserData = gtk_object_get_data(GTK_OBJECT(Data[0]), "ViewData");
  Frame = UserData->FrameView;
  /* Get the grid spin button values and divide by 10 */
  Frame->SnapGrid[0] = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[1]))/10.0;
  Frame->SnapGrid[1] = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[2]))/10.0;
  Frame->SnapGrid[2] = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(Data[3]))/10.0;
  /* Now set the values back */
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(Data[1]),Frame->SnapGrid[0]);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(Data[2]),Frame->SnapGrid[1]);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(Data[3]),Frame->SnapGrid[2]);
}

/****************************************************************************
*  SnapDialog
*****************************************************************************/
void SnapDialogShow(void)
{
  GtkWidget *Table;
  GtkWidget *Label;
  GtkWidget *Spin;
  GtkWidget *CheckButton;
  GtkWidget *Button;
  GtkAdjustment *adj;
  ViewUserDataStruct *UserData;
  FrameStruct *Frame;
  GtkWidget **Data;

  UserData = get_view_user_data();
  Frame = UserData->FrameView;
  if (SnapDialog)
  { /* If it's already on screen, juste raise it... */
    gdk_window_raise(SnapDialog->window);
  } else
  {
    Data = g_malloc(6*sizeof(GtkWidget*));
    Data[0] = UserData->DrawingArea;
    SnapDialog = gtk_dialog_new();
    /* Put SnapDialog Back to NULL when it gets destroyed */
    gtk_signal_connect(GTK_OBJECT(SnapDialog), "destroy",
                       GTK_SIGNAL_FUNC(gtk_widget_destroyed),
                       &SnapDialog);
    Table = gtk_table_new(7, 2, FALSE);
    gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(SnapDialog)->vbox), Table);
    /* The X spin button */
    Label = gtk_label_new(_("X Snap Size: "));
    gtk_table_attach_defaults(GTK_TABLE(Table), Label, 0,1, 0,1);
    adj = (GtkAdjustment *)gtk_adjustment_new(Frame->SnapGrid[0], 
                                              0.00, 50.00, 0.05, 0.25, 0.0);
    Spin = gtk_spin_button_new(adj, 0.2, 2);
    gtk_table_attach_defaults(GTK_TABLE(Table), Spin, 1,2, 0,1);
    Data[1] = Spin;
    /* The Y spin button */
    Label = gtk_label_new(_("Y Snap Size: "));
    gtk_table_attach_defaults(GTK_TABLE(Table), Label, 0,1, 1,2);
    adj = (GtkAdjustment *)gtk_adjustment_new(Frame->SnapGrid[1],
                                              0.00, 50.00, 0.05, 0.25, 0.0);
    Spin = gtk_spin_button_new(adj, 0.2, 2);
    gtk_table_attach_defaults(GTK_TABLE(Table), Spin, 1,2, 1,2);
    Data[2] = Spin;
    /* The Z spin button */
    Label = gtk_label_new(_("Z Snap Size: "));
    gtk_table_attach_defaults(GTK_TABLE(Table), Label, 0,1, 2,3);
    adj = (GtkAdjustment *)gtk_adjustment_new(Frame->SnapGrid[2],
                                              0.00, 50.00, 0.05, 0.25, 0.0);
    Spin = gtk_spin_button_new(adj, 0.2, 2);
    gtk_table_attach_defaults(GTK_TABLE(Table), Spin, 1,2, 2,3);
    Data[3] = Spin;
    /* The Snap to Grid Check Button */
    CheckButton = gtk_check_button_new_with_label(_("Snap to Grid"));
    gtk_table_attach_defaults(GTK_TABLE(Table), CheckButton, 0,2, 3,4);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(CheckButton),
                                 Frame->SnapToGrid);
    Data[4] = CheckButton;
    /* The Show Grid Check Button */
    CheckButton = gtk_check_button_new_with_label(_("Show Grid"));
    gtk_table_attach_defaults(GTK_TABLE(Table), CheckButton, 0,2, 4,5);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(CheckButton),
                                 Frame->ShowGrid);
    Data[5] = CheckButton;
    /* The OK Button */
    Button = gtk_button_new_with_label(_("OK"));
    gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(SnapDialog)->action_area),
                                Button);
    GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
    gtk_signal_connect(GTK_OBJECT(Button), "clicked",
                       (GtkSignalFunc)SnapDialogOK, (gpointer)Data);
    /* The Cancel Button */
    Button = gtk_button_new_with_label(_("Cancel"));
    gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(SnapDialog)->action_area),
                                Button);
    GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(Button);
    gtk_signal_connect(GTK_OBJECT(Button), "clicked",
                       (GtkSignalFunc)SnapDialogCancel, (gpointer)Data);
    
    Button = gtk_button_new_with_label(_("x10"));
    gtk_table_attach_defaults(GTK_TABLE(Table), Button, 0,1, 6,7);
    gtk_signal_connect(GTK_OBJECT(Button), "clicked",
                       (GtkSignalFunc)SnapX10, (gpointer)Data);
    Button = gtk_button_new_with_label(_("/10"));
    gtk_table_attach_defaults(GTK_TABLE(Table), Button, 1,2, 6,7);
    gtk_signal_connect(GTK_OBJECT(Button), "clicked",
                       (GtkSignalFunc)SnapDiv10, (gpointer)Data);

    /* Show the whole thing */
    gtk_widget_show_all(SnapDialog);
  }
}

/****************************************************************************
*  MouseToReal
*****************************************************************************/
void MouseToReal(double mx, double my, GtkWidget *Area, double *px, double *py)
{
  ViewUserDataStruct *UserData;
  double Zoom, Xoff, Yoff;
  double xs = 0.0, ys = 0.0;

  UserData = gtk_object_get_data(GTK_OBJECT(Area), "ViewData");
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;

  *px = Xoff+(mx-Area->allocation.width /2)/Zoom;
  *py = Yoff-(my-Area->allocation.height/2)/Zoom;
  if (!(UserData->FrameView->SnapToGrid))
    return;
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      xs = UserData->FrameView->SnapGrid[0];
      ys = UserData->FrameView->SnapGrid[1];
      break;
    case ORTHO_XZ_CAMERA:
      xs = UserData->FrameView->SnapGrid[0];
      ys = UserData->FrameView->SnapGrid[2];
      break;
    case ORTHO_ZY_CAMERA:
      xs = UserData->FrameView->SnapGrid[2];
      ys = UserData->FrameView->SnapGrid[1];
      break;
  }
  if (xs != 0.0)
    *px = xs*(floor((*px)/xs+0.5*xs));
  if (ys != 0.0)
    *py = ys*(floor((*py)/ys+0.5*xs));
}

