/* sor.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "csgtree.h"
#include "object.h"
#include "utils.h"
#include "sor.h"

#include "giramintl.h"

static void giram_sor_build_triangle_mesh(ObjectStruct *sor);
static gboolean giram_sor_inside(ObjectStruct *sor, double x, double y, double z);
static gboolean giram_sor_find_intersection_segment(ObjectStruct *sor,
                                                    Vector in_point, Vector out_point,
                                                    Vector inter_point, Vector inter_norm);

/*****************************************************************************
*  giram_sor_build_triangle_mesh FIXME: needs to be smoother
******************************************************************************/
static void giram_sor_build_triangle_mesh(ObjectStruct *sor)
{ /* XXX: clipped_by */
  Vector p1, p2, p3, p4;
  Vector n1, n2, n3, n4;
  guint i, j, DetailLevel;
  SorStruct *ssor = (SorStruct *)sor;

  DetailLevel = 16; /* XXX: Should be a private field */

  if (sor->FirstTriangle)
    DestroyObjectTriangleMesh(sor);

  for (i=1 ; i<ssor->NumberOfPoints-2 ; i++)
  {
    for (j=0 ; j<DetailLevel ; j++)
    {
      p1[0] = ssor->Point[i][0]*cos((j*2.0)*M_PI/DetailLevel);
      p1[1] = ssor->Point[i][1];
      p1[2] = ssor->Point[i][0]*sin((j*2.0)*M_PI/DetailLevel);
      n1[0] = n1[1] = n1[2];
      p2[0] = ssor->Point[i][0]*cos(((j+1)*2.0)*M_PI/DetailLevel);
      p2[1] = ssor->Point[i][1];
      p2[2] = ssor->Point[i][0]*sin(((j+1)*2.0)*M_PI/DetailLevel);
      n2[0] = n2[1] = n2[2];
      p3[0] = ssor->Point[i+1][0]*cos((j*2.0)*M_PI/DetailLevel);
      p3[1] = ssor->Point[i+1][1];
      p3[2] = ssor->Point[i+1][0]*sin((j*2.0)*M_PI/DetailLevel);
      n3[0] = n3[1] = n3[2];
      p4[0] = ssor->Point[i+1][0]*cos(((j+1)*2.0)*M_PI/DetailLevel);
      p4[1] = ssor->Point[i+1][1];
      p4[2] = ssor->Point[i+1][0]*sin(((j+1)*2.0)*M_PI/DetailLevel);
      n4[0] = n4[1] = n4[2];
      if (sor->Trans)
      {
        MEvaluatePoint(p1, sor->Trans, p1);
        MEvaluateVector(n1, sor->Trans, n1);
        MEvaluatePoint(p2, sor->Trans, p2);
        MEvaluateVector(n2, sor->Trans, n2);
        MEvaluatePoint(p3, sor->Trans, p3);
        MEvaluateVector(n3, sor->Trans, n3);
        MEvaluatePoint(p4, sor->Trans, p4);
        MEvaluateVector(n4, sor->Trans, n4);
      }
      AddTriangleToObjectMesh(sor, p1,p2,p3, n1,n2,n3);
      AddTriangleToObjectMesh(sor, p2,p3,p4, n2,n3,n4);
    }
  }
}

/*****************************************************************************
*  giram_sor_inside FIXME
******************************************************************************/
static gboolean giram_sor_inside(ObjectStruct *sor, double x, double y, double z)
{
  return FALSE; /* FIXME */
}

/*****************************************************************************
*  FindIntersectSorSegment FIXME
******************************************************************************/
static gboolean giram_sor_find_intersection_segment(ObjectStruct *sor,
                                                    Vector in_point, Vector out_point,
                                                    Vector inter_point, Vector inter_norm)
{
  return FALSE;
}

/*****************************************************************************/
GtkWidget *SorDrawingArea;
GtkWidget *SorLabelCoordX;
GtkWidget *SorLabelCoordY;
GtkWidget *SorOffsetXSpinButton;
GtkWidget *SorOffsetYSpinButton;
GtkWidget *SorSizeXSpinButton;
GtkWidget *SorSizeYSpinButton;
int SorSnapToGrid;
int SorDrawGrid;
int SorNbPoint;
Vector *SorAllPoints = NULL;
int SorSpline;
double SorZoom;
double SorXoffset, SorYoffset;
#define SOR_ADD_POINT    1
#define SOR_DELETE_POINT 2
#define SOR_MOVE_POINT   3
int SorModeTool;
int SorPointIndex;
#define SOR_NONE   1
#define SOR_MOVING 2
int SOR_STATE = SOR_NONE;

/*****************************************************************************
*  SorSetModeAdd
******************************************************************************/
void SorSetModeAdd(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  SorModeTool = SOR_ADD_POINT;
}

/*****************************************************************************
*  SorSetModeDelete
******************************************************************************/
void SorSetModeDelete(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  SorModeTool = SOR_DELETE_POINT;
}

/*****************************************************************************
*  SorSetModeMove
******************************************************************************/
void SorSetModeMove(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  SorModeTool = SOR_MOVE_POINT;
}

/*****************************************************************************
*  SorSetStraightLine
******************************************************************************/
void SorSetStraightLine(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  SorSpline = FALSE;
  gtk_widget_queue_draw(SorDrawingArea);
}

/*****************************************************************************
*  SorSetSpline
******************************************************************************/
void SorSetSpline(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  SorSpline = TRUE;
  gtk_widget_queue_draw(SorDrawingArea);
}

/*****************************************************************************
*  SorToggleDrawGrid
******************************************************************************/
void SorToggleDrawGrid(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  SorDrawGrid = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  gtk_widget_queue_draw(SorDrawingArea);
}

/*****************************************************************************
*  SorToggleSnapToGrid
******************************************************************************/
void SorToggleSnapToGrid(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  SorSnapToGrid = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}

/*****************************************************************************
*  SorDrawSnapGrid
******************************************************************************/
void SorDrawSnapGrid(GtkWidget *Area)
{
  GdkColormap *colormap;
  GdkColor ForeColor;
  GdkGC *gc;
  double x,y;
  double StartX, StartY;
  double EndX, EndY;
  double Xinc, Yinc;

  colormap = gdk_window_get_colormap(Area->window);
  ForeColor.red = 32767;
  ForeColor.green = 32767;
  ForeColor.blue = 32767;
  gdk_color_alloc(colormap, &ForeColor);
  gc = gdk_gc_new(Area->window);
  gdk_gc_set_foreground(gc, &ForeColor);

  Xinc = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorSizeXSpinButton));
  Yinc = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorSizeYSpinButton));
  StartX = SorXoffset-Area->allocation.width/2.0/SorZoom;
  StartY = SorYoffset-Area->allocation.height/2.0/SorZoom;

  StartX = Xinc*((int)(StartX/Xinc))+gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorOffsetXSpinButton));
  StartY = Yinc*((int)(StartY/Yinc))+gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorOffsetYSpinButton));

  EndX = Area->allocation.width/2.0/SorZoom+SorXoffset;
  EndY = SorYoffset+Area->allocation.height/2.0/SorZoom;
  for (x=StartX; x<EndX ; x+=Xinc)
  {
    for (y=StartY ; y<EndY ; y+=Yinc)
    {
      gdk_draw_point(Area->window, gc,
                     Area->allocation.width/2+(x-SorXoffset)*SorZoom,
                     Area->allocation.height/2-(y-SorYoffset)*SorZoom);
    }
  }
  gdk_gc_destroy(gc);
}

/*****************************************************************************
*  SorDrawSpline
******************************************************************************/
void SorDrawSpline(GtkWidget *Area, GdkGC *gc,
                   double x0, double y0, double x1, double y1,
                   double x00, double y00, double x11, double y11)
{
  TransformStruct Mat;
  Vector ABCD, Ord;
  int i, From, To;
  double RealY, RealX;
  double x;

  Mat.Direct[0][0] = x0*x0*x0;
  Mat.Direct[0][1] = x1*x1*x1;
  Mat.Direct[0][2] = 3.0*x0*x0;
  Mat.Direct[0][3] = 3.0*x1*x1;
  Mat.Direct[1][0] = x0*x0;
  Mat.Direct[1][1] = x1*x1;
  Mat.Direct[1][2] = 2*x0;
  Mat.Direct[1][3] = 2*x1;
  Mat.Direct[2][0] = x0;
  Mat.Direct[2][1] = x1;
  Mat.Direct[2][2] = 1.0;
  Mat.Direct[2][3] = 1.0;
  Mat.Direct[3][0] = 1.0;
  Mat.Direct[3][1] = 1.0;
  Mat.Direct[3][2] = 0.0;
  Mat.Direct[3][3] = 0.0;
  MInverse(Mat.Inverse, Mat.Direct);
  Ord[0] = y0*y0;
  Ord[1] = y1*y1;
  Ord[2] = 2*y0*(y1-y00)/(x1-x00);
  Ord[3] = 2*y1*(y11-y0)/(x11-x0);

  for (i=0 ; i<4 ; i++)
  {
    ABCD[i] = Ord[0] * Mat.Inverse[0][i] +
              Ord[1] * Mat.Inverse[1][i] +
              Ord[2] * Mat.Inverse[2][i] +
              Ord[3] * Mat.Inverse[3][i];
  }
  From = (int)(Area->allocation.height/2-(x1-SorXoffset)*SorZoom);
  To = (int)(Area->allocation.height/2-(x0-SorXoffset)*SorZoom);
  for (i=From ; i<To ; i++)
  {
    RealY = SorYoffset-(i-Area->allocation.height/2)/SorZoom;
    RealX = sqrt(((ABCD[0]*RealY+ABCD[1])*RealY+ABCD[2])*RealY+ABCD[3]);
    x = Area->allocation.width/2+(RealX-SorXoffset)*SorZoom;
    gdk_draw_point(Area->window, gc, x,i);
  }
}

/*****************************************************************************
*  ExposeNewSorDrawingArea
******************************************************************************/
gint ExposeNewSorDrawingArea(GtkWidget *DrawingArea, GdkEventExpose *ev)
{
  GdkColormap *colormap;
  GdkColor BackColor;
  GdkGC *gc;
  GdkGC *DashedGC;
  GdkColor ForeColor;
  int i;

  colormap = gdk_window_get_colormap(DrawingArea->window);
  BackColor.red = 65535;
  BackColor.green = 65535;
  BackColor.blue = 65535;
  gdk_color_alloc(colormap, &BackColor);
  gdk_window_set_background(DrawingArea->window, &BackColor);
  gdk_window_clear(DrawingArea->window);

  if (SorDrawGrid)
  {
    SorDrawSnapGrid(DrawingArea);
  }

  DashedGC = gdk_gc_new(DrawingArea->window);
  gdk_gc_set_line_attributes(DashedGC, 1, GDK_LINE_ON_OFF_DASH, 0, 0);
  
  gc = gdk_gc_new(DrawingArea->window);
  ForeColor.red = 0;
  ForeColor.green = 0;
  ForeColor.blue = 0;
  gdk_color_alloc(colormap, &ForeColor);
  gdk_gc_set_foreground(gc, &ForeColor);
  
 
  /* First, we draw the axes */
  gdk_draw_line(DrawingArea->window, gc,
                DrawingArea->allocation.width/2-SorXoffset*SorZoom,
                0,
                DrawingArea->allocation.width/2-SorXoffset*SorZoom,
                DrawingArea->allocation.height-1);
  gdk_draw_line(DrawingArea->window, gc,
                0,
                DrawingArea->allocation.height/2+SorYoffset*SorZoom,
                DrawingArea->allocation.width-1,
                DrawingArea->allocation.height/2+SorYoffset*SorZoom);
  /* Then the lines between the dot */
  gdk_draw_line(DrawingArea->window, DashedGC,
                DrawingArea->allocation.width/2+(SorAllPoints[0][0]-SorXoffset)*SorZoom,
                DrawingArea->allocation.height/2-(SorAllPoints[0][1]-SorYoffset)*SorZoom,
                DrawingArea->allocation.width/2+(SorAllPoints[1][0]-SorXoffset)*SorZoom,
                DrawingArea->allocation.height/2-(SorAllPoints[1][1]-SorYoffset)*SorZoom);
  if (SorSpline)
  {
    for (i=1 ; i<SorNbPoint-2 ; i++)
    {
      SorDrawSpline(DrawingArea, gc,
                    SorAllPoints[i][1], SorAllPoints[i][0],
                    SorAllPoints[i+1][1], SorAllPoints[i+1][0],
                    SorAllPoints[i-1][1], SorAllPoints[i-1][0], 
                    SorAllPoints[i+2][1], SorAllPoints[i+1][0]);
    }
  } else
  {
    for (i=1 ; i<SorNbPoint-2 ; i++)
    {
      gdk_draw_line(DrawingArea->window, gc,
                    DrawingArea->allocation.width/2+(SorAllPoints[i][0]-SorXoffset)*SorZoom,
                    DrawingArea->allocation.height/2-(SorAllPoints[i][1]-SorYoffset)*SorZoom,
                    DrawingArea->allocation.width/2+(SorAllPoints[i+1][0]-SorXoffset)*SorZoom,
                    DrawingArea->allocation.height/2-(SorAllPoints[i+1][1]-SorYoffset)*SorZoom);
    }
  }
  gdk_draw_line(DrawingArea->window, DashedGC,
                DrawingArea->allocation.width/2+(SorAllPoints[SorNbPoint-2][0]-SorXoffset)*SorZoom,
                DrawingArea->allocation.height/2-(SorAllPoints[SorNbPoint-2][1]-SorYoffset)*SorZoom,
                DrawingArea->allocation.width/2+(SorAllPoints[SorNbPoint-1][0]-SorXoffset)*SorZoom,
                DrawingArea->allocation.height/2-(SorAllPoints[SorNbPoint-1][1]-SorYoffset)*SorZoom);
  /* Draw the plots */
  gdk_draw_rectangle(DrawingArea->window, gc, FALSE,
                     DrawingArea->allocation.width/2+
                              (SorAllPoints[0][0]-SorXoffset)*SorZoom-5,
                     DrawingArea->allocation.height/2-
                              (SorAllPoints[0][1]-SorYoffset)*SorZoom-5,
                     11, 11); 
  for (i=1 ; i<SorNbPoint-1 ; i++)
  {
    gdk_draw_rectangle(DrawingArea->window, gc, TRUE,
                       DrawingArea->allocation.width/2+
                                (SorAllPoints[i][0]-SorXoffset)*SorZoom-5,
                       DrawingArea->allocation.height/2-
                                (SorAllPoints[i][1]-SorYoffset)*SorZoom-5,
                       11, 11); 
  }
  gdk_draw_rectangle(DrawingArea->window, gc, FALSE,
                     DrawingArea->allocation.width/2+
                              (SorAllPoints[i][0]-SorXoffset)*SorZoom-5,
                     DrawingArea->allocation.height/2-
                              (SorAllPoints[i][1]-SorYoffset)*SorZoom-5,
                     11, 11); 
  return FALSE;
}

/*****************************************************************************
*  SorUpdateCursorPosition
******************************************************************************/
void SorUpdateCursorPosition(double x, double y)
{
  double RealX, RealY;
  gchar *buf;

  RealX = (x-SorDrawingArea->allocation.width/2)/SorZoom+SorXoffset;
  RealY = SorYoffset-(y-SorDrawingArea->allocation.height/2)/SorZoom;
  buf = g_strdup_printf("x: %g", RealX);
  gtk_label_set_text(GTK_LABEL(SorLabelCoordX), buf);
  g_free(buf);
  buf = g_strdup_printf("y: %g", RealY);
  gtk_label_set_text(GTK_LABEL(SorLabelCoordY), buf);
  g_free(buf);
}

/*****************************************************************************
*  SorAddPoint
******************************************************************************/
void SorAddPoint(GtkWidget *Area, GdkEventButton *bevent)
{
  double RealX, RealY;
  int i;

  RealX = (bevent->x-Area->allocation.width/2)/SorZoom+SorXoffset;
  RealY = SorYoffset-(bevent->y-Area->allocation.height/2)/SorZoom;

  if (SorSnapToGrid)
  {
    double Xinc, Yinc;
    Xinc = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorSizeXSpinButton));
    Yinc = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorSizeYSpinButton));
    RealX = Xinc*((int)(RealX/Xinc))+gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorOffsetXSpinButton));
    RealY = Yinc*((int)(RealY/Yinc))+gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorOffsetYSpinButton));
  }
  SorAllPoints = g_realloc(SorAllPoints, (SorNbPoint+1)*sizeof(Vector));
  for (i=SorNbPoint ; (i>0) && (SorAllPoints[i-1][1]>RealY) ; i--)
  {
    SorAllPoints[i][0] = SorAllPoints[i-1][0];
    SorAllPoints[i][1] = SorAllPoints[i-1][1];
  }
  SorAllPoints[i][0] = RealX;
  SorAllPoints[i][1] = RealY;
  SorNbPoint++;
  gtk_widget_queue_draw(Area);
}

/*****************************************************************************
*  SorBeginMovePoint
******************************************************************************/
void SorBeginMovePoint(GtkWidget *Area, GdkEventButton *bevent)
{
  int i;
  double XD, YD;
  /* Search for the index of the point if it exists */
  SorPointIndex = -1;
  for (i=0 ; i<SorNbPoint ; i++)
  {
    XD = Area->allocation.width/2+(SorAllPoints[i][0]-SorXoffset)*SorZoom;
    YD = Area->allocation.height/2-(SorAllPoints[i][1]-SorYoffset)*SorZoom;
    if (fabs(XD-bevent->x)<5.0 && fabs(YD-bevent->y)<5.0)
      SorPointIndex = i;
  }
  if (SorPointIndex != -1)
  {
    SOR_STATE = SOR_MOVING;
    gdk_pointer_grab(Area->window, FALSE,
                     GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK,
                     NULL, NULL, bevent->time);
  }
}

/*****************************************************************************
*  SorMoveMovePoint
******************************************************************************/
void SorMoveMovePoint(GtkWidget *Area, GdkEventMotion *mevent)
{
  double RealX, RealY;

  RealX = (mevent->x-Area->allocation.width/2)/SorZoom+SorXoffset;
  RealY = SorYoffset-(mevent->y-Area->allocation.height/2)/SorZoom;
  if (SorPointIndex == 0)
  {
    RealY = MIN(RealY, SorAllPoints[1][1]-1/SorZoom);
  } else if (SorPointIndex == SorNbPoint-1)
  {
    RealY = MAX(RealY, SorAllPoints[SorNbPoint-2][1]+1/SorZoom);
  } else
  {
    RealY = MIN(RealY, SorAllPoints[SorPointIndex+1][1]-1/SorZoom);
    RealY = MAX(RealY, SorAllPoints[SorPointIndex-1][1]+1/SorZoom);
  }
  if (SorSnapToGrid)
  {
    double Xinc, Yinc;
    Xinc = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorSizeXSpinButton));
    Yinc = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorSizeYSpinButton));
    RealX = Xinc*((int)(RealX/Xinc))+gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorOffsetXSpinButton));
    RealY = Yinc*((int)(RealY/Yinc))+gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorOffsetYSpinButton));
  }
  SorAllPoints[SorPointIndex][0] = RealX;
  SorAllPoints[SorPointIndex][1] = RealY;
  gtk_widget_queue_draw(Area);
}

/*****************************************************************************
*  SorEndMovePoint
******************************************************************************/
void SorEndMovePoint(GtkWidget *Area, GdkEventButton *bevent)
{
  double RealX, RealY;

  RealX = (bevent->x-Area->allocation.width/2)/SorZoom+SorXoffset;
  RealY = SorYoffset-(bevent->y-Area->allocation.height/2)/SorZoom;
  if (SorPointIndex == 0)
  {
    RealY = MIN(RealY, SorAllPoints[1][1]-1/SorZoom);
  } else if (SorPointIndex == SorNbPoint-1)
  {
    RealY = MAX(RealY, SorAllPoints[SorNbPoint-2][1]+1/SorZoom);
  } else
  {
    RealY = MIN(RealY, SorAllPoints[SorPointIndex+1][1]-1/SorZoom);
    RealY = MAX(RealY, SorAllPoints[SorPointIndex-1][1]+1/SorZoom);
  }
  if (SorSnapToGrid)
  {
    double Xinc, Yinc;
    Xinc = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorSizeXSpinButton));
    Yinc = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorSizeYSpinButton));
    RealX = Xinc*((int)(RealX/Xinc))+gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorOffsetXSpinButton));
    RealY = Yinc*((int)(RealY/Yinc))+gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SorOffsetYSpinButton));
  }
  SorAllPoints[SorPointIndex][0] = RealX;
  SorAllPoints[SorPointIndex][1] = RealY;
  SOR_STATE = SOR_NONE;
  gdk_pointer_ungrab(bevent->time);
  
  gtk_widget_queue_draw(Area);
}

/*****************************************************************************
*  SorDeletePoint
******************************************************************************/
void SorDeletePoint(GtkWidget *Area, GdkEventButton *bevent)
{
  int index, i;
  double XD, YD;
  /* Search for the index of the point if it exists */
  if (SorNbPoint < 5)
    return;
  index = -1;
  for (i=0 ; i<SorNbPoint ; i++)
  {
    XD = Area->allocation.width/2+(SorAllPoints[i][0]-SorXoffset)*SorZoom;
    YD = Area->allocation.height/2-(SorAllPoints[i][1]-SorYoffset)*SorZoom;
    if (fabs(XD-bevent->x)<5.0 && fabs(YD-bevent->y)<5.0)
      index = i;
  }
  if (index != -1)
  {
    for (i=index ; i<SorNbPoint-1 ; i++)
    {
      SorAllPoints[i][0] = SorAllPoints[i+1][0];
      SorAllPoints[i][1] = SorAllPoints[i+1][1];
    }
    SorNbPoint--;
    SorAllPoints = g_realloc(SorAllPoints, SorNbPoint*sizeof(Vector));
  }
  gtk_widget_queue_draw(Area);
}

/*****************************************************************************
*  SorDrawingAreaEventDispatcher
******************************************************************************/
gint SorDrawingAreaEventDispatcher(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventButton *bevent = (GdkEventButton *) event;
  GdkEventMotion *mevent = (GdkEventMotion *) event;
  switch (event->type)
  {
    case GDK_BUTTON_PRESS:
      if (bevent->button == 2) /* Middle Button ==> Begin a pan */
      {
        /* FIXME */
      }
      else if (bevent->button == 1)
      { /* Left Button */
        switch (SorModeTool)
        {
          case SOR_ADD_POINT:
            SorAddPoint(DrawingArea, bevent);
            break;
          case SOR_DELETE_POINT:
            SorDeletePoint(DrawingArea, bevent);
            break;
          case SOR_MOVE_POINT:
            SorBeginMovePoint(DrawingArea, bevent);
            break;
        }
      }
      return TRUE;
    case GDK_MOTION_NOTIFY:
      SorUpdateCursorPosition(mevent->x, mevent->y);
      if ((SorModeTool == SOR_MOVE_POINT) && (SOR_STATE == SOR_MOVING))
      {
        SorMoveMovePoint(DrawingArea, mevent);
      }
      return TRUE;
    case GDK_BUTTON_RELEASE:
      if ((SorModeTool == SOR_MOVE_POINT) && (SOR_STATE == SOR_MOVING))
      {
        SorEndMovePoint(DrawingArea, bevent);
      }
      return TRUE;
    default:
      return FALSE;
  }
  return FALSE;
}

/*****************************************************************************
*  NewSorOK
******************************************************************************/
void NewSorOK(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;
  ObjectStruct       *sor;
  SorStruct          *SSor;
  guint               i;

  UserData = get_view_user_data();

  sor = giram_sor_new(SorNbPoint, SorAllPoints);
  UserData->FrameView->all_objects = g_slist_append(UserData->FrameView->all_objects,
                                                    sor);
  giram_object_build_triangle_mesh(sor);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
}

/*****************************************************************************
*  ToolNewSor
******************************************************************************/
void ToolNewSor(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  GtkWidget *Dialog;
  GtkWidget *Button;
  GtkWidget *Table;
  GtkWidget *Frame;
  GtkWidget *VBox;
  GSList *RadioGroup;
  GtkWidget *Alignment;
  GtkWidget *Pixmap;
  GtkWidget *HBox;
  GtkWidget *RenderVBox;
  GtkWidget *GridVBox;
  GtkWidget *Label;
  GtkWidget *CheckButton;
  GtkAdjustment *adj;

  Dialog = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(Dialog), _("New Sor"));
  /* The OK Button */
  Button = gtk_button_new_with_label("OK");
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button,
                     TRUE, TRUE, 0);
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)NewSorOK, GTK_OBJECT(DrawingArea));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  gtk_widget_grab_default(Button);
  /* The Cancel Button */
  Button = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button,
                     TRUE, TRUE, 0);
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));

  /* The main table */
  Table = gtk_table_new(2, 2, FALSE);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(Dialog)->vbox), Table);
  /* The Drawing Area */
  SorDrawingArea = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(SorDrawingArea), 200, 300);
  gtk_table_attach(GTK_TABLE(Table), SorDrawingArea, 0,1, 0,1, 
                   GTK_FILL | GTK_EXPAND | GTK_SHRINK,
                   GTK_FILL | GTK_EXPAND | GTK_SHRINK, 5, 5);
  gtk_signal_connect(GTK_OBJECT(SorDrawingArea), "event",
                     GTK_SIGNAL_FUNC(SorDrawingAreaEventDispatcher), NULL);

  gtk_signal_connect(GTK_OBJECT(SorDrawingArea), "expose_event",
                     (GtkSignalFunc)ExposeNewSorDrawingArea, NULL);
  gtk_widget_set_events(SorDrawingArea, GDK_EXPOSURE_MASK
                                      | GDK_BUTTON_PRESS_MASK
                                      | GDK_BUTTON_RELEASE_MASK
                                      | GDK_POINTER_MOTION_MASK
                                      | GDK_KEY_PRESS_MASK);
  /* The Frame */
  Frame = gtk_frame_new("SOR parameters");
  gtk_table_attach(GTK_TABLE(Table), Frame, 1,2, 0,1,
                   GTK_FILL,
                   GTK_FILL | GTK_EXPAND, 5, 5);
  /* The Vertical box into the big Frame */
  VBox = gtk_vbox_new(FALSE, 10);
  gtk_container_add(GTK_CONTAINER(Frame), VBox);
  /* The table for the mode Button */
  Table = gtk_table_new(1,3, TRUE);
  gtk_box_pack_start(GTK_BOX(VBox), Table, FALSE, TRUE, 10);
  RadioGroup = NULL;
  SorModeTool = SOR_ADD_POINT;
    /* The "Add point" button */
    Button = gtk_radio_button_new(RadioGroup);
    gtk_container_set_border_width(GTK_CONTAINER(Button), 0);
    RadioGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Button));
    gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(Button), FALSE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Button), TRUE);
    gtk_signal_connect(GTK_OBJECT(Button), "toggled",
                       GTK_SIGNAL_FUNC(SorSetModeAdd), NULL);
    gtk_table_attach(GTK_TABLE(Table), Button, 0,1, 0,1,
                     GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                     GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    Alignment = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
    gtk_container_set_border_width(GTK_CONTAINER(Alignment), 0);
    gtk_container_add(GTK_CONTAINER(Button), Alignment);
    Pixmap = gtk_label_new("Add Point");
    gtk_container_add(GTK_CONTAINER(Alignment), Pixmap);
    /* The "Delete point" button */
    Button = gtk_radio_button_new(RadioGroup);
    gtk_container_set_border_width(GTK_CONTAINER(Button), 0);
    RadioGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Button));
    gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(Button), FALSE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Button), FALSE);
    gtk_signal_connect(GTK_OBJECT(Button), "toggled",
                       GTK_SIGNAL_FUNC(SorSetModeDelete), NULL);
    gtk_table_attach(GTK_TABLE(Table), Button, 1,2, 0,1,
                     GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                     GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    Alignment = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
    gtk_container_set_border_width(GTK_CONTAINER(Alignment), 0);
    gtk_container_add(GTK_CONTAINER(Button), Alignment);
    Pixmap = gtk_label_new("Delete Point");
    gtk_container_add(GTK_CONTAINER(Alignment), Pixmap);
    /* The "Move point" button */
    Button = gtk_radio_button_new(RadioGroup);
    gtk_container_set_border_width(GTK_CONTAINER(Button), 0);
    RadioGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Button));
    gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(Button), FALSE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Button), FALSE);
    gtk_signal_connect(GTK_OBJECT(Button), "toggled",
                       GTK_SIGNAL_FUNC(SorSetModeMove), NULL);
    gtk_table_attach(GTK_TABLE(Table), Button, 2,3, 0,1,
                     GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                     GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    Alignment = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
    gtk_container_set_border_width(GTK_CONTAINER(Alignment), 0);
    gtk_container_add(GTK_CONTAINER(Button), Alignment);
    Pixmap = gtk_label_new("Move Point");
    gtk_container_add(GTK_CONTAINER(Alignment), Pixmap);
  /* The Horizontal Box that contain the two Labels for the pointer coords */
  HBox = gtk_hbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(VBox), HBox, TRUE, TRUE, 0);
    /* The Label that shows the X coordinate of the pointer in the Area */
    SorLabelCoordX = gtk_label_new("x:     ");
    gtk_box_pack_start(GTK_BOX(HBox), SorLabelCoordX, TRUE, TRUE, 0);
    /* The Label that shows the Y coordinate of the pointer in the Area */
    SorLabelCoordY = gtk_label_new("y:     ");
    gtk_box_pack_start(GTK_BOX(HBox), SorLabelCoordY, TRUE, TRUE, 0);
  /* The Frame for the Rendering */
  Frame = gtk_frame_new("Rendering Options");
  gtk_box_pack_start(GTK_BOX(VBox), Frame, FALSE, TRUE, 0);
  RenderVBox = gtk_vbox_new(TRUE, 0);
  gtk_container_add(GTK_CONTAINER(Frame), RenderVBox);
  RadioGroup = NULL;
    /* The Straight line rendering radio button */
    Button = gtk_radio_button_new_with_label(RadioGroup, "Straight Line");
    RadioGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Button));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Button), TRUE);
    gtk_box_pack_start(GTK_BOX(RenderVBox), Button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(Button), "toggled",
                       GTK_SIGNAL_FUNC(SorSetStraightLine), NULL);
    /* The Spline rendering radio button */
    Button = gtk_radio_button_new_with_label(RadioGroup, "Spline");
    RadioGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Button));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Button), FALSE);
    gtk_box_pack_start(GTK_BOX(RenderVBox), Button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(Button), "toggled",
                       GTK_SIGNAL_FUNC(SorSetSpline), NULL);
    SorSpline = FALSE;
  /* The Frame for the Grid Parameter */
  Frame = gtk_frame_new("Grid Parameter");
  gtk_box_pack_start(GTK_BOX(VBox), Frame, TRUE, TRUE, 0);
  GridVBox = gtk_vbox_new(TRUE, 0);
  gtk_container_add(GTK_CONTAINER(Frame), GridVBox);
    /* The Table for the offset and size parameters */
    Table = gtk_table_new(2, 4, FALSE);
    gtk_box_pack_start(GTK_BOX(GridVBox), Table, TRUE, TRUE, 0);
      /* The "offset x" label */
      Label = gtk_label_new("offset x");
      gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 0,1,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
      /* The "offset x" spin */
      adj = (GtkAdjustment *) gtk_adjustment_new(0.0, 0.00, 5.00, 0.05, 0.25, 0.0);
      SorOffsetXSpinButton = gtk_spin_button_new(adj, 0.2, 2);
      gtk_table_attach(GTK_TABLE(Table), SorOffsetXSpinButton, 1,2, 0,1,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
      /* The "offset y" label */
      Label = gtk_label_new("offset y");
      gtk_table_attach(GTK_TABLE(Table), Label, 2,3, 0,1,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
      /* The "offset y" spin */
      adj = (GtkAdjustment *) gtk_adjustment_new(0.00, 0.00, 5.00, 0.05, 0.25, 0.0);
      SorOffsetYSpinButton = gtk_spin_button_new(adj, 0.2, 2);
      gtk_table_attach(GTK_TABLE(Table), SorOffsetYSpinButton, 3,4, 0,1,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
      /* The "size x" label */
      Label = gtk_label_new("size x");
      gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 1,2,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
      /* The "size x" spin */
      adj = (GtkAdjustment *) gtk_adjustment_new(0.25, 0.05, 5.00, 0.05, 0.25, 0.0);
      SorSizeXSpinButton = gtk_spin_button_new(adj, 0.2, 2);
      gtk_table_attach(GTK_TABLE(Table), SorSizeXSpinButton, 1,2, 1,2,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
      /* The "size y" label */
      Label = gtk_label_new("size y");
      gtk_table_attach(GTK_TABLE(Table), Label, 2,3, 1,2,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
      /* The "size y" spin */
      adj = (GtkAdjustment *) gtk_adjustment_new(0.25, 0.05, 5.00, 0.05, 0.25, 0.0);
      SorSizeYSpinButton = gtk_spin_button_new(adj, 0.2, 2);
      gtk_table_attach(GTK_TABLE(Table), SorSizeYSpinButton, 3,4, 1,2,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    /* The check button for the "show grid" option */
    CheckButton = gtk_check_button_new_with_label("Show grid");
    SorDrawGrid = 0;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(CheckButton), FALSE);
    gtk_signal_connect(GTK_OBJECT(CheckButton), "toggled",
                       GTK_SIGNAL_FUNC(SorToggleDrawGrid), NULL);
    gtk_box_pack_start_defaults(GTK_BOX(GridVBox), CheckButton);
    /* The check button for the "snap to grid" option */
    CheckButton = gtk_check_button_new_with_label("Snap to grid");
    SorSnapToGrid = 0;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(CheckButton), FALSE);
    gtk_signal_connect(GTK_OBJECT(CheckButton), "toggled",
                       GTK_SIGNAL_FUNC(SorToggleSnapToGrid), NULL);
    gtk_box_pack_start_defaults(GTK_BOX(GridVBox), CheckButton);
    
  gtk_widget_show_all(Dialog);    
  SorXoffset = SorYoffset = 0.0;
  SorZoom = 50.0;
  
  SorNbPoint = 4;
  if (SorAllPoints)
    g_free(SorAllPoints);
  SorAllPoints = g_malloc(4*sizeof(Vector));
  SorAllPoints[0][0] = 1.0;
  SorAllPoints[0][1] = -1.5;
  SorAllPoints[1][0] = 1.0;
  SorAllPoints[1][1] = -0.5;
  SorAllPoints[2][0] = 1.0;
  SorAllPoints[2][1] = 0.5;
  SorAllPoints[3][0] = 1.0;
  SorAllPoints[3][1] = 1.5;
}

/*****************************************************************************
*  giram_sor_new
******************************************************************************/
ObjectStruct *giram_sor_new(guint nb_points, Vector *points)
{
  ObjectStruct *sor;
  SorStruct *ssor;
  static GiramObjectClass *sor_klass = NULL;
  guint i;

  if (sor_klass == NULL)
  {
    sor_klass = giram_object_class_new();

    sor_klass->build_triangle_mesh       = giram_sor_build_triangle_mesh;
    sor_klass->inside                    = giram_sor_inside;
    sor_klass->find_intersection_segment = giram_sor_find_intersection_segment;
  }
  ssor = g_new(SorStruct, 1);
  sor = (ObjectStruct *)ssor;
  InitObject(sor);
  sor->Type = SOR_OBJECT;
  sor->klass = sor_klass;
  ssor->NumberOfPoints = nb_points;
  if (nb_points)
  {
    ssor->Point = g_malloc(ssor->NumberOfPoints*sizeof(Vector));
    for (i=0 ; i<ssor->NumberOfPoints ; i++)
    {
      ssor->Point[i][0] = points[i][0];
      ssor->Point[i][1] = points[i][1];
    }
  }
  return sor;
}

