/* superellipsoid.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "csgtree.h"
#include "object.h"
#include "utils.h"
#include "superellipsoid.h"

#include "giramintl.h"

static void giram_superellipsoid_build_triangle_mesh(ObjectStruct *object);
static gboolean giram_superellipsoid_inside(ObjectStruct *object, double x, double y, double z);
static gboolean giram_superellipsoid_find_intersection_segment(ObjectStruct *object,
                                                               Vector in_point, Vector out_point,
                                                               Vector inter_point, Vector inter_norm);

/*****************************************************************************
*  SuperEllipsoidMeshNormal FIXME
******************************************************************************/
static void SuperEllipsoidMeshNormal(Vector Result, ObjectStruct *SuperEllipsoid,
                                     Vector Inter)
{ /* FIXME */
  Result[0] = 0.0;
  Result[1] = 0.0;
  Result[2] = 0.0;
}

/*****************************************************************************
*  ComputeSuperEllipsoidTriangleMesh
******************************************************************************/
static void giram_superellipsoid_build_triangle_mesh(ObjectStruct *SuperEllipsoid)
{ /* XXX: clipped_by */
  Vector P1, P2, P3, P4;
  Vector N1, N2, N3, N4;
  int i,j;
  int DetailLevel;
  SuperEllipsoidStruct *SSuperEllipsoid = (SuperEllipsoidStruct *)SuperEllipsoid;
  double t1, s1, c1, t2, s2, c2;

  DetailLevel = 16; /* FIXME: should be a private field */

  if (SuperEllipsoid->FirstTriangle)
    DestroyObjectTriangleMesh(SuperEllipsoid);
  /* First the North Pole */
  P1[0] = 0.0;
  P1[1] = 0.0;
  P1[2] = 1.0;
  SuperEllipsoidMeshNormal(N1, SuperEllipsoid, P1);
  if (SuperEllipsoid->Trans)
  {
    MEvaluatePoint(P1, SuperEllipsoid->Trans, P1);
    MEvaluateVector(N1, SuperEllipsoid->Trans, N1);
  }
  for (i=0; i<DetailLevel; i++)
  { /* The "DetailLevel" triangles adjacents to the pole */
    t1 = M_PI/DetailLevel;
    if (sin(t1) >= 0.0) s1 =  pow( sin(t1), SSuperEllipsoid->N);
    else                s1 = -pow(-sin(t1), SSuperEllipsoid->N);
    if (cos(t1) >= 0.0) c1 =  pow( cos(t1), SSuperEllipsoid->N);
    else                c1 = -pow(-cos(t1), SSuperEllipsoid->N);
    t2 = i*2.0*M_PI/DetailLevel;
    if (sin(t2) >= 0.0) s2 =  pow( sin(t2), SSuperEllipsoid->E);
    else                s2 = -pow(-sin(t2), SSuperEllipsoid->E);
    if (cos(t2) >= 0.0) c2 =  pow( cos(t2), SSuperEllipsoid->E);
    else                c2 = -pow(-cos(t2), SSuperEllipsoid->E);
    P2[0] = s1*c2;
    P2[1] = s1*s2;
    P2[2] = c1;
    SuperEllipsoidMeshNormal(N2, SuperEllipsoid, P2);
    if (SuperEllipsoid->Trans)
    {
      MEvaluatePoint(P2, SuperEllipsoid->Trans, P2);
      MEvaluateVector(N2, SuperEllipsoid->Trans, N2);
    }
    t2 = (i+1)*2.0*M_PI/DetailLevel;
    if (sin(t2) >= 0.0) s2 =  pow( sin(t2), SSuperEllipsoid->E);
    else                s2 = -pow(-sin(t2), SSuperEllipsoid->E);
    if (cos(t2) >= 0.0) c2 =  pow( cos(t2), SSuperEllipsoid->E);
    else                c2 = -pow(-cos(t2), SSuperEllipsoid->E);
    P3[0] = s1*c2;
    P3[1] = s1*s2;
    P3[2] = c1;
    SuperEllipsoidMeshNormal(N3, SuperEllipsoid, P3);
    if (SuperEllipsoid->Trans)
    {
      MEvaluatePoint(P3, SuperEllipsoid->Trans, P3);
      MEvaluateVector(N3, SuperEllipsoid->Trans, N3);
    }
    AddTriangleToObjectMesh(SuperEllipsoid, P1,P2,P3, N1,N2,N3);
  }
  /* Then the Mid Zone */
  for (i=1 ; i<DetailLevel-1 ; i++)
  {
    for (j=0 ; j<DetailLevel ; j++)
    {
      t1 = i*M_PI/DetailLevel;
      if (sin(t1) >= 0.0) s1 =  pow( sin(t1), SSuperEllipsoid->N);
      else                s1 = -pow(-sin(t1), SSuperEllipsoid->N);
      if (cos(t1) >= 0.0) c1 =  pow( cos(t1), SSuperEllipsoid->N);
      else                c1 = -pow(-cos(t1), SSuperEllipsoid->N);
      t2 = j*2.0*M_PI/DetailLevel;
      if (sin(t2) >= 0.0) s2 =  pow( sin(t2), SSuperEllipsoid->E);
      else                s2 = -pow(-sin(t2), SSuperEllipsoid->E);
      if (cos(t2) >= 0.0) c2 =  pow( cos(t2), SSuperEllipsoid->E);
      else                c2 = -pow(-cos(t2), SSuperEllipsoid->E);
      P1[0] = s1*c2;
      P1[1] = s1*s2;
      P1[2] = c1;
      SuperEllipsoidMeshNormal(N1, SuperEllipsoid, P1);
      if (SuperEllipsoid->Trans)
      {
        MEvaluatePoint(P1, SuperEllipsoid->Trans, P1);
        MEvaluateVector(N1, SuperEllipsoid->Trans, N1);
      }
      t1 = (i+1)*M_PI/DetailLevel;
      if (sin(t1) >= 0.0) s1 =  pow( sin(t1), SSuperEllipsoid->N);
      else                s1 = -pow(-sin(t1), SSuperEllipsoid->N);
      if (cos(t1) >= 0.0) c1 =  pow( cos(t1), SSuperEllipsoid->N);
      else                c1 = -pow(-cos(t1), SSuperEllipsoid->N);
      t2 = j*2.0*M_PI/DetailLevel;
      if (sin(t2) >= 0.0) s2 =  pow( sin(t2), SSuperEllipsoid->E);
      else                s2 = -pow(-sin(t2), SSuperEllipsoid->E);
      if (cos(t2) >= 0.0) c2 =  pow( cos(t2), SSuperEllipsoid->E);
      else                c2 = -pow(-cos(t2), SSuperEllipsoid->E);
      P2[0] = s1*c2;
      P2[1] = s1*s2;
      P2[2] = c1;
      SuperEllipsoidMeshNormal(N2, SuperEllipsoid, P2);
      if (SuperEllipsoid->Trans)
      {
        MEvaluatePoint(P2, SuperEllipsoid->Trans, P2);
        MEvaluateVector(N2, SuperEllipsoid->Trans, N2);
      }
      t1 = i*M_PI/DetailLevel;
      if (sin(t1) >= 0.0) s1 =  pow( sin(t1), SSuperEllipsoid->N);
      else                s1 = -pow(-sin(t1), SSuperEllipsoid->N);
      if (cos(t1) >= 0.0) c1 =  pow( cos(t1), SSuperEllipsoid->N);
      else                c1 = -pow(-cos(t1), SSuperEllipsoid->N);
      t2 = (j+1)*2.0*M_PI/DetailLevel;
      if (sin(t2) >= 0.0) s2 =  pow( sin(t2), SSuperEllipsoid->E);
      else                s2 = -pow(-sin(t2), SSuperEllipsoid->E);
      if (cos(t2) >= 0.0) c2 =  pow( cos(t2), SSuperEllipsoid->E);
      else                c2 = -pow(-cos(t2), SSuperEllipsoid->E);
      P3[0] = s1*c2;
      P3[1] = s1*s2;
      P3[2] = c1;
      SuperEllipsoidMeshNormal(N3, SuperEllipsoid, P3);
      if (SuperEllipsoid->Trans)
      {
        MEvaluatePoint(P3, SuperEllipsoid->Trans, P3);
        MEvaluateVector(N3, SuperEllipsoid->Trans, N3);
      }
      t1 = (i+1)*M_PI/DetailLevel;
      if (sin(t1) >= 0.0) s1 =  pow( sin(t1), SSuperEllipsoid->N);
      else                s1 = -pow(-sin(t1), SSuperEllipsoid->N);
      if (cos(t1) >= 0.0) c1 =  pow( cos(t1), SSuperEllipsoid->N);
      else                c1 = -pow(-cos(t1), SSuperEllipsoid->N);
      t2 = (j+1)*2.0*M_PI/DetailLevel;
      if (sin(t2) >= 0.0) s2 =  pow( sin(t2), SSuperEllipsoid->E);
      else                s2 = -pow(-sin(t2), SSuperEllipsoid->E);
      if (cos(t2) >= 0.0) c2 =  pow( cos(t2), SSuperEllipsoid->E);
      else                c2 = -pow(-cos(t2), SSuperEllipsoid->E);
      P4[0] = s1*c2;
      P4[1] = s1*s2;
      P4[2] = c1;
      SuperEllipsoidMeshNormal(N4, SuperEllipsoid, P4);
      if (SuperEllipsoid->Trans)
      {
        MEvaluatePoint(P4, SuperEllipsoid->Trans, P4);
        MEvaluateVector(N4, SuperEllipsoid->Trans, N4);
      }
      AddTriangleToObjectMesh(SuperEllipsoid, P1,P2,P4, N1,N2,N4);
      AddTriangleToObjectMesh(SuperEllipsoid, P1,P3,P4, N1,N3,N4);
    }
  }
  /* And the South Pole */
  P1[0] =  0.0;
  P1[1] =  0.0;
  P1[2] = -1.0;
  SuperEllipsoidMeshNormal(N1, SuperEllipsoid, P1);
  if (SuperEllipsoid->Trans)
  {
    MEvaluatePoint(P1, SuperEllipsoid->Trans, P1);
    MEvaluateVector(N1, SuperEllipsoid->Trans, N1);
  }
  for (i=0 ; i<DetailLevel ; i++)
  { /* The "DetailLevel" triangles adjacents to the south pole */
    t1 = M_PI-M_PI/DetailLevel;
    if (sin(t1) >= 0.0) s1 =  pow( sin(t1), SSuperEllipsoid->N);
    else                s1 = -pow(-sin(t1), SSuperEllipsoid->N);
    if (cos(t1) >= 0.0) c1 =  pow( cos(t1), SSuperEllipsoid->N);
    else                c1 = -pow(-cos(t1), SSuperEllipsoid->N);
    t2 = i*2.0*M_PI/DetailLevel;
    if (sin(t2) >= 0.0) s2 =  pow( sin(t2), SSuperEllipsoid->E);
    else                s2 = -pow(-sin(t2), SSuperEllipsoid->E);
    if (cos(t2) >= 0.0) c2 =  pow( cos(t2), SSuperEllipsoid->E);
    else                c2 = -pow(-cos(t2), SSuperEllipsoid->E);
    P2[0] = s1*c2;
    P2[1] = s1*s2;
    P2[2] = c1;
    SuperEllipsoidMeshNormal(N2, SuperEllipsoid, P2);
    if (SuperEllipsoid->Trans)
    {
      MEvaluatePoint(P2, SuperEllipsoid->Trans, P2);
      MEvaluateVector(N2, SuperEllipsoid->Trans, N2);
    }
    t2 = (i+1)*2.0*M_PI/DetailLevel;
    if (sin(t2) >= 0.0) s2 =  pow( sin(t2), SSuperEllipsoid->E);
    else                s2 = -pow(-sin(t2), SSuperEllipsoid->E);
    if (cos(t2) >= 0.0) c2 =  pow( cos(t2), SSuperEllipsoid->E);
    else                c2 = -pow(-cos(t2), SSuperEllipsoid->E);
    P3[0] = s1*c2;
    P3[1] = s1*s2;
    P3[2] = c1;
    SuperEllipsoidMeshNormal(N3, SuperEllipsoid, P3);
    if (SuperEllipsoid->Trans)
    {
      MEvaluatePoint(P3, SuperEllipsoid->Trans, P3);
      MEvaluateVector(N3, SuperEllipsoid->Trans, N3);
    }
    AddTriangleToObjectMesh(SuperEllipsoid, P1,P2,P3, N1,N2,N3);
  }
}

/*****************************************************************************
*  giram_superellipsoid_inside
******************************************************************************/
static gboolean giram_superellipsoid_inside(ObjectStruct *object, double x, double y, double z)
{
  Vector Vect;

  Vect[0] = x;
  Vect[1] = y;
  Vect[2] = z;
  if (object->Trans)
  {
    MInverseEvaluatePoint(Vect, object->Trans, Vect);
  }
  return FALSE; /* FIXME */
}

/*****************************************************************************
*  giram_superellipsoid_find_intersection_segment
******************************************************************************/
static gboolean giram_superellipsoid_find_intersection_segment(ObjectStruct *object,
                                                               Vector in_point, Vector out_point,
                                                               Vector inter_point, Vector inter_norm)
{
  return FALSE; /* FIXME */
}

/****************************************************************************
*  giram_superellipsoid_dialog_area_expose
*****************************************************************************/
static gboolean giram_superellipsoid_dialog_area_expose(GtkWidget *area,
                                                        GdkEventExpose *ev,
                                                        GtkWidget *dialog)
{
  GtkAdjustment *adj;
  gdouble        ns, ew;
  Vector P1, P2, P3, P4;
  int i,j;
  double t1, s1, c1, t2, s2, c2;

  adj = gtk_object_get_data(GTK_OBJECT(dialog), "east-west");
  ew = adj->value;
  adj = gtk_object_get_data(GTK_OBJECT(dialog), "north-south");
  ns = adj->value;

  gdk_window_clear(area->window);
#define DETAIL_LEVEL (16)

  /* First the North Pole */
  P1[0] = 0.0;
  P1[1] = 0.0;
  P1[2] = 1.0;
  for (i=0; i<DETAIL_LEVEL; i++)
  { /* The triangles which are adjacents to the pole */
    t1 = M_PI / DETAIL_LEVEL;
    if (sin(t1) >= 0.0) s1 =  pow( sin(t1), ns);
    else                s1 = -pow(-sin(t1), ns);
    if (cos(t1) >= 0.0) c1 =  pow( cos(t1), ns);
    else                c1 = -pow(-cos(t1), ns);
    t2 = i * 2.0 * M_PI / DETAIL_LEVEL;
    if (sin(t2) >= 0.0) s2 =  pow( sin(t2), ew);
    else                s2 = -pow(-sin(t2), ew);
    if (cos(t2) >= 0.0) c2 =  pow( cos(t2), ew);
    else                c2 = -pow(-cos(t2), ew);
    P2[0] = s1*c2;
    P2[1] = s1*s2;
    P2[2] = c1;
    t2 = (i+1) * 2.0 * M_PI / DETAIL_LEVEL;
    if (sin(t2) >= 0.0) s2 =  pow( sin(t2), ew);
    else                s2 = -pow(-sin(t2), ew);
    if (cos(t2) >= 0.0) c2 =  pow( cos(t2), ew);
    else                c2 = -pow(-cos(t2), ew);
    P3[0] = s1*c2;
    P3[1] = s1*s2;
    P3[2] = c1;
/*    AddTriangleToObjectMesh(SuperEllipsoid, P1,P2,P3, N1,N2,N3);*/
    gdk_draw_line(area->window, area->style->black_gc,
                  area->allocation.width /2+P1[0]*100,
                  area->allocation.height/2-P1[1]*100,
                  area->allocation.width /2+P2[0]*100,
                  area->allocation.height/2-P2[1]*100);
    gdk_draw_line(area->window, area->style->black_gc,
                  area->allocation.width /2+P1[0]*100,
                  area->allocation.height/2-P1[1]*100,
                  area->allocation.width /2+P3[0]*100,
                  area->allocation.height/2-P3[1]*100);
    gdk_draw_line(area->window, area->style->black_gc,
                  area->allocation.width /2+P3[0]*100,
                  area->allocation.height/2-P3[1]*100,
                  area->allocation.width /2+P2[0]*100,
                  area->allocation.height/2-P2[1]*100);
  }
  /* Then the Mid Zone */
  for (i=1 ; i<DETAIL_LEVEL-1 ; i++)
  {
    for (j=0 ; j<DETAIL_LEVEL ; j++)
    {
      t1 = i * M_PI / DETAIL_LEVEL;
      if (sin(t1) >= 0.0) s1 =  pow( sin(t1), ns);
      else                s1 = -pow(-sin(t1), ns);
      if (cos(t1) >= 0.0) c1 =  pow( cos(t1), ns);
      else                c1 = -pow(-cos(t1), ns);
      t2 = j * 2.0 * M_PI / DETAIL_LEVEL;
      if (sin(t2) >= 0.0) s2 =  pow( sin(t2), ew);
      else                s2 = -pow(-sin(t2), ew);
      if (cos(t2) >= 0.0) c2 =  pow( cos(t2), ew);
      else                c2 = -pow(-cos(t2), ew);
      P1[0] = s1*c2;
      P1[1] = s1*s2;
      P1[2] = c1;
      t1 = (i+1) * M_PI / DETAIL_LEVEL;
      if (sin(t1) >= 0.0) s1 =  pow( sin(t1), ns);
      else                s1 = -pow(-sin(t1), ns);
      if (cos(t1) >= 0.0) c1 =  pow( cos(t1), ns);
      else                c1 = -pow(-cos(t1), ns);
      t2 = j * 2.0 * M_PI / DETAIL_LEVEL;
      if (sin(t2) >= 0.0) s2 =  pow( sin(t2), ew);
      else                s2 = -pow(-sin(t2), ew);
      if (cos(t2) >= 0.0) c2 =  pow( cos(t2), ew);
      else                c2 = -pow(-cos(t2), ew);
      P2[0] = s1*c2;
      P2[1] = s1*s2;
      P2[2] = c1;
      t1 = i * M_PI / DETAIL_LEVEL;
      if (sin(t1) >= 0.0) s1 =  pow( sin(t1), ns);
      else                s1 = -pow(-sin(t1), ns);
      if (cos(t1) >= 0.0) c1 =  pow( cos(t1), ns);
      else                c1 = -pow(-cos(t1), ns);
      t2 = (j+1) * 2.0 * M_PI / DETAIL_LEVEL;
      if (sin(t2) >= 0.0) s2 =  pow( sin(t2), ew);
      else                s2 = -pow(-sin(t2), ew);
      if (cos(t2) >= 0.0) c2 =  pow( cos(t2), ew);
      else                c2 = -pow(-cos(t2), ew);
      P3[0] = s1*c2;
      P3[1] = s1*s2;
      P3[2] = c1;
      t1 = (i+1) * M_PI / DETAIL_LEVEL;
      if (sin(t1) >= 0.0) s1 =  pow( sin(t1), ns);
      else                s1 = -pow(-sin(t1), ns);
      if (cos(t1) >= 0.0) c1 =  pow( cos(t1), ns);
      else                c1 = -pow(-cos(t1), ns);
      t2 = (j+1) * 2.0 * M_PI / DETAIL_LEVEL;
      if (sin(t2) >= 0.0) s2 =  pow( sin(t2), ew);
      else                s2 = -pow(-sin(t2), ew);
      if (cos(t2) >= 0.0) c2 =  pow( cos(t2), ew);
      else                c2 = -pow(-cos(t2), ew);
      P4[0] = s1*c2;
      P4[1] = s1*s2;
      P4[2] = c1;
      gdk_draw_line(area->window, area->style->black_gc,
                    area->allocation.width /2+P1[0]*100,
                    area->allocation.height/2-P1[1]*100,
                    area->allocation.width /2+P2[0]*100,
                    area->allocation.height/2-P2[1]*100);
      gdk_draw_line(area->window, area->style->black_gc,
                    area->allocation.width /2+P2[0]*100,
                    area->allocation.height/2-P2[1]*100,
                    area->allocation.width /2+P4[0]*100,
                    area->allocation.height/2-P4[1]*100);
      gdk_draw_line(area->window, area->style->black_gc,
                    area->allocation.width /2+P4[0]*100,
                    area->allocation.height/2-P4[1]*100,
                    area->allocation.width /2+P3[0]*100,
                    area->allocation.height/2-P3[1]*100);
      gdk_draw_line(area->window, area->style->black_gc,
                    area->allocation.width /2+P3[0]*100,
                    area->allocation.height/2-P3[1]*100,
                    area->allocation.width /2+P1[0]*100,
                    area->allocation.height/2-P1[1]*100);
    }
  }
  /* And the South Pole */
  P1[0] =  0.0;
  P1[1] =  0.0;
  P1[2] = -1.0;
  for (i=0 ; i<DETAIL_LEVEL ; i++)
  { /* The triangles which are adjacents to the south pole */
    t1 = M_PI - M_PI / DETAIL_LEVEL;
    if (sin(t1) >= 0.0) s1 =  pow( sin(t1), ns);
    else                s1 = -pow(-sin(t1), ns);
    if (cos(t1) >= 0.0) c1 =  pow( cos(t1), ns);
    else                c1 = -pow(-cos(t1), ns);
    t2 = i * 2.0 * M_PI / DETAIL_LEVEL;
    if (sin(t2) >= 0.0) s2 =  pow( sin(t2), ew);
    else                s2 = -pow(-sin(t2), ew);
    if (cos(t2) >= 0.0) c2 =  pow( cos(t2), ew);
    else                c2 = -pow(-cos(t2), ew);
    P2[0] = s1*c2;
    P2[1] = s1*s2;
    P2[2] = c1;
    t2 = (i+1) * 2.0 * M_PI / DETAIL_LEVEL;
    if (sin(t2) >= 0.0) s2 =  pow( sin(t2), ew);
    else                s2 = -pow(-sin(t2), ew);
    if (cos(t2) >= 0.0) c2 =  pow( cos(t2), ew);
    else                c2 = -pow(-cos(t2), ew);
    P3[0] = s1*c2;
    P3[1] = s1*s2;
    P3[2] = c1;
    gdk_draw_line(area->window, area->style->black_gc,
                  area->allocation.width /2+P1[0]*100,
                  area->allocation.height/2-P1[1]*100,
                  area->allocation.width /2+P2[0]*100,
                  area->allocation.height/2-P2[1]*100);
    gdk_draw_line(area->window, area->style->black_gc,
                  area->allocation.width /2+P1[0]*100,
                  area->allocation.height/2-P1[1]*100,
                  area->allocation.width /2+P3[0]*100,
                  area->allocation.height/2-P3[1]*100);
    gdk_draw_line(area->window, area->style->black_gc,
                  area->allocation.width /2+P3[0]*100,
                  area->allocation.height/2-P3[1]*100,
                  area->allocation.width /2+P2[0]*100,
                  area->allocation.height/2-P2[1]*100);
  }
  return TRUE;
}

/*****************************************************************************
*  NewSuperEllipsoidOK
******************************************************************************/
static void NewSuperEllipsoidOK(GtkWidget *dialog)
{
  GtkWidget             *area;
  ViewUserDataStruct    *UserData;
  ObjectStruct          *SuperEllipsoid;
  SuperEllipsoidStruct  *SSuperEllipsoid;
  GSList                *tmp_list;
  FrameViewStruct       *TmpView;
  GtkAdjustment         *adj;
  gdouble                ns, ew;

  area = gtk_object_get_data(GTK_OBJECT(dialog), "area");
  UserData = gtk_object_get_data(GTK_OBJECT(area), "ViewData");
  adj = gtk_object_get_data(GTK_OBJECT(dialog), "east-west");
  ew = adj->value;
  adj = gtk_object_get_data(GTK_OBJECT(dialog), "north-south");
  ns = adj->value;
  SuperEllipsoid = giram_superellipsoid_new(ns, ew);
  UserData->FrameView->all_objects = g_slist_append(UserData->FrameView->all_objects,
                                                    SuperEllipsoid);
  giram_object_build_triangle_mesh(SuperEllipsoid);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
  gtk_widget_destroy(dialog);
}

/*****************************************************************************
*  ToolNewSuperEllipsoid
******************************************************************************/
void ToolNewSuperEllipsoid(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  GtkWidget *dialog, *button, *table, *label;
  GtkWidget *canvas;
  GtkAdjustment *adj;
  GtkWidget *scale;

  dialog = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dialog), _("New SuperEllipsoid"));
  gtk_object_set_data(GTK_OBJECT(dialog), "area", DrawingArea);

  /* The Ok Button */
  button = gtk_button_new_with_label(_("OK"));
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                            (GtkSignalFunc)NewSuperEllipsoidOK,
                            GTK_OBJECT(dialog));
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(dialog)->action_area), button);
  gtk_widget_grab_default(button);

  /* The Close Button */
  button = gtk_button_new_with_label(_("Close"));
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(dialog));
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(dialog)->action_area), button);

  /* The canvas, for user feedback */
  canvas = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(canvas), 200, 200);
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(dialog)->vbox), canvas);
  gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                     (GtkSignalFunc)giram_superellipsoid_dialog_area_expose,
                     dialog);

  /* The East-West Parameter */
  table = gtk_table_new(2, 2, FALSE);
  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(dialog)->vbox), table);
  
  label = gtk_label_new(_("East-West Parameter: "));
  gtk_table_attach(GTK_TABLE(table), label, 0,1, 0,1,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);

  adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 5.0, 0.1, 0.1, 1.0);
  gtk_signal_connect_object(GTK_OBJECT(adj), "value_changed",
                            (GtkSignalFunc)gtk_widget_queue_draw,
                            GTK_OBJECT(canvas));
  scale = gtk_hscale_new(adj);
  gtk_widget_set_usize(scale, 150, 30);
  gtk_range_set_update_policy(GTK_RANGE(scale), GTK_UPDATE_CONTINUOUS);
  gtk_scale_set_draw_value(GTK_SCALE(scale), TRUE);
  gtk_table_attach(GTK_TABLE(table), scale, 1,2, 0,1,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  gtk_object_set_data(GTK_OBJECT(dialog), "east-west", adj);

  /* The North-South Parameter */
  label = gtk_label_new(_("North-South Parameter: "));
  gtk_table_attach(GTK_TABLE(table), label, 0,1, 1,2,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    
  adj= (GtkAdjustment*)gtk_adjustment_new(1.0, 0.1, 5.0, 0.1, 0.1, 1.0);
  gtk_signal_connect_object(GTK_OBJECT(adj), "value_changed",
                            (GtkSignalFunc)gtk_widget_queue_draw,
                            GTK_OBJECT(canvas));
  scale = gtk_hscale_new(adj);
  gtk_widget_set_usize(scale, 150, 30);
  gtk_range_set_update_policy(GTK_RANGE(scale), GTK_UPDATE_CONTINUOUS);
  gtk_scale_set_draw_value(GTK_SCALE(scale), TRUE);
  gtk_table_attach(GTK_TABLE(table), scale, 1,2, 1,2,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  gtk_object_set_data(GTK_OBJECT(dialog), "north-south", adj);
  
  gtk_widget_show_all(dialog);
}

/*****************************************************************************
*  giram_superellipsoid_new
******************************************************************************/
ObjectStruct *giram_superellipsoid_new(double N, double E)
{
  ObjectStruct *SuperEllipsoid;
  SuperEllipsoidStruct *SSuperEllipsoid;
  static GiramObjectClass *superellipsoid_klass = NULL;

  if (superellipsoid_klass == NULL)
  {
    superellipsoid_klass = giram_object_class_new();

    superellipsoid_klass->build_triangle_mesh       = giram_superellipsoid_build_triangle_mesh;
    superellipsoid_klass->inside                    = giram_superellipsoid_inside;
    superellipsoid_klass->find_intersection_segment = giram_superellipsoid_find_intersection_segment;
  }
  SSuperEllipsoid = g_new(SuperEllipsoidStruct, 1);
  SuperEllipsoid = (ObjectStruct *)SSuperEllipsoid;
  InitObject(SuperEllipsoid);
  SuperEllipsoid->Type = SUPERELLIPSOID_OBJECT;
  SuperEllipsoid->klass = superellipsoid_klass;
  SSuperEllipsoid->E = E;
  SSuperEllipsoid->N = N;
  return SuperEllipsoid;
}

/*****************************************************************************
*  PDBCreateSuperEllipsoid
******************************************************************************/
int PDBCreateSuperEllipsoid(int *Id, double *N, double *E)
{
  ObjectStruct         *SuperEllipsoid;
  SuperEllipsoidStruct *SSuperEllipsoid;
  FrameStruct          *LocalFrame;
  GSList               *tmp_list;
  FrameViewStruct      *TmpView;

  SuperEllipsoid = giram_superellipsoid_new(*N, *E);

  for (LocalFrame = FirstFrame ;
       LocalFrame->Id != *Id ;
       LocalFrame = LocalFrame->Next)
                  ;
  LocalFrame->all_objects = g_slist_append(LocalFrame->all_objects, SuperEllipsoid);
  giram_object_build_triangle_mesh(SuperEllipsoid);
  for (tmp_list = LocalFrame->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, LocalFrame);
  return SuperEllipsoid->Id;
}

