/* tool_move.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "view.h"
#include "object.h"
#include "snap.h"
#include "utils.h"
#include "tool_move.h"

#include "giramintl.h"

TriangleListStruct *AllSelectionTri, *InitialSelectionTri;
double Xorg, Yorg;
GdkColormap *colormap;
GdkGC *gcPurple, *gcBlack;
GdkColor Purple, Black;
guint id;

#define NONE   1
#define MOVING 2
int TRANSLATE_STATE = NONE;

/*****************************************************************************
*  ToolInitMoveSelection
******************************************************************************/
void ToolInitMoveSelection(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  ViewUserDataStruct *UserData;
  double Zoom, Xoff, Yoff;
  GtkWidget *StatusBar;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  id = gtk_statusbar_get_context_id(GTK_STATUSBAR(StatusBar), "translate");
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  MouseToReal(bevent->x, bevent->y, DrawingArea, &Xorg, &Yorg);
  
  if (UserData->FrameView->Selection == NULL)
    return;
  DrawUnSelected(DrawingArea);
  colormap = gdk_window_get_colormap(DrawingArea->window);
  gcPurple = gdk_gc_new(DrawingArea->window);
  Purple.red = 65535; Purple.green = 0; Purple.blue = 65535;
  /* XXX: the color should be settable */
  gdk_color_alloc(colormap, &Purple);
  gdk_gc_set_foreground(gcPurple, &Purple);

  gcBlack = gdk_gc_new(DrawingArea->window);
  Black.red = 0; Black.green = 0; Black.blue = 0;
  gdk_color_alloc(colormap, &Black);
  gdk_gc_set_foreground(gcBlack, &Black);
  
  AllSelectionTri = ComputeSelectionTriangleMesh(UserData->FrameView->Selection);
  InitialSelectionTri = ComputeSelectionTriangleMesh(UserData->FrameView->Selection);

  DrawSelection(DrawingArea, AllSelectionTri, gcPurple, Xoff, Yoff, Zoom);
                          
  TRANSLATE_STATE = MOVING;
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, _("Translation vector: "));
  gdk_pointer_grab(DrawingArea->window, FALSE,
                   GDK_POINTER_MOTION_MASK | 
                   GDK_POINTER_MOTION_HINT_MASK |
                   GDK_BUTTON_RELEASE_MASK,
                   NULL, NULL, bevent->time);
}

/*****************************************************************************
*  ToolMotionMoveSelection
******************************************************************************/
void ToolMotionMoveSelection(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventMotion *Mev = (GdkEventMotion *)event;
  ViewUserDataStruct *UserData;
  GtkWidget *StatusBar;
  double Zoom, Xoff, Yoff, Xold, Yold;
  double XTrans, YTrans;
  Vector Trans;
  TriangleListStruct *TmpTri, *TmpOrigTri;
  gchar *Message;

  if (TRANSLATE_STATE != MOVING)
    return;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;

  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);

  DrawSelection(DrawingArea, AllSelectionTri, gcBlack, Xoff, Yoff, Zoom);
  MouseToReal(Mev->x, Mev->y, DrawingArea, &Xold, &Yold);
    
  XTrans = Xold-Xorg;
  YTrans = Yold-Yorg;
  if (Mev->state & GDK_SHIFT_MASK)
  {
    if (fabs(XTrans) > fabs(YTrans))
      YTrans = 0.0;
    else
      XTrans = 0.0;
  }
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Trans[0] = XTrans;
      Trans[1] = YTrans;
      Trans[2] = 0.0;
      break;
    case ORTHO_XZ_CAMERA:
      Trans[0] = XTrans;
      Trans[1] = 0.0;
      Trans[2] = YTrans;
      break;
    case ORTHO_ZY_CAMERA:
      Trans[0] = 0.0;
      Trans[1] = YTrans;
      Trans[2] = XTrans;
      break;
  }
  Message = g_strdup_printf(_("Translation vector: <%g, %g, %g>"),
                            Trans[0], Trans[1], Trans[2]);
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, Message);
  g_free(Message);
  for (TmpTri = AllSelectionTri, TmpOrigTri = InitialSelectionTri ;
       TmpTri ;
       TmpTri = TmpTri->Next, TmpOrigTri = TmpOrigTri->Next)
  {
    TmpTri->P1[0] = TmpOrigTri->P1[0] + Trans[0];
    TmpTri->P1[1] = TmpOrigTri->P1[1] + Trans[1];
    TmpTri->P1[2] = TmpOrigTri->P1[2] + Trans[2];
    TmpTri->P2[0] = TmpOrigTri->P2[0] + Trans[0];
    TmpTri->P2[1] = TmpOrigTri->P2[1] + Trans[1];
    TmpTri->P2[2] = TmpOrigTri->P2[2] + Trans[2];
    TmpTri->P3[0] = TmpOrigTri->P3[0] + Trans[0];
    TmpTri->P3[1] = TmpOrigTri->P3[1] + Trans[1];
    TmpTri->P3[2] = TmpOrigTri->P3[2] + Trans[2];
  }
  DrawSelection(DrawingArea, AllSelectionTri, gcPurple, Xoff, Yoff, Zoom);
}
 
/*****************************************************************************
*  ToolReleaseMoveSelection
******************************************************************************/
void ToolReleaseMoveSelection(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventButton *Bev = (GdkEventButton *)event;
  ViewUserDataStruct *UserData;
  GtkWidget          *StatusBar;
  double              Xold, Yold;
  double              Zoom;
  double              XTrans, YTrans;
  Vector              Trans;
  ObjectSelection    *TmpSel;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;

  if (TRANSLATE_STATE != MOVING)
    return;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;

  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);
  MouseToReal(Bev->x, Bev->y, DrawingArea, &Xold, &Yold);
  
  XTrans = Xold-Xorg;
  YTrans = Yold-Yorg;
  if (Bev->state & GDK_SHIFT_MASK)
  {
    if (fabs(XTrans) > fabs(YTrans))
      YTrans = 0.0;
    else
      XTrans = 0.0;
  }
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Trans[0] = XTrans;
      Trans[1] = YTrans;
      Trans[2] = 0.0;
      break;
    case ORTHO_XZ_CAMERA:
      Trans[0] = XTrans;
      Trans[1] = 0.0;
      Trans[2] = YTrans;
      break;
    case ORTHO_ZY_CAMERA:
      Trans[0] = 0.0;
      Trans[1] = YTrans;
      Trans[2] = XTrans;
      break;
  }
  for (TmpSel = UserData->FrameView->Selection ; 
       TmpSel->Next != UserData->FrameView->Selection ; 
       TmpSel = TmpSel->Next)
  {
    giram_object_translate(TmpSel->Object, Trans);
    giram_object_build_triangle_mesh(TmpSel->Object);
  }
  giram_object_translate(TmpSel->Object, Trans);
  giram_object_build_triangle_mesh(TmpSel->Object);
  TRANSLATE_STATE = NONE;
  { /* Freeing the two temporary mesh */
    TriangleListStruct *Tmp1, *Tmp2;

    Tmp1 = AllSelectionTri;
    AllSelectionTri = NULL;
    while(Tmp1)
    {
      Tmp2=Tmp1->Next;
      g_free(Tmp1);
      Tmp1=Tmp2;
    }
    Tmp1 = InitialSelectionTri;
    InitialSelectionTri = NULL;
    while(Tmp1)
    {
      Tmp2=Tmp1->Next;
      g_free(Tmp1);
      Tmp1=Tmp2;
    }
  }
  gdk_gc_destroy(gcPurple);
  gdk_gc_destroy(gcBlack);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  gdk_pointer_ungrab(Bev->time);
}
