/* tool_options.c 
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Many ideas in this modules are borrowed from gimp/app/tools.c
 * (C) 1995 Spencer Kimball and Peter Mattis
 */
#include "giram.h"
#include "tool_options.h"
#include "tool_rotate.h"
#include "tool_scale.h"
#include "cone.h"
#include "cylinder.h"
#include "disc.h"
#include "tools.h"

#include "giramintl.h"

GtkWidget *ToolOptionsDialog;

GtkWidget *ToolOptionsBox[MT_LAST];

/*****************************************************************************
*  ShowToolOptionsDialog
******************************************************************************/
void ShowToolOptionsDialog(GtkWidget *DrawingArea)
{
  if (!GTK_WIDGET_VISIBLE(ToolOptionsDialog))
  {
    gtk_widget_show(ToolOptionsDialog);
  }
  else
  {
    gdk_window_raise(ToolOptionsDialog->window);
  }
}

/*****************************************************************************
*  ToolsOptionsDeleteFunc
******************************************************************************/
static gint ToolsOptionsDeleteFunc(GtkWidget *Wid, GdkEvent *ev, GtkWidget *Dialog)
{
  gtk_widget_hide(Dialog);
  return TRUE;
}

/*****************************************************************************
*  ToolsOptionsCloseFunc
******************************************************************************/
static gint ToolsOptionsCloseFunc(GtkWidget *Wid, GtkWidget *Dialog)
{
  gtk_widget_hide(Dialog);
  return TRUE;
}

/*****************************************************************************
*  BuildNoOptions
******************************************************************************/
GtkWidget *BuildNoOptions(GtkWidget *VBox, char *ToolName)
{
  GtkWidget *vbox;
  GtkWidget *label;
  /*  the main vbox  */
  vbox = gtk_vbox_new (FALSE, 1);
  /*  the main label  */
  label = gtk_label_new(ToolName);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  /*  this tool has no special options  */
  label = gtk_label_new(_("This tool has no options."));
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(VBox), vbox, TRUE, TRUE, 0);
  gtk_widget_show(vbox);
  return vbox;
}

/*****************************************************************************
*  ShowOptionsForTool
******************************************************************************/
void ShowOptionsForTool(GiramModeToolType Tool)
{
  GiramModeToolType i;

  for (i=MT_SELECT ; i<MT_LAST ; i++)
  {
    if (i==Tool)
      gtk_widget_show(ToolOptionsBox[i]);
    else
      gtk_widget_hide(ToolOptionsBox[i]);
  }
}

/*****************************************************************************
*  CreateToolOptionsDialog
******************************************************************************/
void CreateToolOptionsDialog(void)
{
  GtkWidget *VBox;
  GtkWidget *Button;

  ToolOptionsDialog = gtk_dialog_new();
  gtk_window_set_wmclass(GTK_WINDOW(ToolOptionsDialog), "tool_options", "Giram");
  gtk_window_set_title(GTK_WINDOW(ToolOptionsDialog), _("Tool Options"));
  gtk_window_set_policy(GTK_WINDOW(ToolOptionsDialog), FALSE, TRUE, TRUE);

  VBox = gtk_vbox_new(FALSE, 2);
  gtk_container_border_width(GTK_CONTAINER(VBox), 2);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(ToolOptionsDialog)->vbox), VBox, TRUE, TRUE, 0);
  gtk_widget_show(VBox);
  /* handle the window manager trying to close the window */
  gtk_signal_connect(GTK_OBJECT(ToolOptionsDialog), "delete_event",
                     GTK_SIGNAL_FUNC(ToolsOptionsDeleteFunc),
                     ToolOptionsDialog);
          
  gtk_container_border_width(GTK_CONTAINER(GTK_DIALOG(ToolOptionsDialog)->action_area), 2);
  Button = gtk_button_new_with_label("Close");
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(ToolOptionsDialog)->action_area), Button, TRUE, TRUE, 0);
  gtk_signal_connect(GTK_OBJECT(Button), "clicked",
                     GTK_SIGNAL_FUNC(ToolsOptionsCloseFunc),
                     ToolOptionsDialog);
        
  gtk_widget_grab_default(Button);
  gtk_widget_show(Button);
    
  ToolOptionsBox[MT_SELECT] = BuildSelectOptions(VBox);
  ToolOptionsBox[MT_MOVE] = BuildNoOptions(VBox, _("Move"));
  ToolOptionsBox[MT_ROTATE] = BuildRotateOptions(VBox);
  ToolOptionsBox[MT_SCALE] = BuildScaleOptions(VBox);
/*  ToolOptionsBox[MT_NEW_BICUBIC_PATCH] = BuildNoOptions(VBox, _("New Bicubic Patch"));
  ToolOptionsBox[MT_NEW_BLOB] = BuildNoOptions(VBox, _("New Blob"));*/
  ToolOptionsBox[MT_NEW_BOX] = BuildNoOptions(VBox, _("New Box"));
  ToolOptionsBox[MT_NEW_CONE] = BuildConeOptions(VBox);
/*  ToolOptionsBox[MT_NEW_CSG] = BuildNoOptions(VBox, _("New CSG"));
  ToolOptionsBox[MT_NEW_CUBIC] = BuildNoOptions(VBox, _("New Cubic"));*/
  ToolOptionsBox[MT_NEW_CYLINDER] = BuildCylinderOptions(VBox);
  ToolOptionsBox[MT_NEW_DISC] = BuildDiscOptions(VBox);
/*  ToolOptionsBox[MT_NEW_HEIGHT_FIELD] = BuildNoOptions(VBox, _("New Height Field"));
  ToolOptionsBox[MT_NEW_JULIA_FRACTAL] = BuildNoOptions(VBox, _("New Julia Fractal"));
  ToolOptionsBox[MT_NEW_LATHE] = BuildNoOptions(VBox, _("New Lathe"));
  ToolOptionsBox[MT_NEW_MESH] = BuildNoOptions(VBox, _("New Mesh"));*/
  ToolOptionsBox[MT_NEW_PLANE] = BuildNoOptions(VBox, _("New Plane"));
/*  ToolOptionsBox[MT_NEW_POLY] = BuildNoOptions(VBox, _("New Poly"));
  ToolOptionsBox[MT_NEW_POLYGON] = BuildNoOptions(VBox, _("New Polygon"));
  ToolOptionsBox[MT_NEW_PRISM] = BuildNoOptions(VBox, _("New Prism"));
  ToolOptionsBox[MT_NEW_QUADRIC] = BuildNoOptions(VBox, _("New Quadric"));
  ToolOptionsBox[MT_NEW_QUARTIC] = BuildNoOptions(VBox, _("New Quartic"));*/
  ToolOptionsBox[MT_NEW_SOR] = BuildNoOptions(VBox, _("New Sor"));
  ToolOptionsBox[MT_NEW_SPHERE] = BuildNoOptions(VBox, _("New Sphere"));
  ToolOptionsBox[MT_NEW_SUPERELLIPSOID] = BuildNoOptions(VBox, _("New SuperEllipsoid"));
/*  ToolOptionsBox[MT_NEW_TEXT] = BuildNoOptions(VBox, _("New Text"));*/
  ToolOptionsBox[MT_NEW_TORUS] = BuildNoOptions(VBox, _("New Torus"));
  ToolOptionsBox[MT_NEW_TRIANGLE] = BuildNoOptions(VBox, _("New Triangle"));
  ToolOptionsBox[MT_NEW_LIGHTSOURCE] = BuildNoOptions(VBox, _("New Light Source"));
}

