/* tool_rotate.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "view.h"
#include "object.h"
#include "utils.h"
#include "tool_rotate.h"

#include "giramintl.h"

TriangleListStruct *AllSelectionTri, *InitialSelectionTri;
double Xorg, Yorg, OrgNorme;
GdkColormap *colormap;
GdkGC *gcPurple, *gcBlack;
GdkColor Purple, Black;
guint id;

#define NONE   1
#define MOVING 2
int ROTATE_STATE = NONE;

int SnapToAngleFlag;
double SnapAngle;

/*****************************************************************************
*  ToggleSnapToAngle
******************************************************************************/
void ToggleSnapToAngle(GtkWidget *SpinButton)
{
  SnapToAngleFlag = !SnapToAngleFlag;
  gtk_widget_set_sensitive(SpinButton, SnapToAngleFlag);
}

/*****************************************************************************
*  SnapAngleChange
******************************************************************************/
void SnapAngleChange(GtkWidget *SpinButton)
{
  SnapAngle = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(SpinButton));
}

/*****************************************************************************
*  BuildRotateOptions
******************************************************************************/
GtkWidget *BuildRotateOptions(GtkWidget *VBox)
{
  GtkWidget *vbox;
  GtkWidget *SnapToAngle;
  GtkWidget *Label;
  GtkWidget *HBox;
  GtkWidget *SpinButton;
  GtkAdjustment *adj;

  /*  the main vbox  */
  vbox = gtk_vbox_new(FALSE, 1);
  /*  the main label  */
  Label = gtk_label_new(_("Rotate"));
  gtk_box_pack_start(GTK_BOX(vbox), Label, FALSE, FALSE, 0);
  gtk_widget_show(Label);

  /* The Toggle button for enabling/disabling snaping */
  SnapToAngle = gtk_check_button_new_with_label(_("Snap to angle"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(SnapToAngle), SnapToAngleFlag);
  gtk_box_pack_start(GTK_BOX(vbox), SnapToAngle, FALSE, FALSE, 0);
  gtk_widget_show(SnapToAngle);

  /* The horizontal box for the angle label/spin button pair */
  HBox = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX(vbox), HBox, TRUE, TRUE, 0);
  gtk_widget_show(HBox);
  /* The Label for the angle */
  Label = gtk_label_new(_("Angle:"));
  gtk_box_pack_start(GTK_BOX(HBox), Label, FALSE, FALSE, 0);
  gtk_widget_show(Label);
  /* The spin button for the angle */
  adj = (GtkAdjustment *) gtk_adjustment_new(SnapAngle, 1.0, 90.0, 1.0, 5.0, 0.0);
  SpinButton = gtk_spin_button_new(adj, 0.2, 2);
  gtk_widget_set_sensitive(SpinButton, SnapToAngleFlag);
  gtk_box_pack_start(GTK_BOX(HBox), SpinButton, TRUE, TRUE, 0);
  gtk_widget_show(SpinButton);

  gtk_signal_connect_object(GTK_OBJECT(SnapToAngle), "toggled",
                            GTK_SIGNAL_FUNC(ToggleSnapToAngle), 
                            GTK_OBJECT(SpinButton));

  gtk_signal_connect_object(GTK_OBJECT(adj), "value_changed",
                            GTK_SIGNAL_FUNC(SnapAngleChange),
                            GTK_OBJECT(SpinButton));

  gtk_box_pack_start(GTK_BOX(VBox), vbox, TRUE, TRUE, 0);
  gtk_widget_show(vbox);
  return vbox;
}

/*****************************************************************************
*  ToolInitRotateSelection
******************************************************************************/
void ToolInitRotateSelection(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  ViewUserDataStruct *UserData;
  double              Zoom, Xoff, Yoff;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;
  GtkWidget          *StatusBar;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  id = gtk_statusbar_get_context_id(GTK_STATUSBAR(StatusBar), "rotate");
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  Xorg = Xoff+(bevent->x-DrawingArea->allocation.width / 2)/Zoom; 
  Yorg = Yoff-(bevent->y-DrawingArea->allocation.height/ 2)/Zoom;

  if (bevent->state & 4) /* Ctrl */
  { /* The user only want to move the Rotation Center */
    switch (UserData->CameraStyle)
    {
      case ORTHO_XY_CAMERA:
        UserData->FrameView->RotationCenter[0] = Xorg;
        UserData->FrameView->RotationCenter[1] = Yorg;
        break;
      case ORTHO_XZ_CAMERA:
        UserData->FrameView->RotationCenter[0] = Xorg;
        UserData->FrameView->RotationCenter[2] = Yorg;
        break;
      case ORTHO_ZY_CAMERA:
        UserData->FrameView->RotationCenter[2] = Xorg;
        UserData->FrameView->RotationCenter[1] = Yorg;
        break;
    }
    for (tmp_list = UserData->FrameView->all_views ;
         tmp_list ;
         tmp_list = tmp_list->next)
    {
      TmpView = tmp_list->data;
      gtk_widget_queue_draw(TmpView->drawing_area);
    }
    return;
  }
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Xorg -= UserData->FrameView->RotationCenter[0];
      Yorg -= UserData->FrameView->RotationCenter[1];
      break;
    case ORTHO_XZ_CAMERA:
      Xorg -= UserData->FrameView->RotationCenter[0];
      Yorg -= UserData->FrameView->RotationCenter[2];
      break;
    case ORTHO_ZY_CAMERA:
      Xorg -= UserData->FrameView->RotationCenter[2];
      Yorg -= UserData->FrameView->RotationCenter[1];
      break;
  }
  OrgNorme = sqrt(Xorg*Xorg+Yorg*Yorg);
  if (OrgNorme == 0.0)
    return;
  if (UserData->FrameView->Selection == NULL)
    return;
  DrawUnSelected(DrawingArea);
  colormap = gdk_window_get_colormap(DrawingArea->window);
  gcPurple = gdk_gc_new(DrawingArea->window);
  Purple.red = 65535; Purple.green = 0; Purple.blue = 65535;
  /* XXX: that color should be settable */
  gdk_color_alloc(colormap, &Purple);
  gdk_gc_set_foreground(gcPurple, &Purple);

  gcBlack = gdk_gc_new(DrawingArea->window);
  Black.red = 0; Black.green = 0; Black.blue = 0;
  gdk_color_alloc(colormap, &Black);
  gdk_gc_set_foreground(gcBlack, &Black);

  AllSelectionTri = ComputeSelectionTriangleMesh(UserData->FrameView->Selection);
  InitialSelectionTri = ComputeSelectionTriangleMesh(UserData->FrameView->Selection);

  DrawSelection(DrawingArea, AllSelectionTri, gcPurple, Xoff, Yoff, Zoom);

  ROTATE_STATE = MOVING;
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, _("Angle: "));
  gdk_pointer_grab(DrawingArea->window, FALSE,
                   GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK,
                   NULL, NULL, bevent->time);
}

/*****************************************************************************
*  ToolMotionRotateSelection
******************************************************************************/
void ToolMotionRotateSelection(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventMotion *Mev = (GdkEventMotion *)event;
  ViewUserDataStruct *UserData;
  GtkWidget *StatusBar;
  double Zoom, Xoff, Yoff;
  double XTrans, YTrans;
  double Norme, CosAngle, Angle;
  Matrix MatTrans;
  TriangleListStruct *TmpTri, *TmpOrigTri;
  gchar *Message;

  if (ROTATE_STATE != MOVING)
    return;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
      
  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);
  DrawSelection(DrawingArea, AllSelectionTri, gcBlack, Xoff, Yoff, Zoom);
  XTrans = Xoff+(Mev->x-DrawingArea->allocation.width /2)/Zoom;
  YTrans = Yoff-(Mev->y-DrawingArea->allocation.height/2)/Zoom;
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      XTrans -= UserData->FrameView->RotationCenter[0];
      YTrans -= UserData->FrameView->RotationCenter[1];
      break;
    case ORTHO_XZ_CAMERA:
      XTrans -= UserData->FrameView->RotationCenter[0];
      YTrans -= UserData->FrameView->RotationCenter[2];
      break;
    case ORTHO_ZY_CAMERA:
      XTrans -= UserData->FrameView->RotationCenter[2];
      YTrans -= UserData->FrameView->RotationCenter[1];
      break;
  }
  Norme = sqrt(XTrans*XTrans+YTrans*YTrans);
  if (Norme == 0.0)
    Angle = 0.0;
  else
  {
    CosAngle = (Xorg*XTrans+Yorg*YTrans) / (OrgNorme * Norme);
    Angle = acos(CosAngle);
    if (XTrans*Yorg-Xorg*YTrans > 0.0)
    {
      Angle = -Angle;
    }
  }
  if (SnapToAngleFlag)
  {
    Angle = SnapAngle*(floor(Angle*180.0/M_PI/SnapAngle))/180.0*M_PI;
  }
  Message = g_strdup_printf(_("Angle: %g"), Angle*180.0/M_PI);
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, Message);
  g_free(Message);
  MIdentity(MatTrans);
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      MatTrans[0][0] = cos(Angle);
      MatTrans[1][1] = cos(Angle);
      MatTrans[0][1] = sin(Angle);
      MatTrans[1][0] = -sin(Angle);
      MatTrans[3][0] = -UserData->FrameView->RotationCenter[0]*cos(Angle)
                       +UserData->FrameView->RotationCenter[1]*sin(Angle)
                       +UserData->FrameView->RotationCenter[0];
      MatTrans[3][1] = -UserData->FrameView->RotationCenter[0]*sin(Angle)
                       -UserData->FrameView->RotationCenter[1]*cos(Angle)
                       +UserData->FrameView->RotationCenter[1];
      break;
    case ORTHO_XZ_CAMERA:
      MatTrans[0][0] = cos(Angle);
      MatTrans[2][2] = cos(Angle);
      MatTrans[0][2] = sin(Angle);
      MatTrans[2][0] = -sin(Angle);
      MatTrans[3][0] = -UserData->FrameView->RotationCenter[0]*cos(Angle)
                       +UserData->FrameView->RotationCenter[2]*sin(Angle)
                       +UserData->FrameView->RotationCenter[0];
      MatTrans[3][2] = -UserData->FrameView->RotationCenter[0]*sin(Angle)
                       -UserData->FrameView->RotationCenter[2]*cos(Angle)
                       +UserData->FrameView->RotationCenter[2];
      break;
    case ORTHO_ZY_CAMERA:
      MatTrans[2][2] = cos(Angle);
      MatTrans[1][1] = cos(Angle);
      MatTrans[2][1] = sin(Angle);
      MatTrans[1][2] = -sin(Angle);
      MatTrans[3][1] = -UserData->FrameView->RotationCenter[1]*cos(Angle)
                       -UserData->FrameView->RotationCenter[2]*sin(Angle)
                       +UserData->FrameView->RotationCenter[1];
      MatTrans[3][2] =  UserData->FrameView->RotationCenter[1]*sin(Angle)
                       -UserData->FrameView->RotationCenter[2]*cos(Angle)
                       +UserData->FrameView->RotationCenter[2];
      break;
  }
  for (TmpTri = AllSelectionTri, TmpOrigTri = InitialSelectionTri ;
       TmpTri ;
       TmpTri = TmpTri->Next, TmpOrigTri = TmpOrigTri->Next)
  {
    TmpTri->P1[0] = TmpOrigTri->P1[0]*MatTrans[0][0] + TmpOrigTri->P1[1]*MatTrans[1][0] + TmpOrigTri->P1[2]*MatTrans[2][0] + MatTrans[3][0];
    TmpTri->P1[1] = TmpOrigTri->P1[0]*MatTrans[0][1] + TmpOrigTri->P1[1]*MatTrans[1][1] + TmpOrigTri->P1[2]*MatTrans[2][1] + MatTrans[3][1];
    TmpTri->P1[2] = TmpOrigTri->P1[0]*MatTrans[0][2] + TmpOrigTri->P1[1]*MatTrans[1][2] + TmpOrigTri->P1[2]*MatTrans[2][2] + MatTrans[3][2];
    TmpTri->P2[0] = TmpOrigTri->P2[0]*MatTrans[0][0] + TmpOrigTri->P2[1]*MatTrans[1][0] + TmpOrigTri->P2[2]*MatTrans[2][0] + MatTrans[3][0];
    TmpTri->P2[1] = TmpOrigTri->P2[0]*MatTrans[0][1] + TmpOrigTri->P2[1]*MatTrans[1][1] + TmpOrigTri->P2[2]*MatTrans[2][1] + MatTrans[3][1];
    TmpTri->P2[2] = TmpOrigTri->P2[0]*MatTrans[0][2] + TmpOrigTri->P2[1]*MatTrans[1][2] + TmpOrigTri->P2[2]*MatTrans[2][2] + MatTrans[3][2];
    TmpTri->P3[0] = TmpOrigTri->P3[0]*MatTrans[0][0] + TmpOrigTri->P3[1]*MatTrans[1][0] + TmpOrigTri->P3[2]*MatTrans[2][0] + MatTrans[3][0];
    TmpTri->P3[1] = TmpOrigTri->P3[0]*MatTrans[0][1] + TmpOrigTri->P3[1]*MatTrans[1][1] + TmpOrigTri->P3[2]*MatTrans[2][1] + MatTrans[3][1];
    TmpTri->P3[2] = TmpOrigTri->P3[0]*MatTrans[0][2] + TmpOrigTri->P3[1]*MatTrans[1][2] + TmpOrigTri->P3[2]*MatTrans[2][2] + MatTrans[3][2];
  }
  DrawSelection(DrawingArea, AllSelectionTri, gcPurple, Xoff, Yoff, Zoom);
}

/*****************************************************************************
*  ToolReleaseRotateSelection
******************************************************************************/
void ToolReleaseRotateSelection(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventButton *Bev = (GdkEventButton *)event;
  ViewUserDataStruct *UserData;
  GtkWidget          *StatusBar;
  double              Zoom, Xoff, Yoff;
  double              XTrans, YTrans;
  double              Norme, CosAngle, Angle;
  Vector              RotateVect, TransVect;
  ObjectSelection    *TmpSel;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;

  if (ROTATE_STATE != MOVING)
    return;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;

  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);
  XTrans = Xoff+(Bev->x-DrawingArea->allocation.width /2)/Zoom;
  YTrans = Yoff-(Bev->y-DrawingArea->allocation.height/2)/Zoom;
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      XTrans -= UserData->FrameView->RotationCenter[0];
      YTrans -= UserData->FrameView->RotationCenter[1];
      break;
    case ORTHO_XZ_CAMERA:
      XTrans -= UserData->FrameView->RotationCenter[0];
      YTrans -= UserData->FrameView->RotationCenter[2];
      break;
    case ORTHO_ZY_CAMERA:
      XTrans -= UserData->FrameView->RotationCenter[2];
      YTrans -= UserData->FrameView->RotationCenter[1];
      break;
  }
  Norme = sqrt(XTrans*XTrans+YTrans*YTrans);
  if (Norme == 0.0)
    Angle = 0.0;
  else
  {
    CosAngle = (Xorg*XTrans+Yorg*YTrans) / (OrgNorme * Norme);
    Angle = acos(CosAngle);
    if (XTrans*Yorg-Xorg*YTrans > 0.0)
    {
      Angle = -Angle;
    }
  }
  if (SnapToAngleFlag)
  {
    Angle = SnapAngle*(floor(Angle*180.0/M_PI/SnapAngle))/180.0*M_PI;
  }
  Angle = Angle*180/M_PI;
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      RotateVect[0] = 0.0;
      RotateVect[1] = 0.0;
      RotateVect[2] = Angle;
      break;
    case ORTHO_XZ_CAMERA:
      RotateVect[0] = 0.0;
      RotateVect[1] = -Angle;
      RotateVect[2] = 0.0;
      break;
    case ORTHO_ZY_CAMERA:
      RotateVect[0] = -Angle;
      RotateVect[1] = 0.0;
      RotateVect[2] = 0.0;
      break;
  }
  TransVect[0] = -UserData->FrameView->RotationCenter[0];
  TransVect[1] = -UserData->FrameView->RotationCenter[1];
  TransVect[2] = -UserData->FrameView->RotationCenter[2];
  for (TmpSel = UserData->FrameView->Selection ;
       TmpSel->Next != UserData->FrameView->Selection ;
       TmpSel = TmpSel->Next)
  {
    giram_object_translate(TmpSel->Object, TransVect);
    giram_object_rotate(TmpSel->Object, RotateVect);
    giram_object_translate(TmpSel->Object, UserData->FrameView->RotationCenter);
    giram_object_build_triangle_mesh(TmpSel->Object);
  }
  giram_object_translate(TmpSel->Object, TransVect);
  giram_object_rotate(TmpSel->Object, RotateVect);
  giram_object_translate(TmpSel->Object, UserData->FrameView->RotationCenter);
  giram_object_build_triangle_mesh(TmpSel->Object);
  ROTATE_STATE = NONE;
  { /* Freeing the two temporary mesh */
    TriangleListStruct *Tmp1, *Tmp2;
    Tmp1 = AllSelectionTri;
    AllSelectionTri = NULL;
    while(Tmp1)
    {
      Tmp2=Tmp1->Next;
      g_free(Tmp1);
      Tmp1=Tmp2;
    }
    Tmp1 = InitialSelectionTri;
    InitialSelectionTri = NULL;
    while(Tmp1)
    {
      Tmp2=Tmp1->Next;
      g_free(Tmp1);
      Tmp1=Tmp2;
    }
  }
  gdk_gc_destroy(gcPurple);
  gdk_gc_destroy(gcBlack);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  gdk_pointer_ungrab(Bev->time);
}

