/* tool_scale.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2001 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "view.h"
#include "object.h"
#include "snap.h"
#include "utils.h"
#include "tool_scale.h"

#include "giramintl.h"

TriangleListStruct *AllSelectionTri, *InitialSelectionTri;
double Xorg, Yorg, OrgNorme;
GdkColormap *colormap;
GdkGC *gcPurple, *gcBlack;
GdkColor Purple, Black;
guint id;

#define NONE   1
#define MOVING 2
int SCALE_STATE = NONE;

int EvenlyScaleFlag;

/*****************************************************************************
*  ToggleEvenlyScale
******************************************************************************/
void ToggleEvenlyScale(GtkWidget *Check)
{
  EvenlyScaleFlag = !EvenlyScaleFlag;
}

/*****************************************************************************
*  BuildScaleOptions
******************************************************************************/
GtkWidget *BuildScaleOptions(GtkWidget *VBox)
{
  GtkWidget *vbox;
  GtkWidget *CheckEvenly;
  GtkWidget *label;

  /*  the main vbox  */
  vbox = gtk_vbox_new (FALSE, 1);
  /*  the main label  */
  label = gtk_label_new(_("Scale"));
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  
  CheckEvenly = gtk_check_button_new_with_label(_("Scale evenly"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(CheckEvenly), EvenlyScaleFlag);
  gtk_signal_connect(GTK_OBJECT(CheckEvenly), "toggled",
                     GTK_SIGNAL_FUNC(ToggleEvenlyScale), NULL);
  gtk_box_pack_start(GTK_BOX(vbox), CheckEvenly, FALSE, FALSE, 0);
  gtk_widget_show(CheckEvenly);

  gtk_box_pack_start(GTK_BOX(VBox), vbox, TRUE, TRUE, 0);
  gtk_widget_show(vbox);
  return vbox;
}

/*****************************************************************************
*  ToolInitScaleSelection
******************************************************************************/
void ToolInitScaleSelection(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  ViewUserDataStruct *UserData;
  double              Zoom, Xoff, Yoff;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;
  GtkWidget          *StatusBar;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  id = gtk_statusbar_get_context_id(GTK_STATUSBAR(StatusBar), "scale");
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  MouseToReal(bevent->x, bevent->y, DrawingArea, &Xorg, &Yorg);
  
  if (bevent->state & 4) /* Ctrl */
  { /* The user only want to move the Scale Center */
    switch (UserData->CameraStyle)
    {
      case ORTHO_XY_CAMERA:
        UserData->FrameView->ScaleCenter[0] = Xorg;
        UserData->FrameView->ScaleCenter[1] = Yorg;
        break;
      case ORTHO_XZ_CAMERA:
        UserData->FrameView->ScaleCenter[0] = Xorg;
        UserData->FrameView->ScaleCenter[2] = Yorg;
        break;
      case ORTHO_ZY_CAMERA:
        UserData->FrameView->ScaleCenter[2] = Xorg;
        UserData->FrameView->ScaleCenter[1] = Yorg;
        break;
    }
    for (tmp_list = UserData->FrameView->all_views ;
         tmp_list ;
         tmp_list = tmp_list->next)
    {
      TmpView = tmp_list->data;
      gtk_widget_queue_draw(TmpView->drawing_area);
    }
    return;
  }
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Xorg -= UserData->FrameView->ScaleCenter[0];
      Yorg -= UserData->FrameView->ScaleCenter[1];
      break;
    case ORTHO_XZ_CAMERA:
      Xorg -= UserData->FrameView->ScaleCenter[0];
      Yorg -= UserData->FrameView->ScaleCenter[2];
      break;
    case ORTHO_ZY_CAMERA:
      Xorg -= UserData->FrameView->ScaleCenter[2];
      Yorg -= UserData->FrameView->ScaleCenter[1];
      break;
  }
  OrgNorme = sqrt(Xorg*Xorg+Yorg*Yorg);
  if (EvenlyScaleFlag && OrgNorme == 0.0)
    return;
  if ((!EvenlyScaleFlag) && (Xorg == 0.0 || Yorg == 0.0))
    return;
  if (UserData->FrameView->Selection == NULL)
    return;
  DrawUnSelected(DrawingArea);
  colormap = gdk_window_get_colormap(DrawingArea->window);
  gcPurple = gdk_gc_new(DrawingArea->window);
  Purple.red = 65535; Purple.green = 0; Purple.blue = 65535;
  /* XXX: that color should be settable */
  gdk_color_alloc(colormap, &Purple);
  gdk_gc_set_foreground(gcPurple, &Purple);

  gcBlack = gdk_gc_new(DrawingArea->window);
  Black.red = 0; Black.green = 0; Black.blue = 0;
  gdk_color_alloc(colormap, &Black);
  gdk_gc_set_foreground(gcBlack, &Black);

  AllSelectionTri = ComputeSelectionTriangleMesh(UserData->FrameView->Selection);
  InitialSelectionTri = ComputeSelectionTriangleMesh(UserData->FrameView->Selection);

  DrawSelection(DrawingArea, AllSelectionTri, gcPurple, Xoff, Yoff, Zoom);
  SCALE_STATE = MOVING;
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, _("Scale Factor: "));
  gdk_pointer_grab(DrawingArea->window, FALSE,
                   GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK,
                   NULL, NULL, bevent->time);
}

/*****************************************************************************
*  ToolMotionScaleSelection
******************************************************************************/
void ToolMotionScaleSelection(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventMotion *Mev = (GdkEventMotion *)event;
  ViewUserDataStruct *UserData;
  GtkWidget *StatusBar;
  double Zoom, Xoff, Yoff;
  double XTrans, YTrans;
  double Norme, Ratio;
  Matrix MatTrans;
  TriangleListStruct *TmpTri, *TmpOrigTri;
  gchar *Message;
  
  if (SCALE_STATE != MOVING)
    return;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;

  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);
  DrawSelection(DrawingArea, AllSelectionTri, gcBlack, Xoff, Yoff, Zoom);
  MouseToReal(Mev->x, Mev->y, DrawingArea, &XTrans, &YTrans);
  
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      XTrans -= UserData->FrameView->ScaleCenter[0];
      YTrans -= UserData->FrameView->ScaleCenter[1];
      break;
    case ORTHO_XZ_CAMERA:
      XTrans -= UserData->FrameView->ScaleCenter[0];
      YTrans -= UserData->FrameView->ScaleCenter[2];
      break;
    case ORTHO_ZY_CAMERA:
      XTrans -= UserData->FrameView->ScaleCenter[2];
      YTrans -= UserData->FrameView->ScaleCenter[1];
      break;
  }
  Norme = sqrt(XTrans*XTrans+YTrans*YTrans);
  if (Norme == 0.0)
    Ratio = 1.0;
  else
    Ratio = Norme / OrgNorme;
  MIdentity(MatTrans);
  if (EvenlyScaleFlag)
  {
    MatTrans[0][0] = MatTrans[1][1] = MatTrans[2][2] = Ratio;
  }
  else
  {
    switch (UserData->CameraStyle)
    {
      case ORTHO_XY_CAMERA:
        MatTrans[0][0] = XTrans / Xorg;
        if (MatTrans[0][0] == 0.0) MatTrans[0][0] = 1.0;
        MatTrans[1][1] = YTrans / Yorg;
        if (MatTrans[1][1] == 0.0) MatTrans[1][1] = 1.0;
        MatTrans[2][2] = 1.0;
        break;
      case ORTHO_XZ_CAMERA:
        MatTrans[0][0] = XTrans / Xorg;
        if (MatTrans[0][0] == 0.0) MatTrans[0][0] = 1.0;
        MatTrans[1][1] = 1.0;
        MatTrans[2][2] = YTrans / Yorg;
        if (MatTrans[2][2] == 0.0) MatTrans[2][2] = 1.0;
        break;
      case ORTHO_ZY_CAMERA:
        MatTrans[0][0] = 1.0;
        MatTrans[1][1] = YTrans / Yorg;
        if (MatTrans[1][1] == 0.0) MatTrans[1][1] = 1.0;
        MatTrans[2][2] = XTrans / Xorg;
        if (MatTrans[2][2] == 0.0) MatTrans[2][2] = 1.0;
        break;
    }
  }
  MatTrans[3][0] = (1.0-MatTrans[0][0])*UserData->FrameView->ScaleCenter[0];
  MatTrans[3][1] = (1.0-MatTrans[1][1])*UserData->FrameView->ScaleCenter[1];
  MatTrans[3][2] = (1.0-MatTrans[2][2])*UserData->FrameView->ScaleCenter[2];
  Message = g_strdup_printf(_("Scale Factor: <%g, %g, %g>"), 
                            MatTrans[0][0], MatTrans[1][1], MatTrans[2][2]);
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, Message);
  g_free(Message);
  for (TmpTri = AllSelectionTri, TmpOrigTri = InitialSelectionTri ;
       TmpTri ;
       TmpTri = TmpTri->Next, TmpOrigTri = TmpOrigTri->Next)
  {
    TmpTri->P1[0] = TmpOrigTri->P1[0]*MatTrans[0][0] + TmpOrigTri->P1[1]*MatTrans[1][0] + TmpOrigTri->P1[2]*MatTrans[2][0] + MatTrans[3][0];
    TmpTri->P1[1] = TmpOrigTri->P1[0]*MatTrans[0][1] + TmpOrigTri->P1[1]*MatTrans[1][1] + TmpOrigTri->P1[2]*MatTrans[2][1] + MatTrans[3][1];
    TmpTri->P1[2] = TmpOrigTri->P1[0]*MatTrans[0][2] + TmpOrigTri->P1[1]*MatTrans[1][2] + TmpOrigTri->P1[2]*MatTrans[2][2] + MatTrans[3][2];
    TmpTri->P2[0] = TmpOrigTri->P2[0]*MatTrans[0][0] + TmpOrigTri->P2[1]*MatTrans[1][0] + TmpOrigTri->P2[2]*MatTrans[2][0] + MatTrans[3][0];
    TmpTri->P2[1] = TmpOrigTri->P2[0]*MatTrans[0][1] + TmpOrigTri->P2[1]*MatTrans[1][1] + TmpOrigTri->P2[2]*MatTrans[2][1] + MatTrans[3][1];
    TmpTri->P2[2] = TmpOrigTri->P2[0]*MatTrans[0][2] + TmpOrigTri->P2[1]*MatTrans[1][2] + TmpOrigTri->P2[2]*MatTrans[2][2] + MatTrans[3][2];
    TmpTri->P3[0] = TmpOrigTri->P3[0]*MatTrans[0][0] + TmpOrigTri->P3[1]*MatTrans[1][0] + TmpOrigTri->P3[2]*MatTrans[2][0] + MatTrans[3][0];
    TmpTri->P3[1] = TmpOrigTri->P3[0]*MatTrans[0][1] + TmpOrigTri->P3[1]*MatTrans[1][1] + TmpOrigTri->P3[2]*MatTrans[2][1] + MatTrans[3][1];
    TmpTri->P3[2] = TmpOrigTri->P3[0]*MatTrans[0][2] + TmpOrigTri->P3[1]*MatTrans[1][2] + TmpOrigTri->P3[2]*MatTrans[2][2] + MatTrans[3][2];
  }
  DrawSelection(DrawingArea, AllSelectionTri, gcPurple, Xoff, Yoff, Zoom);
}

/*****************************************************************************
*  ToolReleaseScaleSelection
******************************************************************************/
void ToolReleaseScaleSelection(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventButton *Bev = (GdkEventButton *)event;
  ViewUserDataStruct *UserData;
  GtkWidget          *StatusBar;
  double              Zoom, Xoff, Yoff;
  double              XTrans, YTrans;
  ObjectSelection    *TmpSel;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;
  double              Norme, Ratio;
  Vector              ScaleVect, TransVect;

  if (SCALE_STATE != MOVING)
    return;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;

  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);
  MouseToReal(Bev->x, Bev->y, DrawingArea, &XTrans, &YTrans);
    
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      XTrans -= UserData->FrameView->ScaleCenter[0];
      YTrans -= UserData->FrameView->ScaleCenter[1];
      break;
    case ORTHO_XZ_CAMERA:
      XTrans -= UserData->FrameView->ScaleCenter[0];
      YTrans -= UserData->FrameView->ScaleCenter[2];
      break;
    case ORTHO_ZY_CAMERA:
      XTrans -= UserData->FrameView->ScaleCenter[2];
      YTrans -= UserData->FrameView->ScaleCenter[1];
      break;
  }

  Norme = sqrt(XTrans*XTrans+YTrans*YTrans);
  if (Norme == 0.0)
    Ratio = 1.0;
  else
    Ratio = Norme / OrgNorme;
  if (EvenlyScaleFlag)
  {
    ScaleVect[0] = ScaleVect[1] = ScaleVect[2] = Ratio;
  } else
  {
    switch (UserData->CameraStyle)
    {
      case ORTHO_XY_CAMERA:
        ScaleVect[0] = XTrans / Xorg;
        if (ScaleVect[0] == 0.0) ScaleVect[0] = 1.0;
        ScaleVect[1] = YTrans / Yorg;
        if (ScaleVect[1] == 0.0) ScaleVect[1] = 1.0;
        ScaleVect[2] = 1.0;
        break;
      case ORTHO_XZ_CAMERA:
        ScaleVect[0] = XTrans / Xorg;
        if (ScaleVect[0] == 0.0) ScaleVect[0] = 1.0;
        ScaleVect[1] = 1.0;
        ScaleVect[2] = YTrans / Yorg;
        if (ScaleVect[2] == 0.0) ScaleVect[2] = 1.0;
        break;
      case ORTHO_ZY_CAMERA:
        ScaleVect[0] = 1.0;
        ScaleVect[1] = YTrans / Yorg;
        if (ScaleVect[1] == 0.0) ScaleVect[1] = 1.0;
        ScaleVect[2] = XTrans / Xorg;
        if (ScaleVect[2] == 0.0) ScaleVect[2] = 1.0;
        break;
    }
  }
  TransVect[0] = (1.0-ScaleVect[0])*UserData->FrameView->ScaleCenter[0];
  TransVect[1] = (1.0-ScaleVect[1])*UserData->FrameView->ScaleCenter[1];
  TransVect[2] = (1.0-ScaleVect[2])*UserData->FrameView->ScaleCenter[2];
  for (TmpSel = UserData->FrameView->Selection ;
       TmpSel->Next != UserData->FrameView->Selection ;
       TmpSel = TmpSel->Next)
  {
    giram_object_scale(TmpSel->Object, ScaleVect);
    giram_object_translate(TmpSel->Object, TransVect);
    giram_object_build_triangle_mesh(TmpSel->Object);
  }
  giram_object_scale(TmpSel->Object, ScaleVect);
  giram_object_translate(TmpSel->Object, TransVect);
  giram_object_build_triangle_mesh(TmpSel->Object);
 
  SCALE_STATE = NONE;
  { /* Freeing the two temporary mesh */
    TriangleListStruct *Tmp1, *Tmp2;
    Tmp1 = AllSelectionTri;
    AllSelectionTri = NULL;
    while(Tmp1)
    {
      Tmp2=Tmp1->Next;
      g_free(Tmp1);
      Tmp1=Tmp2;
    }
    Tmp1 = InitialSelectionTri;
    InitialSelectionTri = NULL;
    while(Tmp1)
    {
      Tmp2=Tmp1->Next;
      g_free(Tmp1);
      Tmp1=Tmp2;
    }
  }
  gdk_gc_destroy(gcPurple);
  gdk_gc_destroy(gcBlack);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  gdk_pointer_ungrab(Bev->time);
}

