/* tools.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include "giram.h"
#include "icon_pixmap.h"
#include "tool_rotate.h"
#include "tool_scale.h"
#include "cone.h"
#include "cylinder.h"
#include "disc.h"
#include "tool_options.h"
#include "tools.h"
#include "giramrc.h"
#include "giramintl.h"

typedef GtkWidget *(*OptionsFuncType)(GtkWidget *Box, char **Name);

typedef struct
{
  char       *Path;
  int         ToolNumber;
  GtkWidget  *Button;
  int         nb_columns, nb_rows;
  GtkWidget  *PopWindow;
  GtkWidget  *Table;
  GSList     *Group;
  GtkWidget **PixmapInBox;
  GtkWidget  *Box;
  int         pos;
  int        *IDTable;
} TableDescriptor;

struct {
  char  *ToolTip;
  char **Icon;
  char  *Path;
  int    ID;
  OptionsFuncType OptionsFunc;
}
Tools[]= 
{
  { N_("Selection"),          selection_icon,   "<ToolBar>",         MT_SELECT,             NULL },
  { N_("Move"),               move_icon,        "<ToolBar>",         MT_MOVE,               NULL },
  { N_("Rotate"),             rotate_icon,      "<ToolBar>",         MT_ROTATE,             (OptionsFuncType)BuildRotateOptions },
  { N_("Scale"),              scale_icon,       "<ToolBar>",         MT_SCALE,              (OptionsFuncType)BuildScaleOptions },
/*  { N_("New Bicubic Patch"),  bicubic_icon,     "<ToolBar/Patches>", MT_NEW_BICUBIC_PATCH,  NULL },*/
  { N_("New Box"),            box_icon,         "<ToolBar/Shape3D>", MT_NEW_BOX,            NULL },
/*  { N_("New Blob"),           blob_icon,        "<ToolBar/Shape3D>", MT_NEW_BLOB,           NULL },*/
  { N_("New Cone"),           cone_icon,        "<ToolBar/Shape3D>", MT_NEW_CONE,           (OptionsFuncType)BuildConeOptions },
/*  { N_("New CSG"),            csg_icon,         "<ToolBar/Shape3D>", MT_NEW_CSG,            NULL },*/
  { N_("New Cylinder"),       cylinder_icon,    "<ToolBar/Shape3D>", MT_NEW_CYLINDER,       (OptionsFuncType)BuildCylinderOptions },
  { N_("New Disc"),           disc_icon,        "<ToolBar/Shape2D>", MT_NEW_DISC,           (OptionsFuncType)BuildDiscOptions },
/*  { N_("New Height Field"),   heightfield_icon, "<ToolBar/Patches>", MT_NEW_HEIGHT_FIELD,   NULL },*/
/*  { N_("New Julia Fractal"),  julia_icon,       "<ToolBar/Shape3D>", MT_NEW_JULIA_FRACTAL,  NULL },*/
/*  { N_("New Lathe"),          lathe_icon,       "<ToolBar/Lathes>",  MT_NEW_LATHE,          NULL },*/
/*  { N_("New Mesh"),           mesh_icon,        "<ToolBar/Patches>", MT_NEW_MESH,           NULL },*/
  { N_("New Plane"),          plane_icon,       "<ToolBar/Shape2D>", MT_NEW_PLANE,          NULL },
/*  { N_("New Polygon"),        polygon_icon,     "<ToolBar/Shape2D>", MT_NEW_POLYGON,        NULL },*/
/*  { N_("New Prism"),          prism_icon,       "<ToolBar/Shape3D>", MT_NEW_PRISM,          NULL },*/
/*  { N_("New Quadric"),        quadric_icon,     "<ToolBar/Patches>", MT_NEW_QUADRIC,        NULL },*/
/*  { N_("New Cubic"),          cubic_icon,       "<ToolBar/Patches>", MT_NEW_CUBIC,          NULL },*/
/*  { N_("New Quartic"),        quartic_icon,     "<ToolBar/Patches>", MT_NEW_QUARTIC,        NULL },*/
/*  { N_("New Poly"),           poly_icon,        "<ToolBar/Patches>", MT_NEW_POLY,           NULL },*/
  { N_("New Sor"),            sor_icon,         "<ToolBar/Lathes>",  MT_NEW_SOR,            NULL },
  { N_("New Sphere"),         sphere_icon,      "<ToolBar/Shape3D>", MT_NEW_SPHERE,         NULL },
  { N_("New SuperEllipsoid"), ellipsoid_icon,   "<ToolBar/Shape3D>", MT_NEW_SUPERELLIPSOID, NULL },
/*  { N_("New Text"),           text_icon,        "<ToolBar/Shape3D>", MT_NEW_TEXT,           NULL },*/
  { N_("New Torus"),          torus_icon,       "<ToolBar/Shape3D>", MT_NEW_TORUS,          NULL },
  { N_("New Triangle"),       triangle_icon,    "<ToolBar/Shape2D>", MT_NEW_TRIANGLE,       NULL },
  { N_("New Light Source"),   lightsource_icon, "<ToolBar>",         MT_NEW_LIGHTSOURCE,    NULL },
};

#define VERTICAL_TOOLBOX   0
#define HORIZONTAL_TOOLBOX 1
#define SQUARE_TOOLBOX     2
gint DynamicToolBoxShape = SQUARE_TOOLBOX;

/*****************************************************************************
*  ToolSelect
******************************************************************************/
void ToolSelect(GtkWidget *widget, gpointer data)
{
  ModeTool = (int)data;
  ShowOptionsForTool(ModeTool);
}

/*****************************************************************************
*  SubToolSelect
******************************************************************************/
void SubToolSelect(GtkObject *button)
{
  ModeTool = (int)gtk_object_get_user_data(button);
  ShowOptionsForTool(ModeTool);
}

/*****************************************************************************
*  SecondaryToolSelect
******************************************************************************/
gint SecondaryToolSelect(GtkObject *widget, gpointer data)
{
  TableDescriptor *TD;
  int i, ToolIndex;

  TD = gtk_object_get_user_data(widget);
  ToolIndex = (int)data;
  for (i=0 ; i < TD->ToolNumber ; i++)
  {
    if (i==ToolIndex)
      gtk_widget_show(TD->PixmapInBox[i]);
    else
      gtk_widget_hide(TD->PixmapInBox[i]);
  }
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(TD->Button)))
  {
    ModeTool = TD->IDTable[ToolIndex];
    ShowOptionsForTool(ModeTool);
  }
  gtk_object_set_user_data(GTK_OBJECT(TD->Button),
                           (gpointer)(TD->IDTable[ToolIndex]));
  return FALSE;
}

/*****************************************************************************
*  ToolsDoubleClick
******************************************************************************/
int ToolsDoubleClick(GtkWidget *w, GdkEventButton *event)
{
  if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1))
    ShowToolOptionsDialog(NULL);
  return FALSE;
}

/*****************************************************************************
*  SubToolButton
******************************************************************************/
int SubToolButton(GtkWidget *w, GdkEventButton *event, GtkWidget *PopWin)
{
  if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1))
    ShowToolOptionsDialog(NULL);
  else if (event->button == 3)
  {
    gtk_widget_popup(PopWin, CLAMP(event->x_root-PopWin->allocation.width/2,  0, gdk_screen_width() -PopWin->allocation.width),
                             CLAMP(event->y_root-PopWin->allocation.height/2, 0, gdk_screen_height()-PopWin->allocation.height));
  }
  return FALSE;
}

/*****************************************************************************
*  CreateTools
******************************************************************************/
GtkWidget *CreateTools(GtkWidget *Box)
{
  GtkWidget *MainTable;
  GtkTooltips *ToolTips;
  GSList *RadioGroup;
  GtkWidget *Button;
  guint i, number_of_tools;
  guint number_of_main_tools;
  GdkBitmap *MaskIcon;
  GdkColor  transparent;
  GdkPixmap *PixmapIcon;
  GtkWidget *Pixmap;
  GSList *SecondaryTables = NULL;
  guint pos;

  number_of_tools = sizeof(Tools)/sizeof(Tools[0]);
  /* First Pass: Counting the number of icons on the Main Table */
  number_of_main_tools = 0;
  for (i=0; i < number_of_tools ; i++)
  {
    if (!strcmp(Tools[i].Path, "<ToolBar>") || toolbox_style == FALSE)
      number_of_main_tools++;
    else
    {
      GSList *TmpList;
      for (TmpList = SecondaryTables ; 
           TmpList && strcmp(((TableDescriptor *)(TmpList->data))->Path, Tools[i].Path) ;
           TmpList = TmpList->next)
        ;
      if (TmpList)
      {
        ((TableDescriptor *)(TmpList->data))->ToolNumber++;
      } else
      {
        TableDescriptor *NewTable;
        NewTable = g_malloc(sizeof(TableDescriptor));
        NewTable->Path = g_strdup(Tools[i].Path);
        NewTable->ToolNumber = 1;
        SecondaryTables = g_slist_append(SecondaryTables, NewTable);
        number_of_main_tools++;
        NewTable->Button = NULL;
      }
    }
  }
  /* Second pass: the real creation */  
  MainTable = gtk_table_new((number_of_main_tools-1)/TOOLBOX_WIDTH+1, TOOLBOX_WIDTH, TRUE);
  gtk_box_pack_start_defaults(GTK_BOX(Box), MainTable);
  gtk_widget_realize(MainTable);
  ToolTips = gtk_tooltips_new();
  RadioGroup = NULL;
  pos = 0;
  for (i=0 ; i<number_of_tools ; i++)
  {
    if (!strcmp(Tools[i].Path, "<ToolBar>") || toolbox_style == FALSE)
    { /* We've found a lonely tool, that only belongs to the main ToolBar... */
      Button = gtk_radio_button_new(RadioGroup);
      gtk_container_set_border_width(GTK_CONTAINER(Button), 0);
      RadioGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(Button));
      gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(Button), FALSE);
      gtk_table_attach(GTK_TABLE(MainTable), Button, pos%TOOLBOX_WIDTH,pos%TOOLBOX_WIDTH+1, pos/TOOLBOX_WIDTH,pos/TOOLBOX_WIDTH+1,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                       GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
      gtk_signal_connect(GTK_OBJECT(Button), "toggled",
                         GTK_SIGNAL_FUNC(ToolSelect), (gpointer)(Tools[i].ID));
      gtk_tooltips_set_tip(ToolTips, Button, _(Tools[i].ToolTip), NULL);
      if (Tools[i].Icon)
      {
        PixmapIcon = gdk_pixmap_create_from_xpm_d(MainTable->window,
                                                  &MaskIcon,
                                                  &transparent,
                                                  Tools[i].Icon);
        Pixmap = gtk_pixmap_new(PixmapIcon, MaskIcon);
      } else
      {
        Pixmap = gtk_label_new(Tools[i].ToolTip);
      }
      gtk_container_add(GTK_CONTAINER(Button), Pixmap);
      gtk_signal_connect(GTK_OBJECT(Button), "button_press_event",
                       (GtkSignalFunc)ToolsDoubleClick, NULL);
      gtk_widget_show(Pixmap);
      gtk_widget_show(Button);
      pos++;
    } else
    { /* This tools belongs to a secondary toolbox */
      GSList *TmpList;

      for (TmpList = SecondaryTables ; 
           TmpList && strcmp(((TableDescriptor *)(TmpList->data))->Path, Tools[i].Path) ;
           TmpList = TmpList->next)
        ;
      if (((TableDescriptor *)(TmpList->data))->Button)
      { /* Does the toolbox exists yet ? yes */
        TableDescriptor *TD = (TableDescriptor *)TmpList->data;

        Button = gtk_radio_button_new(TD->Group);
        gtk_container_set_border_width(GTK_CONTAINER(Button), 0);
        TD->Group = gtk_radio_button_group(GTK_RADIO_BUTTON(Button));
        gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(Button), FALSE);
        gtk_table_attach(GTK_TABLE(TD->Table), Button, 
                                   TD->pos%TD->nb_columns, TD->pos%TD->nb_columns+1,
                                   TD->pos/TD->nb_columns, TD->pos/TD->nb_columns+1,
                                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
        gtk_signal_connect(GTK_OBJECT(Button), "clicked",
                           (GtkSignalFunc)SecondaryToolSelect,
                           (gpointer)(TD->pos));
        gtk_signal_connect_object_after(GTK_OBJECT(Button), "clicked",
                                        GTK_SIGNAL_FUNC(gtk_widget_hide),
                                        GTK_OBJECT(TD->PopWindow));
        gtk_object_set_user_data(GTK_OBJECT(Button), TD);
        gtk_widget_show(Button);
        if (Tools[i].Icon)
        {
          GdkGC *gc;
          GdkColor color;
          int off;

          PixmapIcon = gdk_pixmap_create_from_xpm_d(MainTable->window,
                                                    &MaskIcon,
                                                    &transparent,
                                                    Tools[i].Icon);
          Pixmap = gtk_pixmap_new(PixmapIcon, MaskIcon);
          PixmapIcon = gdk_pixmap_create_from_xpm_d(MainTable->window,
                                                    &MaskIcon,
                                                    &transparent,
                                                    Tools[i].Icon);
          TD->PixmapInBox[TD->pos] = gtk_pixmap_new(PixmapIcon, MaskIcon);
          /* Draw the little black triangle */
          gc = gdk_gc_new((GTK_PIXMAP(TD->PixmapInBox[TD->pos]))->mask);
          gdk_color_white(gdk_colormap_get_system(), &color);
          gdk_gc_set_foreground(gc, &color);
          for (off=0 ; off<5 ; off++)
          {
            gdk_draw_line((GTK_PIXMAP(TD->PixmapInBox[TD->pos]))->pixmap,
                          MainTable->style->black_gc,
                          19+off,off, 23,off);
            gdk_draw_line((GTK_PIXMAP(TD->PixmapInBox[TD->pos]))->mask,
                          gc, 19+off,off, 23,off);
          }
          gdk_gc_destroy(gc);
        } else
        {
          Pixmap = gtk_label_new(Tools[i].ToolTip);
          TD->PixmapInBox[TD->pos] = gtk_label_new(Tools[i].ToolTip);
        }
        gtk_container_add(GTK_CONTAINER(Button), Pixmap);
        gtk_box_pack_start_defaults(GTK_BOX(TD->Box), TD->PixmapInBox[TD->pos]);
        gtk_widget_show(Pixmap);
        gtk_widget_show(TD->PixmapInBox[TD->pos]);
        gtk_widget_hide(TD->PixmapInBox[TD->pos]);
        TD->IDTable[TD->pos] = Tools[i].ID;
        TD->pos++;
      } else
      { /* We have to create the subtable */
        TableDescriptor *TD = (TableDescriptor *)TmpList->data;

        TD->Button = gtk_radio_button_new(RadioGroup);
        gtk_container_set_border_width(GTK_CONTAINER(TD->Button), 0);
        RadioGroup = gtk_radio_button_group(GTK_RADIO_BUTTON(TD->Button));
        gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(TD->Button), FALSE);
        gtk_table_attach(GTK_TABLE(MainTable), TD->Button, pos%TOOLBOX_WIDTH,pos%TOOLBOX_WIDTH+1, pos/TOOLBOX_WIDTH,pos/TOOLBOX_WIDTH+1,
                         GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                         GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
        gtk_widget_show(TD->Button);
        pos++;
        TD->Box = gtk_hbox_new(TRUE, 0);
        gtk_container_add(GTK_CONTAINER(TD->Button), TD->Box);
        gtk_widget_show(TD->Box);
        TD->PixmapInBox = g_malloc(TD->ToolNumber * sizeof(GtkWidget*));
        switch (DynamicToolBoxShape)
        {
          case VERTICAL_TOOLBOX:
            TD->nb_columns = 1;
            TD->nb_rows = TD->ToolNumber;
            break;
          case HORIZONTAL_TOOLBOX:
            TD->nb_columns = TD->ToolNumber;
            TD->nb_rows = 1;
            break;
          case SQUARE_TOOLBOX:
            TD->nb_columns = 1+(int)(sqrt(TD->ToolNumber-1.0));
            TD->nb_rows = (TD->ToolNumber-1) / TD->nb_columns + 1;
            break;
        }
        TD->PopWindow = gtk_window_new(GTK_WINDOW_POPUP);
        gtk_window_set_modal(GTK_WINDOW(TD->PopWindow), TRUE);
        TD->Table = gtk_table_new(TD->nb_rows, TD->nb_columns, TRUE);
        gtk_container_add(GTK_CONTAINER(TD->PopWindow), TD->Table);
        gtk_widget_show(TD->Table);
        gtk_widget_realize(TD->PopWindow);
        
        if (Tools[i].Icon)
        {
          GdkGC *gc;
          GdkColor color;
          int off;

          PixmapIcon = gdk_pixmap_create_from_xpm_d(MainTable->window,
                                                    &MaskIcon,
                                                    &transparent,
                                                    Tools[i].Icon);
          Pixmap = gtk_pixmap_new(PixmapIcon, MaskIcon);
          PixmapIcon = gdk_pixmap_create_from_xpm_d(MainTable->window,
                                                    &MaskIcon,
                                                    &transparent,
                                                    Tools[i].Icon);
          TD->pos = 0;
          TD->PixmapInBox[TD->pos] = gtk_pixmap_new(PixmapIcon, MaskIcon);
          /* Draw the little black triangle */
          gc = gdk_gc_new((GTK_PIXMAP(TD->PixmapInBox[TD->pos]))->mask);
          gdk_color_white(gdk_colormap_get_system(), &color);
          gdk_gc_set_foreground(gc, &color);
          for (off=0 ; off<5 ; off++)
          {
            gdk_draw_line((GTK_PIXMAP(TD->PixmapInBox[TD->pos]))->pixmap,
                          MainTable->style->black_gc,
                          19+off,off, 23,off);
            gdk_draw_line((GTK_PIXMAP(TD->PixmapInBox[TD->pos]))->mask,
                          gc, 19+off,off, 23,off);
          }
          gdk_gc_destroy(gc);
        } else
        {
          Pixmap = gtk_label_new(Tools[i].ToolTip);
          TD->PixmapInBox[TD->pos] = gtk_label_new(Tools[i].ToolTip);
        }
        gtk_box_pack_start_defaults(GTK_BOX(TD->Box), TD->PixmapInBox[TD->pos]);
        gtk_widget_show(TD->PixmapInBox[TD->pos]);
        TD->Group = NULL;
        Button = gtk_radio_button_new(TD->Group);
        gtk_container_set_border_width(GTK_CONTAINER(Button), 0);
        TD->Group = gtk_radio_button_group(GTK_RADIO_BUTTON(Button));
        gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(Button), FALSE);
        gtk_table_attach(GTK_TABLE(TD->Table), Button, 0,1, 0,1,
                         GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                         GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
        gtk_object_set_user_data(GTK_OBJECT(Button), TD);
        gtk_widget_show(Button);
        gtk_container_add(GTK_CONTAINER(Button), Pixmap);
        gtk_widget_show(Pixmap);
        gtk_signal_connect(GTK_OBJECT(TD->Button), "clicked",
                           (GtkSignalFunc)SubToolSelect, NULL);
        gtk_object_set_user_data(GTK_OBJECT(TD->Button),
                                 (gpointer)(Tools[i].ID));
        gtk_signal_connect(GTK_OBJECT(TD->Button), "button_press_event",
                           (GtkSignalFunc)SubToolButton, TD->PopWindow);
        
        gtk_signal_connect(GTK_OBJECT(Button), "clicked",
                           (GtkSignalFunc)SecondaryToolSelect,
                           (gpointer)0);
        gtk_signal_connect_object_after(GTK_OBJECT(Button), "clicked",
                                        GTK_SIGNAL_FUNC(gtk_widget_hide),
                                        GTK_OBJECT(TD->PopWindow));
        gtk_widget_show(TD->PopWindow);
        gtk_widget_hide(TD->PopWindow);
        TD->IDTable = g_malloc(TD->ToolNumber*sizeof(int));
        TD->IDTable[0] = Tools[i].ID;
        TD->pos++;
      }
    }
  }

  return MainTable;
}

