/* trimesh.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "giram.h"
#include "trimesh.h"

/*****************************************************************************
*  DestroyObjectTriangleMesh
******************************************************************************/
void DestroyObjectTriangleMesh(ObjectStruct *Object)
{
  TriangleListStruct *Tmp1, *Tmp2;

  Tmp1 = Object->FirstTriangle;
  Object->FirstTriangle = NULL;
  while(Tmp1)
  {
    Tmp2=Tmp1->Next;
    g_free(Tmp1);
    Tmp1=Tmp2;
  }
}

/*****************************************************************************
*  AddTriangleToObjectMesh
******************************************************************************/
void AddTriangleToObjectMesh(ObjectStruct *Object,
                             Vector P1, Vector P2, Vector P3,
                             Vector N1, Vector N2, Vector N3)
{
  TriangleListStruct *Tmp;

  Tmp = g_malloc(sizeof(TriangleListStruct));
  Tmp->Next = Object->FirstTriangle;
  Object->FirstTriangle = Tmp;
  Tmp->P1[0] = P1[0]; Tmp->P1[1] = P1[1]; Tmp->P1[2] = P1[2];
  Tmp->P2[0] = P2[0]; Tmp->P2[1] = P2[1]; Tmp->P2[2] = P2[2];
  Tmp->P3[0] = P3[0]; Tmp->P3[1] = P3[1]; Tmp->P3[2] = P3[2];
  Tmp->N1[0] = N1[0]; Tmp->N1[1] = N1[1]; Tmp->N1[2] = N1[2];
  Tmp->N2[0] = N2[0]; Tmp->N2[1] = N2[1]; Tmp->N2[2] = N2[2];
  Tmp->N3[0] = N3[0]; Tmp->N3[1] = N3[1]; Tmp->N3[2] = N3[2];
}

/*****************************************************************************
*  ComputeSelectionTriangleMesh
******************************************************************************/
TriangleListStruct *ComputeSelectionTriangleMesh(ObjectSelection *Sel)
{
  TriangleListStruct *SelTri, *TmpTri, *Triangle;
  ObjectSelection *TmpSelection;

  SelTri = NULL;
  for (TmpSelection = Sel ;
       TmpSelection->Next != Sel ;
       TmpSelection = TmpSelection->Next)
  {
    for (TmpTri = TmpSelection->Object->FirstTriangle ;
         TmpTri ;
         TmpTri = TmpTri->Next)
    { /* XXX: May use memcpy? */
      Triangle = g_malloc(sizeof(TriangleListStruct));
      Triangle->Next = SelTri;
      SelTri = Triangle;
      Triangle->P1[0] = TmpTri->P1[0];
      Triangle->P1[1] = TmpTri->P1[1];
      Triangle->P1[2] = TmpTri->P1[2];
      Triangle->P2[0] = TmpTri->P2[0];
      Triangle->P2[1] = TmpTri->P2[1];
      Triangle->P2[2] = TmpTri->P2[2];
      Triangle->P3[0] = TmpTri->P3[0];
      Triangle->P3[1] = TmpTri->P3[1];
      Triangle->P3[2] = TmpTri->P3[2];
    }
  }
  for (TmpTri = TmpSelection->Object->FirstTriangle ;
       TmpTri ;
       TmpTri = TmpTri->Next)
  { /* XXX: May use memcpy? */
    Triangle = g_malloc(sizeof(TriangleListStruct));
    Triangle->Next = SelTri;
    SelTri = Triangle;
    Triangle->P1[0] = TmpTri->P1[0];
    Triangle->P1[1] = TmpTri->P1[1];
    Triangle->P1[2] = TmpTri->P1[2];
    Triangle->P2[0] = TmpTri->P2[0];
    Triangle->P2[1] = TmpTri->P2[1];
    Triangle->P2[2] = TmpTri->P2[2];
    Triangle->P3[0] = TmpTri->P3[0];
    Triangle->P3[1] = TmpTri->P3[1];
    Triangle->P3[2] = TmpTri->P3[2];
  }
  return SelTri;
}

