/* view.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include "giram.h"
#include "tool_select.h"
#include "tool_move.h"
#include "tool_rotate.h"
#include "tool_scale.h"
#include "tool_options.h"
#include "box.h"
#include "cone.h"
#include "csg.h"
#include "cylinder.h"
#include "disc.h"
#include "mesh.h"
#include "plane.h"
#include "sor.h"
#include "sphere.h"
#include "superellipsoid.h"
#include "torus.h"
#include "triangle.h"
#include "light_source.h"
#include "object.h"
#include "csgtree.h"
#include "viewXY.h"
#include "viewXZ.h"
#include "viewZY.h"
#include "view_camera.h"
#include "utils.h"
#include "camera.h"
#include "SavePov.h"
#include "plugins.h"
#include "edit.h"
#include "snap.h"
#include "dxf.h"
#include "gtkcolorbutton.h"
#include "texture.h"
#ifdef WITH_S3D
#include "s3d.h"
#endif

#include "giramintl.h"

#include "pixmaps/navbutton.xpm"


/*****************************************************************************
*  SetWireFrameExpose
******************************************************************************/
void SetWireFrameExpose(void)
{
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  gtk_signal_disconnect(GTK_OBJECT(UserData->DrawingArea),
                        UserData->RenderStyle);
  switch (UserData->CameraStyle)
  {
    case FRAME_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewCamera, NULL);
      break;
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(GTK_WIDGET(UserData->DrawingArea));
}

/*****************************************************************************
*  SetHiddenFacesExpose
******************************************************************************/
void SetHiddenFacesExpose(void)
{
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  gtk_signal_disconnect(GTK_OBJECT(UserData->DrawingArea),
                        UserData->RenderStyle);
  switch (UserData->CameraStyle)
  {
    case FRAME_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewCamera, NULL);
      break;
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(GTK_WIDGET(UserData->DrawingArea));
}

/*****************************************************************************
*  SetGouraudExpose
******************************************************************************/
void SetGouraudExpose(void)
{
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  gtk_signal_disconnect(GTK_OBJECT(UserData->DrawingArea),
                        UserData->RenderStyle);
  switch (UserData->CameraStyle)
  {
    case FRAME_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewCamera, NULL);
      break;
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(GTK_WIDGET(UserData->DrawingArea));
}

#ifdef HAVE_GL
/*****************************************************************************
*  SetOpenGLExpose
******************************************************************************/
void SetOpenGLExpose(void)
{
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  gtk_signal_disconnect(GTK_OBJECT(UserData->DrawingArea),
                        UserData->RenderStyle);
  switch (UserData->CameraStyle)
  {
    case FRAME_CAMERA:
      break;
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGLViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGLViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(UserData->DrawingArea), "expose_event",
                           (GtkSignalFunc)ExposeGLViewZY, NULL);
      break;
  }
  /* OpenGL functions can be called only if make_current returns true */
  if (glXMakeCurrent(GDK_DISPLAY(),
                     GDK_WINDOW_XWINDOW(UserData->DrawingArea->window),
                     UserData->glxcontext))
  {
    glViewport(0,0, UserData->DrawingArea->allocation.width,
                    UserData->DrawingArea->allocation.height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0,100.0, 100.0,0.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
  }
  gtk_widget_queue_draw(GTK_WIDGET(UserData->DrawingArea));
}
#endif

/*****************************************************************************
*  CallPovFront
******************************************************************************/
void CallPovFront(void)
{
  ViewUserDataStruct *UserData;
  FrameStruct *LocalFrame;
  char *TmpFileName;
  gchar *CommandLine;

  TmpFileName = g_malloc(15);
  strcpy(TmpFileName, "/tmp/GiramXXXXXX");
  UserData = get_view_user_data();
  LocalFrame = UserData->FrameView;    
  mktemp(TmpFileName);
  SavePovByName(LocalFrame, TmpFileName);
  CommandLine = g_strdup_printf("povfront -p -s %s &", TmpFileName);
  system(CommandLine);
  g_free(CommandLine);
  g_free(TmpFileName);
}

/*****************************************************************************
*  ZoomIn
******************************************************************************/
void ZoomIn(void)
{
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();

  UserData->ZoomFactor *= 1.5;
  gtk_widget_queue_draw(GTK_WIDGET(UserData->DrawingArea));
  /* FIXME: Rulers */
}

/*****************************************************************************
*  ZoomOut
******************************************************************************/
void ZoomOut(void)
{
  ViewUserDataStruct *UserData;

  UserData = get_view_user_data();
  
  UserData->ZoomFactor /= 1.5;
  gtk_widget_queue_draw(GTK_WIDGET(UserData->DrawingArea));
  /* FIXME: Rulers */
}

/*****************************************************************************
*  FitToSelectionFunc
******************************************************************************/
void FitToSelectionFunc(void)
{
  ViewUserDataStruct *UserData;
  ObjectSelection *TmpSelect;
  TriangleListStruct *TmpTri;
  Vector MaxExt, MinExt;
  double ZoomX=1.0, ZoomY=1.0;

  UserData = get_view_user_data();

  MinExt[0] = MinExt[1] = MinExt[2] = 1e100;
  MaxExt[0] = MaxExt[1] = MaxExt[2] = -1e100;
  if (UserData->FrameView->Selection == NULL)
    return;
  for (TmpSelect = UserData->FrameView->Selection;
       TmpSelect->Next != UserData->FrameView->Selection ;
       TmpSelect = TmpSelect->Next)
  {
    /* Find the extents */
    for (TmpTri = TmpSelect->Object->FirstTriangle ; 
         TmpTri ; 
         TmpTri = TmpTri->Next)
    {
      if (MinExt[0] > TmpTri->P1[0]) MinExt[0] = TmpTri->P1[0];
      if (MinExt[0] > TmpTri->P2[0]) MinExt[0] = TmpTri->P2[0];
      if (MinExt[0] > TmpTri->P3[0]) MinExt[0] = TmpTri->P3[0];
      if (MinExt[1] > TmpTri->P1[1]) MinExt[1] = TmpTri->P1[1];
      if (MinExt[1] > TmpTri->P2[1]) MinExt[1] = TmpTri->P2[1];
      if (MinExt[1] > TmpTri->P3[1]) MinExt[1] = TmpTri->P3[1];
      if (MinExt[2] > TmpTri->P1[2]) MinExt[2] = TmpTri->P1[2];
      if (MinExt[2] > TmpTri->P2[2]) MinExt[2] = TmpTri->P2[2];
      if (MinExt[2] > TmpTri->P3[2]) MinExt[2] = TmpTri->P3[2];

      if (MaxExt[0] < TmpTri->P1[0]) MaxExt[0] = TmpTri->P1[0];
      if (MaxExt[0] < TmpTri->P2[0]) MaxExt[0] = TmpTri->P2[0];
      if (MaxExt[0] < TmpTri->P3[0]) MaxExt[0] = TmpTri->P3[0];
      if (MaxExt[1] < TmpTri->P1[1]) MaxExt[1] = TmpTri->P1[1];
      if (MaxExt[1] < TmpTri->P2[1]) MaxExt[1] = TmpTri->P2[1];
      if (MaxExt[1] < TmpTri->P3[1]) MaxExt[1] = TmpTri->P3[1];
      if (MaxExt[2] < TmpTri->P1[2]) MaxExt[2] = TmpTri->P1[2];
      if (MaxExt[2] < TmpTri->P2[2]) MaxExt[2] = TmpTri->P2[2];
      if (MaxExt[2] < TmpTri->P3[2]) MaxExt[2] = TmpTri->P3[2];
    }
  }
  for (TmpTri = TmpSelect->Object->FirstTriangle ; 
       TmpTri ; 
       TmpTri = TmpTri->Next)
  {
    if (MinExt[0] > TmpTri->P1[0]) MinExt[0] = TmpTri->P1[0];
    if (MinExt[0] > TmpTri->P2[0]) MinExt[0] = TmpTri->P2[0];
    if (MinExt[0] > TmpTri->P3[0]) MinExt[0] = TmpTri->P3[0];
    if (MinExt[1] > TmpTri->P1[1]) MinExt[1] = TmpTri->P1[1];
    if (MinExt[1] > TmpTri->P2[1]) MinExt[1] = TmpTri->P2[1];
    if (MinExt[1] > TmpTri->P3[1]) MinExt[1] = TmpTri->P3[1];
    if (MinExt[2] > TmpTri->P1[2]) MinExt[2] = TmpTri->P1[2];
    if (MinExt[2] > TmpTri->P2[2]) MinExt[2] = TmpTri->P2[2];
    if (MinExt[2] > TmpTri->P3[2]) MinExt[2] = TmpTri->P3[2];

    if (MaxExt[0] < TmpTri->P1[0]) MaxExt[0] = TmpTri->P1[0];
    if (MaxExt[0] < TmpTri->P2[0]) MaxExt[0] = TmpTri->P2[0];
    if (MaxExt[0] < TmpTri->P3[0]) MaxExt[0] = TmpTri->P3[0];
    if (MaxExt[1] < TmpTri->P1[1]) MaxExt[1] = TmpTri->P1[1];
    if (MaxExt[1] < TmpTri->P2[1]) MaxExt[1] = TmpTri->P2[1];
    if (MaxExt[1] < TmpTri->P3[1]) MaxExt[1] = TmpTri->P3[1];
    if (MaxExt[2] < TmpTri->P1[2]) MaxExt[2] = TmpTri->P1[2];
    if (MaxExt[2] < TmpTri->P2[2]) MaxExt[2] = TmpTri->P2[2];
    if (MaxExt[2] < TmpTri->P3[2]) MaxExt[2] = TmpTri->P3[2];
  }
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      ZoomX = UserData->DrawingArea->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = UserData->DrawingArea->allocation.height / (MaxExt[1]-MinExt[1]);
      UserData->Xoffset = (MaxExt[0]+MinExt[0]) / 2.0;
      UserData->Yoffset = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
    case ORTHO_XZ_CAMERA:
      ZoomX = UserData->DrawingArea->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = UserData->DrawingArea->allocation.height / (MaxExt[2]-MinExt[2]);
      UserData->Xoffset = (MaxExt[0]+MinExt[0]) / 2.0;
      UserData->Yoffset = (MaxExt[2]+MinExt[2]) / 2.0;
      break;
    case ORTHO_ZY_CAMERA:
      ZoomX = UserData->DrawingArea->allocation.width / (MaxExt[2]-MinExt[2]);
      ZoomY = UserData->DrawingArea->allocation.height / (MaxExt[1]-MinExt[1]);
      UserData->Xoffset = (MaxExt[2]+MinExt[2]) / 2.0;
      UserData->Yoffset = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
  }
  UserData->ZoomFactor = MIN(ZoomX, ZoomY) * 0.95;
  gtk_widget_queue_draw(GTK_WIDGET(UserData->DrawingArea));
}

/*****************************************************************************
*  FitToSceneFunc
******************************************************************************/
void FitToSceneFunc(void)
{
  ViewUserDataStruct *UserData;
  GSList *tmp_list;
  ObjectStruct *TmpObject;
  TriangleListStruct *TmpTri;
  Vector MaxExt, MinExt;
  double ZoomX=1.0, ZoomY=1.0;

  UserData = get_view_user_data();

  MinExt[0] = MinExt[1] = MinExt[2] = 1e100;
  MaxExt[0] = MaxExt[1] = MaxExt[2] = -1e100;
  if (!(UserData->FrameView->all_objects))
    return;
  for (tmp_list = UserData->FrameView->all_objects;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    TmpObject = (ObjectStruct*)(tmp_list->data);
    /* Find the extents */
    for (TmpTri = TmpObject->FirstTriangle ; 
         TmpTri ; 
         TmpTri = TmpTri->Next)
    {
      if (MinExt[0] > TmpTri->P1[0]) MinExt[0] = TmpTri->P1[0];
      if (MinExt[0] > TmpTri->P2[0]) MinExt[0] = TmpTri->P2[0];
      if (MinExt[0] > TmpTri->P3[0]) MinExt[0] = TmpTri->P3[0];
      if (MinExt[1] > TmpTri->P1[1]) MinExt[1] = TmpTri->P1[1];
      if (MinExt[1] > TmpTri->P2[1]) MinExt[1] = TmpTri->P2[1];
      if (MinExt[1] > TmpTri->P3[1]) MinExt[1] = TmpTri->P3[1];
      if (MinExt[2] > TmpTri->P1[2]) MinExt[2] = TmpTri->P1[2];
      if (MinExt[2] > TmpTri->P2[2]) MinExt[2] = TmpTri->P2[2];
      if (MinExt[2] > TmpTri->P3[2]) MinExt[2] = TmpTri->P3[2];

      if (MaxExt[0] < TmpTri->P1[0]) MaxExt[0] = TmpTri->P1[0];
      if (MaxExt[0] < TmpTri->P2[0]) MaxExt[0] = TmpTri->P2[0];
      if (MaxExt[0] < TmpTri->P3[0]) MaxExt[0] = TmpTri->P3[0];
      if (MaxExt[1] < TmpTri->P1[1]) MaxExt[1] = TmpTri->P1[1];
      if (MaxExt[1] < TmpTri->P2[1]) MaxExt[1] = TmpTri->P2[1];
      if (MaxExt[1] < TmpTri->P3[1]) MaxExt[1] = TmpTri->P3[1];
      if (MaxExt[2] < TmpTri->P1[2]) MaxExt[2] = TmpTri->P1[2];
      if (MaxExt[2] < TmpTri->P2[2]) MaxExt[2] = TmpTri->P2[2];
      if (MaxExt[2] < TmpTri->P3[2]) MaxExt[2] = TmpTri->P3[2];
    }
  }
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      ZoomX = UserData->DrawingArea->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = UserData->DrawingArea->allocation.height / (MaxExt[1]-MinExt[1]);
      UserData->Xoffset = (MaxExt[0]+MinExt[0]) / 2.0;
      UserData->Yoffset = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
    case ORTHO_XZ_CAMERA:
      ZoomX = UserData->DrawingArea->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = UserData->DrawingArea->allocation.height / (MaxExt[2]-MinExt[2]);
      UserData->Xoffset = (MaxExt[0]+MinExt[0]) / 2.0;
      UserData->Yoffset = (MaxExt[2]+MinExt[2]) / 2.0;
      break;
    case ORTHO_ZY_CAMERA:
      ZoomX = UserData->DrawingArea->allocation.width / (MaxExt[2]-MinExt[2]);
      ZoomY = UserData->DrawingArea->allocation.height / (MaxExt[1]-MinExt[1]);
      UserData->Xoffset = (MaxExt[2]+MinExt[2]) / 2.0;
      UserData->Yoffset = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
  }
  UserData->ZoomFactor = MIN(ZoomX, ZoomY) * 0.95;
  gtk_widget_queue_draw(GTK_WIDGET(UserData->DrawingArea));
}

/*****************************************************************************
*  SelectAll
******************************************************************************/
void SelectAll(void)
{
  ViewUserDataStruct *UserData;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;

  UserData = get_view_user_data();
  DestroySelection(&(UserData->FrameView->Selection));

  for (tmp_list = UserData->FrameView->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
    AddObjectToSelection(&(UserData->FrameView->Selection),
                         (ObjectStruct*)(tmp_list->data));
  /* FIXME: should update the CSG Tree selection */
  /* We redraw all the Views */
  for (tmp_list = UserData->FrameView->all_views ; 
       tmp_list ; 
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
}

/*****************************************************************************
*  SelectNone
******************************************************************************/
void SelectNone(void)
{
  ViewUserDataStruct *UserData;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;

  UserData = get_view_user_data();
  DestroySelection(&(UserData->FrameView->Selection));
  /* FIXME: should update the CSG Tree selection */
  /* We redraw all the Views */
  for (tmp_list = UserData->FrameView->all_views ; 
       tmp_list ; 
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
}

/*****************************************************************************
*  InvertSelection
******************************************************************************/
void InvertSelection(void)
{
  ViewUserDataStruct *UserData;
  FrameViewStruct    *TmpView;
  ObjectSelection    *TmpSelection;
  ObjectSelection    *Prev;
  ObjectStruct       *Object;
  GSList             *tmp_list;

  UserData = get_view_user_data();
  TmpSelection = NULL;
  for (tmp_list = UserData->FrameView->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
    AddObjectToSelection(&(TmpSelection), (ObjectStruct*)(tmp_list->data));
  if (UserData->FrameView->Selection)
  {
    for (Prev = UserData->FrameView->Selection ;
         Prev->Next != UserData->FrameView->Selection ;
         Prev = Prev->Next)
      RemoveObjectFromSelection(&(TmpSelection), Prev->Object);
    RemoveObjectFromSelection(&(TmpSelection), Prev->Object);
  }
  DestroySelection(&(UserData->FrameView->Selection));
  UserData->FrameView->Selection = TmpSelection;
  /* Update the CSG Tree selection. A little drastic, perhaps, but it'll
   * do until someone comes up with a better way. */
  RebuildCSGTree(NULL, UserData->FrameView);
  /* We redraw all the Views */
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
}

/*****************************************************************************
*  TranslateSelection
******************************************************************************/
void TranslateSelection(GtkWidget *Dialog)
{
  Vector               Vect;
  GtkWidget          **Data;
  GtkWidget           *DrawingArea;
  ViewUserDataStruct  *UserData;
  ObjectSelection     *ObjectSel;
  GSList              *tmp_list;
  FrameViewStruct     *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  DrawingArea = Data[3];
  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (UserData->FrameView->Selection)
  {
    giram_object_translate(UserData->FrameView->Selection->Object, Vect);
    giram_object_build_triangle_mesh(UserData->FrameView->Selection->Object);
    for (ObjectSel = UserData->FrameView->Selection ; 
         ObjectSel->Next != UserData->FrameView->Selection ; 
         ObjectSel = ObjectSel->Next)
    {
      giram_object_translate(ObjectSel->Object, Vect);
      giram_object_build_triangle_mesh(ObjectSel->Object);
    }
    for (tmp_list = UserData->FrameView->all_views ; 
         tmp_list ; 
         tmp_list = tmp_list->next)
    {
      TmpView = tmp_list->data;
      gtk_widget_queue_draw(TmpView->drawing_area);
    }
  }
}

/*****************************************************************************
*  RotateSelection
******************************************************************************/
void RotateSelection(GtkWidget *Dialog)
{
  Vector               Vect;
  GtkWidget          **Data;
  GtkWidget           *DrawingArea;
  ViewUserDataStruct  *UserData;
  ObjectSelection     *ObjectSel;
  GSList              *tmp_list;
  FrameViewStruct     *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  DrawingArea = Data[3];
  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (UserData->FrameView->Selection)
  {
    giram_object_rotate(UserData->FrameView->Selection->Object, Vect);
    giram_object_build_triangle_mesh(UserData->FrameView->Selection->Object);
    for (ObjectSel = UserData->FrameView->Selection ; 
         ObjectSel->Next != UserData->FrameView->Selection ; 
         ObjectSel = ObjectSel->Next)
    {
      giram_object_rotate(ObjectSel->Object, Vect);
      giram_object_build_triangle_mesh(ObjectSel->Object);
    }
    for (tmp_list = UserData->FrameView->all_views ; 
         tmp_list ; 
         tmp_list = tmp_list->next)
    {
      TmpView = tmp_list->data;
      gtk_widget_queue_draw(TmpView->drawing_area);
    }
  }
}

/*****************************************************************************
*  ScaleSelection
******************************************************************************/
void ScaleSelection(GtkWidget *Dialog)
{
  Vector               Vect;
  GtkWidget          **Data;
  GtkWidget           *DrawingArea;
  ViewUserDataStruct  *UserData;
  ObjectSelection     *ObjectSel;
  GSList              *tmp_list;
  FrameViewStruct     *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  DrawingArea = Data[3];
  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (fabs(Vect[0]) < 10e-8)
  {
    Vect[0] = 1.0;
    g_message(_("You can't scale by 0, changed to 1"));
  }
  if (fabs(Vect[1]) < 10e-8)
  {
    Vect[1] = 1.0;
    g_message(_("You can't scale by 0, changed to 1"));
  }
  if (fabs(Vect[2]) < 10e-8)
  {
    Vect[2] = 1.0;
    g_message(_("You can't scale by 0, changed to 1"));
  }
  if (UserData->FrameView->Selection)
  {
    giram_object_scale(UserData->FrameView->Selection->Object, Vect);
    giram_object_build_triangle_mesh(UserData->FrameView->Selection->Object);
    for (ObjectSel = UserData->FrameView->Selection ; 
         ObjectSel->Next != UserData->FrameView->Selection ; 
         ObjectSel = ObjectSel->Next)
    {
      giram_object_scale(ObjectSel->Object, Vect);
      giram_object_build_triangle_mesh(ObjectSel->Object);
    }
    for (tmp_list = UserData->FrameView->all_views ; 
         tmp_list ; 
         tmp_list = tmp_list->next)
    {
      TmpView = tmp_list->data;
      gtk_widget_queue_draw(TmpView->drawing_area);
    }
  }
}

/*****************************************************************************
*  TRSSelectionDialog
******************************************************************************/
void TRSSelectionDialog(GtkWidget *DrawingArea, 
                        gchar *Title, 
                        GtkSignalFunc CallBack)
{
  GtkWidget *Dialog;
  GtkWidget *Button;
  GtkWidget *Table;
  GtkWidget *Label;
  GtkWidget *Entry;
  GtkWidget **Data;

  Dialog = gtk_dialog_new();
  Data = g_malloc(4*sizeof(GtkWidget*));
  gtk_object_set_data(GTK_OBJECT(Dialog), "WidgetData", Data);
  gtk_window_set_title(GTK_WINDOW(Dialog), Title);
  /* The "Close" Button */
  Button = gtk_button_new_with_label(_("Close"));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE, TRUE, 0);
  /* The "Ok" Button */
  Button = gtk_button_new_with_label(_("OK"));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            CallBack,
                            GTK_OBJECT(Dialog));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE, TRUE, 0);
  gtk_widget_grab_default(Button);
  
  Table = gtk_table_new(2, 3, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(Dialog)->vbox), 5);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(Dialog)->vbox), Table);
  /* X */
  Label = gtk_label_new(" x: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 0,1,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);
  
  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 0,1,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  Data[0] = Entry;

  /* Y */
  Label = gtk_label_new(" y: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 1,2,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);
  
  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 1,2,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  Data[1] = Entry;
  
  /* Z */
  Label = gtk_label_new(" z: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 2,3,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);
  
  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 2,3,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  Data[2] = Entry;
  
  Data[3] = DrawingArea;
  
  gtk_widget_show_all(Dialog);
}

/*****************************************************************************
*  TranslateSelectionDialog
******************************************************************************/
void TranslateSelectionDialog(GtkWidget *DrawingArea)
{
  TRSSelectionDialog(DrawingArea, _("Translate Selection"), TranslateSelection);
}

/*****************************************************************************
*  RotateSelectionDialog
******************************************************************************/
void RotateSelectionDialog(GtkWidget *DrawingArea)
{
  TRSSelectionDialog(DrawingArea, _("Rotate Selection"), RotateSelection);
}

/*****************************************************************************
*  ScaleSelectionDialog
******************************************************************************/
void ScaleSelectionDialog(GtkWidget *DrawingArea)
{
  TRSSelectionDialog(DrawingArea, _("Scale Selection"), ScaleSelection);
}

/*****************************************************************************
*  DrawSelection
******************************************************************************/
void DrawSelection(GtkWidget *DrawingArea, TriangleListStruct *Tri,
                   GdkGC *gc, double Xoff, double Yoff, double Zoom)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      DrawSelectionXY(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;
    case ORTHO_XZ_CAMERA:
      DrawSelectionXZ(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;
    case ORTHO_ZY_CAMERA:
      DrawSelectionZY(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;
  }
}

/*****************************************************************************
*  DrawUnSelected
******************************************************************************/
void DrawUnSelected(GtkWidget *DrawingArea)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      DrawUnSelectedXY(DrawingArea);
      break;
    case ORTHO_XZ_CAMERA:
      DrawUnSelectedXZ(DrawingArea);
      break;
    case ORTHO_ZY_CAMERA:
      DrawUnSelectedZY(DrawingArea);
      break;
  }
}

/*****************************************************************************
*  event_dispatcher
******************************************************************************/
gint event_dispatcher(GtkWidget *DrawingArea, GdkEvent *event, gpointer Menu)
{
  CurrentUserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  switch (event->type)
  {
    case GDK_BUTTON_PRESS:
    {
      GdkEventButton *bevent = (GdkEventButton *) event;
      if (bevent->button == 3) /* Right Button ==> Popup Menu */
      {
        CurrentUserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
        gtk_menu_popup(GTK_MENU(Menu), NULL, NULL, NULL, NULL,
                                              bevent->button, bevent->time);
      }
      else if (bevent->button == 1)
      { /* Left Button : this is the big game */
        switch (ModeTool)
        {
          case MT_SELECT:       /* The Select Tool */
            ToggleSelect(DrawingArea, bevent);
            break;
          case MT_MOVE:         /* The Move Tool */
            ToolInitMoveSelection(DrawingArea, bevent);
            break;
          case MT_ROTATE:       /* The Rotate Tool */
            ToolInitRotateSelection(DrawingArea, bevent);
            break;
          case MT_SCALE:        /* The Scale Tool */
            ToolInitScaleSelection(DrawingArea, bevent);
            break;
          case MT_NEW_BOX:      /* The New Box Tool */
            ToolInitNewBox(DrawingArea, bevent);
            break;
          case MT_NEW_CONE:     /* The New Cone Tool */
            ToolInitNewCone(DrawingArea, bevent);
            break;
/* FIXME          case MT_NEW_CSG:
            ToolNewCSG(DrawingArea, bevent);
            break;*/
          case MT_NEW_CYLINDER: /* The New Cylinder Tool */
            ToolInitNewCylinder(DrawingArea, bevent);
            break;
          case MT_NEW_DISC:
            ToolInitNewDisc(DrawingArea, bevent);
            break;
          case MT_NEW_PLANE:    /* The New Plane Tool */
            ToolNewPlane(DrawingArea, bevent);
            break;
          case MT_NEW_SOR:
            ToolNewSor(DrawingArea, bevent);
            break;
          case MT_NEW_SPHERE:   /* The New Sphere Tool */
            ToolInitNewSphere(DrawingArea, bevent);
            break;
          case MT_NEW_SUPERELLIPSOID: /* The New SuperEllipsoid Tool */
            ToolNewSuperEllipsoid(DrawingArea, bevent);
            break;
          case MT_NEW_TORUS:    /* The New Torus Tool */
            ToolInitNewTorus(DrawingArea, bevent);
            break;
          case MT_NEW_TRIANGLE:
            ToolNewTriangle(DrawingArea, bevent);
            break;
          case MT_NEW_LIGHTSOURCE:
            ToolNewLightSource(DrawingArea, bevent);
            break;
          case MT_LAST:
            break;
        }   
      }
      return TRUE;
    }
    case GDK_MOTION_NOTIFY:
    {
      switch (ModeTool)
      {
        case MT_MOVE:
          ToolMotionMoveSelection(DrawingArea, event);
          break;
        case MT_ROTATE:
          ToolMotionRotateSelection(DrawingArea, event);
          break;
        case MT_SCALE:
          ToolMotionScaleSelection(DrawingArea, event);
          break;
        case MT_NEW_BOX:
          ToolMotionNewBox(DrawingArea, event);
          break;
        case MT_NEW_CONE:
          ToolMotionNewCone(DrawingArea, event);
          break;
        case MT_NEW_CYLINDER:
          ToolMotionNewCylinder(DrawingArea, event);
          break;
        case MT_NEW_DISC:
          ToolMotionNewDisc(DrawingArea, event);
          break;
        case MT_NEW_SPHERE:
          ToolMotionNewSphere(DrawingArea, event);
          break;
        case MT_NEW_TORUS:
          ToolMotionNewTorus(DrawingArea, event);
          break;
        default:
          break;
      }
      return FALSE;
    }
    case GDK_BUTTON_RELEASE:
    {
      switch (ModeTool)
      {
        case MT_MOVE:
          ToolReleaseMoveSelection(DrawingArea, event);
          break;
        case MT_ROTATE:
          ToolReleaseRotateSelection(DrawingArea, event);
          break;
        case MT_SCALE:
          ToolReleaseScaleSelection(DrawingArea, event);
          break;
        case MT_NEW_BOX:
          ToolReleaseNewBox(DrawingArea, event);
          break;
        case MT_NEW_CONE:
          ToolReleaseNewCone(DrawingArea, event);
          break;
        case MT_NEW_CYLINDER:
          ToolReleaseNewCylinder(DrawingArea, event);
          break;
        case MT_NEW_DISC:
          ToolReleaseNewDisc(DrawingArea, event);
          break;
        case MT_NEW_SPHERE:
          ToolReleaseNewSphere(DrawingArea, event);
          break;
        case MT_NEW_TORUS:
          ToolReleaseNewTorus(DrawingArea, event);
          break;
        default:
          break;
      }
      return TRUE;
    }
    case GDK_KEY_PRESS:
      printf("a key pressed event has been catched...\n");
      break;
    default:
      return FALSE;
  }
  return FALSE;
}

/*****************************************************************************
*  HRulerConfigure
******************************************************************************/
void HRulerConfigure(GtkWidget *DrawingArea, gpointer data, GtkWidget *HRuler)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_ruler_set_range(GTK_RULER(HRuler),
                      UserData->Xoffset-DrawingArea->allocation.width / 2.0 / UserData->ZoomFactor,
                      UserData->Xoffset+DrawingArea->allocation.width / 2.0 / UserData->ZoomFactor,
                      0.0, 50.0);
}

/*****************************************************************************
*  VRulerConfigure
******************************************************************************/
void VRulerConfigure(GtkWidget *DrawingArea, gpointer data, GtkWidget *VRuler)
{
  ViewUserDataStruct *UserData;

  UserData = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_ruler_set_range(GTK_RULER(VRuler),
                      UserData->Yoffset-DrawingArea->allocation.height / 2.0 / UserData->ZoomFactor,
                      UserData->Yoffset+DrawingArea->allocation.height / 2.0 / UserData->ZoomFactor,
                      0.0, 50.0);
}

#ifdef HAVE_GL
/*****************************************************************************
*  XVisualInfo: borrowed from gtkglarea.
******************************************************************************/
static XVisualInfo *get_xvisualinfo(GdkVisual *visual)
{
  Display *dpy;
  XVisualInfo vinfo_template;
  XVisualInfo *vi;
  int nitems_return;

  dpy = GDK_DISPLAY();

  /* TODO: is this right way to get VisualInfo from Visual ?? */
  /* AFAIK VisualID and depth should be enough to uniquely identify visual */
  vinfo_template.visual   = GDK_VISUAL_XVISUAL(visual);
  vinfo_template.visualid = XVisualIDFromVisual(vinfo_template.visual);
  vinfo_template.depth    = visual->depth;
  vi = XGetVisualInfo(dpy, VisualIDMask|VisualDepthMask, &vinfo_template, &nitems_return);
  g_assert( vi!=0  && nitems_return==1 ); /* visualinfo needs to be unique */

  /* remember to XFree returned XVisualInfo !!! */
  return vi;
}
#endif

/*****************************************************************************
*  gview_delete
******************************************************************************/
static gint gview_delete(GtkWidget *window, GdkEvent *ev, FrameViewStruct *FrameView)
{
  FrameViewStruct *TmpFrameView;
  ViewUserDataStruct *UserData;

  g_print("The deletion of an XYZ view has been catched!\n");
  g_free(FrameView->name);
  g_slist_remove(FrameView->Frame->all_views, FrameView);
  g_free(FrameView);
  UserData = gtk_object_get_data(GTK_OBJECT(window), "ViewData");
#ifdef HAVE_GL
  glXMakeCurrent(GDK_DISPLAY(), None, NULL);
  glXDestroyContext(GDK_DISPLAY(), UserData->glxcontext);
#endif
  g_free(UserData);
  return FALSE;
}

static GtkItemFactoryEntry ViewMenu[] =
{
  {"/_File",                         NULL, NULL,                     0, "<Branch>"},
  {"/File/Load _Pov",                NULL, LoadPovFunc,              0, "<Item>"},
  {"/File/Load AutoCAD _DXF",        NULL, LoadDXFFunc,              0, "<Item>"},
#ifdef USE_LIB3DS
  {"/File/Load 3DStudio _3DS",       NULL, Load3DSFunc,              0, "<Item>"},
#endif
  {"/File/Dum1",                     NULL, NULL,                     0, "<Separator>"},
  {"/File/Save P_ov",                NULL, SavePovFunc,              0, "<Item>"},
  {"/File/Save AutoCAD D_XF",        NULL, SaveDXFFunc,              0, "<Item>"},
#ifdef WITH_S3D
  {"/File/Save _Bishop S3D",         NULL, SaveS3DFunc,              0, "<Item>"},
#endif
  {"/File/Dum2",                     NULL, NULL,                     0, "<Separator>"},
  {"/File/_Quit",                    NULL, ExitFunc,                 0, "<Item>"},
  {"/_Edit",                         NULL, NULL,                     0, "<Branch>"},
  {"/Edit/_Cut",                     NULL, CutSelection,             0, "<Item>"},
  {"/Edit/C_opy",                    NULL, CopySelection,            0, "<Item>"},
  {"/Edit/_Paste",                   NULL, PasteSelection,           0, "<Item>"},
  {"/Edit/_Snap",                    NULL, SnapDialogShow,           0, "<Item>"},
  {"/_View",                         NULL, NULL,                     0, "<Branch>"},
  {"/View/Zoom _in",                 NULL, ZoomIn,                   0, "<Item>"},
  {"/View/Zoom _out",                NULL, ZoomOut,                  0, "<Item>"},
  {"/View/Fit To Se_lection",        NULL, FitToSelectionFunc,       0, "<Item>"},
  {"/View/Fit To _Scene",            NULL, FitToSceneFunc,           0, "<Item>"},
  {"/_Selection",                    NULL, NULL,                     0, "<Branch>"},
  {"/Selection/Select _All",         NULL, SelectAll,                0, "<Item>"},
  {"/Selection/Select _None",        NULL, SelectNone,               0, "<Item>"},
  {"/Selection/_Invert Selection",   NULL, InvertSelection,          0, "<Item>"},
  {"/Selection/Export to _DXF",      NULL, SaveSelectedDXFFunc,      0, "<Item>"},
#ifdef WITH_S3D
  {"/Selection/Export to _S3D",      NULL, SaveSelectedS3DFunc,      0, "<Item>"},
#endif
  {"/Selection/Import DXF _Objects", NULL, InsertDXFFunc,            0, "<Item>"},
#ifdef USE_LIB3DS
  {"/Selection/Import 3DS Objects", NULL, Insert3DSFunc,            0, "<Item>"},
#endif
  {"/_Operations",                   NULL, NULL,                     0, "<Branch>"},
  {"/Operations/_Translate",         NULL, TranslateSelectionDialog, 0, "<Item>"},
  {"/Operations/_Rotate",            NULL, RotateSelectionDialog,    0, "<Item>"},
  {"/Operations/_Scale",             NULL, ScaleSelectionDialog,     0, "<Item>"},
  {"/Operations/Change _Color",      NULL, ChangeColorDialog,        0, "<Item>"},
  {"/Operations/_Edit Camera",       NULL, EditCamera,               0, "<Item>"},
  {"/_Render",                       NULL, NULL,                     0, "<Branch>"},
  {"/Render/_Wire Frame",            NULL, SetWireFrameExpose,       0, "<Item>"},
  {"/Render/_Hidden Faces",          NULL, SetHiddenFacesExpose,     0, "<Item>"},
  {"/Render/_Gouraud",               NULL, SetGouraudExpose,         0, "<Item>"},
#ifdef HAVE_GL
  {"/Render/_OpenGL WireFrame",      NULL, SetOpenGLExpose,          0, "<Item>"},
#endif
  {"/Render/_PovRay Rendering",      NULL, CallPovFront,             0, "<Item>"},
  {"/_Dialogs",                      NULL, NULL,                     0, "<Branch>"},
  {"/Dialogs/_CSG Tree",             NULL, ShowCSGTree,              0, "<Item>"},
  {"/Dialogs/_Tool Options",         NULL, ShowToolOptionsDialog,    0, "<Item>"},
  {"/_Plug-Ins",                     NULL, NULL,                     0, "<Branch>"}
};
static guint NbViewEntries = sizeof(ViewMenu) /
                               sizeof(GtkItemFactoryEntry);

static GtkAccelGroup *ViewAccelGroup;
static GtkItemFactory *ViewFactory;
  
                               
/*****************************************************************************
*  CreateViewMenu
******************************************************************************/
void CreateViewMenu(void)
{
  /* New Keyboard accel group */
  ViewAccelGroup = gtk_accel_group_new();

  /* Creation of the item factory */
  ViewFactory = gtk_item_factory_new(GTK_TYPE_MENU, "<View>", ViewAccelGroup);
  gtk_item_factory_create_items(ViewFactory, NbViewEntries, ViewMenu, 0);
}

gboolean gdisplay_origin_button_press(GtkWidget      *widget,
                                      GdkEventButton *event,
                                      gpointer        data)
{
#if 0
  GDisplay *gdisp;

  gdisp = (GDisplay *) data;

  if (! gimp_busy && event->button == 1)
  {
    gint x, y;

    gdisplay_origin_menu_position(GTK_MENU(gdisp->ifactory->widget),
                                  &x, &y, widget);
    gtk_item_factory_popup_with_data(gdisp->ifactory,
                                     gdisp->gimage, NULL,
                                     x, y,
                                     1, event->time);
  }

  /*  Stop the signal emission so the button doesn't grab the
   *  pointer away from us
   */
  gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_press_event");
#endif
  return FALSE;
}



#define CANVAS_EVENT_MASK (GDK_EXPOSURE_MASK            | \
                           GDK_POINTER_MOTION_MASK      | \
                           GDK_POINTER_MOTION_HINT_MASK | \
                           GDK_BUTTON_PRESS_MASK        | \
                           GDK_BUTTON_RELEASE_MASK      | \
                           GDK_STRUCTURE_MASK           | \
                           GDK_ENTER_NOTIFY_MASK        | \
                           GDK_LEAVE_NOTIFY_MASK        | \
                           GDK_KEY_PRESS_MASK           | \
                           GDK_KEY_RELEASE_MASK         | \
                           GDK_PROXIMITY_OUT_MASK)

/*****************************************************************************
*  NewView
******************************************************************************/
void NewView(FrameStruct *LocalFrame, int CameraStyle)
{
  static GdkPixmap *navbutton_pixmap  = NULL;
  static GdkBitmap *navbutton_mask    = NULL;

  GtkWidget *window, *main_vbox, *disp_vbox, *upper_hbox, *inner_table;
  GtkWidget *lower_hbox, *hsb, *vsb;
  GtkWidget *origin, *arrow, *hrule, *vrule;
  GtkAdjustment *hsbdata, *vsbdata;
  GtkWidget *canvas, *nav_ebox, *label_frame;
  
  GtkWidget *statusarea, *status_hbox, *cursor_label, *statusbar;
  GtkWidget *progressbar, *cancelbutton;
  
  GtkWidget *pixmap;
  
  GtkWidget *ruler;

  gint contextid;
  
  /*  GtkWidget *Arrow; */
  GtkWidget *DrawingArea;
  ViewUserDataStruct *UserData;
  GtkWidget *menu;
  FrameViewStruct *View;
#ifdef HAVE_GL
  XVisualInfo *vi;
#endif

  View = g_malloc(sizeof(FrameViewStruct));
  
  LocalFrame->all_views = g_slist_append(LocalFrame->all_views, View);
  
  switch(CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      View->name = g_strdup_printf("(X-Y) %s", LocalFrame->Name);
      break;
    case ORTHO_XZ_CAMERA:
      View->name = g_strdup_printf("(X-Z) %s", LocalFrame->Name);
      break;
    case ORTHO_ZY_CAMERA:
      View->name = g_strdup_printf("(Z-Y) %s", LocalFrame->Name);
      break;
  }
  
  View->Frame = LocalFrame;
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);
  
  gtk_window_set_title(GTK_WINDOW(window), View->name);
  gtk_window_set_wmclass(GTK_WINDOW(window), "view_window", "Giram");
  gtk_widget_set_usize(window, 512, 384); /* XXX: should be setable */
  gtk_widget_set_events(window, GDK_POINTER_MOTION_MASK
                              | GDK_POINTER_MOTION_HINT_MASK
                              | GDK_BUTTON_PRESS_MASK
                              | GDK_KEY_PRESS_MASK
                              | GDK_KEY_RELEASE_MASK);
  gtk_container_set_border_width(GTK_CONTAINER(window), 0);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
                     (GtkSignalFunc)gview_delete, View);
  
  /*  main_vbox
   *     |
   *     +-- disp_vbox
   *     |      |
   *     |      +-- upper_hbox
   *     |      |      |
   *     |      |      +-- inner_table
   *     |      |      |      |
   *     |      |      |      +-- origin
   *     |      |      |      +-- hruler
   *     |      |      |      +-- vruler
   *     |      |      |      +-- canvas
   *     |      |      |
   *     |      |      +-- vscrollbar
   *     |      |
   *     |      +-- lower_hbox
   *     |             |
   *     |             +-- qmaskoff
   *     |             +-- qmaskon
   *     |             +-- hscrollbar
   *     |             +-- navbutton
   *     |
   *     +-- statusarea
   *            |
   *            +-- cursorlabel
   *            +-- statusbar
   *            +-- progressbar
   *            +-- cancelbutton
   */

  /*  first, set up the container hierarchy  *********************************/

  /*  the vbox containing all widgets  */
  main_vbox = gtk_vbox_new(FALSE, 2);
  gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 2);
  gtk_container_add(GTK_CONTAINER(window), main_vbox);
  
  /*  another vbox for everything except the statusbar  */
  disp_vbox = gtk_vbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX(main_vbox), disp_vbox, TRUE, TRUE, 0);
  gtk_widget_show(disp_vbox);

  /*  a hbox for the inner_table and the vertical scrollbar  */
  upper_hbox = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX(disp_vbox), upper_hbox, TRUE, TRUE, 0);
  gtk_widget_show(upper_hbox);

  /*  the table containing origin, rulers and the canvas  */
  inner_table = gtk_table_new(2, 2, FALSE);
  gtk_table_set_col_spacing(GTK_TABLE(inner_table), 0, 1);
  gtk_table_set_row_spacing(GTK_TABLE(inner_table), 0, 1);
  gtk_box_pack_start(GTK_BOX(upper_hbox), inner_table, TRUE, TRUE, 0);
  gtk_widget_show(inner_table);

  /*  the hbox containing qmask buttons, vertical scrollbar and nav button  */
  lower_hbox = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX(disp_vbox), lower_hbox, FALSE, FALSE, 0);
  gtk_widget_show(lower_hbox);

  /*  eventbox and hbox for status area  */
  /*gdisp->*/statusarea = gtk_event_box_new();
  gtk_box_pack_start(GTK_BOX(main_vbox), /*gdisp->*/statusarea, FALSE, FALSE, 0);

  status_hbox = gtk_hbox_new(FALSE, 2);
  gtk_container_add(GTK_CONTAINER(/*gdisp->*/statusarea), status_hbox);
  gtk_widget_show(status_hbox);

  gtk_container_set_resize_mode(GTK_CONTAINER(status_hbox),
                                GTK_RESIZE_QUEUE);

  /*  create the scrollbars  *************************************************/

  /*  the horizontal scrollbar  */
  /*gdisp->*/hsbdata = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 512/*width*/, 1, 1, 512/*width*/));
  /*gdisp->*/hsb = gtk_hscrollbar_new(/*gdisp->*/hsbdata);
  GTK_WIDGET_UNSET_FLAGS(/*gdisp->*/hsb, GTK_CAN_FOCUS);

  /*  the vertical scrollbar  */
  /*gdisp->*/vsbdata = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 384/*height*/, 1, 1, 384/*height*/));
  /*gdisp->*/vsb = gtk_vscrollbar_new(/*gdisp->*/vsbdata);
  GTK_WIDGET_UNSET_FLAGS(/*gdisp->*/vsb, GTK_CAN_FOCUS);

  /*  create the contents of the inner_table  ********************************/

  /*  the menu popup button  */
  /*gdisp->*/origin = gtk_button_new();
  GTK_WIDGET_UNSET_FLAGS(/*gdisp->*/origin, GTK_CAN_FOCUS);
  gtk_widget_set_events(GTK_WIDGET(/*gdisp->*/origin),
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
#if 0
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/origin), "button_press_event",
                     (GtkSignalFunc)gview_origin_button_press, /*gdisp*/NULL);
#endif

  arrow = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_OUT);
  gtk_container_set_border_width(GTK_CONTAINER(/*gdisp->*/origin), 0);
  gtk_container_add(GTK_CONTAINER(/*gdisp->*/origin), arrow);
  gtk_widget_show(arrow);

  /*  the horizontal ruler  */
  /*gdisp->*/hrule = gtk_hruler_new();
  gtk_widget_set_events(/*gdisp->*/hrule,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  gtk_signal_connect_object(GTK_OBJECT(window), "motion_notify_event",
                            (GtkSignalFunc)GTK_WIDGET_CLASS(GTK_OBJECT(/*gdisp->*/hrule)->klass)->motion_notify_event,
                            GTK_OBJECT(/*gdisp->*/hrule));
#if 0
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/hrule), "button_press_event",
                     (GtkSignalFunc)gview_hruler_button_press, /*gdisp*/NULL);
#endif

  /*  the vertical ruler  */
  /*gdisp->*/vrule = gtk_vruler_new();
  gtk_widget_set_events(/*gdisp->*/vrule,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  gtk_signal_connect_object(GTK_OBJECT(window), "motion_notify_event",
                            (GtkSignalFunc)GTK_WIDGET_CLASS(GTK_OBJECT(/*gdisp->*/vrule)->klass)->motion_notify_event,
                            GTK_OBJECT(/*gdisp->*/vrule));
#if 0
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/vrule), "button_press_event",
                     (GtkSignalFunc)gdisplay_vruler_button_press, /*gdisp*/NULL);
#endif

  /*  the canvas  */
  /*gdisp->*/canvas = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(/*gdisp->*/canvas), 512, 384);
  gtk_widget_set_events(/*gdisp->*/canvas, CANVAS_EVENT_MASK);
  gtk_widget_set_extension_events(/*gdisp->*/canvas, GDK_EXTENSION_EVENTS_ALL);
  GTK_WIDGET_SET_FLAGS(/*gdisp->*/canvas, GTK_CAN_FOCUS);
/*  gtk_object_set_user_data(GTK_OBJECT(gdisp->canvas), (gpointer)gdisp);*/

  /*  set the active display before doing any other canvas event processing  */
#if 0
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/canvas), "event",
                     (GtkSignalFunc)gdisplay_shell_events, gdisp);
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/canvas), "event",
                     GTK_SIGNAL_FUNC (gdisplay_canvas_events), gdisp);
#endif

  /*  the navigation window button  */
  nav_ebox = gtk_event_box_new();
#if 0
  gtk_signal_connect(GTK_OBJECT(nav_ebox), "button_press_event",
                     (GtkSignalFunc)nav_popup_click_handler, /*gdisp*/NULL);
#endif

  /* We need to realize the shell so that we have a GdkWindow for
   * the pixmap creation.  */
  gtk_widget_realize(window);

  /*  create the pixmaps  ****************************************************/
  if (!navbutton_pixmap)
  {
    GtkStyle *style;

    style = gtk_widget_get_style(window);

    navbutton_pixmap = gdk_pixmap_create_from_xpm_d(window->window,
                                                    &navbutton_mask,
                                                    &style->bg[GTK_STATE_NORMAL],
                                                    navbutton_xpm);
  }
  
  /*  create the GtkPixmaps  */
  pixmap = gtk_pixmap_new(navbutton_pixmap, navbutton_mask);
  gtk_container_add(GTK_CONTAINER(nav_ebox), pixmap);
  gtk_widget_show(pixmap);

  /*  create the contents of the status area *********************************/

  /*  the cursor label  */
  label_frame = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(label_frame), GTK_SHADOW_IN);

  /*gdisp->*/cursor_label = gtk_label_new(" ");
  gtk_container_add(GTK_CONTAINER(label_frame), /*gdisp->*/cursor_label);
  gtk_widget_show(/*gdisp->*/cursor_label);

  /*  the statusbar  */
  /*gdisp->*/statusbar = gtk_statusbar_new();
  gtk_widget_set_usize(/*gdisp->*/statusbar, 1, -1);
  gtk_container_set_resize_mode(GTK_CONTAINER(/*gdisp->*/statusbar),
                                GTK_RESIZE_QUEUE);
  contextid = gtk_statusbar_get_context_id(GTK_STATUSBAR(/*gdisp->*/statusbar),
                                           "title");
  gtk_statusbar_push(GTK_STATUSBAR(/*gdisp->*/statusbar),
                     contextid,
                     View->name);

  /*  the progress bar  */
  /*gdisp->*/progressbar = gtk_progress_bar_new();
  gtk_widget_set_usize(/*gdisp->*/progressbar, 80, -1);

  /*  the cancel button  */
  /*gdisp->*/cancelbutton = gtk_button_new_with_label(_("Cancel"));
  gtk_widget_set_sensitive(/*gdisp->*/cancelbutton, FALSE);

  /*  pack all the widgets  **************************************************/

  /*  fill the upper_hbox  */
  gtk_box_pack_start(GTK_BOX(upper_hbox), /*gdisp->*/vsb, FALSE, FALSE, 0);

  /*  fill the inner_table  */
  gtk_table_attach(GTK_TABLE(inner_table), /*gdisp->*/origin, 0, 1, 0, 1,
                   GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(inner_table), /*gdisp->*/hrule, 1, 2, 0, 1,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(inner_table), /*gdisp->*/vrule, 0, 1, 1, 2,
                   GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(inner_table), /*gdisp->*/canvas, 1, 2, 1, 2,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);

  /*  fill the lower_hbox  */
  gtk_box_pack_start(GTK_BOX(lower_hbox), /*gdisp->*/hsb, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(lower_hbox), nav_ebox, FALSE, FALSE, 0);

  /*  fill the status area  */
  gtk_box_pack_start(GTK_BOX(status_hbox), label_frame, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(status_hbox), /*gdisp->*/statusbar, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(status_hbox), /*gdisp->*/progressbar, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(status_hbox), /*gdisp->*/cancelbutton, FALSE, FALSE, 0);
  
  /*  show everything  *******************************************************/

/*  if (giramrc.show_rulers)
  {*/
    gtk_widget_show(/*gdisp->*/origin);
    gtk_widget_show(/*gdisp->*/hrule);
    gtk_widget_show(/*gdisp->*/vrule);
/*  }*/
  gtk_widget_show(/*gdisp->*/canvas);

  gtk_widget_show(/*gdisp->*/vsb);
  gtk_widget_show(/*gdisp->*/hsb);

  gtk_widget_show(nav_ebox);

  gtk_widget_show(label_frame);
  gtk_widget_show(/*gdisp->*/statusbar);
  gtk_widget_show(/*gdisp->*/progressbar);
  gtk_widget_show(/*gdisp->*/cancelbutton);
/*  if (giramrc.show_statusbar)
  {*/
    gtk_widget_show(/*gdisp->*/statusarea);
/*  }*/

  gtk_widget_realize(/*gdisp->*/canvas);
  gdk_window_set_back_pixmap(/*gdisp->*/canvas->window, NULL, FALSE);

  /*  we need to realize the cursor_label widget here, so the size gets
   *  computed correctly
   */
  gtk_widget_realize(/*gdisp->*/cursor_label);
/*  gdisplay_resize_cursor_label (gdisp);*/

  gtk_widget_show(main_vbox);
  gtk_widget_show(window);

  /*  set the focus to the canvas area  */
  gtk_widget_grab_focus(/*gdisp->*/canvas);


  
  UserData = g_malloc(sizeof(ViewUserDataStruct));
#if 0 /* def HAVE_GL */
  if (ReallyHaveGL)
  {
    vi = get_xvisualinfo(GLVisual);  
    UserData->glxcontext = glXCreateContext(GDK_DISPLAY(), vi, 0, True);
    XFree(vi);
/*    gtk_widget_push_colormap(gdk_colormap_new(GLVisual,TRUE));
    gtk_widget_push_visual(GLVisual);
    This should be done so the drawing area is created with a suitable
    visual and colormap for OpenGL rendering, but this cause Giram to
    segfault on the second close of an XYZ window. Don't know why...
 */
  }
#endif
  DrawingArea = gtk_drawing_area_new();

#if 0 /* def HAVE_GL */
  if (ReallyHaveGL)
  {
/*    gtk_widget_pop_visual();
    gtk_widget_pop_colormap();*/
  }
#endif
  View->drawing_area = canvas;
  switch (CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      UserData->RenderStyle = 
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXY, NULL);
      break;
    case ORTHO_XZ_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXZ, NULL);
      break;
    case ORTHO_ZY_CAMERA:
      UserData->RenderStyle =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewZY, NULL);
      break;
  }
  UserData->CameraStyle = CameraStyle;
  UserData->FrameView = LocalFrame;
  UserData->ZoomFactor = 50.0;
  UserData->Xoffset = 0.0;
  UserData->Yoffset = 0.0;
  UserData->DrawingArea = canvas;
  gtk_object_set_data(GTK_OBJECT(window), "ViewData", UserData);
  gtk_object_set_data(GTK_OBJECT(canvas), "ViewData", UserData);
  UserData->StatusBar = statusbar;


  gtk_accel_group_attach(ViewAccelGroup, GTK_OBJECT(window));

  /* We get the menu from the factory */
  menu = gtk_item_factory_get_widget(ViewFactory, "<View>");
        
  /* The "Plugins" SubMenu */
/*  gtk_menu_append(GTK_MENU(menu), CreatePlugInsMenu(DrawingArea));*/
  gtk_signal_connect(GTK_OBJECT(canvas), "event",
                     GTK_SIGNAL_FUNC(event_dispatcher), GTK_OBJECT(menu));

}

