/* viewXY.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "giram.h"
#include "utils.h"

/****************************************************************************
*  DrawSnapGridXY
*****************************************************************************/
void DrawSnapGridXY(GtkWidget *Area)
{
  GdkColormap *colormap;
  GdkColor ForeColor;
  GdkGC *gc;
  double x,y;
  double Zoom, Xoff, Yoff;
  double StartX, StartY;
  double EndX, EndY;
  double Xinc, Yinc;
  ViewUserDataStruct *UserData;
  FrameStruct *Frame;

  UserData = gtk_object_get_data(GTK_OBJECT(Area), "ViewData");
  Frame = UserData->FrameView;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;

  if (!Frame->ShowGrid)
    return;
  colormap = gdk_window_get_colormap(Area->window);
  ForeColor.red = 65535;
  ForeColor.green = 65535;
  ForeColor.blue = 65535;
  gdk_color_alloc(colormap, &ForeColor);
  gc = gdk_gc_new(Area->window);
  gdk_gc_set_foreground(gc, &ForeColor);
 
  Xinc = Frame->SnapGrid[0];
  Yinc = Frame->SnapGrid[1];
  StartX = Xoff-Area->allocation.width/2.0/Zoom;
  StartY = Yoff-Area->allocation.height/2.0/Zoom;

  StartX = Xinc*((int)(StartX/Xinc));
  StartY = Yinc*((int)(StartY/Yinc));
  EndX = Area->allocation.width/2.0/Zoom;
  EndY = Area->allocation.height/2.0/Zoom;

  for (x=StartX; x<EndX ; x+=Xinc)
  {
    for (y=StartY ; y<EndY ; y+=Yinc)
    {
      gdk_draw_point(Area->window, gc,
                     floor(Area->allocation.width/2.0+(x-Xoff)*Zoom),
                     floor(Area->allocation.height/2.0-(y-Yoff)*Zoom));
    }
  }
  gdk_gc_destroy(gc);
}

/*****************************************************************************
*  DrawCenterXY
******************************************************************************/
void DrawCenterXY(GtkWidget *View, double Xoff, double Yoff, double Zoom, 
                  Vector Center)
{
  GdkColor gdk_color;
  GdkColormap *colormap;
  GdkGC *gc;

  gc = gdk_gc_new(View->window);
  colormap = gdk_window_get_colormap(View->window);
  gdk_color.red = 0;
  gdk_color.green = 65535; /* XXX may be settable... */
  gdk_color.blue = 0;
  gdk_color_alloc(colormap, &gdk_color);
  gdk_gc_set_foreground(gc, &gdk_color);
  gdk_draw_line(View->window, gc,
                View->allocation.width/2+(Center[0]-Xoff)*Zoom-10,
                View->allocation.height/2-(Center[1]-Yoff)*Zoom-10,
                View->allocation.width/2+(Center[0]-Xoff)*Zoom+10,
                View->allocation.height/2-(Center[1]-Yoff)*Zoom+10);
  gdk_draw_line(View->window, gc,
                View->allocation.width/2+(Center[0]-Xoff)*Zoom-10,
                View->allocation.height/2-(Center[1]-Yoff)*Zoom+10,
                View->allocation.width/2+(Center[0]-Xoff)*Zoom+10,
                View->allocation.height/2-(Center[1]-Yoff)*Zoom-10);
  gdk_gc_destroy(gc);
}

/*****************************************************************************
*  ExposeWireFrameViewXY
******************************************************************************/
gint ExposeWireFrameViewXY(GtkWidget *View, GdkEventExpose *ev)
{
  ObjectStruct *Tmp;
  LightSourceStruct *TmpLight;
  TriangleListStruct *TmpTri;
  double Zoom, Xoff, Yoff;
  ViewUserDataStruct *UserData;
  GdkColor gdk_color;
  GdkColormap *colormap;
  GdkGC *gc;
  GSList *tmp_list;

  gdk_color.pixel = 0;
  gdk_color.red = 0;
  gdk_color.green = 0;
  gdk_color.blue = 0;
  gdk_window_set_background(View->window, &gdk_color);

  colormap = gdk_window_get_colormap(View->window);
  gc = gdk_gc_new(View->window);

  gdk_window_clear(View->window);
  UserData = get_view_user_data();
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  DrawSnapGridXY(View);
  for (tmp_list = UserData->FrameView->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    Tmp = (ObjectStruct*)(tmp_list->data);
    if (Tmp->Visible)
    {
      if (Tmp->Selected)
      { /* Selected Objects are red (XXX this should be settable) */
        gdk_color.red = 65535;
        gdk_color.green = 0;
        gdk_color.blue = 0;
      }
      else
      { /* UnSelected Objects are white (XXX this should be settable) */
        gdk_color.red = 65535;
        gdk_color.green = 65535;
        gdk_color.blue = 65535;
      }
      /* Allocate color */
      gdk_color_alloc(colormap, &gdk_color);
      gdk_gc_set_foreground(gc, &gdk_color);
      for (TmpTri = Tmp->FirstTriangle ; TmpTri ; TmpTri = TmpTri->Next)
      {
        gdk_draw_line(View->window, gc,
                      View->allocation.width/2+(TmpTri->P1[0]-Xoff)*Zoom,
                      View->allocation.height/2-(TmpTri->P1[1]-Yoff)*Zoom,
                      View->allocation.width/2+(TmpTri->P2[0]-Xoff)*Zoom,
                      View->allocation.height/2-(TmpTri->P2[1]-Yoff)*Zoom);
        gdk_draw_line(View->window, gc,
                      View->allocation.width/2+(TmpTri->P2[0]-Xoff)*Zoom,
                      View->allocation.height/2-(TmpTri->P2[1]-Yoff)*Zoom,
                      View->allocation.width/2+(TmpTri->P3[0]-Xoff)*Zoom,
                      View->allocation.height/2-(TmpTri->P3[1]-Yoff)*Zoom);
        gdk_draw_line(View->window, gc,
                      View->allocation.width/2+(TmpTri->P3[0]-Xoff)*Zoom,
                      View->allocation.height/2-(TmpTri->P3[1]-Yoff)*Zoom,
                      View->allocation.width/2+(TmpTri->P1[0]-Xoff)*Zoom,
                      View->allocation.height/2-(TmpTri->P1[1]-Yoff)*Zoom);
      }
    }
  }
  gdk_color.red = 65535;
  gdk_color.green = 65535;
  gdk_color.blue = 0;
  gdk_color_alloc(colormap, &gdk_color);
  gdk_gc_set_foreground(gc, &gdk_color);
  for (TmpLight = UserData->FrameView->FirstLightSource ;
       TmpLight ;
       TmpLight = TmpLight->Next)
  {
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10);
  }
  /* Draw the Rotation/Scale Center if any... */
  if (ModeTool == MT_ROTATE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->RotationCenter);
  }
  if (ModeTool == MT_SCALE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->ScaleCenter);
  }
  gdk_gc_destroy(gc);
  return(FALSE);
}

/*****************************************************************************
*  ExposeHiddenFacesViewXY
******************************************************************************/
gint ExposeHiddenFacesViewXY(GtkWidget *View, GdkEventExpose *ev)
{
  ObjectStruct *Tmp;
  LightSourceStruct *TmpLight;
  TriangleListStruct *TmpTri;
  double Zoom, Xoff, Yoff;
  ViewUserDataStruct *UserData;
  GdkColor gdk_color;
  GdkColormap *colormap;
  GdkGC *gc, *BlackGC;
  Triangle2D *TabTri;
  int nbTri;
  register int i;
  GdkPoint points[3];
  int max_triangles;
  GSList *tmp_list;

  gdk_color.pixel = 0;
  gdk_color.red = 0;
  gdk_color.green = 0;
  gdk_color.blue = 0;
  gdk_window_set_background(View->window, &gdk_color);

  colormap = gdk_window_get_colormap(View->window);
  gc = gdk_gc_new(View->window);
  BlackGC = gdk_gc_new(View->window);
  gdk_window_clear(View->window);
  UserData = get_view_user_data();
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  DrawSnapGridXY(View);
  nbTri = 0;
  max_triangles = 256;
  TabTri = g_malloc(max_triangles*sizeof(Triangle2D));
  for (tmp_list = UserData->FrameView->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    Tmp = (ObjectStruct*)(tmp_list->data);
    if (Tmp->Visible)
    {
      if (Tmp->Selected)
      { /* Selected Objects are red (XXX this should be settable) */
        gdk_color.red = 65535;
        gdk_color.green = 0;
        gdk_color.blue = 0;
      }
      else
      { /* Selected Objects are white (XXX this should be settable) */
        gdk_color.red = 65535;
        gdk_color.green = 65535;
        gdk_color.blue = 65535;
      }
      /* Allocate color */
      gdk_color_alloc(colormap, &gdk_color);
      for (TmpTri = Tmp->FirstTriangle ; TmpTri ; TmpTri = TmpTri->Next)
      {
        if (nbTri >= max_triangles)
        {
          max_triangles *= 2;
          TabTri = g_realloc(TabTri,max_triangles*sizeof(Triangle2D));
        }
        TabTri[nbTri].P1[0] = View->allocation.width/2+(TmpTri->P1[0]-Xoff)*Zoom;
        TabTri[nbTri].P1[1] = View->allocation.height/2-(TmpTri->P1[1]-Yoff)*Zoom;
        TabTri[nbTri].P2[0] = View->allocation.width/2+(TmpTri->P2[0]-Xoff)*Zoom;
        TabTri[nbTri].P2[1] = View->allocation.height/2-(TmpTri->P2[1]-Yoff)*Zoom;
        TabTri[nbTri].P3[0] = View->allocation.width/2+(TmpTri->P3[0]-Xoff)*Zoom;
        TabTri[nbTri].P3[1] = View->allocation.height/2-(TmpTri->P3[1]-Yoff)*Zoom;
        TabTri[nbTri].Color = gdk_color;
        TabTri[nbTri].distance = TmpTri->P1[2]+TmpTri->P2[2]+TmpTri->P3[2];
        nbTri++;
      }
    } /* if (Tmp->Visible) */
  }
  /* Sorting ... */
  qsort(TabTri, nbTri, sizeof(Triangle2D), comparTri);
  /* Displaying */
  for (i=0 ; i<nbTri ; i++)
  {
    points[0].x=TabTri[i].P1[0]; points[0].y=TabTri[i].P1[1];
    points[1].x=TabTri[i].P2[0]; points[1].y=TabTri[i].P2[1];
    points[2].x=TabTri[i].P3[0]; points[2].y=TabTri[i].P3[1];
    gdk_draw_polygon(View->window, BlackGC, TRUE, points, 3);
    gdk_gc_set_foreground(gc, &(TabTri[i].Color));
    gdk_draw_line(View->window, gc, points[0].x, points[0].y,
                                    points[1].x, points[1].y);
    gdk_draw_line(View->window, gc, points[1].x, points[1].y,
                                    points[2].x, points[2].y);
    gdk_draw_line(View->window, gc, points[2].x, points[2].y,
                                    points[0].x, points[0].y);
  }
  /* Freeing ... */
  g_free(TabTri);
  /* Draw the Light Sources */
  gdk_color.red = 65535;
  gdk_color.green = 65535;
  gdk_color.blue = 0;
  gdk_color_alloc(colormap, &gdk_color);
  gdk_gc_set_foreground(gc, &gdk_color);
  for (TmpLight = UserData->FrameView->FirstLightSource ;
       TmpLight ;
       TmpLight = TmpLight->Next)
  {
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10);
  }
  /* Draw the Rotation/Scale Center if any... */
  if (ModeTool == MT_ROTATE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->RotationCenter);
  }
  if (ModeTool == MT_SCALE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->ScaleCenter);
  }
  gdk_gc_destroy(gc);
  gdk_gc_destroy(BlackGC);
  return(FALSE);
}

/*****************************************************************************
*  ExposeGouraudViewXY
******************************************************************************/
gint ExposeGouraudViewXY(GtkWidget *View, GdkEventExpose *ev)
{
  ObjectStruct *Tmp;
  LightSourceStruct *TmpLight;
  TriangleListStruct *TmpTri;
  double Zoom, Xoff, Yoff;
  ViewUserDataStruct *UserData;
  GdkColor gdk_color;
  GdkColormap *colormap;
  GdkGC *gc, *BlackGC;
  Triangle2DNormCol *TabTri;
  int nbTri;
  register int i,j;
  int max_triangles;
  double Color[3], ColorResult[3];
  Vector NormalLight;
  LightSourceStruct *Light;
  double alpha;
  double *Zbuffer;
  double Norme;
  guchar *rgb_buf;
  GSList *tmp_list;

  gdk_color.pixel = 0;
  gdk_color.red = 0;
  gdk_color.green = 0;
  gdk_color.blue = 0;
  gdk_window_set_background(View->window, &gdk_color);

  colormap = gdk_window_get_colormap(View->window);
  gc = gdk_gc_new(View->window);
  BlackGC = gdk_gc_new(View->window);
  gdk_window_clear(View->window);
  UserData = get_view_user_data();
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  DrawSnapGridXY(View);
  nbTri = 0;
  max_triangles = 256;
  TabTri = g_malloc(max_triangles*sizeof(Triangle2DNormCol));
  for (tmp_list = UserData->FrameView->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    Tmp = (ObjectStruct*)(tmp_list->data);
    if (Tmp->Visible)
    {
      if (Tmp->Selected)
      { /* Selected Objects are red (XXX this should be settable) */
        Color[0] = 1.0;
        Color[1] = 0.0;
        Color[2] = 0.0;
      }
      else
      { /* Selected Objects are white (XXX this should be settable) */
        Color[0] = 1.0;
        Color[1] = 1.0;
        Color[2] = 1.0;
      }
      for (TmpTri=Tmp->FirstTriangle ; TmpTri ; TmpTri=TmpTri->Next)
      {
        if (nbTri >= max_triangles)
        {
          max_triangles *= 2;
          TabTri = g_realloc(TabTri, max_triangles*sizeof(Triangle2DNormCol));
        }
        TabTri[nbTri].P[0][0] = View->allocation.width/2+(TmpTri->P1[0]-Xoff)*Zoom;
        TabTri[nbTri].P[0][1] = View->allocation.height/2-(TmpTri->P1[1]-Yoff)*Zoom;
        TabTri[nbTri].P[1][0] = View->allocation.width/2+(TmpTri->P2[0]-Xoff)*Zoom;
        TabTri[nbTri].P[1][1] = View->allocation.height/2-(TmpTri->P2[1]-Yoff)*Zoom;
        TabTri[nbTri].P[2][0] = View->allocation.width/2+(TmpTri->P3[0]-Xoff)*Zoom;
        TabTri[nbTri].P[2][1] = View->allocation.height/2-(TmpTri->P3[1]-Yoff)*Zoom;
        /* Calcul des distances des trois sommets a la camera */
        TabTri[nbTri].Distance[0] = TmpTri->P1[2];
        TabTri[nbTri].Distance[1] = TmpTri->P2[2];
        TabTri[nbTri].Distance[2] = TmpTri->P3[2];
        /* Calcul des couleurs des trois sommets */
        /* 1er sommet */
        ColorResult[0] = Color[0] * 0.6;
        ColorResult[1] = Color[1] * 0.6;
        ColorResult[2] = Color[2] * 0.6;
        for (Light = UserData->FrameView->FirstLightSource ;
             Light ;
             Light = Light->Next)
        {
          NormalLight[0] = Light->Location[0]-TmpTri->P1[0];
          NormalLight[1] = Light->Location[1]-TmpTri->P1[1];
          NormalLight[2] = Light->Location[2]-TmpTri->P1[2];
          Norme = VLength(NormalLight);
          NormalLight[0] /= Norme;
          NormalLight[1] /= Norme;
          NormalLight[2] /= Norme;
          alpha = VDot(TmpTri->N1, NormalLight);
          if (Tmp->Type == PLANE_OBJECT)
            alpha = fabs(alpha);
          if (alpha > 0.0)
          {
            ColorResult[0] += alpha*Color[0]*Light->Color[0];
            ColorResult[1] += alpha*Color[1]*Light->Color[1];
            ColorResult[2] += alpha*Color[2]*Light->Color[2];
          }
        }
        if (ColorResult[0]<0.0) ColorResult[0]=0.0;
        if (ColorResult[0]>1.0) ColorResult[0]=1.0;
        if (ColorResult[1]<0.0) ColorResult[1]=0.0;
        if (ColorResult[1]>1.0) ColorResult[1]=1.0;
        if (ColorResult[2]<0.0) ColorResult[2]=0.0;
        if (ColorResult[2]>1.0) ColorResult[2]=1.0;
        TabTri[nbTri].Color[0][0] = ColorResult[0];
        TabTri[nbTri].Color[0][1] = ColorResult[1];
        TabTri[nbTri].Color[0][2] = ColorResult[2];
        /* 2eme sommet */
        ColorResult[0] = Color[0] * 0.6;
        ColorResult[1] = Color[1] * 0.6;
        ColorResult[2] = Color[2] * 0.6;
        for (Light = UserData->FrameView->FirstLightSource ;
             Light ;
             Light = Light->Next)
        {
          NormalLight[0] = Light->Location[0]-TmpTri->P2[0];
          NormalLight[1] = Light->Location[1]-TmpTri->P2[1];
          NormalLight[2] = Light->Location[2]-TmpTri->P2[2];
          Norme = VLength(NormalLight);
          NormalLight[0] /= Norme;
          NormalLight[1] /= Norme;
          NormalLight[2] /= Norme;
          alpha = VDot(TmpTri->N2, NormalLight);
          if (Tmp->Type == PLANE_OBJECT)
            alpha = fabs(alpha);
          if (alpha > 0.0)
          {
            ColorResult[0] += alpha*Color[0]*Light->Color[0];
            ColorResult[1] += alpha*Color[1]*Light->Color[1];
            ColorResult[2] += alpha*Color[2]*Light->Color[2];
          }
        }
        if (ColorResult[0]<0.0) ColorResult[0]=0.0;
        if (ColorResult[0]>1.0) ColorResult[0]=1.0;
        if (ColorResult[1]<0.0) ColorResult[1]=0.0;
        if (ColorResult[1]>1.0) ColorResult[1]=1.0;
        if (ColorResult[2]<0.0) ColorResult[2]=0.0;
        if (ColorResult[2]>1.0) ColorResult[2]=1.0;
        TabTri[nbTri].Color[1][0] = ColorResult[0];
        TabTri[nbTri].Color[1][1] = ColorResult[1];
        TabTri[nbTri].Color[1][2] = ColorResult[2];
        /* 3eme sommet */
        ColorResult[0] = Color[0] * 0.6;
        ColorResult[1] = Color[1] * 0.6;
        ColorResult[2] = Color[2] * 0.6;
        for (Light = UserData->FrameView->FirstLightSource ;
             Light ;
             Light = Light->Next)
        {
          NormalLight[0] = Light->Location[0]-TmpTri->P3[0];
          NormalLight[1] = Light->Location[1]-TmpTri->P3[1];
          NormalLight[2] = Light->Location[2]-TmpTri->P3[2];
          Norme = VLength(NormalLight);
          NormalLight[0] /= Norme;
          NormalLight[1] /= Norme;
          NormalLight[2] /= Norme;
          alpha = VDot(TmpTri->N3, NormalLight);
          if (Tmp->Type == PLANE_OBJECT)
            alpha = fabs(alpha);
          if (alpha > 0.0)
          {
            ColorResult[0] += alpha*Color[0]*Light->Color[0];
            ColorResult[1] += alpha*Color[1]*Light->Color[1];
            ColorResult[2] += alpha*Color[2]*Light->Color[2];
          }
        }
        if (ColorResult[0]<0.0) ColorResult[0]=0.0;
        if (ColorResult[0]>1.0) ColorResult[0]=1.0;
        if (ColorResult[1]<0.0) ColorResult[1]=0.0;
        if (ColorResult[1]>1.0) ColorResult[1]=1.0;
        if (ColorResult[2]<0.0) ColorResult[2]=0.0;
        if (ColorResult[2]>1.0) ColorResult[2]=1.0;
        TabTri[nbTri].Color[2][0] = ColorResult[0];
        TabTri[nbTri].Color[2][1] = ColorResult[1];
        TabTri[nbTri].Color[2][2] = ColorResult[2];
        nbTri++;
      } /* for */
    } /* if (Tmp->Visible) */
  }
  /* Displaying */
  Zbuffer = g_malloc(View->allocation.width *
                     View->allocation.height * sizeof(double));
  for (i=0 ; i<View->allocation.width ; i++)
    for (j=0 ; j<View->allocation.height ; j++)
      Zbuffer[j*View->allocation.width+i] = 1E20;
  rgb_buf = g_malloc0(3*View->allocation.width*View->allocation.height);
  for (i=0 ; i<nbTri ; i++)
    DrawGouraudTriangleZBuffer(View, Zbuffer, &(TabTri[i]), rgb_buf);
  gdk_draw_rgb_image(View->window, gc, 0,0,
                     View->allocation.width,
                     View->allocation.height,
                     GDK_RGB_DITHER_MAX,
                     rgb_buf, 3*View->allocation.width);
  /* Freeing ... */
  g_free(rgb_buf);
  g_free(TabTri);
  g_free(Zbuffer);
  /* Draw the Light Sources */
  gdk_color.red = 65535;
  gdk_color.green = 65535;
  gdk_color.blue = 0;
  gdk_color_alloc(colormap, &gdk_color);
  gdk_gc_set_foreground(gc, &gdk_color);
  for (TmpLight = UserData->FrameView->FirstLightSource ;
       TmpLight ;
       TmpLight = TmpLight->Next)
  {
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10);
    gdk_draw_line(View->window, gc,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom-10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom+10,
                  View->allocation.width/2+(TmpLight->Location[0]-Xoff)*Zoom+10,
                  View->allocation.height/2-(TmpLight->Location[1]-Yoff)*Zoom-10);
  }
  /* Draw the Rotation/Scale Center if any... */
  if (ModeTool == MT_ROTATE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->RotationCenter);
  }
  if (ModeTool == MT_SCALE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->ScaleCenter);
  }
  return (FALSE);
}

#ifdef HAVE_GL
/*****************************************************************************
*  ExposeGLViewXY
******************************************************************************/
gint ExposeGLViewXY(GtkWidget *View, GdkEventExpose *ev)
{
  ViewUserDataStruct *UserData;
  ObjectStruct *Tmp;
  TriangleListStruct *TmpTri;
  double left, right, top, bottom;
  GSList *tmp_list;

  UserData = get_view_user_data();
  /* Draw only last expose. */
  if (ev->count > 0)
    return TRUE;
  DrawSnapGridXY(View);
  /* OpenGL functions can be called only if make_current returns true */
  if (glXMakeCurrent(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(View->window),
                     UserData->glxcontext))
  {
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glViewport(0,0, View->allocation.width,View->allocation.height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    left   = UserData->Xoffset-View->allocation.width/2.0/UserData->ZoomFactor;
    right  = UserData->Xoffset+View->allocation.width/2.0/UserData->ZoomFactor;
    bottom = UserData->Yoffset-View->allocation.height/2.0/UserData->ZoomFactor;
    top    = UserData->Yoffset+View->allocation.height/2.0/UserData->ZoomFactor;
    glOrtho(left, right, bottom, top, -10e6, 10e6);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    for (tmp_list = UserData->FrameView->all_objects ;
         tmp_list ;
         tmp_list = g_slist_next(tmp_list))
    {
      Tmp = (ObjectStruct*)(tmp_list->data);
      if (Tmp->Visible)
      {
        if (Tmp->Selected)
          glColor3f(1.0, 0.0, 0.0);
        else
          glColor3f(1.0, 1.0, 1.0);
        glPushMatrix();
        glBegin(GL_TRIANGLES);
        for (TmpTri = Tmp->FirstTriangle ; TmpTri ; TmpTri = TmpTri->Next)
        {
          glVertex3f(TmpTri->P1[0], TmpTri->P1[1], TmpTri->P1[2]);
          glVertex3f(TmpTri->P2[0], TmpTri->P2[1], TmpTri->P2[2]);
          glVertex3f(TmpTri->P3[0], TmpTri->P3[1], TmpTri->P3[2]);
        }
        glEnd();
        glPopMatrix();
      }
    }
    glFlush();
    /* Swap backbuffer to front */
    glXSwapBuffers(GDK_WINDOW_XDISPLAY(View->window),
                   GDK_WINDOW_XWINDOW(View->window));
  }
  return TRUE;
}
#endif

/*****************************************************************************
*  DrawUnSelectedXY
******************************************************************************/
void DrawUnSelectedXY(GtkWidget *DrawingArea)
{
  ObjectStruct *Tmp;
  TriangleListStruct *TmpTri;
  double Zoom, Xoff, Yoff;
  ViewUserDataStruct *UserData;
  GdkColor gdk_color;
  GdkColormap *colormap;
  GdkGC *gc;
  GSList *tmp_list;
  
  gdk_color.pixel = 0;
  gdk_color.red = 0;
  gdk_color.green = 0;
  gdk_color.blue = 0;
  gdk_window_set_background(DrawingArea->window, &gdk_color);

  colormap = gdk_window_get_colormap(DrawingArea->window);
  gc = gdk_gc_new(DrawingArea->window);

  gdk_window_clear(DrawingArea->window);
  UserData = get_view_user_data();
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  DrawSnapGridXY(DrawingArea);
  for (tmp_list = UserData->FrameView->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    Tmp = (ObjectStruct*)(tmp_list->data);
    if (Tmp->Visible && !(Tmp->Selected))
    { /* UnSelected Objects are white (XXX this should be settable) */
      gdk_color.red = 65535;
      gdk_color.green = 65535;
      gdk_color.blue = 65535;

      /* Allocate color */
      gdk_color_alloc(colormap, &gdk_color);
      gdk_gc_set_foreground(gc, &gdk_color);
      for (TmpTri = Tmp->FirstTriangle ; TmpTri ; TmpTri = TmpTri->Next)
      {
        gdk_draw_line(DrawingArea->window, gc,
                      DrawingArea->allocation.width/2+(TmpTri->P1[0]-Xoff)*Zoom,
                      DrawingArea->allocation.height/2-(TmpTri->P1[1]-Yoff)*Zoom,
                      DrawingArea->allocation.width/2+(TmpTri->P2[0]-Xoff)*Zoom,
                      DrawingArea->allocation.height/2-(TmpTri->P2[1]-Yoff)*Zoom);
        gdk_draw_line(DrawingArea->window, gc,
                      DrawingArea->allocation.width/2+(TmpTri->P2[0]-Xoff)*Zoom,
                      DrawingArea->allocation.height/2-(TmpTri->P2[1]-Yoff)*Zoom,
                      DrawingArea->allocation.width/2+(TmpTri->P3[0]-Xoff)*Zoom,
                      DrawingArea->allocation.height/2-(TmpTri->P3[1]-Yoff)*Zoom);
        gdk_draw_line(DrawingArea->window, gc,
                      DrawingArea->allocation.width/2+(TmpTri->P3[0]-Xoff)*Zoom,
                      DrawingArea->allocation.height/2-(TmpTri->P3[1]-Yoff)*Zoom,
                      DrawingArea->allocation.width/2+(TmpTri->P1[0]-Xoff)*Zoom,
                      DrawingArea->allocation.height/2-(TmpTri->P1[1]-Yoff)*Zoom);
      }
    }
  }
  /* Draw the Rotation/Scale Center if any... */
  if (ModeTool == MT_ROTATE)
  {
    DrawCenterXY(DrawingArea, Xoff, Yoff, Zoom, UserData->FrameView->RotationCenter);
  }
  if (ModeTool == MT_SCALE)
  {
    DrawCenterXY(DrawingArea, Xoff, Yoff, Zoom, UserData->FrameView->ScaleCenter);
  }
  gdk_gc_destroy(gc);
}

/*****************************************************************************
*  DrawSelectionXY
******************************************************************************/
void DrawSelectionXY(GtkWidget *DrawingArea, TriangleListStruct *Tri,
                     GdkGC *gc, double Xoff, double Yoff, double Zoom)
{
  TriangleListStruct *TmpTri;
  ViewUserDataStruct *UserData;

  DrawSnapGridXY(DrawingArea);
  for (TmpTri = Tri ; TmpTri ; TmpTri = TmpTri->Next)
  {
    gdk_draw_line(DrawingArea->window, gc,
                  DrawingArea->allocation.width/2+(TmpTri->P1[0]-Xoff)*Zoom,
                  DrawingArea->allocation.height/2-(TmpTri->P1[1]-Yoff)*Zoom,
                  DrawingArea->allocation.width/2+(TmpTri->P2[0]-Xoff)*Zoom,
                  DrawingArea->allocation.height/2-(TmpTri->P2[1]-Yoff)*Zoom);
    gdk_draw_line(DrawingArea->window, gc,
                  DrawingArea->allocation.width/2+(TmpTri->P2[0]-Xoff)*Zoom,
                  DrawingArea->allocation.height/2-(TmpTri->P2[1]-Yoff)*Zoom,
                  DrawingArea->allocation.width/2+(TmpTri->P3[0]-Xoff)*Zoom,
                  DrawingArea->allocation.height/2-(TmpTri->P3[1]-Yoff)*Zoom);
    gdk_draw_line(DrawingArea->window, gc,
                  DrawingArea->allocation.width/2+(TmpTri->P3[0]-Xoff)*Zoom,
                  DrawingArea->allocation.height/2-(TmpTri->P3[1]-Yoff)*Zoom,
                  DrawingArea->allocation.width/2+(TmpTri->P1[0]-Xoff)*Zoom,
                  DrawingArea->allocation.height/2-(TmpTri->P1[1]-Yoff)*Zoom);
  }
  /* Draw the Rotation/Scale Center if any... */
  if (ModeTool == MT_ROTATE)
  {
    UserData = get_view_user_data();
    DrawCenterXY(DrawingArea, Xoff, Yoff, Zoom, UserData->FrameView->RotationCenter);
  }
  if (ModeTool == MT_SCALE)
  {
    UserData = get_view_user_data();
    DrawCenterXY(DrawingArea, Xoff, Yoff, Zoom, UserData->FrameView->ScaleCenter);
  }
}

