/* view_camera.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "giram.h"
#include "view.h"
#include "csgtree.h"
#include "tool_options.h"
#include "camera.h"
#include "utils.h"
#include "view_camera.h"

#include "giramintl.h"

/*****************************************************************************
*  Persp3dTo2d
******************************************************************************/
void  Persp3dTo2d(GtkWidget *DrawingArea, Vector Point3d, Vector Point2d, int width, int height)
{
  double xd,yd,zd;
  double norme;
  double vd,v0,t;
  double RiX, RiY, RiZ;
  double Ka, Kb;
  double u,v;
  ViewUserDataStruct *UserData;
  CameraStruct *Camera;

  UserData = get_view_user_data();
  Camera = UserData->FrameView->Camera;

  xd = Point3d[0] - Camera->Location[0];
  yd = Point3d[1] - Camera->Location[1];
  zd = Point3d[2] - Camera->Location[2];
  norme = sqrt(xd*xd+yd*yd+zd*zd);
  xd=xd / norme;
  yd=yd / norme;
  zd=zd / norme;
  vd = Camera->A*xd+Camera->B*yd+Camera->C*zd;
  if (fabs(vd)<1e-6) vd = 1e-6;
  v0 = -(Camera->A*Camera->x0+Camera->B*Camera->y0+Camera->C*Camera->z0+
         Camera->D);
  t = v0 / vd;
  RiX = Camera->Location[0]+xd*t;
  RiY = Camera->Location[1]+yd*t;
  RiZ = Camera->Location[2]+zd*t;
  
  if (fabs(Camera->Du2)<=1e-6)
  {
    u=(Camera->Nc[0]*RiX+Camera->Nc[1]*RiY+Camera->Nc[2]*RiZ-Camera->Du0)/
      (Camera->Du1-Camera->Na[0]*RiX-Camera->Na[1]*RiY-Camera->Na[2]*RiZ);
  } else
  {
    Ka = Camera->Dux + (Camera->Qux[0]*RiX+Camera->Qux[1]*RiY+Camera->Qux[2]*RiZ);
    Kb = Camera->Duy + (Camera->Quy[0]*RiX+Camera->Quy[1]*RiY+Camera->Quy[2]*RiZ);
    u = Ka - sqrt(Ka*Ka-Kb);
    if ((u>1.0) || (u<0.0))
      u = Ka + sqrt(Ka*Ka-Kb);
  }
  if (fabs(Camera->Dv2)<=1e-6)
  {
    v=(Camera->Nb[0]*RiX+Camera->Nb[1]*RiY+Camera->Nb[2]*RiZ-Camera->Dv0)/
      (Camera->Dv1-Camera->Na[0]*RiX-Camera->Na[1]*RiY-Camera->Na[2]*RiZ);
  } else
  {
    Ka = Camera->Dvx + (Camera->Qvx[0]*RiX+Camera->Qvx[1]*RiY+Camera->Qvx[2]*RiZ);
    Kb = Camera->Dvy + (Camera->Qvy[0]*RiX+Camera->Qvy[1]*RiY+Camera->Qvy[2]*RiZ);
    v = Ka - sqrt(Ka*Ka-Kb);
    if ((v>1.0) || (v<0.0))
      v = Ka + sqrt(Ka*Ka-Kb);
  }
  Point2d[0]=(u*width);
  Point2d[1]=((1.0-v)*height);
}

/*****************************************************************************
*  ExposeWireFrameViewCamera
******************************************************************************/
gint ExposeWireFrameViewCamera(GtkWidget *View, GdkEventExpose *ev)
{
  ObjectStruct *Tmp;
  TriangleListStruct *TmpTri;
  ViewUserDataStruct *UserData;
  GdkColor gdk_color;
  GdkColormap *colormap;
  GdkGC *gc;
  Vector P1, P2, P3;
  GSList *tmp_list;

  gdk_color.pixel = 0;
  gdk_color.red = 0;
  gdk_color.green = 0;
  gdk_color.blue = 0;
  gdk_window_set_background(View->window, &gdk_color);

  colormap = gdk_window_get_colormap(View->window);
  gc = gdk_gc_new(View->window);

  gdk_window_clear(View->window);
  UserData = get_view_user_data();
  for (tmp_list = UserData->FrameView->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    Tmp = (ObjectStruct*)(tmp_list->data);
    if (Tmp->Visible)
    {
      if (Tmp->Selected)
      { /* Selected Objects are red (XXX this should be settable) */
        gdk_color.red = 65535;
        gdk_color.green = 0;
        gdk_color.blue = 0;
      }
      else
      { /* UnSelected Objects are white (XXX this should be settable) */
        gdk_color.red = 65535;
        gdk_color.green = 65535;
        gdk_color.blue = 65535;
      }
      /* Allocate color */
      gdk_color_alloc(colormap, &gdk_color);
      gdk_gc_set_foreground(gc, &gdk_color);
      for (TmpTri = Tmp->FirstTriangle ; TmpTri ; TmpTri = TmpTri->Next)
      {
        Persp3dTo2d(View, TmpTri->P1, P1, View->allocation.width, View->allocation.height);
        Persp3dTo2d(View, TmpTri->P2, P2, View->allocation.width, View->allocation.height);
        Persp3dTo2d(View, TmpTri->P3, P3, View->allocation.width, View->allocation.height);
        gdk_draw_line(View->window, gc, P1[0], P1[1], P2[0], P2[1]);
        gdk_draw_line(View->window, gc, P2[0], P2[1], P3[0], P3[1]);
        gdk_draw_line(View->window, gc, P3[0], P3[1], P1[0], P1[1]);
      }
    } /* if (Tmp->Visible) */
  }
  /* Draw the Rotation/Scale Center if any... */
/* FIXME if (ModeTool == MT_ROTATE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->RotationCenter);
  }
  if (ModeTool == MT_SCALE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->ScaleCenter);
  }*/
  gdk_gc_destroy(gc);
  return(FALSE);
}

/*****************************************************************************
*  ExposeHiddenFacesViewCamera
******************************************************************************/
gint ExposeHiddenFacesViewCamera(GtkWidget *View, GdkEventExpose *ev)
{
  ObjectStruct *Tmp;
  TriangleListStruct *TmpTri;
  ViewUserDataStruct *UserData;
  GdkColor gdk_color;
  GdkColormap *colormap;
  GdkGC *gc, *BlackGC;
  Triangle2D *TabTri;
  int nbTri;
  register int i;
  GdkPoint points[3];
  int max_triangles;
  Vector Vect;
  CameraStruct *Camera;
  GSList *tmp_list;

  gdk_color.pixel = 0;
  gdk_color.red = 0;
  gdk_color.green = 0;
  gdk_color.blue = 0;
  gdk_window_set_background(View->window, &gdk_color);

  colormap = gdk_window_get_colormap(View->window);
  gc = gdk_gc_new(View->window);
  BlackGC = gdk_gc_new(View->window);
  gdk_window_clear(View->window);
  UserData = get_view_user_data();
  Camera = UserData->FrameView->Camera;
  nbTri = 0;
  max_triangles = 256;
  TabTri = g_malloc(max_triangles*sizeof(Triangle2D));
  for (tmp_list = UserData->FrameView->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    Tmp = (ObjectStruct*)(tmp_list->data);
    if (Tmp->Visible)
    {
      if (Tmp->Selected)
      { /* Selected Objects are red (XXX this should be settable) */
        gdk_color.red = 65535;
        gdk_color.green = 0;
        gdk_color.blue = 0;
      }
      else
      { /* Selected Objects are white (XXX this should be settable) */
        gdk_color.red = 65535;
        gdk_color.green = 65535;
        gdk_color.blue = 65535;
      }
      /* Allocate color */
      gdk_color_alloc(colormap, &gdk_color);
      for (TmpTri = Tmp->FirstTriangle ; TmpTri ; TmpTri = TmpTri->Next)
      {
        if (nbTri >= max_triangles)
        {
          max_triangles *= 2;
          TabTri = g_realloc(TabTri,max_triangles*sizeof(Triangle2D));
        }
        Persp3dTo2d(View, TmpTri->P1, TabTri[nbTri].P1, View->allocation.width, View->allocation.height);
        Persp3dTo2d(View, TmpTri->P2, TabTri[nbTri].P2, View->allocation.width, View->allocation.height);
        Persp3dTo2d(View, TmpTri->P3, TabTri[nbTri].P3, View->allocation.width, View->allocation.height);
        TabTri[nbTri].Color = gdk_color;
        Vect[0]=TmpTri->P1[0]+TmpTri->P2[0]+TmpTri->P3[0]-3.0*Camera->Location[0];
        Vect[1]=TmpTri->P1[1]+TmpTri->P2[1]+TmpTri->P3[1]-3.0*Camera->Location[1];
        Vect[2]=TmpTri->P1[2]+TmpTri->P2[2]+TmpTri->P3[2]-3.0*Camera->Location[2];
        TabTri[nbTri].distance = VLength(Vect);
        nbTri++;
      }
    } /* if (Tmp->Visible) */
  }
  /* Sorting ... */
  qsort(TabTri, nbTri, sizeof(Triangle2D), comparTri);
  /* Displaying */
  for (i=0 ; i<nbTri ; i++)
  {
    points[0].x=TabTri[i].P1[0]; points[0].y=TabTri[i].P1[1];
    points[1].x=TabTri[i].P2[0]; points[1].y=TabTri[i].P2[1];
    points[2].x=TabTri[i].P3[0]; points[2].y=TabTri[i].P3[1];
    gdk_draw_polygon(View->window, BlackGC, TRUE, points, 3);
    gdk_gc_set_foreground(gc, &(TabTri[i].Color));
    gdk_draw_line(View->window, gc, points[0].x, points[0].y, 
                                    points[1].x, points[1].y);
    gdk_draw_line(View->window, gc, points[1].x, points[1].y,
                                    points[2].x, points[2].y);
    gdk_draw_line(View->window, gc, points[2].x, points[2].y,
                                    points[0].x, points[0].y);
  }
  /* Freeing ... */
  g_free(TabTri);
  /* Draw the Rotation/Scale Center if any... */
/*  if (ModeTool == MT_ROTATE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->RotationCenter);
  }
  if (ModeTool == MT_SCALE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->ScaleCenter);
  }*/
  gdk_gc_destroy(gc);
  gdk_gc_destroy(BlackGC);
  return(FALSE);
}                  

/*****************************************************************************
*  ExposeGouraudViewCamera
******************************************************************************/
gint ExposeGouraudViewCamera(GtkWidget *View, GdkEventExpose *ev)
{
#if 0
  ObjectStruct *Tmp;
  TriangleListStruct *TmpTri;
  ViewUserDataStruct *UserData;
  GdkColor gdk_color;
  GdkColormap *colormap;
  GdkGC *gc, *BlackGC;
  Triangle2DNormCol *TabTri;
  int nbTri;
  register int i,j;
  int max_triangles;
  double Color[3], ColorResult[3];
  Vector NormalLight;
  LightSourceStruct *Light;
  double alpha;
  double *Zbuffer;
  double Norme;
  Vector P1, P2, P3;
  CameraStruct *Camera;

  gdk_color.pixel = 0;
  gdk_color.red = 0;
  gdk_color.green = 0;
  gdk_color.blue = 0;
  gdk_window_set_background(View->window, &gdk_color);

  colormap = gdk_window_get_colormap(View->window);
  gc = gdk_gc_new(View->window);
  BlackGC = gdk_gc_new(View->window);
  gdk_window_clear(View->window);
  UserData = get_view_user_data();
  Camera = UserData->FrameView->Camera;
  nbTri = 0;
  max_triangles = 256;
  TabTri = g_malloc(max_triangles*sizeof(Triangle2DNormCol));
  for (Tmp = UserData->FrameView->FirstObject ; Tmp ; Tmp = Tmp->Next)
  {
    if (Tmp->Visible)
    {
      if (Tmp->Selected)
      { /* Selected Objects are red (XXX this should be settable) */
        Color[0] = 1.0;
        Color[1] = 0.0;
        Color[2] = 0.0;
      }
      else
      { /* Selected Objects are white (XXX this should be settable) */
        Color[0] = 1.0;
        Color[1] = 1.0;
        Color[2] = 1.0;
      }
      for (TmpTri=Tmp->FirstTriangle ; TmpTri ; TmpTri=TmpTri->Next)
      {
        if (nbTri >= max_triangles)
        {
          max_triangles *= 2;
          TabTri = g_realloc(TabTri, max_triangles*sizeof(Triangle2DNormCol));
        }
        Persp3dTo2d(View, TmpTri->P1, P1, View->allocation.width, View->allocation.height);
        TabTri[nbTri].P[0][0] = (int)P1[0];
        TabTri[nbTri].P[0][1] = (int)P1[1];
        Persp3dTo2d(View, TmpTri->P2, P2, View->allocation.width, View->allocation.height);
        TabTri[nbTri].P[1][0] = (int)P2[0];
        TabTri[nbTri].P[1][1] = (int)P2[1];
        Persp3dTo2d(View, TmpTri->P3, P3, View->allocation.width, View->allocation.height);
        TabTri[nbTri].P[2][0] = (int)P3[0];
        TabTri[nbTri].P[2][1] = (int)P3[1];
        /* Calcul des distances des trois sommets a la camera */
        P1[0] = TmpTri->P1[0]-Camera->Location[0];
        P1[1] = TmpTri->P1[1]-Camera->Location[1];
        P1[2] = TmpTri->P1[2]-Camera->Location[2];
        TabTri[nbTri].Distance[0] = VLength(P1);
        P2[0] = TmpTri->P2[0]-Camera->Location[0];
        P2[1] = TmpTri->P2[1]-Camera->Location[1];
        P2[2] = TmpTri->P2[2]-Camera->Location[2];
        TabTri[nbTri].Distance[1] = VLength(P2);
        P3[0] = TmpTri->P3[0]-Camera->Location[0];
        P3[1] = TmpTri->P3[1]-Camera->Location[1];
        P3[2] = TmpTri->P3[2]-Camera->Location[2];
        TabTri[nbTri].Distance[2] = VLength(P3);
        /* Calcul des couleurs des trois sommets */
        /* 1er sommet */
        ColorResult[0] = Color[0] * 0.6;
        ColorResult[1] = Color[1] * 0.6;
        ColorResult[2] = Color[2] * 0.6;
        for (Light = UserData->FrameView->FirstLightSource ;
             Light ;
             Light = Light->Next)
        {
          NormalLight[0] = Light->Location[0]-TmpTri->P1[0];
          NormalLight[1] = Light->Location[1]-TmpTri->P1[1];
          NormalLight[2] = Light->Location[2]-TmpTri->P1[2];
          Norme = VLength(NormalLight);
          NormalLight[0] /= Norme;
          NormalLight[1] /= Norme;
          NormalLight[2] /= Norme;
          alpha = VDot(TmpTri->N1, NormalLight);
          if (Tmp->Type == PLANE_OBJECT)
            alpha = fabs(alpha);
          if (alpha > 0.0)
          {
            ColorResult[0] += alpha*Color[0]*Light->Color[0];
            ColorResult[1] += alpha*Color[1]*Light->Color[1];
            ColorResult[2] += alpha*Color[2]*Light->Color[2];
          }
        }
        if (ColorResult[0]<0.0) ColorResult[0]=0.0;
        if (ColorResult[0]>1.0) ColorResult[0]=1.0;
        if (ColorResult[1]<0.0) ColorResult[1]=0.0;
        if (ColorResult[1]>1.0) ColorResult[1]=1.0;
        if (ColorResult[2]<0.0) ColorResult[2]=0.0;
        if (ColorResult[2]>1.0) ColorResult[2]=1.0;
        TabTri[nbTri].Color[0][0] = ColorResult[0];
        TabTri[nbTri].Color[0][1] = ColorResult[1];
        TabTri[nbTri].Color[0][2] = ColorResult[2];
        /* 2eme sommet */
        ColorResult[0] = Color[0] * 0.6;
        ColorResult[1] = Color[1] * 0.6;
        ColorResult[2] = Color[2] * 0.6;
        for (Light = UserData->FrameView->FirstLightSource ;
             Light ;
             Light = Light->Next)
        {
          NormalLight[0] = Light->Location[0]-TmpTri->P2[0];
          NormalLight[1] = Light->Location[1]-TmpTri->P2[1];
          NormalLight[2] = Light->Location[2]-TmpTri->P2[2];
          Norme = VLength(NormalLight);
          NormalLight[0] /= Norme;
          NormalLight[1] /= Norme;
          NormalLight[2] /= Norme;
          alpha = VDot(TmpTri->N2, NormalLight);
          if (Tmp->Type == PLANE_OBJECT)
            alpha = fabs(alpha);
          if (alpha > 0.0)
          {
            ColorResult[0] += alpha*Color[0]*Light->Color[0];
            ColorResult[1] += alpha*Color[1]*Light->Color[1];
            ColorResult[2] += alpha*Color[2]*Light->Color[2];
          }
        }
        if (ColorResult[0]<0.0) ColorResult[0]=0.0;
        if (ColorResult[0]>1.0) ColorResult[0]=1.0;
        if (ColorResult[1]<0.0) ColorResult[1]=0.0;
        if (ColorResult[1]>1.0) ColorResult[1]=1.0;
        if (ColorResult[2]<0.0) ColorResult[2]=0.0;
        if (ColorResult[2]>1.0) ColorResult[2]=1.0;
        TabTri[nbTri].Color[1][0] = ColorResult[0];
        TabTri[nbTri].Color[1][1] = ColorResult[1];
        TabTri[nbTri].Color[1][2] = ColorResult[2];
        /* 3eme sommet */
        ColorResult[0] = Color[0] * 0.6;
        ColorResult[1] = Color[1] * 0.6;
        ColorResult[2] = Color[2] * 0.6;
        for (Light = UserData->FrameView->FirstLightSource ;
             Light ;
             Light = Light->Next)
        {
          NormalLight[0] = Light->Location[0]-TmpTri->P3[0];
          NormalLight[1] = Light->Location[1]-TmpTri->P3[1];
          NormalLight[2] = Light->Location[2]-TmpTri->P3[2];
          Norme = VLength(NormalLight);
          NormalLight[0] /= Norme;
          NormalLight[1] /= Norme;
          NormalLight[2] /= Norme;
          alpha = VDot(TmpTri->N3, NormalLight);
          if (Tmp->Type == PLANE_OBJECT)
            alpha = fabs(alpha);
          if (alpha > 0.0)
          {
            ColorResult[0] += alpha*Color[0]*Light->Color[0];
            ColorResult[1] += alpha*Color[1]*Light->Color[1];
            ColorResult[2] += alpha*Color[2]*Light->Color[2];
          }
        }
        if (ColorResult[0]<0.0) ColorResult[0]=0.0;
        if (ColorResult[0]>1.0) ColorResult[0]=1.0;
        if (ColorResult[1]<0.0) ColorResult[1]=0.0;
        if (ColorResult[1]>1.0) ColorResult[1]=1.0;
        if (ColorResult[2]<0.0) ColorResult[2]=0.0;
        if (ColorResult[2]>1.0) ColorResult[2]=1.0;
        TabTri[nbTri].Color[2][0] = ColorResult[0];
        TabTri[nbTri].Color[2][1] = ColorResult[1];
        TabTri[nbTri].Color[2][2] = ColorResult[2];
        nbTri++;
      } /* for */
    } /* if (Tmp->Visible) */
  }
  /* Displaying */
  Zbuffer = g_malloc(View->allocation.width *
                     View->allocation.height * sizeof(double));
  for (i=0 ; i<View->allocation.width ; i++)
    for (j=0 ; j<View->allocation.height ; j++)
      Zbuffer[j*View->allocation.width+i] = 1E20;
  for (i=0 ; i<nbTri ; i++)
    DrawGouraudTriangleZBuffer(View, Zbuffer, &(TabTri[i]));
  /* Freeing ... */
  g_free(TabTri);
  g_free(Zbuffer);
  /* Draw the Rotation/Scale Center if any... */
/*  if (ModeTool == MT_ROTATE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->RotationCenter);
  }
  if (ModeTool == MT_SCALE)
  {
    DrawCenterXY(View, Xoff, Yoff, Zoom, UserData->FrameView->ScaleCenter);
  }*/
#endif
  return (FALSE);
}

/*****************************************************************************
*  CameraViewEventDispatcher
******************************************************************************/
gint CameraViewEventDispatcher(GtkWidget *DrawingArea, GdkEvent *event, 
                               gpointer Menu)
{ 
  switch (event->type)
  {
    case GDK_BUTTON_PRESS:
      {
        GdkEventButton *bevent = (GdkEventButton *) event;
        if (bevent->button == 3) /* Right Button ==> Popup Menu */
        {
          CurrentUserData = gtk_object_get_data(GTK_OBJECT(DrawingArea),
                                                "ViewData");
          gtk_menu_popup(GTK_MENU(Menu), NULL, NULL, NULL, NULL,
                         bevent->button, bevent->time);
        }
        return TRUE;
      }
    default:
      return FALSE;
  }
}

static GtkItemFactoryEntry CameraViewMenu[] =
{
  {"/_File",                       NULL, NULL,                     0, "<Branch>"},
  {"/File/Load _Pov",              NULL, LoadPovFunc,              0, "<Item>"},
  {"/File/Save P_ov",              NULL, SavePovFunc,              0, "<Item>"},
  {"/File/Dummy",                  NULL, NULL,                     0, "<Separator>"},
  {"/File/_Quit",                  NULL, ExitFunc,                 0, "<Item>"},
  {"/_Selection",                  NULL, NULL,                     0, "<Branch>"},
  {"/Selection/Select _All",       NULL, SelectAll,                0, "<Item>"},
  {"/Selection/Select _None",      NULL, SelectNone,               0, "<Item>"},
  {"/Selection/_Invert Selection", NULL, InvertSelection,          0, "<Item>"},
  {"/_Operations",                 NULL, NULL,                     0, "<Branch>"},
  {"/Operations/_Translate",       NULL, TranslateSelectionDialog, 0, "<Item>"},
  {"/Operations/_Rotate",          NULL, RotateSelectionDialog,    0, "<Item>"},
  {"/Operations/_Scale",           NULL, ScaleSelectionDialog,     0, "<Item>"},
  {"/Operations/_Edit Camera",     NULL, EditCamera,               0, "<Item>"},
  {"/_Render",                     NULL, NULL,                     0, "<Branch>"},
  {"/Render/_Wire Frame",          NULL, SetWireFrameExpose,       0, "<Item>"},
  {"/Render/_Hidden Faces",        NULL, SetHiddenFacesExpose,     0, "<Item>"},
  {"/Render/_Gouraud",             NULL, SetGouraudExpose,         0, "<Item>"},
  {"/Render/_PovRay Rendering",    NULL, CallPovFront,             0, "<Item>"},
  {"/_Dialogs",                    NULL, NULL,                     0, "<Branch>"},
  {"/Dialogs/_CSG Tree",           NULL, ShowCSGTree,              0, "<Item>"},
  {"/Dialogs/_Tool Options",       NULL, ShowToolOptionsDialog,    0, "<Item>"}
};


/*****************************************************************************
*  NewViewCamera
******************************************************************************/
void NewViewCamera(FrameStruct *LocalFrame)
{
  FrameViewStruct *View;
  GtkWidget *window;
  GtkWidget *DrawingArea;
  GtkWidget *table;
  ViewUserDataStruct *UserData;
  GtkWidget *StatusBar;
  GtkWidget *menu;
  GtkWidget *SubMenu;
  GtkWidget *menu_items;
  GtkWidget *FileMenu;
  GtkWidget *SelectionMenu;
  GtkWidget *OperationMenu;
  GtkWidget *RenderMenu;
  GtkWidget *DialogMenu;

  View = g_malloc(sizeof(FrameViewStruct));
  LocalFrame->all_views = g_slist_append(LocalFrame->all_views, View);
  View->name = g_strdup_printf("Camera %s", LocalFrame->Name);

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);

  gtk_window_set_title(GTK_WINDOW(window), View->name);
  gtk_widget_set_events(window, GDK_POINTER_MOTION_MASK
                              | GDK_POINTER_MOTION_HINT_MASK
                              | GDK_KEY_PRESS_MASK);
  gtk_container_set_border_width(GTK_CONTAINER(window), 0);
  /* The Table */
  table = gtk_table_new(4, 5, FALSE);
  gtk_container_add(GTK_CONTAINER(window), table);
  gtk_widget_show(table);
  DrawingArea = gtk_drawing_area_new();
  gtk_widget_set_usize(DrawingArea, 512, 384); /* XXX should be setable */
  View->drawing_area = DrawingArea;
  gtk_table_attach(GTK_TABLE(table), DrawingArea, 0,4, 0,4,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0,0);
  UserData = g_malloc(sizeof(ViewUserDataStruct));
  UserData->FrameView = LocalFrame;
  gtk_object_set_data(GTK_OBJECT(DrawingArea), "ViewData", UserData);
  gtk_object_set_data(GTK_OBJECT(window), "ViewData", UserData);
  UserData->RenderStyle =
            gtk_signal_connect(GTK_OBJECT(DrawingArea), "expose_event",
                               (GtkSignalFunc)ExposeWireFrameViewCamera, 
                               NULL);
  UserData->CameraStyle = FRAME_CAMERA;
  gtk_widget_set_events(DrawingArea, GDK_EXPOSURE_MASK
                                   | GDK_LEAVE_NOTIFY_MASK
                                   | GDK_BUTTON_PRESS_MASK
                                   | GDK_BUTTON_RELEASE_MASK
                                   | GDK_POINTER_MOTION_MASK
                                   | GDK_KEY_PRESS_MASK);
  StatusBar = gtk_statusbar_new();
  gtk_table_attach(GTK_TABLE(table), StatusBar, 0,4, 4,5,
                   GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_widget_show(StatusBar);
  UserData->StatusBar = StatusBar;
  UserData->DrawingArea = DrawingArea;

  /* Create the Menu */
  menu = gtk_menu_new();
  /* The "File" SubMenu */
  SubMenu = gtk_menu_new();
  /* The Load Pov Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Load Pov"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(LoadPovFunc), NULL);
  gtk_widget_show(menu_items);
  /* The Save Pov Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Save Pov"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect(GTK_OBJECT(menu_items), "activate",
                     GTK_SIGNAL_FUNC(SavePovFunc), LocalFrame);
  gtk_widget_show(menu_items);
  /* Separator */
  menu_items = gtk_menu_item_new();
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_widget_show(menu_items);
  /* The Quit Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Quit"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ExitFunc), NULL);
  gtk_widget_show(menu_items);
  FileMenu = gtk_menu_item_new_with_label(_("Files"));
  gtk_widget_show(FileMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(FileMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), FileMenu);
  /* The "Selection" SubMenu */
  SubMenu = gtk_menu_new();
  /* The "Select All" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Select All"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SelectAll),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Select None" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Select None"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SelectNone),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Invert Selection" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Invert Selection"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(InvertSelection),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  SelectionMenu = gtk_menu_item_new_with_label(_("Selection"));
  gtk_widget_show(SelectionMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(SelectionMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), SelectionMenu);
  /* The "Operations" SubMenu */
  SubMenu = gtk_menu_new();
  /* The "Translate" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Translate"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(TranslateSelectionDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Rotate" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Rotate"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(RotateSelectionDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Scale" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Scale"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ScaleSelectionDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The "Edit Camera" Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Edit Camera"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(EditCamera),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  OperationMenu = gtk_menu_item_new_with_label(_("Operations"));
  gtk_widget_show(OperationMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(OperationMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), OperationMenu);
  /* The "Render" SubMenu */
  /* Create the SubMenu */
  SubMenu = gtk_menu_new();
  /* The Wire Frame Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Wire Frame"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SetWireFrameExpose),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Hidden Faces Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Hidden Faces"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SetHiddenFacesExpose),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Gouraud Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Gouraud"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(SetGouraudExpose),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Povfront MenuItem */
  menu_items = gtk_menu_item_new_with_label(_("POVRay Rendering"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(CallPovFront),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  RenderMenu = gtk_menu_item_new_with_label(_("Render"));
  gtk_widget_show(RenderMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(RenderMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), RenderMenu);
  /* The "Dialogs" SubMenu */
  /* Create the SubMenu */
  SubMenu = gtk_menu_new();
  /* The CSG Tree Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("CSG Tree"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ShowCSGTree),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* The Tool Options Menu Item */
  menu_items = gtk_menu_item_new_with_label(_("Tool Options"));
  gtk_menu_append(GTK_MENU(SubMenu), menu_items);
  gtk_signal_connect_object(GTK_OBJECT(menu_items), "activate",
                            GTK_SIGNAL_FUNC(ShowToolOptionsDialog),
                            GTK_OBJECT(DrawingArea));
  gtk_widget_show(menu_items);
  /* */
  DialogMenu = gtk_menu_item_new_with_label(_("Dialogs"));
  gtk_widget_show(DialogMenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(DialogMenu), SubMenu);
  gtk_menu_append(GTK_MENU(menu), DialogMenu);
  /*  */
       
  gtk_signal_connect(GTK_OBJECT(DrawingArea), "event",
                     GTK_SIGNAL_FUNC(CameraViewEventDispatcher), menu);
  gtk_widget_show(DrawingArea);
  gtk_widget_show(window);
}
