/* copy.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include "giram.h"
#include "object.h"
#include "copy.h"

/*****************************************************************************
*  CopyObject
******************************************************************************/
ObjectStruct *CopyObject(ObjectStruct *OrgObject)
{
  ObjectStruct *LocalObject;

  if (OrgObject)
  {
    switch (OrgObject->Type)
    {
      case BOX_OBJECT:
        LocalObject = g_malloc(sizeof(BoxStruct));
        {
          BoxStruct *BBox = (BoxStruct *)LocalObject;
          BoxStruct *OrgBox = (BoxStruct *)OrgObject;

          BBox->MinCorner[0] = OrgBox->MinCorner[0];
          BBox->MinCorner[1] = OrgBox->MinCorner[1];
          BBox->MinCorner[2] = OrgBox->MinCorner[2];
          BBox->MaxCorner[0] = OrgBox->MaxCorner[0];
          BBox->MaxCorner[1] = OrgBox->MaxCorner[1];
          BBox->MaxCorner[2] = OrgBox->MaxCorner[2];
        }
        break;
      case CONE_OBJECT:
        LocalObject = g_malloc(sizeof(ConeStruct));
        {
          ConeStruct *CCone = (ConeStruct *)LocalObject;
          ConeStruct *OrgCone = (ConeStruct *)OrgObject;

          CCone->Open = OrgCone->Open;
          CCone->Radius1 = OrgCone->Radius1;
          CCone->Radius2 = OrgCone->Radius2;
        }
        break;
      case CSG_OBJECT:
        LocalObject = g_malloc(sizeof(CSGStruct));
        {
          CSGStruct *CCSG = (CSGStruct *)LocalObject;
          CSGStruct *OrgCSG = (CSGStruct *)OrgObject;
          ObjectStruct *Object1, *Object2;

          CCSG->Type = OrgCSG->Type;
/*          CCSG->FirstObject = CopyObject(OrgCSG->FirstObject);
          Object1 = CCSG->FirstObject;
          Object2 = OrgCSG->FirstObject;
          while (Object2)
          {
            Object1->Next = CopyObject(Object2->Next);
            Object1 = Object1->Next;
            Object2 = Object2->Next;
          }*/
        }
        break;
      case CYLINDER_OBJECT:
        LocalObject = g_malloc(sizeof(CylinderStruct));
        {
          CylinderStruct *CCylinder = (CylinderStruct *)LocalObject;
          CylinderStruct *OrgCylinder = (CylinderStruct *)OrgObject;

          CCylinder->Open = OrgCylinder->Open;
        }
        break;
      case DISC_OBJECT:
        LocalObject = g_malloc(sizeof(DiscStruct));
        {
          DiscStruct *DDisc = (DiscStruct *)LocalObject;
          DiscStruct *OrgDisc = (DiscStruct *)OrgObject;

          DDisc->Radius = OrgDisc->Radius;
          DDisc->HoleRadius = DDisc->HoleRadius;
        }
        break;
      case MESH_OBJECT:
        LocalObject = g_malloc(sizeof(MeshStruct));
        {
          MeshStruct *MMesh = (MeshStruct *)LocalObject;
          MeshStruct *OrgMesh = (MeshStruct *)OrgObject;
          TriangleListStruct *TmpTri, *NewTri;

          MMesh->FirstTriangle = NULL;
          for (TmpTri = OrgMesh->FirstTriangle ; TmpTri ; TmpTri=TmpTri->Next)
          {
            NewTri = g_malloc(sizeof(TriangleListStruct));
            NewTri->Next = MMesh->FirstTriangle;
            MMesh->FirstTriangle = NewTri;
            NewTri->P1[0] = TmpTri->P1[0]; NewTri->N1[0] = TmpTri->N1[0];
            NewTri->P1[1] = TmpTri->P1[1]; NewTri->N1[1] = TmpTri->N1[1];
            NewTri->P1[2] = TmpTri->P1[2]; NewTri->N1[2] = TmpTri->N1[2];
            NewTri->P2[0] = TmpTri->P2[0]; NewTri->N2[0] = TmpTri->N2[0];
            NewTri->P2[1] = TmpTri->P2[1]; NewTri->N2[1] = TmpTri->N2[1];
            NewTri->P2[2] = TmpTri->P2[2]; NewTri->N2[2] = TmpTri->N2[2];
            NewTri->P3[0] = TmpTri->P3[0]; NewTri->N3[0] = TmpTri->N3[0];
            NewTri->P3[1] = TmpTri->P3[1]; NewTri->N3[1] = TmpTri->N3[1];
            NewTri->P3[2] = TmpTri->P3[2]; NewTri->N3[2] = TmpTri->N3[2];
          }
        }
        break;
      case PLANE_OBJECT:
        LocalObject = g_malloc(sizeof(PlaneStruct));
        {
          PlaneStruct *PPlane = (PlaneStruct *)LocalObject;
          PlaneStruct *OrgPlane = (PlaneStruct *)OrgObject;

          PPlane->Normal[0] = OrgPlane->Normal[0];
          PPlane->Normal[1] = OrgPlane->Normal[1];
          PPlane->Normal[2] = OrgPlane->Normal[2];
          PPlane->Distance = OrgPlane->Distance;
          PPlane->NbSquare = OrgPlane->NbSquare;
          PPlane->SquareSize = OrgPlane->SquareSize;
        }
        break;
      case SOR_OBJECT:
        LocalObject = g_malloc(sizeof(SorStruct));
        {
          SorStruct *SSor = (SorStruct *)LocalObject;
          SorStruct *OrgSor = (SorStruct *)OrgObject;
          guint i;

          SSor->NumberOfPoints = OrgSor->NumberOfPoints;
          SSor->Point = g_malloc(SSor->NumberOfPoints*sizeof(Vector));
          for (i=0 ; i<SSor->NumberOfPoints ; i++)
          {
            SSor->Point[i][0] = OrgSor->Point[i][0];
            SSor->Point[i][1] = OrgSor->Point[i][1];
          }
        }
        break;
      case SPHERE_OBJECT:
        LocalObject = g_malloc(sizeof(SphereStruct));
        {
          SphereStruct *SSphere = (SphereStruct *)LocalObject;
          SphereStruct *OrgSphere = (SphereStruct *)OrgObject;

          SSphere->Center[0] = OrgSphere->Center[0];
          SSphere->Center[1] = OrgSphere->Center[1];
          SSphere->Center[2] = OrgSphere->Center[2];
          SSphere->Radius = OrgSphere->Radius;
        }
        break;
      case SUPERELLIPSOID_OBJECT:
        LocalObject = g_malloc(sizeof(SuperEllipsoidStruct));
        {
          SuperEllipsoidStruct *SSuperEllipsoid = (SuperEllipsoidStruct *)LocalObject;
          SuperEllipsoidStruct *OrgSuperEllipsoid = (SuperEllipsoidStruct *)OrgObject;

          SSuperEllipsoid->E = OrgSuperEllipsoid->E;
          SSuperEllipsoid->N = OrgSuperEllipsoid->N;
        }
        break;
      case TORUS_OBJECT:
        LocalObject = g_malloc(sizeof(TorusStruct));
        {
          TorusStruct *TTorus = (TorusStruct *)LocalObject;
          TorusStruct *OrgTorus = (TorusStruct *)OrgObject;

          TTorus->Major = OrgTorus->Major;
          TTorus->Minor = OrgTorus->Minor;
        }
        break;
      case TRIANGLE_OBJECT:
        LocalObject = g_malloc(sizeof(TriangleStruct));
        {
          TriangleStruct *TTriangle = (TriangleStruct *)LocalObject;
          TriangleStruct *OrgTriangle = (TriangleStruct *)OrgObject;

          TTriangle->P1[0] = OrgTriangle->P1[0];
          TTriangle->P1[1] = OrgTriangle->P1[1];
          TTriangle->P1[2] = OrgTriangle->P1[2];
          TTriangle->P2[0] = OrgTriangle->P2[0];
          TTriangle->P2[1] = OrgTriangle->P2[1];
          TTriangle->P2[2] = OrgTriangle->P2[2];
          TTriangle->P3[0] = OrgTriangle->P3[0];
          TTriangle->P3[1] = OrgTriangle->P3[1];
          TTriangle->P3[2] = OrgTriangle->P3[2];
          TTriangle->N1[0] = OrgTriangle->N1[0];
          TTriangle->N1[1] = OrgTriangle->N1[1];
          TTriangle->N1[2] = OrgTriangle->N1[2];
          TTriangle->N2[0] = OrgTriangle->N2[0];
          TTriangle->N2[1] = OrgTriangle->N2[1];
          TTriangle->N2[2] = OrgTriangle->N2[2];
          TTriangle->N3[0] = OrgTriangle->N3[0];
          TTriangle->N3[1] = OrgTriangle->N3[1];
          TTriangle->N3[2] = OrgTriangle->N3[2];
        }
        break;
      default:
        LocalObject = NULL;
        printf("Severe bug in CopyObject\n");
        exit(1);
    }
    LocalObject->Type = OrgObject->Type;
    LocalObject->FirstTriangle = NULL;
    LocalObject->Trans = CopyTransform(OrgObject->Trans);
    LocalObject->Texture = CopyTexture(OrgObject->Texture);
    LocalObject->Inverse = OrgObject->Inverse;
    LocalObject->NoShadow = OrgObject->NoShadow;
    LocalObject->Hollow = OrgObject->Hollow;
    LocalObject->Selected = OrgObject->Selected;
    LocalObject->Visible = OrgObject->Visible;

    LocalObject->klass = OrgObject->klass;
    
    giram_object_build_triangle_mesh(LocalObject);
    return LocalObject;
  } else
    return NULL;
}

/*****************************************************************************
*  CopyTexture
******************************************************************************/
TextureStruct *CopyTexture(TextureStruct *OrgTexture)
{
  TextureStruct *LocalTexture;

  if (OrgTexture)
  {
    LocalTexture = g_malloc(sizeof(TextureStruct));
    LocalTexture->Pigment = CopyPigment(OrgTexture->Pigment);
    LocalTexture->Trans = CopyTransform(OrgTexture->Trans);
    return LocalTexture;
  } else
    return NULL;
}

/*****************************************************************************
*  CopyPigment
******************************************************************************/
PigmentStruct *CopyPigment(PigmentStruct *OrgPigment)
{
  PigmentStruct *LocalPigment;

  if (OrgPigment)
  {
    LocalPigment = g_malloc(sizeof(PigmentStruct));
    LocalPigment->Type = OrgPigment->Type;
    LocalPigment->Color[0] = OrgPigment->Color[0];
    LocalPigment->Color[1] = OrgPigment->Color[1];
    LocalPigment->Color[2] = OrgPigment->Color[2];
    LocalPigment->Color[3] = OrgPigment->Color[3];
    LocalPigment->Color[4] = OrgPigment->Color[4];
    LocalPigment->Trans = CopyTransform(OrgPigment->Trans);
    return LocalPigment;
  } else
    return NULL;
}

/*****************************************************************************
*  CopyCamera
******************************************************************************/
CameraStruct *CopyCamera(CameraStruct *OrgCamera)
{
  CameraStruct *LocalCamera;

  if (OrgCamera)
  {
    LocalCamera = g_malloc(sizeof(CameraStruct));
    LocalCamera->Type = OrgCamera->Type;
    LocalCamera->CylinderType = OrgCamera->CylinderType;
    LocalCamera->Location[0] = OrgCamera->Location[0];
    LocalCamera->Location[1] = OrgCamera->Location[1];
    LocalCamera->Location[2] = OrgCamera->Location[2];
    LocalCamera->LookAt[0] = OrgCamera->LookAt[0];
    LocalCamera->LookAt[1] = OrgCamera->LookAt[1];
    LocalCamera->LookAt[2] = OrgCamera->LookAt[2];
    LocalCamera->Right[0] = OrgCamera->Right[0];
    LocalCamera->Right[1] = OrgCamera->Right[1];
    LocalCamera->Right[2] = OrgCamera->Right[2];
    LocalCamera->Up[0] = OrgCamera->Up[0];
    LocalCamera->Up[1] = OrgCamera->Up[1];
    LocalCamera->Up[2] = OrgCamera->Up[2];
    LocalCamera->Direction[0] = OrgCamera->Direction[0];
    LocalCamera->Direction[1] = OrgCamera->Direction[1];
    LocalCamera->Direction[2] = OrgCamera->Direction[2];
    LocalCamera->Sky[0] = OrgCamera->Sky[0];
    LocalCamera->Sky[1] = OrgCamera->Sky[1];
    LocalCamera->Sky[2] = OrgCamera->Sky[2];
    LocalCamera->Angle = OrgCamera->Angle;
    LocalCamera->BlurSamples = OrgCamera->BlurSamples;
    LocalCamera->Aperture = OrgCamera->Aperture;
    LocalCamera->FocalPoint[0] = OrgCamera->FocalPoint[0];
    LocalCamera->FocalPoint[1] = OrgCamera->FocalPoint[1];
    LocalCamera->FocalPoint[2] = OrgCamera->FocalPoint[2];
    LocalCamera->Confidence = OrgCamera->Confidence;
    LocalCamera->Variance = OrgCamera->Variance;
    return LocalCamera;
  } else
    return NULL;
}

/*****************************************************************************
*  CopyLightSource
******************************************************************************/
LightSourceStruct *CopyLightSource(LightSourceStruct *OrgLightSource)
{
  LightSourceStruct *LocalLightSource;

  if (OrgLightSource)
  {
    LocalLightSource = g_malloc(sizeof(LightSourceStruct));
    LocalLightSource->Location[0] = OrgLightSource->Location[0];
    LocalLightSource->Location[1] = OrgLightSource->Location[1];
    LocalLightSource->Location[2] = OrgLightSource->Location[2];
    LocalLightSource->Color[0] = OrgLightSource->Color[0];
    LocalLightSource->Color[1] = OrgLightSource->Color[1];
    LocalLightSource->Color[2] = OrgLightSource->Color[2];
    LocalLightSource->Color[3] = OrgLightSource->Color[3];
    LocalLightSource->Color[4] = OrgLightSource->Color[4];
    LocalLightSource->Type = OrgLightSource->Type;
    LocalLightSource->PointAt[0] = OrgLightSource->PointAt[0];
    LocalLightSource->PointAt[1] = OrgLightSource->PointAt[1];
    LocalLightSource->PointAt[2] = OrgLightSource->PointAt[2];
    LocalLightSource->Radius = OrgLightSource->Radius;
    LocalLightSource->FallOff = OrgLightSource->FallOff;
    LocalLightSource->Tightness = OrgLightSource->Tightness;
    LocalLightSource->AreaLight = OrgLightSource->AreaLight;
    LocalLightSource->Area1[0] = OrgLightSource->Area1[0];
    LocalLightSource->Area1[1] = OrgLightSource->Area1[1];
    LocalLightSource->Area1[2] = OrgLightSource->Area1[2];
    LocalLightSource->Area2[0] = OrgLightSource->Area2[0];
    LocalLightSource->Area2[1] = OrgLightSource->Area2[1];
    LocalLightSource->Area2[2] = OrgLightSource->Area2[2];
    LocalLightSource->AreaSize1 = OrgLightSource->AreaSize1;
    LocalLightSource->AreaSize2 = OrgLightSource->AreaSize2;
    LocalLightSource->Adaptive = OrgLightSource->Adaptive;
    LocalLightSource->Jitter = OrgLightSource->Jitter;
    LocalLightSource->LooksLike = CopyObject(OrgLightSource->LooksLike);
    LocalLightSource->FadeDistance = OrgLightSource->FadeDistance;
    LocalLightSource->FadePower = OrgLightSource->FadePower;
    LocalLightSource->AtmosphericAttenuation = OrgLightSource->AtmosphericAttenuation;
    LocalLightSource->Atmosphere = LocalLightSource->Atmosphere;
    LocalLightSource->Next = NULL;
    return LocalLightSource;
  } else
    return NULL;
}

/*****************************************************************************
*  CopyTransform
******************************************************************************/
TransformStruct *CopyTransform(TransformStruct *OrgTransform)
{
  TransformStruct *LocalTransform;
  int i,j;

  if (OrgTransform)
  {
    LocalTransform = g_malloc(sizeof(TransformStruct));
    for (i=0 ; i<4 ; i++)
      for (j=0 ; j<4 ; j++)
      {
        LocalTransform->Direct[i][j] = OrgTransform->Direct[i][j];
        LocalTransform->Inverse[i][j] = OrgTransform->Inverse[i][j];
      }
    return LocalTransform;
  } else
    return NULL;
}

