/* marching_cube.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "giram.h"
#include "trimesh.h"
#include "csg.h"
#include "marching_cube.h"
/*
 *            6-----------------7
 *           /|                /|
 *          / |               / |
 *         /  |              /  |
 *        2---+-------------3   |
 *        |   |             |   |
 *        |   |             |   |
 *        |   |             |   |
 *        |   4-------------+---5
 *        |  /              |  /
 *        | /               | /
 *        |/                |/
 *        0-----------------1
 *
 */

/*****************************************************************************
*  AddTriangleToCSGMesh
******************************************************************************/
void AddTriangleToCSGMesh(ObjectStruct *CSG, Vector P1, Vector P2, Vector P3,
                                             Vector N1, Vector N2, Vector N3)
{
  Vector P11, P22, P33, N11, N22, N33;

  if (CSG->Trans)
  {
    MEvaluatePoint(P11, CSG->Trans, P1);
    MEvaluatePoint(P22, CSG->Trans, P2);
    MEvaluatePoint(P33, CSG->Trans, P3);
    MEvaluateVector(N11, CSG->Trans, N1);
    MEvaluateVector(N22, CSG->Trans, N2);
    MEvaluateVector(N33, CSG->Trans, N3);
  } else
  {
    P11[0] = P1[0]; P11[1] = P1[1]; P11[2] = P1[2];
    P22[0] = P2[0]; P22[1] = P2[1]; P22[2] = P2[2];
    P33[0] = P3[0]; P33[1] = P3[1]; P33[2] = P3[2];
    N11[0] = N1[0]; N11[1] = N1[1]; N11[2] = N1[2];
    N22[0] = N2[0]; N22[1] = N2[1]; N22[2] = N2[2];
    N33[0] = N3[0]; N33[1] = N3[1]; N33[2] = N3[2];
  }
  AddTriangleToObjectMesh(CSG, P11,P22,P33, N11,N22,N33);
}

/*****************************************************************************
*  MarchingCube
******************************************************************************/
void MarchingCube(ObjectStruct *CSG, Vector Cube[8])
{
  int i,n;
  Vector P1,P2,P3,P4, N1,N2,N3,N4;

  n=0;
  for (i=0; i<8 ; i++)
  {
    if (CSG->Trans)
    {
      MEvaluatePoint(P1, CSG->Trans, Cube[i]);
    } else
    {
      P1[0] = Cube[i][0]; P1[1] = Cube[i][1]; P1[2] = Cube[i][2];
    }
    if (giram_object_inside(CSG, P1[0], P1[1], P1[2]))
      n |= (1 << i);
  }
  N1[0] = N1[1] = N1[2] = 0.0;
  N2[0] = N2[1] = N2[2] = 0.0;
  N3[0] = N3[1] = N3[2] = 0.0;
  switch (n)
  {
    case 0: /* case 0 */
      break;
    case 1: /* (0) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 2: /* (1) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 3: /* (0, 1) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 4: /* (2) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 5: /* (0, 2) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 6: /* (1, 2) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 7: /* (0, 1, 2) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 8: /* (3) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 9: /* (0, 3) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 10: /* (1, 3) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 11: /* (0, 1, 3) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 12: /* (2, 3) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 13: /* (0, 2, 3) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 14: /* (1, 2, 3) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 15: /* (0, 1, 2, 3) case 8 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 16: /* (4) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 17: /* (0, 4) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 18: /* (1, 4) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 19: /* (0, 1, 4) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 20: /* (2, 4) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 21: /* (0, 2, 4) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 22: /* (1, 2, 4) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 23: /* (0, 1, 2, 4) case 9 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 24: /* (3, 4) case 4 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 25: /* (0, 3, 4) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 26: /* (1, 3, 4) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 27: /* (0, 1, 3, 4) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 28: /* (2, 3, 4) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 29: /* (0, 2, 3, 4) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 30: /* (1, 2, 3, 4) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 31: /* (0, 1, 2, 3, 4) = (5, 6, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 32: /* (5) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 33: /* (0, 5) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 34: /* (1, 5) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 35: /* (0, 1, 5) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 36: /* (2, 5) case 4 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 37: /* (0, 2, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 38: /* (1, 2, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 39: /* (0, 1, 2, 5) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 40: /* (3, 5) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 41: /* (0, 3, 5) 7 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 42: /* (1, 3, 5) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 43: /* (0, 1, 3, 5) case 9 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 44: /* (2, 3, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 45: /* (0, 2, 3, 5) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 46: /* (1, 2, 3, 5) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 47: /* (0, 1, 2, 3, 5) = (4, 6, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 48: /* (4, 5) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 49: /* (0, 4, 5) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 50: /* (1, 4, 5) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 51: /* (0, 1, 4, 5) case 8 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 52: /* (2, 4, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 53: /* (0, 2, 4, 5) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 54: /* (1, 2, 4, 5) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 55: /* (0, 1, 2, 4, 5) = (3, 6, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 56: /* (3, 4, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 57: /* (0, 3, 4, 5) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 58: /* (1, 3, 4, 5) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 59: /* (0, 1, 3, 4, 5) = (2, 6, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 60: /* (2, 3, 4, 5) case 10 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 61: /* (0, 2, 3, 4, 5) = (1, 6, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 62: /* (1, 2, 3, 4, 5) = (0, 6, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 63: /* (0, 1, 2, 3, 4, 5) = (6, 7) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 64: /* (6) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 65: /* (0, 6) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 66: /* (1, 6) case 4 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 67: /* (0, 1, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 68: /* (2, 6) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 69: /* (0, 2, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 70: /* (1, 2, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 71: /* (0, 1, 2, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 72: /* (3, 6) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 73: /* (0, 3, 6) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 74: /* (1, 3, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 75: /* (0, 1, 3, 6) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 76: /* (2, 3, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 77: /* (0, 2, 3, 6) case 9 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 78: /* (1, 2, 3, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1); 
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 79: /* (0, 1, 2, 3, 6) = (4, 5, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 80: /* (4, 6) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 81: /* (0, 4, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 82: /* (1, 4, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 83: /* (0, 1, 4, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 84: /* (2, 4, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 85: /* (0, 2, 4, 6) case 8 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 86: /* (1, 2, 4, 6) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 87: /* (0, 1, 2, 4, 6) = (3, 5, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 88: /* (3, 4, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 89: /* (0, 3, 4, 6) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 90: /* (1, 3, 4, 6) case 10 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 91: /* (0, 1, 3, 4, 6) = (2, 5, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 92: /* (2, 3, 4, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 93: /* (0, 2, 3, 4, 6) = (1, 5, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 94: /* (1, 2, 3, 4, 6) = (0, 5, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 95: /* (0, 1, 2, 3, 4, 6) = (5, 7) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 96: /* (5, 6) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 97: /* (0, 5, 6) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 98: /* (1, 5, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 99: /* (0, 1, 5, 6) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 100: /* (2, 5, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 101: /* (0, 2, 5, 6) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 102: /* (1, 2, 5, 6) case 10 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 103: /* (0, 1, 2, 5, 6) = (3, 4, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 104: /* (3, 5, 6) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 105: /* (0, 3, 5, 6) case 13 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 106: /* (1, 3, 5, 6) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 107: /* (0, 1, 3, 5, 6) = (2, 4, 7) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 108: /* (2, 3, 5, 6) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 109: /* (0, 2, 3, 5, 6) = (1, 4, 7) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 110: /* (1, 2, 3, 5, 6) = (0, 4, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 111: /* (0, 1, 2, 3, 5, 6) = (4, 7) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 112: /* (4, 5, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 113: /* (0, 4, 5, 6) case 9 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 114: /* (1, 4, 5, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 115: /* (0, 1, 4, 5, 6) = (2, 3, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 116: /* (2, 4, 5, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 117: /* (0, 2, 4, 5, 6) = (1, 3, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 118: /* (1, 2, 4, 5, 6) = (0, 3, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 119: /* (0, 1, 2, 4, 5, 6) = (3, 7) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 120: /* (3, 4, 5, 6) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 121: /* (0, 3, 4, 5, 6) = (1, 2, 7) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 122: /* (1, 3, 4, 5, 6) = (0, 2, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 123: /* (0, 1, 3, 4, 5, 6) = (2, 7) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 124: /* (2, 3, 4, 5, 6) = (0, 1, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 125: /* (0, 2, 3, 4, 5, 6) = (1, 7) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 126: /* (1, 2, 3, 4, 5, 6) = (0, 7) case 4 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 127: /* (0, 1, 2, 3, 4, 5, 6) = (7) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 128: /* (7) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 129: /* (0, 7) case 4 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 130: /* (1, 7) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 131: /* (0, 1, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 132: /* (2, 7) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 133: /* (0, 2, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 134: /* (1, 2, 7) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 135: /* (0, 1, 2, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 136: /* (3, 7) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 137: /* (0, 3, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 138: /* (1, 3, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 139: /* (0, 1, 3, 7)  = (2, 4, 5, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 140: /* (2, 3, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 141: /* (0, 2, 3, 7) = (1, 4, 5, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 142: /* (1, 2, 3, 7) case 9 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 143: /* (0, 1, 2, 3, 7) = (4, 5, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 144: /* (4, 7) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 145: /* (0, 4, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 146: /* (1, 4, 7) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 147: /* (0, 1, 4, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 148: /* (2, 4, 7) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 149: /* (0, 2, 4, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 150: /* (1, 2, 4, 7) case 13 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 151: /* (0, 1, 2, 4, 7) = (3, 5, 6) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 152: /* (3, 4, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 153: /* (0, 3, 4, 7) case 10 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 154: /* (1, 3, 4, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 155: /* (0, 1, 3, 4, 7) = (2, 5, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 156: /* (2, 3, 4, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 157: /* (0, 2, 3, 4, 7) = (1, 5, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 158: /* (1, 2, 3, 4, 7) = (0, 5, 6) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 159: /* (0, 1, 2, 3, 4, 7) = (5, 6) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 160: /* (5, 7) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 161: /* (0, 5, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 162: /* (1, 5, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 163: /* (0, 1, 5, 7) = (2, 3, 4, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 164: /* (2, 5, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 165: /* (0, 2, 5, 7) case 10 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 166: /* (1, 2, 5, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 167: /* (0, 1, 2, 5, 7) = (3, 4, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 168: /* (3, 5, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 169: /* (0, 3, 5, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 170: /* (1, 3, 5, 7) case 8 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 171: /* (0, 1, 3, 5, 7) = (2, 4, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 172: /* (2, 3, 5, 7) = (0, 1, 4, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 173: /* (0, 2, 3, 5, 7) = (1, 4, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 174: /* (1, 2, 3, 5, 7) = (0, 4, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 175: /* (0, 1, 2, 3, 5, 7) = (4, 6) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 176: /* (4, 5, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 177: /* (0, 4, 5, 7) = (1, 2, 3, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1); 
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 178: /* (1, 4, 5, 7) case 9 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 179: /* (0, 1, 4, 5, 7) = (2, 3, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 180: /* (2, 4, 5, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 181: /* (0, 2, 4, 5, 7) = (1, 3, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 182: /* (1, 2, 4, 5, 7) = (0, 3, 6) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 183: /* (0, 1, 2, 4, 5, 7) = (3, 6) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 184: /* (3, 4, 5, 7) = (0, 1, 2, 6) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 185: /* (0, 3, 4, 5, 7) = (1, 2, 6) case 6*/
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 186: /* (1, 3, 4, 5, 7) = (0, 2, 6) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 187: /* (0, 1, 3, 4, 5, 7) = (2, 6) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 188: /* (2, 3, 4, 5, 7) = (0, 1, 6) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 189: /* (0, 2, 3, 4, 5, 7) = (1, 6) case 4 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 190: /* (1, 2, 3, 4, 5, 7) = (0, 6) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 191: /* (0, 1, 2, 3, 4, 5, 7) = (6) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 192: /* (6, 7) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 193: /* (0, 6, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 194: /* (1, 6, 7) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 195: /* (0, 1, 6, 7) case 10 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[6], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[7], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 196: /* (2, 6, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 197: /* (0, 2, 6, 7) = (1, 3, 4, 5) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 198: /* (1, 2, 6, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 199: /* (0, 1, 2, 6, 7) = (3, 4, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 200: /* (3, 6, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 201: /* (0, 3, 6, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 202: /* (1, 3, 6, 7) = (0, 2, 4, 5) ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 203: /* (0, 1, 3, 6, 7) = (2, 4, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 204: /* (2, 3, 6, 7) case 8 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 205: /* (0, 2, 3, 6, 7) = (1, 4, 5) case 5*/
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 206: /* (1, 2, 3, 6, 7) = (0, 4, 5) case 5*/
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 207: /* (0, 1, 2, 3, 6, 7) = (4, 5) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 208: /* (4, 6, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 209: /* (0, 4, 6, 7) = (1, 2, 3, 5) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 210: /* (1, 4, 6, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 211: /* (0, 1, 4, 6, 7) = (2, 3, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 212: /* (2, 4, 6, 7) case 9 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 213: /* (0, 2, 4, 6, 7) = (1, 3, 5) case 5*/
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 214: /* (1, 2, 4, 6, 7) = (0, 3, 5) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 215: /* (0, 1, 2, 4, 6, 7) = (3, 5) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 216: /* (3, 4, 6, 7) = (0, 1, 2, 5) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 217: /* (0, 3, 4, 6, 7) = (1, 2, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 218: /* (1, 3, 4, 6, 7) = (0, 2, 5) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 219: /* (0, 1, 3, 4, 6, 7) = (2, 5) case 4 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 220: /* (2, 3, 4, 6, 7) = (0, 1, 5) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 221: /* (0, 2, 3, 4, 6, 7) = (1, 5) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 222: /* (1, 2, 3, 4, 6, 7) = (0, 5) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 223: /* (0, 1, 2, 3, 4, 6, 7) = (5) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[5], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 224: /* (5, 6, 7) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 225: /* (0, 5, 6, 7) case 12 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 226: /* (1, 5, 6, 7) = (0, 2, 3, 4) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 227: /* (0, 1, 5, 6, 7) = (2, 3, 4) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 228: /* (2, 5, 6, 7) = (0, 1, 3, 4) case ?? */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 229: /* (0, 2, 5, 6, 7) = (1, 3, 4) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 230: /* (1, 2, 5, 6, 7) = (0, 3, 4) case 6 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 231: /* (0, 1, 2, 5, 6, 7) = (3, 4) case 4 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 232: /* (3, 5, 6, 7) case 9 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 233: /* (0, 3, 5, 6, 7) = (1, 2, 4) case 7 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 234: /* (1, 3, 5, 6, 7) = (0, 2, 4) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 235: /* (0, 1, 3, 5, 6, 7) = (2, 4) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 236: /* (2, 3, 5, 6, 7) = (0, 1, 4) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 237: /* (0, 2, 3, 5, 6, 7) = (1, 4) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 238: /* (1, 2, 3, 5, 6, 7) = (0, 4) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 239: /* (0, 1, 2, 3, 5, 6, 7) = (4) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[4], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 240: /* (4, 5, 6, 7) case 8 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 241: /* (0, 4, 5, 6, 7) = (1, 2, 3) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 242: /* (1, 4, 5, 6, 7) = (0, 2, 3) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 243: /* (0, 1, 4, 5, 6, 7) = (2, 3) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 244: /* (2, 4, 5, 6, 7) = (0, 1, 3) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 245: /* (0, 2, 4, 5, 6, 7) = (1, 3) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 246: /* (1, 2, 4, 5, 6, 7) = (0, 3) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 247: /* (0, 1, 2, 4, 5, 6, 7) = (3) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[3], Cube[7], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 248: /* (3, 4, 5, 6, 7) = (0, 1, 2) case 5 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P1, N1);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 249: /* (0, 3, 4, 5, 6, 7) = (1, 2) case 3 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 250: /* (1, 3, 4, 5, 6, 7) = (0, 2) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 251: /* (0, 1, 3, 4, 5, 6, 7) = (2) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[2], Cube[6], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 252: /* (2, 3, 4, 5, 6, 7) = (0, 1) case 2 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P3, N3);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P4, N4);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      AddTriangleToCSGMesh(CSG, P2,P3,P4, N2,N3,N4);
      break;
    case 253: /* (0, 2, 3, 4, 5, 6, 7) = (1) case 1 */
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[0], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[3], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[1], Cube[5], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 254: /* (1, 2, 3, 4, 5, 6, 7) = (0) case 0 */
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[1], P1, N1);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[2], P2, N2);
      giram_object_find_intersection_segment(CSG, Cube[0], Cube[4], P3, N3);
      AddTriangleToCSGMesh(CSG, P1,P2,P3, N1,N2,N3);
      break;
    case 255: /* case 0: nothing to do */
      break;
  }
}

